;==============================================================================

; file: "_io#.scm"

; Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.

;==============================================================================

; Representation of ports.

; There are 5 kinds of ports, each providing a set of operations.  All
; port objects have the capability of being both an input port and an
; output port.  The "none-port" kind provides no operation and is
; mainly for internal use to indicate that no input operation is
; available or that no output operation is available.
;
; 1) An "object-port" provides operations to read and write Scheme
;    data (i.e. any Scheme object) to/from the port.  It also provides
;    operations to get the name of the port, to force output to occur,
;    and to close the port.  This kind of port need not be connected
;    to a character based device or file (it could for example be a
;    FIFO queue linking two threads that need to communicate Scheme
;    objects).
;
; 2) A "char-port" provides all the operations of an "object-port",
;    and also operations to read and write individual characters
;    to/from the port.  When a Scheme object is written to a
;    char-port, it is converted into the sequence of characters that
;    corresponds to its "external-representation".  When reading a
;    Scheme object, an inverse conversion occurs.
;
; 3) A "byte-port" provides all the operations of a "char-port", and
;    also operations to read and write individual bytes to/from the
;    port.  When a **character** is written to a byte-port, some
;    encoding of that character into a sequence of bytes will occur
;    (for example, #\newline might be encoded as the 2 bytes CR-LF
;    when using LATIN-1 encoding, or a non-ASCII character will
;    generate more than 1 byte when using UTF8 encoding).  When
;    reading a character, a similar decoding occurs.
;
; 4) A "device-port" provides all the operations of a "byte-port", and
;    also operations to control the device (file, tty, etc) that is
;    connected to the port, such as changing the tty settings.

(##define-macro (macro-port? obj)
  `(macro-structure-object? ,obj ##port-tag))

(##define-macro (macro-check-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-port ',(car form) ,@(cdr form))
          `(##trap-check-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-trap-open-file form)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  (if (list? form)
    `(##trap-open-file ',(car form) ,@(cdr form))
    `(##trap-open-file* ',(car form) ,@(flat (cdr form)))))

(##define-macro (macro-port-of-rkind? obj kind)
  `(let ((obj ,obj))
     (and (macro-port? obj)
          (##fixnum.= (##fixnum.logand (macro-port-rkind obj) ,kind)
                      ,kind))))

(##define-macro (macro-port-of-wkind? obj kind)
  `(let ((obj ,obj))
     (and (macro-port? obj)
          (##fixnum.= (##fixnum.logand (macro-port-wkind obj) ,kind)
                      ,kind))))

(##define-macro (macro-none-kind)    0) ; allows nothing
(##define-macro (macro-object-kind)  1) ; can read and write objects
(##define-macro (macro-char-kind)    3) ; can also read and write chars
(##define-macro (macro-byte-kind)    7) ; can also read and write bytes
(##define-macro (macro-device-kind) 15) ; can also do device operations

(##define-macro (macro-file-device-kind)       (+ 15 16))
(##define-macro (macro-tty-device-kind)        (+ 15 32))
(##define-macro (macro-tcp-client-device-kind) (+ 15 64))
(##define-macro (macro-tcp-server-kind)        (+ 1 128))
(##define-macro (macro-pipe-device-kind)       (+ 15 256))
(##define-macro (macro-serial-device-kind)     (+ 15 512))
(##define-macro (macro-directory-kind)         (+ 1 1024))
(##define-macro (macro-event-queue-kind)       (+ 1 2048))
(##define-macro (macro-timer-kind)             (+ 1 4096))
(##define-macro (macro-vector-kind)            (+ 1 8192))
(##define-macro (macro-string-kind)            (+ 3 16384))
(##define-macro (macro-u8vector-kind)          (+ 7 32768))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of object ports.

(##define-macro
  (macro-make-object-port

    mutex              ; access to the port is controlled with this mutex

    rkind              ; port kind for reading (none-port if can't read)
    wkind              ; port kind for writing (none-port if can't write)

    name               ; procedure which returns the name of the port
    read-datum         ; procedure to read a datum
    write-datum        ; procedure to write a datum
    newline            ; procedure to write a datum separator
    force-output       ; procedure to force output to occur on target device
    close              ; procedure to close the port
    roptions           ; options for reading (buffering type, encoding, etc)
    rtimeout           ; time at which a read that would block times out
    rtimeout-thunk     ; thunk called when a read timeout occurs
    set-rtimeout       ; procedure to set rtimeout and rtimeout-thunk
    woptions           ; options for writing (buffering type, encoding, etc)
    wtimeout           ; time at which a write that would block times out
    wtimeout-thunk     ; thunk called when a write timeout occurs
    set-wtimeout       ; procedure to set wtimeout and wtimeout-thunk

    . others           ; other slots
  )

  `(macro-make-structure-object

    ##port-tag

    ,mutex

    ,rkind
    ,wkind

    ,name
    ,read-datum
    ,write-datum
    ,newline
    ,force-output
    ,close
    ,roptions
    ,rtimeout
    ,rtimeout-thunk
    ,set-rtimeout
    ,woptions
    ,wtimeout
    ,wtimeout-thunk
    ,set-wtimeout

    ,@others))

(##define-macro (macro-port-mutex p)                `(macro-slot 1 ,p))
(##define-macro (macro-port-mutex-set! p x)         `(macro-slot 1 ,p ,x))
(##define-macro (macro-port-rkind p)                `(macro-slot 2 ,p))
(##define-macro (macro-port-rkind-set! p x)         `(macro-slot 2 ,p ,x))
(##define-macro (macro-port-wkind p)                `(macro-slot 3 ,p))
(##define-macro (macro-port-wkind-set! p x)         `(macro-slot 3 ,p ,x))
(##define-macro (macro-port-name p)                 `(macro-slot 4 ,p))
(##define-macro (macro-port-name-set! p x)          `(macro-slot 4 ,p ,x))
(##define-macro (macro-port-read-datum p)           `(macro-slot 5 ,p))
(##define-macro (macro-port-read-datum-set! p x)    `(macro-slot 5 ,p ,x))
(##define-macro (macro-port-write-datum p)          `(macro-slot 6 ,p))
(##define-macro (macro-port-write-datum-set! p x)   `(macro-slot 6 ,p ,x))
(##define-macro (macro-port-newline p)              `(macro-slot 7 ,p))
(##define-macro (macro-port-newline-set! p x)       `(macro-slot 7 ,p ,x))
(##define-macro (macro-port-force-output p)         `(macro-slot 8 ,p))
(##define-macro (macro-port-force-output-set! p x)  `(macro-slot 8 ,p ,x))
(##define-macro (macro-port-close p)                `(macro-slot 9 ,p))
(##define-macro (macro-port-close-set! p x)         `(macro-slot 9 ,p ,x))
(##define-macro (macro-port-roptions p)             `(macro-slot 10 ,p))
(##define-macro (macro-port-roptions-set! p x)      `(macro-slot 10 ,p ,x))
(##define-macro (macro-port-rtimeout p)             `(macro-slot 11 ,p))
(##define-macro (macro-port-rtimeout-set! p x)      `(macro-slot 11 ,p ,x))
(##define-macro (macro-port-rtimeout-thunk p)       `(macro-slot 12 ,p))
(##define-macro (macro-port-rtimeout-thunk-set! p x)`(macro-slot 12 ,p ,x))
(##define-macro (macro-port-set-rtimeout p)         `(macro-slot 13 ,p))
(##define-macro (macro-port-set-rtimeout-set! p x)  `(macro-slot 13 ,p ,x))
(##define-macro (macro-port-woptions p)             `(macro-slot 14 ,p))
(##define-macro (macro-port-woptions-set! p x)      `(macro-slot 14 ,p ,x))
(##define-macro (macro-port-wtimeout p)             `(macro-slot 15 ,p))
(##define-macro (macro-port-wtimeout-set! p x)      `(macro-slot 15 ,p ,x))
(##define-macro (macro-port-wtimeout-thunk p)       `(macro-slot 16 ,p))
(##define-macro (macro-port-wtimeout-thunk-set! p x)`(macro-slot 16 ,p ,x))
(##define-macro (macro-port-set-wtimeout p)         `(macro-slot 17 ,p))
(##define-macro (macro-port-set-wtimeout-set! p x)  `(macro-slot 17 ,p ,x))

(##define-macro (macro-port-object-other1 p)        `(macro-slot 18 ,p))
(##define-macro (macro-port-object-other1-set! p x) `(macro-slot 18 ,p ,x))
(##define-macro (macro-port-object-other2 p)        `(macro-slot 19 ,p))
(##define-macro (macro-port-object-other2-set! p x) `(macro-slot 19 ,p ,x))
(##define-macro (macro-port-object-other3 p)        `(macro-slot 20 ,p))
(##define-macro (macro-port-object-other3-set! p x) `(macro-slot 20 ,p ,x))
(##define-macro (macro-port-object-other4 p)        `(macro-slot 21 ,p))
(##define-macro (macro-port-object-other4-set! p x) `(macro-slot 21 ,p ,x))
(##define-macro (macro-port-object-other5 p)        `(macro-slot 22 ,p))
(##define-macro (macro-port-object-other5-set! p x) `(macro-slot 22 ,p ,x))
(##define-macro (macro-port-object-other6 p)        `(macro-slot 23 ,p))
(##define-macro (macro-port-object-other6-set! p x) `(macro-slot 23 ,p ,x))
(##define-macro (macro-port-object-other7 p)        `(macro-slot 24 ,p))
(##define-macro (macro-port-object-other7-set! p x) `(macro-slot 24 ,p ,x))
(##define-macro (macro-port-object-other8 p)        `(macro-slot 25 ,p))
(##define-macro (macro-port-object-other8-set! p x) `(macro-slot 25 ,p ,x))
(##define-macro (macro-port-object-other9 p)        `(macro-slot 26 ,p))
(##define-macro (macro-port-object-other9-set! p x) `(macro-slot 26 ,p ,x))
(##define-macro (macro-port-object-other10 p)       `(macro-slot 27 ,p))
(##define-macro (macro-port-object-other10-set! p x)`(macro-slot 27 ,p ,x))
(##define-macro (macro-port-object-other11 p)       `(macro-slot 28 ,p))
(##define-macro (macro-port-object-other11-set! p x)`(macro-slot 28 ,p ,x))
(##define-macro (macro-port-object-other12 p)       `(macro-slot 29 ,p))
(##define-macro (macro-port-object-other12-set! p x)`(macro-slot 29 ,p ,x))

(##define-macro (macro-input-port? obj)
  `(macro-port-of-rkind? ,obj (macro-object-kind)))

(##define-macro (macro-output-port? obj)
  `(macro-port-of-wkind? ,obj (macro-object-kind)))

(##define-macro (macro-check-input-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-input-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-input-port ',(car form) ,@(cdr form))
          `(##trap-check-input-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-check-output-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-output-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-output-port ',(car form) ,@(cdr form))
          `(##trap-check-output-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of char ports.

(##define-macro
  (macro-make-char-port

    mutex              ; access to the port is controlled with this mutex

    rkind              ; port kind for reading (none-port if can't read)
    wkind              ; port kind for writing (none-port if can't write)

    name               ; procedure which returns the name of the port
    read-datum         ; procedure to read a datum
    write-datum        ; procedure to write a datum
    newline            ; procedure to write a datum separator
    force-output       ; procedure to force output to occur on target device
    close              ; procedure to close the port
    roptions           ; options for reading (buffering type, encoding, etc)
    rtimeout           ; time at which a read that would block times out
    rtimeout-thunk     ; thunk called when a read timeout occurs
    set-rtimeout       ; procedure to set rtimeout and rtimeout-thunk
    woptions           ; options for writing (buffering type, encoding, etc)
    wtimeout           ; time at which a write that would block times out
    wtimeout-thunk     ; thunk called when a write timeout occurs
    set-wtimeout       ; procedure to set wtimeout and wtimeout-thunk

    char-rbuf          ; character read buffer (a string)
    char-rlo           ; low pointer (start of unread characters)
    char-rhi           ; high pointer (end of unread characters)
    char-rchars        ; number of characters read at start of read buffer
    char-rlines        ; number of lines read up to low pointer
    char-rcurline      ; absolute character position where current line starts
    char-rbuf-fill     ; procedure to read characters into the read buffer
    char-peek-eof?     ; peeking the next character should return end-of-file?

    char-wbuf          ; character write buffer (a string)
    char-wlo           ; low pointer (start of unwritten characters)
    char-whi           ; high pointer (end of unwritten characters)
    char-wchars        ; number of characters written at start of write buffer
    char-wlines        ; number of lines written up to high pointer
    char-wcurline      ; absolute character position where current line starts
    char-wbuf-drain    ; procedure to write characters from the write buffer

    output-width       ; procedure to get the output width in characters

    . others           ; other slots
  )

  `(macro-make-object-port

    ,mutex

    ,rkind
    ,wkind

    ,name
    ,read-datum
    ,write-datum
    ,newline
    ,force-output
    ,close
    ,roptions
    ,rtimeout
    ,rtimeout-thunk
    ,set-rtimeout
    ,woptions
    ,wtimeout
    ,wtimeout-thunk
    ,set-wtimeout

    ,char-rbuf
    ,char-rlo
    ,char-rhi
    ,char-rchars
    ,char-rlines
    ,char-rcurline
    ,char-rbuf-fill
    ,char-peek-eof?

    ,char-wbuf
    ,char-wlo
    ,char-whi
    ,char-wchars
    ,char-wlines
    ,char-wcurline
    ,char-wbuf-drain

    ,output-width

    ,@others))

(##define-macro (macro-port-char-rbuf p)              `(macro-slot 18 ,p))
(##define-macro (macro-port-char-rbuf-set! p x)       `(macro-slot 18 ,p ,x))
(##define-macro (macro-port-char-rlo p)               `(macro-slot 19 ,p))
(##define-macro (macro-port-char-rlo-set! p x)        `(macro-slot 19 ,p ,x))
(##define-macro (macro-port-char-rhi p)               `(macro-slot 20 ,p))
(##define-macro (macro-port-char-rhi-set! p x)        `(macro-slot 20 ,p ,x))
(##define-macro (macro-port-char-rchars p)            `(macro-slot 21 ,p))
(##define-macro (macro-port-char-rchars-set! p x)     `(macro-slot 21 ,p ,x))
(##define-macro (macro-port-char-rlines p)            `(macro-slot 22 ,p))
(##define-macro (macro-port-char-rlines-set! p x)     `(macro-slot 22 ,p ,x))
(##define-macro (macro-port-char-rcurline p)          `(macro-slot 23 ,p))
(##define-macro (macro-port-char-rcurline-set! p x)   `(macro-slot 23 ,p ,x))
(##define-macro (macro-port-char-rbuf-fill p)         `(macro-slot 24 ,p))
(##define-macro (macro-port-char-rbuf-fill-set! p x)  `(macro-slot 24 ,p ,x))
(##define-macro (macro-port-char-peek-eof? p)         `(macro-slot 25 ,p))
(##define-macro (macro-port-char-peek-eof?-set! p x)  `(macro-slot 25 ,p ,x))

(##define-macro (macro-port-char-wbuf p)              `(macro-slot 26 ,p))
(##define-macro (macro-port-char-wbuf-set! p x)       `(macro-slot 26 ,p ,x))
(##define-macro (macro-port-char-wlo p)               `(macro-slot 27 ,p))
(##define-macro (macro-port-char-wlo-set! p x)        `(macro-slot 27 ,p ,x))
(##define-macro (macro-port-char-whi p)               `(macro-slot 28 ,p))
(##define-macro (macro-port-char-whi-set! p x)        `(macro-slot 28 ,p ,x))
(##define-macro (macro-port-char-wchars p)            `(macro-slot 29 ,p))
(##define-macro (macro-port-char-wchars-set! p x)     `(macro-slot 29 ,p ,x))
(##define-macro (macro-port-char-wlines p)            `(macro-slot 30 ,p))
(##define-macro (macro-port-char-wlines-set! p x)     `(macro-slot 30 ,p ,x))
(##define-macro (macro-port-char-wcurline p)          `(macro-slot 31 ,p))
(##define-macro (macro-port-char-wcurline-set! p x)   `(macro-slot 31 ,p ,x))
(##define-macro (macro-port-char-wbuf-drain p)        `(macro-slot 32 ,p))
(##define-macro (macro-port-char-wbuf-drain-set! p x) `(macro-slot 32 ,p ,x))
(##define-macro (macro-port-output-width p)           `(macro-slot 33 ,p))
(##define-macro (macro-port-output-width-set! p x)    `(macro-slot 33 ,p ,x))

(##define-macro (macro-port-char-other1 p)            `(macro-slot 34 ,p))
(##define-macro (macro-port-char-other1-set! p x)     `(macro-slot 34 ,p ,x))
(##define-macro (macro-port-char-other2 p)            `(macro-slot 35 ,p))
(##define-macro (macro-port-char-other2-set! p x)     `(macro-slot 35 ,p ,x))
(##define-macro (macro-port-char-other3 p)            `(macro-slot 36 ,p))
(##define-macro (macro-port-char-other3-set! p x)     `(macro-slot 36 ,p ,x))
(##define-macro (macro-port-char-other4 p)            `(macro-slot 37 ,p))
(##define-macro (macro-port-char-other4-set! p x)     `(macro-slot 37 ,p ,x))
(##define-macro (macro-port-char-other5 p)            `(macro-slot 38 ,p))
(##define-macro (macro-port-char-other5-set! p x)     `(macro-slot 38 ,p ,x))

(##define-macro (macro-char-input-port? obj)
  `(macro-port-of-rkind? ,obj (macro-char-kind)))

(##define-macro (macro-char-output-port? obj)
  `(macro-port-of-wkind? ,obj (macro-char-kind)))

(##define-macro (macro-check-char-input-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-char-input-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-char-input-port ',(car form) ,@(cdr form))
          `(##trap-check-char-input-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-check-char-output-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-char-output-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-char-output-port ',(car form) ,@(cdr form))
          `(##trap-check-char-output-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of byte ports.

(##define-macro
  (macro-make-byte-port

    mutex              ; access to the port is controlled with this mutex

    rkind              ; port kind for reading (none-port if can't read)
    wkind              ; port kind for writing (none-port if can't write)

    name               ; procedure which returns the name of the port
    read-datum         ; procedure to read a datum
    write-datum        ; procedure to write a datum
    newline            ; procedure to write a datum separator
    force-output       ; procedure to force output to occur on target device
    close              ; procedure to close the port
    roptions           ; options for reading (buffering type, encoding, etc)
    rtimeout           ; time at which a read that would block times out
    rtimeout-thunk     ; thunk called when a read timeout occurs
    set-rtimeout       ; procedure to set rtimeout and rtimeout-thunk
    woptions           ; options for writing (buffering type, encoding, etc)
    wtimeout           ; time at which a write that would block times out
    wtimeout-thunk     ; thunk called when a write timeout occurs
    set-wtimeout       ; procedure to set wtimeout and wtimeout-thunk

    char-rbuf          ; character read buffer (a string)
    char-rlo           ; low pointer (start of unread characters)
    char-rhi           ; high pointer (end of unread characters)
    char-rchars        ; number of characters read at start of read buffer
    char-rlines        ; number of lines read up to low pointer
    char-rcurline      ; absolute character position where current line starts
    char-rbuf-fill     ; procedure to read characters into the read buffer
    char-peek-eof?     ; peeking the next character should return end-of-file?

    char-wbuf          ; character write buffer (a string)
    char-wlo           ; low pointer (start of unwritten characters)
    char-whi           ; high pointer (end of unwritten characters)
    char-wchars        ; number of characters written at start of write buffer
    char-wlines        ; number of lines written up to high pointer
    char-wcurline      ; absolute character position where current line starts
    char-wbuf-drain    ; procedure to write characters from the write buffer

    output-width       ; procedure to get the output width in characters

    byte-rbuf          ; byte read buffer (a u8vector)
    byte-rlo           ; low pointer (start of unread bytes)
    byte-rhi           ; high pointer (end of unread bytes)
    byte-rbuf-fill     ; procedure to read bytes into the read buffer

    byte-wbuf          ; byte write buffer (a u8vector)
    byte-wlo           ; low pointer (start of unwritten bytes)
    byte-whi           ; high pointer (end of unwritten bytes)
    byte-wbuf-drain    ; procedure to write bytes from the write buffer

    . others           ; other slots
  )

  `(macro-make-char-port

    ,mutex

    ,rkind
    ,wkind

    ,name
    ,read-datum
    ,write-datum
    ,newline
    ,force-output
    ,close
    ,roptions
    ,rtimeout
    ,rtimeout-thunk
    ,set-rtimeout
    ,woptions
    ,wtimeout
    ,wtimeout-thunk
    ,set-wtimeout

    ,char-rbuf
    ,char-rlo
    ,char-rhi
    ,char-rchars
    ,char-rlines
    ,char-rcurline
    ,char-rbuf-fill
    ,char-peek-eof?

    ,char-wbuf
    ,char-wlo
    ,char-whi
    ,char-wchars
    ,char-wlines
    ,char-wcurline
    ,char-wbuf-drain

    ,output-width

    ,byte-rbuf
    ,byte-rlo
    ,byte-rhi
    ,byte-rbuf-fill

    ,byte-wbuf
    ,byte-wlo
    ,byte-whi
    ,byte-wbuf-drain

    ,@others))

(##define-macro (macro-port-byte-rbuf p)              `(macro-slot 34 ,p))
(##define-macro (macro-port-byte-rbuf-set! p x)       `(macro-slot 34 ,p ,x))
(##define-macro (macro-port-byte-rlo p)               `(macro-slot 35 ,p))
(##define-macro (macro-port-byte-rlo-set! p x)        `(macro-slot 35 ,p ,x))
(##define-macro (macro-port-byte-rhi p)               `(macro-slot 36 ,p))
(##define-macro (macro-port-byte-rhi-set! p x)        `(macro-slot 36 ,p ,x))
(##define-macro (macro-port-byte-rbuf-fill p)         `(macro-slot 37 ,p))
(##define-macro (macro-port-byte-rbuf-fill-set! p x)  `(macro-slot 37 ,p ,x))

(##define-macro (macro-port-byte-wbuf p)              `(macro-slot 38 ,p))
(##define-macro (macro-port-byte-wbuf-set! p x)       `(macro-slot 38 ,p ,x))
(##define-macro (macro-port-byte-wlo p)               `(macro-slot 39 ,p))
(##define-macro (macro-port-byte-wlo-set! p x)        `(macro-slot 39 ,p ,x))
(##define-macro (macro-port-byte-whi p)               `(macro-slot 40 ,p))
(##define-macro (macro-port-byte-whi-set! p x)        `(macro-slot 40 ,p ,x))
(##define-macro (macro-port-byte-wbuf-drain p)        `(macro-slot 41 ,p))
(##define-macro (macro-port-byte-wbuf-drain-set! p x) `(macro-slot 41 ,p ,x))

(##define-macro (macro-port-byte-other1 p)            `(macro-slot 42 ,p))
(##define-macro (macro-port-byte-other1-set! p x)     `(macro-slot 42 ,p ,x))
(##define-macro (macro-port-byte-other2 p)            `(macro-slot 43 ,p))
(##define-macro (macro-port-byte-other2-set! p x)     `(macro-slot 43 ,p ,x))
(##define-macro (macro-port-byte-other3 p)            `(macro-slot 44 ,p))
(##define-macro (macro-port-byte-other3-set! p x)     `(macro-slot 44 ,p ,x))
(##define-macro (macro-port-byte-other4 p)            `(macro-slot 45 ,p))
(##define-macro (macro-port-byte-other4-set! p x)     `(macro-slot 45 ,p ,x))
(##define-macro (macro-port-byte-other5 p)            `(macro-slot 46 ,p))
(##define-macro (macro-port-byte-other5-set! p x)     `(macro-slot 46 ,p ,x))

(##define-macro (macro-byte-input-port? obj)
  `(macro-port-of-rkind? ,obj (macro-byte-kind)))

(##define-macro (macro-byte-output-port? obj)
  `(macro-port-of-wkind? ,obj (macro-byte-kind)))

(##define-macro (macro-check-byte-input-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-byte-input-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-byte-input-port ',(car form) ,@(cdr form))
          `(##trap-check-byte-input-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-check-byte-output-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-byte-output-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-byte-output-port ',(car form) ,@(cdr form))
          `(##trap-check-byte-output-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of device ports.

(##define-macro
  (macro-make-device-port

    mutex              ; access to the port is controlled with this mutex

    rkind              ; port kind for reading (none-port if can't read)
    wkind              ; port kind for writing (none-port if can't write)

    name               ; procedure which returns the name of the port
    read-datum         ; procedure to read a datum
    write-datum        ; procedure to write a datum
    newline            ; procedure to write a datum separator
    force-output       ; procedure to force output to occur on target device
    close              ; procedure to close the port
    roptions           ; options for reading (buffering type, encoding, etc)
    rtimeout           ; time at which a read that would block times out
    rtimeout-thunk     ; thunk called when a read timeout occurs
    set-rtimeout       ; procedure to set rtimeout and rtimeout-thunk
    woptions           ; options for writing (buffering type, encoding, etc)
    wtimeout           ; time at which a write that would block times out
    wtimeout-thunk     ; thunk called when a write timeout occurs
    set-wtimeout       ; procedure to set wtimeout and wtimeout-thunk

    char-rbuf          ; character read buffer (a string)
    char-rlo           ; low pointer (start of unread characters)
    char-rhi           ; high pointer (end of unread characters)
    char-rchars        ; number of characters read at start of read buffer
    char-rlines        ; number of lines read up to low pointer
    char-rcurline      ; absolute character position where current line starts
    char-rbuf-fill     ; procedure to read characters into the read buffer
    char-peek-eof?     ; peeking the next character should return end-of-file?

    char-wbuf          ; character write buffer (a string)
    char-wlo           ; low pointer (start of unwritten characters)
    char-whi           ; high pointer (end of unwritten characters)
    char-wchars        ; number of characters written at start of write buffer
    char-wlines        ; number of lines written up to high pointer
    char-wcurline      ; absolute character position where current line starts
    char-wbuf-drain    ; procedure to write characters from the write buffer

    output-width       ; procedure to get the output width in characters

    byte-rbuf          ; byte read buffer (a u8vector)
    byte-rlo           ; low pointer (start of unread bytes)
    byte-rhi           ; high pointer (end of unread bytes)
    byte-rbuf-fill     ; procedure to read bytes into the read buffer

    byte-wbuf          ; byte write buffer (a u8vector)
    byte-wlo           ; low pointer (start of unwritten bytes)
    byte-whi           ; high pointer (end of unwritten bytes)
    byte-wbuf-drain    ; procedure to write bytes from the write buffer

    rdevice-condvar    ; device condvar from which bytes are read
    wdevice-condvar    ; device condvar to which bytes are written

    . others           ; other slots
  )

  `(macro-make-byte-port

    ,mutex

    ,rkind
    ,wkind

    ,name
    ,read-datum
    ,write-datum
    ,newline
    ,force-output
    ,close
    ,roptions
    ,rtimeout
    ,rtimeout-thunk
    ,set-rtimeout
    ,woptions
    ,wtimeout
    ,wtimeout-thunk
    ,set-wtimeout

    ,char-rbuf
    ,char-rlo
    ,char-rhi
    ,char-rchars
    ,char-rlines
    ,char-rcurline
    ,char-rbuf-fill
    ,char-peek-eof?

    ,char-wbuf
    ,char-wlo
    ,char-whi
    ,char-wchars
    ,char-wlines
    ,char-wcurline
    ,char-wbuf-drain

    ,output-width

    ,byte-rbuf
    ,byte-rlo
    ,byte-rhi
    ,byte-rbuf-fill

    ,byte-wbuf
    ,byte-wlo
    ,byte-whi
    ,byte-wbuf-drain

    ,rdevice-condvar
    ,wdevice-condvar

    ,@others))

(##define-macro (macro-port-rdevice-condvar p)        `(macro-slot 42 ,p))
(##define-macro (macro-port-rdevice-condvar-set! p x) `(macro-slot 42 ,p ,x))
(##define-macro (macro-port-wdevice-condvar p)        `(macro-slot 43 ,p))
(##define-macro (macro-port-wdevice-condvar-set! p x) `(macro-slot 43 ,p ,x))
(##define-macro (macro-port-device-name p)            `(macro-slot 44 ,p))
(##define-macro (macro-port-device-name-set! p x)     `(macro-slot 44 ,p ,x))

(##define-macro (macro-port-device-other1 p)          `(macro-slot 45 ,p))
(##define-macro (macro-port-device-other1-set! p x)   `(macro-slot 45 ,p ,x))
(##define-macro (macro-port-device-other2 p)          `(macro-slot 46 ,p))
(##define-macro (macro-port-device-other2-set! p x)   `(macro-slot 46 ,p ,x))

(##define-macro (macro-device-input-port? obj)
  `(macro-port-of-rkind? ,obj (macro-device-kind)))

(##define-macro (macro-device-output-port? obj)
  `(macro-port-of-wkind? ,obj (macro-device-kind)))

(##define-macro (macro-check-device-input-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-device-input-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-device-input-port ',(car form) ,@(cdr form))
          `(##trap-check-device-input-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-check-device-output-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-device-output-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-device-output-port ',(car form) ,@(cdr form))
          `(##trap-check-device-output-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of vector, string and u8vector ports.

(##define-macro (macro-port-vector-rbuf p)
  `(macro-port-object-other1 ,p))
(##define-macro (macro-port-vector-rbuf-set! p x)
  `(macro-port-object-other1-set! ,p ,x))
(##define-macro (macro-port-vector-rlo p)
  `(macro-port-object-other2 ,p))
(##define-macro (macro-port-vector-rlo-set! p x)
  `(macro-port-object-other2-set! ,p ,x))
(##define-macro (macro-port-vector-rhi p)
  `(macro-port-object-other3 ,p))
(##define-macro (macro-port-vector-rhi-set! p x)
  `(macro-port-object-other3-set! ,p ,x))
(##define-macro (macro-port-vector-rbuf-fill p)
  `(macro-port-object-other4 ,p))
(##define-macro (macro-port-vector-rbuf-fill-set! p x)
  `(macro-port-object-other4-set! ,p ,x))
(##define-macro (macro-port-vector-wbuf p)
  `(macro-port-object-other5 ,p))
(##define-macro (macro-port-vector-wbuf-set! p x)
  `(macro-port-object-other5-set! ,p ,x))
(##define-macro (macro-port-vector-wlo p)
  `(macro-port-object-other6 ,p))
(##define-macro (macro-port-vector-wlo-set! p x)
  `(macro-port-object-other6-set! ,p ,x))
(##define-macro (macro-port-vector-whi p)
  `(macro-port-object-other7 ,p))
(##define-macro (macro-port-vector-whi-set! p x)
  `(macro-port-object-other7-set! ,p ,x))
(##define-macro (macro-port-vector-wbuf-drain p)
  `(macro-port-object-other8 ,p))
(##define-macro (macro-port-vector-wbuf-drain-set! p x)
  `(macro-port-object-other8-set! ,p ,x))
(##define-macro (macro-port-vector-fifo p)
  `(macro-port-object-other9 ,p))
(##define-macro (macro-port-vector-fifo-set! p x)
  `(macro-port-object-other9-set! ,p ,x))
(##define-macro (macro-port-vector-rcondvar p)
  `(macro-port-object-other10 ,p))
(##define-macro (macro-port-vector-rcondvar-set! p x)
  `(macro-port-object-other10-set! ,p ,x))
(##define-macro (macro-port-vector-wcondvar p)
  `(macro-port-object-other11 ,p))
(##define-macro (macro-port-vector-wcondvar-set! p x)
  `(macro-port-object-other11-set! ,p ,x))
(##define-macro (macro-port-vector-buffering-limit p)
  `(macro-port-object-other12 ,p))
(##define-macro (macro-port-vector-buffering-limit-set! p x)
  `(macro-port-object-other12-set! ,p ,x))

(##define-macro (macro-vector-input-port? obj)
  `(macro-port-of-rkind? ,obj (macro-vector-kind)))

(##define-macro (macro-vector-output-port? obj)
  `(macro-port-of-wkind? ,obj (macro-vector-kind)))

(##define-macro (macro-check-vector-input-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-vector-input-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-vector-input-port ',(car form) ,@(cdr form))
          `(##trap-check-vector-input-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-check-vector-output-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-vector-output-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-vector-output-port ',(car form) ,@(cdr form))
          `(##trap-check-vector-output-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-port-string-rbuf p)
  `(macro-port-char-rbuf ,p))
(##define-macro (macro-port-string-rbuf-set! p x)
  `(macro-port-char-rbuf-set! ,p ,x))
(##define-macro (macro-port-string-rlo p)
  `(macro-port-char-rlo ,p))
(##define-macro (macro-port-string-rlo-set! p x)
  `(macro-port-char-rlo-set! ,p ,x))
(##define-macro (macro-port-string-rhi p)
  `(macro-port-char-rhi ,p))
(##define-macro (macro-port-string-rhi-set! p x)
  `(macro-port-char-rhi-set! ,p ,x))
(##define-macro (macro-port-string-rbuf-fill p)
  `(macro-port-char-rbuf-fill ,p))
(##define-macro (macro-port-string-rbuf-fill-set! p x)
  `(macro-port-char-rbuf-fill-set! ,p ,x))
(##define-macro (macro-port-string-wbuf p)
  `(macro-port-char-wbuf ,p))
(##define-macro (macro-port-string-wbuf-set! p x)
  `(macro-port-char-wbuf-set! ,p ,x))
(##define-macro (macro-port-string-wlo p)
  `(macro-port-char-wlo ,p))
(##define-macro (macro-port-string-wlo-set! p x)
  `(macro-port-char-wlo-set! ,p ,x))
(##define-macro (macro-port-string-whi p)
  `(macro-port-char-whi ,p))
(##define-macro (macro-port-string-whi-set! p x)
  `(macro-port-char-whi-set! ,p ,x))
(##define-macro (macro-port-string-wbuf-drain p)
  `(macro-port-char-wbuf-drain ,p))
(##define-macro (macro-port-string-wbuf-drain-set! p x)
  `(macro-port-char-wbuf-drain-set! ,p ,x))
(##define-macro (macro-port-string-fifo p)
  `(macro-port-char-other1 ,p))
(##define-macro (macro-port-string-fifo-set! p x)
  `(macro-port-char-other1-set! ,p ,x))
(##define-macro (macro-port-string-rcondvar p)
  `(macro-port-char-other2 ,p))
(##define-macro (macro-port-string-rcondvar-set! p x)
  `(macro-port-char-other2-set! ,p ,x))
(##define-macro (macro-port-string-wcondvar p)
  `(macro-port-char-other3 ,p))
(##define-macro (macro-port-string-wcondvar-set! p x)
  `(macro-port-char-other3-set! ,p ,x))
(##define-macro (macro-port-string-width p)
  `(macro-port-char-other4 ,p))
(##define-macro (macro-port-string-width-set! p x)
  `(macro-port-char-other4-set! ,p ,x))
(##define-macro (macro-port-string-buffering-limit p)
  `(macro-port-char-other5 ,p))
(##define-macro (macro-port-string-buffering-limit-set! p x)
  `(macro-port-char-other5-set! ,p ,x))

(##define-macro (macro-string-input-port? obj)
  `(macro-port-of-rkind? ,obj (macro-string-kind)))

(##define-macro (macro-string-output-port? obj)
  `(macro-port-of-wkind? ,obj (macro-string-kind)))

(##define-macro (macro-check-string-input-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-string-input-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-string-input-port ',(car form) ,@(cdr form))
          `(##trap-check-string-input-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-check-string-output-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-string-output-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-string-output-port ',(car form) ,@(cdr form))
          `(##trap-check-string-output-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-port-u8vector-rbuf p)
  `(macro-port-byte-rbuf ,p))
(##define-macro (macro-port-u8vector-rbuf-set! p x)
  `(macro-port-byte-rbuf-set! ,p ,x))
(##define-macro (macro-port-u8vector-rlo p)
  `(macro-port-byte-rlo ,p))
(##define-macro (macro-port-u8vector-rlo-set! p x)
  `(macro-port-byte-rlo-set! ,p ,x))
(##define-macro (macro-port-u8vector-rhi p)
  `(macro-port-byte-rhi ,p))
(##define-macro (macro-port-u8vector-rhi-set! p x)
  `(macro-port-byte-rhi-set! ,p ,x))
(##define-macro (macro-port-u8vector-rbuf-fill p)
  `(macro-port-byte-rbuf-fill ,p))
(##define-macro (macro-port-u8vector-rbuf-fill-set! p x)
  `(macro-port-byte-rbuf-fill-set! ,p ,x))
(##define-macro (macro-port-u8vector-wbuf p)
  `(macro-port-byte-wbuf ,p))
(##define-macro (macro-port-u8vector-wbuf-set! p x)
  `(macro-port-byte-wbuf-set! ,p ,x))
(##define-macro (macro-port-u8vector-wlo p)
  `(macro-port-byte-wlo ,p))
(##define-macro (macro-port-u8vector-wlo-set! p x)
  `(macro-port-byte-wlo-set! ,p ,x))
(##define-macro (macro-port-u8vector-whi p)
  `(macro-port-byte-whi ,p))
(##define-macro (macro-port-u8vector-whi-set! p x)
  `(macro-port-byte-whi-set! ,p ,x))
(##define-macro (macro-port-u8vector-wbuf-drain p)
  `(macro-port-byte-wbuf-drain ,p))
(##define-macro (macro-port-u8vector-wbuf-drain-set! p x)
  `(macro-port-byte-wbuf-drain-set! ,p ,x))
(##define-macro (macro-port-u8vector-fifo p)
  `(macro-port-byte-other1 ,p))
(##define-macro (macro-port-u8vector-fifo-set! p x)
  `(macro-port-byte-other1-set! ,p ,x))
(##define-macro (macro-port-u8vector-rcondvar p)
  `(macro-port-byte-other2 ,p))
(##define-macro (macro-port-u8vector-rcondvar-set! p x)
  `(macro-port-byte-other2-set! ,p ,x))
(##define-macro (macro-port-u8vector-wcondvar p)
  `(macro-port-byte-other3 ,p))
(##define-macro (macro-port-u8vector-wcondvar-set! p x)
  `(macro-port-byte-other3-set! ,p ,x))
(##define-macro (macro-port-u8vector-width p)
  `(macro-port-byte-other4 ,p))
(##define-macro (macro-port-u8vector-width-set! p x)
  `(macro-port-byte-other4-set! ,p ,x))
(##define-macro (macro-port-u8vector-buffering-limit p)
  `(macro-port-byte-other5 ,p))
(##define-macro (macro-port-u8vector-buffering-limit-set! p x)
  `(macro-port-byte-other5-set! ,p ,x))

(##define-macro (macro-u8vector-input-port? obj)
  `(macro-port-of-rkind? ,obj (macro-u8vector-kind)))

(##define-macro (macro-u8vector-output-port? obj)
  `(macro-port-of-wkind? ,obj (macro-u8vector-kind)))

(##define-macro (macro-check-u8vector-input-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-u8vector-input-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-u8vector-input-port ',(car form) ,@(cdr form))
          `(##trap-check-u8vector-input-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-check-u8vector-output-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-u8vector-output-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-u8vector-output-port ',(car form) ,@(cdr form))
          `(##trap-check-u8vector-output-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of file device ports.

(##define-macro (macro-file-device-port? obj)
  `(##port-of-kind? ,obj (macro-file-device-kind)))

(##define-macro (macro-check-file-device-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-file-device-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-file-device-port ',(car form) ,@(cdr form))
          `(##trap-check-file-device-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of tty device ports.

(##define-macro (macro-tty-device-port? obj)
  `(##port-of-kind? ,obj (macro-tty-device-kind)))

(##define-macro (macro-check-tty-device-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-tty-device-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-tty-device-port ',(car form) ,@(cdr form))
          `(##trap-check-tty-device-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of TCP client device ports.

(##define-macro (macro-tcp-client-device-port? obj)
  `(##port-of-kind? ,obj (macro-tcp-client-device-kind)))

(##define-macro (macro-check-tcp-client-device-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-tcp-client-device-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-tcp-client-device-port ',(car form) ,@(cdr form))
          `(##trap-check-tcp-client-device-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of TCP server ports.

(##define-macro (macro-port-tcp-server-rdevice-condvar p)
  `(macro-port-object-other1 ,p))
(##define-macro (macro-port-tcp-server-rdevice-condvar-set! p x)
  `(macro-port-object-other1-set! ,p ,x))
(##define-macro (macro-port-tcp-server-client-psettings p)
  `(macro-port-object-other2 ,p))
(##define-macro (macro-port-tcp-server-client-psettings-set! p x)
  `(macro-port-object-other2-set! ,p ,x))

(##define-macro (macro-tcp-server-port? obj)
  `(##port-of-kind? ,obj (macro-tcp-server-kind)))

(##define-macro (macro-check-tcp-server-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-tcp-server-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-tcp-server-port ',(car form) ,@(cdr form))
          `(##trap-check-tcp-server-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of pipe device ports.

(##define-macro (macro-pipe-device-port? obj)
  `(##port-of-kind? ,obj (macro-pipe-device-kind)))

(##define-macro (macro-check-pipe-device-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-pipe-device-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-pipe-device-port ',(car form) ,@(cdr form))
          `(##trap-check-pipe-device-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of serial device ports.

(##define-macro (macro-serial-device-port? obj)
  `(##port-of-kind? ,obj (macro-serial-device-kind)))

(##define-macro (macro-check-serial-device-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-serial-device-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-serial-device-port ',(car form) ,@(cdr form))
          `(##trap-check-serial-device-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of directory ports.

(##define-macro (macro-port-directory-rdevice-condvar p)
  `(macro-port-object-other1 ,p))
(##define-macro (macro-port-directory-rdevice-condvar-set! p x)
  `(macro-port-object-other1-set! ,p ,x))
(##define-macro (macro-port-directory-path p)
  `(macro-port-object-other2 ,p))
(##define-macro (macro-port-directory-path-set! p x)
  `(macro-port-object-other2-set! ,p ,x))

(##define-macro (macro-directory-port? obj)
  `(##port-of-kind? ,obj (macro-directory-kind)))

(##define-macro (macro-check-directory-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-directory-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-directory-port ',(car form) ,@(cdr form))
          `(##trap-check-directory-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of event queue ports.

(##define-macro (macro-event-queue-port? obj)
  `(##port-of-kind? ,obj (macro-event-queue-kind)))

(##define-macro (macro-check-event-queue-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-event-queue-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-event-queue-port ',(car form) ,@(cdr form))
          `(##trap-check-event-queue-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of timer ports.

(##define-macro (macro-timer-port? obj)
  `(##port-of-kind? ,obj (macro-timer-kind)))

(##define-macro (macro-check-timer-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-timer-port? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-timer-port ',(car form) ,@(cdr form))
          `(##trap-check-timer-port* ',(car form) ,@(flat (cdr form)))))
     ,expr))

;------------------------------------------------------------------------------

; Representation of port mutexes.

(##define-macro (macro-make-port-mutex)
  `(##make-mutex #f))

(##define-macro (macro-port-mutex-lock! port)
  `(macro-mutex-lock! (macro-port-mutex ,port) #f (macro-current-thread)))

(##define-macro (macro-port-mutex-unlock! port)
  `(macro-mutex-unlock! (macro-port-mutex ,port)))

(##define-macro (macro-port-mutex-unlocked-not-abandoned-and-not-multiprocessor? port)
  `(macro-mutex-unlocked-not-abandoned-and-not-multiprocessor? (macro-port-mutex ,port)))

;------------------------------------------------------------------------------

; Representation of port settings.

(##define-macro (macro-psettings-direction ps)          `(macro-slot 0 ,ps))
(##define-macro (macro-psettings-direction-set! ps x)   `(macro-slot 0 ,ps ,x))
(##define-macro (macro-psettings-roptions ps)           `(macro-slot 1 ,ps))
(##define-macro (macro-psettings-roptions-set! ps x)    `(macro-slot 1 ,ps ,x))
(##define-macro (macro-psettings-woptions ps)           `(macro-slot 2 ,ps))
(##define-macro (macro-psettings-woptions-set! ps x)    `(macro-slot 2 ,ps ,x))
(##define-macro (macro-psettings-append ps)             `(macro-slot 3 ,ps))
(##define-macro (macro-psettings-append-set! ps x)      `(macro-slot 3 ,ps ,x))
(##define-macro (macro-psettings-create ps)             `(macro-slot 4 ,ps))
(##define-macro (macro-psettings-create-set! ps x)      `(macro-slot 4 ,ps ,x))
(##define-macro (macro-psettings-truncate ps)           `(macro-slot 5 ,ps))
(##define-macro (macro-psettings-truncate-set! ps x)    `(macro-slot 5 ,ps ,x))
(##define-macro (macro-psettings-permissions ps)        `(macro-slot 6 ,ps))
(##define-macro (macro-psettings-permissions-set! ps x) `(macro-slot 6 ,ps ,x))
(##define-macro (macro-psettings-output-width ps)       `(macro-slot 7 ,ps))
(##define-macro (macro-psettings-output-width-set! ps x)`(macro-slot 7 ,ps ,x))

(##define-macro (macro-options-char-encoding x)        `(macro-slot 0 ,x))
(##define-macro (macro-options-char-encoding-set! x y) `(macro-slot 0 ,x ,y))
(##define-macro (macro-options-eol-encoding x)         `(macro-slot 1 ,x))
(##define-macro (macro-options-eol-encoding-set! x y)  `(macro-slot 1 ,x ,y))
(##define-macro (macro-options-buffering x)            `(macro-slot 2 ,x))
(##define-macro (macro-options-buffering-set! x y)     `(macro-slot 2 ,x ,y))

(##define-macro (macro-char-encoding-shift)   1)
(##define-macro (macro-char-encoding-range)   16)
(##define-macro (macro-char-encoding-ascii)   0)
(##define-macro (macro-char-encoding-latin1)  1)
(##define-macro (macro-char-encoding-utf8)    2)
(##define-macro (macro-char-encoding-ucs2)    3)
(##define-macro (macro-char-encoding-ucs2le)  4)
(##define-macro (macro-char-encoding-ucs2be)  5)
(##define-macro (macro-char-encoding-ucs4)    6)
(##define-macro (macro-char-encoding-ucs4le)  7)
(##define-macro (macro-char-encoding-ucs4be)  8)
(##define-macro (macro-char-encoding-native)  9)
(##define-macro (macro-default-char-encoding) `(macro-char-encoding-range))

(##define-macro (macro-eol-encoding-shift)   16)
(##define-macro (macro-eol-encoding-range)   8)
(##define-macro (macro-eol-encoding-lf)      0)
(##define-macro (macro-eol-encoding-cr)      1)
(##define-macro (macro-eol-encoding-crlf)    2)
(##define-macro (macro-eol-encoding-lfcr)    3)
(##define-macro (macro-default-eol-encoding) `(macro-eol-encoding-range))

(##define-macro (macro-decode-state-shift)  128)
(##define-macro (macro-decode-state-range)  3)
(##define-macro (macro-decode-state-none)   0)
(##define-macro (macro-decode-state-lf)     1)
(##define-macro (macro-decode-state-cr)     2)

(##define-macro (macro-open-state-shift)  512)
(##define-macro (macro-open-state-range)  2)
(##define-macro (macro-open-state-open)   0)
(##define-macro (macro-open-state-closed) 1)

(##define-macro (macro-closed? options)
  `(##not (##fixnum.= (##fixnum.logand ,options 512) 0)))

(##define-macro (macro-close! options)
  `(##fixnum.logior ,options 512))

(##define-macro (macro-buffering-shift)   4096)
(##define-macro (macro-buffering-range)   4)
(##define-macro (macro-no-buffering)      0)
(##define-macro (macro-line-buffering)    1)
(##define-macro (macro-buffering)         2)
(##define-macro (macro-default-buffering) `(macro-buffering-range))

(##define-macro (macro-unbuffered? options)
  `(##fixnum.< ,options 4096))

(##define-macro (macro-buffered? options)
  `(##not (##fixnum.< ,options 8192)))

(##define-macro (macro-direction-in)    1)
(##define-macro (macro-direction-out)   2)
(##define-macro (macro-direction-inout) 3)

(##define-macro (macro-no-append)      0)
(##define-macro (macro-append)         1)
(##define-macro (macro-default-append) 2)

(##define-macro (macro-no-create)      0)
(##define-macro (macro-maybe-create)   1)
(##define-macro (macro-create)         2)
(##define-macro (macro-default-create) 3)

(##define-macro (macro-no-truncate)      0)
(##define-macro (macro-truncate)         1)
(##define-macro (macro-default-truncate) 2)

(##define-macro (macro-default-permissions)  -1)

(##define-macro (macro-default-output-width) -1)

;------------------------------------------------------------------------------

; Representation of write environments.

; A writeenv structure maintains the "write environment" throughout
; the writing of a Scheme datum.  It includes the write style
; (display, write, pretty-print, mark), the port on which to write,
; the readtable, the marktable (for detecting cycles), the force flag,
; the pretty-print width, the number of closing parentheses to follow
; the datum, the current nesting level, and the character count limit.

(##define-macro (macro-make-writeenv
                 style
                 port
                 readtable
                 marktable
                 force?
                 width
                 close-parens
                 level
                 limit)
  `(##vector ,style
             ,port
             ,readtable
             ,marktable
             ,force?
             ,width
             ,close-parens
             ,level
             ,limit))

(##define-macro (macro-writeenv-style we)               `(macro-slot 0 ,we))
(##define-macro (macro-writeenv-style-set! we x)        `(macro-slot 0 ,we ,x))
(##define-macro (macro-writeenv-port we)                `(macro-slot 1 ,we))
(##define-macro (macro-writeenv-port-set! we x)         `(macro-slot 1 ,we ,x))
(##define-macro (macro-writeenv-readtable we)           `(macro-slot 2 ,we))
(##define-macro (macro-writeenv-readtable-set! we x)    `(macro-slot 2 ,we ,x))
(##define-macro (macro-writeenv-marktable we)           `(macro-slot 3 ,we))
(##define-macro (macro-writeenv-marktable-set! we x)    `(macro-slot 3 ,we ,x))
(##define-macro (macro-writeenv-force? we)              `(macro-slot 4 ,we))
(##define-macro (macro-writeenv-force?-set! we x)       `(macro-slot 4 ,we ,x))
(##define-macro (macro-writeenv-width we)               `(macro-slot 5 ,we))
(##define-macro (macro-writeenv-width-set! we x)        `(macro-slot 5 ,we ,x))
(##define-macro (macro-writeenv-close-parens we)        `(macro-slot 6 ,we))
(##define-macro (macro-writeenv-close-parens-set! we x) `(macro-slot 6 ,we ,x))
(##define-macro (macro-writeenv-level we)               `(macro-slot 7 ,we))
(##define-macro (macro-writeenv-level-set! we x)        `(macro-slot 7 ,we ,x))
(##define-macro (macro-writeenv-limit we)               `(macro-slot 8 ,we))
(##define-macro (macro-writeenv-limit-set! we x)        `(macro-slot 8 ,we ,x))

;------------------------------------------------------------------------------

; Representation of read environments.

; A readenv structure maintains the "read environment" throughout the
; reading of a Scheme datum.  It includes the port from which to read,
; the readtable, the error procedure, the wrap and unwrap procedures,
; the table of labels (i.e. "#n#"), and the position where the
; currently being read datum started.

(##define-macro (macro-make-readenv
                 port
                 readtable
                 error-proc
                 wrapper
                 unwrapper
                 labels
                 filepos)
  `(##vector ,port
             ,readtable
             ,error-proc
             ,wrapper
             ,unwrapper
             ,labels
             ,filepos))

(##define-macro (macro-readenv-port re)              `(macro-slot 0 ,re))
(##define-macro (macro-readenv-port-set! re x)       `(macro-slot 0 ,re ,x))
(##define-macro (macro-readenv-readtable re)         `(macro-slot 1 ,re))
(##define-macro (macro-readenv-readtable-set! re x)  `(macro-slot 1 ,re ,x))
(##define-macro (macro-readenv-error-proc re)        `(macro-slot 2 ,re))
(##define-macro (macro-readenv-error-proc-set! re x) `(macro-slot 2 ,re ,x))
(##define-macro (macro-readenv-wrapper re)           `(macro-slot 3 ,re))
(##define-macro (macro-readenv-wrapper-set! re x)    `(macro-slot 3 ,re ,x))
(##define-macro (macro-readenv-unwrapper re)         `(macro-slot 4 ,re))
(##define-macro (macro-readenv-unwrapper-set! re x)  `(macro-slot 4 ,re ,x))
(##define-macro (macro-readenv-labels re)            `(macro-slot 5 ,re))
(##define-macro (macro-readenv-labels-set! re x)     `(macro-slot 5 ,re ,x))
(##define-macro (macro-readenv-filepos re)           `(macro-slot 6 ,re))
(##define-macro (macro-readenv-filepos-set! re x)    `(macro-slot 6 ,re ,x))

(##define-macro (macro-readenv-wrap re x)
  `(let ((re ,re)
         (x ,x))
     ((macro-readenv-wrapper re) re x)))

(##define-macro (macro-readenv-unwrap re x)
  `(let ((re ,re)
         (x ,x))
     ((macro-readenv-unwrapper re) re x)))

;------------------------------------------------------------------------------

; Generic char port procedures.

(##define-macro (macro-peek-char port)
  `(let ((port ,port))

     (##declare (not interrupts-enabled))

     ; try to get exclusive access to port and if successful perform
     ; operation inline

     (if (macro-port-mutex-unlocked-not-abandoned-and-not-multiprocessor? port)

       (let ((char-rlo (macro-port-char-rlo port))
             (char-rhi (macro-port-char-rhi port)))
         (if (##fixnum.< char-rlo char-rhi)

           ; the next character is in the character read buffer

           (##string-ref (macro-port-char-rbuf port) char-rlo)

           ; more characters are needed, do this out-of-line

           (##peek-char port)))

       ; couldn't easily get exclusive access to port, handle this out-of-line

       (##peek-char port))))

(##define-macro (macro-read-char port)
  `(let ((port ,port))

     (##declare (not interrupts-enabled))

     ; try to get exclusive access to port and if successful perform
     ; operation inline

     (if (macro-port-mutex-unlocked-not-abandoned-and-not-multiprocessor? port)

       (let ((char-rlo (macro-port-char-rlo port))
             (char-rhi (macro-port-char-rhi port)))
         (if (##fixnum.< char-rlo char-rhi)

           ; the next character is in the character read buffer

           (let ((c (##string-ref (macro-port-char-rbuf port) char-rlo)))
             (if (##not (##char=? c #\newline))

               ; frequent simple case, just advance rlo

               (begin
                 (macro-port-char-rlo-set! port (##fixnum.+ char-rlo 1))
                 c)

               ; end-of-line processing is complex, so do it out-of-line

               (##read-char port)))

           ; more characters are needed, do this out-of-line

           (##read-char port)))

       ; couldn't easily get exclusive access to port, handle this out-of-line

       (##read-char port))))

(##define-macro (macro-write-char c port)
  `(let ((c ,c)
         (port ,port))

     (##declare (not interrupts-enabled))

     ; try to get exclusive access to port and if successful perform
     ; operation inline

     (if (and (##not (##char=? c #\newline))
              (macro-port-mutex-unlocked-not-abandoned-and-not-multiprocessor? port))

       (let ((char-wbuf (macro-port-char-wbuf port))
             (char-whi+1 (##fixnum.+ (macro-port-char-whi port) 1)))
         (if (##fixnum.< char-whi+1 (##string-length char-wbuf))

           ; adding this character would not make the character write
           ; buffer full, so add character and increment whi

           (begin
             (##string-set! char-wbuf (##fixnum.- char-whi+1 1) c)
             (macro-port-char-whi-set! port char-whi+1)
             (##void))

           ; the character write buffer would become full, so handle
           ; this out-of-line

           (##write-char c port)))

       ; end-of-line processing is needed or exclusive access to port
       ; cannot be obtained easily, so handle this out-of-line

       (##write-char c port))))

;------------------------------------------------------------------------------

; Representation of readtables.

(##define-macro (macro-readtable? obj)
  `(macro-structure-object? ,obj ##readtable-tag))

(##define-macro (macro-check-readtable var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (##readtable? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-readtable ',(car form) ,@(cdr form))
          `(##trap-check-readtable* ',(car form) ,@(flat (cdr form)))))
     ,expr))

;------------------------------------------------------------------------------

;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(##define-macro (macro-current-error-port)
  `(##dynamic-ref ##error-port-sym ##stderr))

(##define-macro (macro-current-readtable)
  `(##dynamic-ref ##readtable-sym ##main-readtable))

;==============================================================================
