;==============================================================================

; file: "_thread#.scm"

; Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.

;------------------------------------------------------------------------------

; Priority queue generator macro.

; The red-black tree implementation used here is inspired from the
; code in the MIT-Scheme runtime (file "rbtree.scm").  That code is
; based on the algorithms presented in the book "Introduction to
; Algorithms" by Cormen, Leiserson, and Rivest.
;
; The main differences with the MIT-Scheme code are:
;
;   1) Nil pointers are replaced by a special sentinel that is also
;      the cell that contains a pointer (in the "left child" field)
;      to the root of the red-black tree.  The "right child" field of
;      the sentinel is never accessed.  The sentinel is black.
;
;   2) The color field contains #f when the node is red and a
;      reference to the sentinel when the node is black.  It is thus
;      possible to find the sentinel from any node in constant time
;      (if the node is black extract the color field, otherwise
;      extract the color field of the parent, which must be black).
;
;   3) One field of the sentinel always points to the leftmost node of
;      the red-black tree.  This allows constant time access to the
;      "minimum" node, which is a frequent operation of priority queues.
;
;   4) Several cases are handled specially (see the code for details).
;
;   5) Macros are used to generate code specialized for each case of
;      symmetrical operations (e.g. left and right rotation).
;
;   6) Nodes are assumed to be preallocated.  Inserting and deleting a
;      node from a red-black tree does not perform any heap
;      allocation.  Moreover, all algorithms consume a constant amount
;      of stack space.

(##define-macro (macro-generate-rbtree
                 rbtree-init!
                 node->rbtree
                 insert!
                 remove!
                 reposition!
                 singleton?
                 color
                 color-set!
                 parent
                 parent-set!
                 left
                 left-set!
                 right
                 right-set!
                 before?
                 leftmost
                 leftmost-set!
                 rightmost
                 rightmost-set!)

  (define (black rbtree)
    rbtree)

  (define (black? rbtree)
    `(lambda (node)
       (,color node)))

  (define (blacken! rbtree)
    `(lambda (node)
       (,color-set! node ,(black rbtree))))

  (define (red)
    #f)

  (define (red?)
    `(lambda (node)
       (##not (,color node))))

  (define (reden!)
    `(lambda (node)
       (,color-set! node ,(red))))

  (define (copy-color!)
    `(lambda (node1 node2)
       (,color-set! node1 (,color node2))))

  (define (exchange-color!)
    `(lambda (node1 node2)
       (let ((color-node1 (,color node1)))
         (,color-set! node1 (,color node2))
         (,color-set! node2 color-node1))))

  (define (update-parent!)
    `(lambda (parent-node old-node new-node)
       (if (##eq? old-node (,left parent-node))
         (,left-set! parent-node new-node)
         (,right-set! parent-node new-node))))

  (define (rotate! side1 side1-set! side2 side2-set!)
    `(lambda (node)
       (let ((side2-node (,side2 node)))
         (let ((side1-side2-node (,side1 side2-node)))
           (,side2-set! node side1-side2-node)
           (,parent-set! side1-side2-node node))
         (let ((parent-node (,parent node)))
           (,side1-set! side2-node node)
           (,parent-set! node side2-node)
           (,parent-set! side2-node parent-node)
           (,(update-parent!) parent-node node side2-node)))))

  (define (rotate-left!)
    (rotate! left left-set! right right-set!))

  (define (rotate-right!)
    (rotate! right right-set! left left-set!))

  (define (neighbor side other-side)
    `(lambda (node rbtree)
       (let ((side-node (,side node)))
         (if (##eq? side-node rbtree)
           (let ((parent-node (,parent node)))
             (if (or (##eq? parent-node rbtree)
                     (##eq? node (,side parent-node)))
               rbtree
               parent-node))
           (let loop ((x side-node))
             (let ((other-side-x (,other-side x)))
               (if (##eq? other-side-x rbtree)
                 x
                 (loop other-side-x))))))))

  `(begin

     (##define-macro (,rbtree-init! rbtree)
       `(let ((rbtree ,rbtree))

          (##declare (not interrupts-enabled))

          ,',@(if leftmost
                `((,leftmost-set! rbtree rbtree))
                `())

          (,',(blacken! 'rbtree) rbtree)
          (,',parent-set! rbtree rbtree)
          (,',left-set! rbtree rbtree)))

     (##define-macro (,node->rbtree node)
       `(let ((node ,node))

          (##declare (not interrupts-enabled))

          (or (,',color node)
              (,',color (,',parent node)))))

     (define ,insert!
       (let ()

         (##declare (not inline) (not interrupts-enabled))

         (lambda (rbtree node)

           (define (fixup!)

             (let loop ((x node))
               (let ((parent-x (,parent x)))
                 (if (,(red?) parent-x)
                   (let ((parent-parent-x (,parent parent-x)))

                     (##define-macro (body
                                      side1
                                      rotate-side1!
                                      side2
                                      rotate-side2!)
                       `(let ((side2-parent-parent-x (,side2 parent-parent-x)))
                          (if (,',(red?) side2-parent-parent-x)
                            (begin
                              (set! case1 (##fixnum.+ case1 1));;;;;;;;;;;;;
                              (,',(blacken! 'rbtree) parent-x)
                              (,',(blacken! 'rbtree) side2-parent-parent-x)
                              (,',(reden!) parent-parent-x)
                              (loop parent-parent-x))
                            (let ((y
                                   (if (##eq? x (,side2 parent-x))
                                     (begin
                                       (set! case2 (##fixnum.+ case2 1));;;;;;;
                                       (,rotate-side1! parent-x)
                                       (,',parent parent-x))
                                     (,',parent x))))
                              (set! case3 (##fixnum.+ case3 1));;;;;;;;;
                              (,',(blacken! 'rbtree) y)
                              (let ((parent-y (,',parent y)))
                                (,',(reden!) parent-y)
                                (,rotate-side2! parent-y))))))

                     (if (##eq? parent-x (,left parent-parent-x))
                       (body ,left
                             ,(rotate-left!)
                             ,right
                             ,(rotate-right!))
                       (body ,right
                             ,(rotate-right!)
                             ,left
                             ,(rotate-left!)))))))

             (,(blacken! 'rbtree) (,left rbtree))
             (##void))

           (##define-macro (insert-below! x)
             `(let ((x ,x))
                (if (,',before? node x)
                  (insert-left! (,',left x) x)
                  (insert-right! (,',right x) x))))

           (define (insert-left! left-x x)
             (if (##eq? left-x rbtree)
               (begin
                 (,left-set! x node)
                 (,parent-set! node x)

                 ; check if leftmost must be updated

                 ,@(if leftmost
                     `((if (##eq? x (,leftmost rbtree))
                         (,leftmost-set! rbtree node)))
                     `())

                 (fixup!))
               (insert-below! left-x)))

           (define (insert-right! right-x x)
             (if (##eq? right-x rbtree)
               (begin
                 (,right-set! x node)
                 (,parent-set! node x)

                 ; check if rightmost must be updated

                 ,@(if rightmost
                     `((if (##eq? x (,rightmost rbtree))
                         (,rightmost-set! rbtree node)))
                     `())

                 (fixup!))
               (insert-below! right-x)))

           (,(reden!) node)
           (,left-set! node rbtree)
           (,right-set! node rbtree)

           (insert-left! (,left rbtree) rbtree)

           (,parent-set! rbtree rbtree))))

     (define ,remove!
       (let ()

         (##declare (not inline) (not interrupts-enabled))

         (lambda (node)
           (let ((rbtree (,node->rbtree node)))

             (define (fixup! parent-node node)

               (##define-macro (body side1 rotate-side1! side2 rotate-side2!)
                 `(let ((x
                         (let ((side2-parent-node (,side2 parent-node)))
                           (if (,',(red?) side2-parent-node)
                             (begin
                               (set! dcase1 (##fixnum.+ dcase1 1));;;;;;;
                               (,',(blacken! 'rbtree) side2-parent-node)
                               (,',(reden!) parent-node)
                               (,rotate-side1! parent-node)
                               (,side2 parent-node))
                             side2-parent-node))))

                    (define (common-case y)
                      (,',(copy-color!) y parent-node)
                      (,',(blacken! 'rbtree) parent-node)
                      (,',(blacken! 'rbtree) (,side2 y))
                      (,rotate-side1! parent-node)
                      (,',(blacken! 'rbtree) (,',left rbtree)))

                    (if (,',(red?) (,side2 x))
                        (begin (set! dcase4 (##fixnum.+ dcase4 1));;;;;;;
                      (common-case x)
                      );;;;;;;;
                      (let ((side1-x (,side1 x)))
                        (if (,',(black? 'rbtree) side1-x)
                          (begin
                            (set! dcase2 (##fixnum.+ dcase2 1));;;;;;;;
                            (,',(reden!) x)
                            (fixup! (,',parent parent-node) parent-node))
                          (begin
                            (set! dcase3 (##fixnum.+ dcase3 1));;;;;;;;
                            (,',(blacken! 'rbtree) side1-x)
                            (,',(reden!) x)
                            (,rotate-side2! x)
                            (common-case (,side2 parent-node))))))))

               (cond ((or (##eq? parent-node rbtree) (,(red?) node))
                      (,(blacken! 'rbtree) node))
                     ((##eq? node (,left parent-node))
                      (body ,left
                            ,(rotate-left!)
                            ,right
                            ,(rotate-right!)))
                     (else
                      (body ,right
                            ,(rotate-right!)
                            ,left
                            ,(rotate-left!)))))

             (let ((parent-node (,parent node))
                   (left-node (,left node))
                   (right-node (,right node)))

               (,parent-set! node #f) ; to avoid leaks
               (,left-set! node #f)
               (,right-set! node #f)

               (cond ((##eq? left-node rbtree)
                      (set! ddcase1 (##fixnum.+ ddcase1 1));;;;;;;;;

                      ; check if leftmost must be updated

                      ,@(if leftmost
                          `((if (##eq? node (,leftmost rbtree))
                              (,leftmost-set!
                               rbtree
                               (if (##eq? right-node rbtree)
                                 parent-node
                                 right-node))))
                          `())

                      (,parent-set! right-node parent-node)
                      (,(update-parent!) parent-node node right-node)

                      (if (,(black? 'rbtree) node)
                        (begin
                          (,(reden!) node) ; to avoid leaks
                          (fixup! parent-node right-node))))

                     ((##eq? right-node rbtree)
                      (set! ddcase2 (##fixnum.+ ddcase2 1));;;;;;;;;;

                      ; check if rightmost must be updated

                      ,@(if rightmost
                          `((if (##eq? node (,rightmost rbtree))
                              (,rightmost-set!
                               rbtree
                               left-node)))
                          `())

                      (,parent-set! left-node parent-node)
                      (,(update-parent!) parent-node node left-node)

                      ; At this point we know that the node is black.
                      ; This is because the right child is nil and the
                      ; left child is red (if the left child was black
                      ; the tree would not be balanced)

                      (,(reden!) node) ; to avoid leaks
                      (fixup! parent-node left-node))

                     (else
                      (set! ddcase3 (##fixnum.+ ddcase3 1));;;;;;;;;;;

                      (let loop ((x right-node) (parent-x node))
                        (let ((left-x (,left x)))
                          (if (##eq? left-x rbtree)
                            (begin
                              (,(exchange-color!) x node)
                              (,parent-set! left-node x)
                              (,left-set! x left-node)
                              (,parent-set! x parent-node)
                              (,(update-parent!) parent-node node x)
                              (if (##eq? x right-node)
                                (if (,(black? 'rbtree) node)
                                  (begin
                                    (,(reden!) node) ; to avoid leaks
                                    (fixup! x (,right x))))
                                (let ((right-x (,right x)))
                                  (,parent-set! right-x parent-x)
                                  (,left-set! parent-x right-x)
                                  (,parent-set! right-node x)
                                  (,right-set! x right-node)
                                  (if (,(black? 'rbtree) node)
                                    (begin
                                      (,(reden!) node) ; to avoid leaks
                                      (fixup! parent-x right-x))))))
                            (loop left-x x)))))))

             (,parent-set! rbtree rbtree)))))

     (##define-macro (,singleton? rbtree)
       `(let ((rbtree ,rbtree))

          (##declare (not interrupts-enabled))

          (let ((root (,',left rbtree)))
            (and (##not (##eq? root rbtree))
                 (##eq? (,',left root) rbtree)
                 (##eq? (,',right root) rbtree)
                 root))))

     (define ,reposition!
       (let ()

         (##declare (not inline) (not interrupts-enabled))

         (lambda (node)
           (let* ((rbtree
                   (,node->rbtree node))
                  (predecessor-node
                   (,(neighbor left right) node rbtree))
                  (successor-node
                   (,(neighbor right left) node rbtree)))
             (if (or (and (##not (##eq? predecessor-node rbtree))
                          (,before? node predecessor-node))
                     (and (##not (##eq? successor-node rbtree))
                          (,before? successor-node node)))
               (begin
                 (,remove! node)
                 (,insert! rbtree node)))))))))

;------------------------------------------------------------------------------

; Representation of dynamic environments.

; The dynamic environment contains the set of dynamic variable
; bindings (such as the current input port and current exception
; handler) that is attached to a continuation, and thereby to a
; thread.

; The dynamic environment is separated in a global environment and a
; local environment.  Each of these is represented as a binary tree
; ordered using the hash of the variable name.  Some frequently used
; dynamic variable (such as the current input port and exception
; handler) are stored separately.  Different dynamic environments may
; share a given global environment (new bindings are added to this
; shared global environment with "dynamic-define").

(##define-macro (macro-make-denv global local im eh ds ip op)
  `(##vector ,global ,local ,im ,eh ,ds ,ip ,op))
(##define-macro (macro-denv-global d)                   `(macro-slot 0 ,d))
(##define-macro (macro-denv-global-set! d x)            `(macro-slot 0 ,d ,x))
(##define-macro (macro-denv-local d)                    `(macro-slot 1 ,d))
(##define-macro (macro-denv-local-set! d x)             `(macro-slot 1 ,d ,x))
(##define-macro (macro-denv-interrupt-mask d)           `(macro-slot 2 ,d))
(##define-macro (macro-denv-interrupt-mask-set! d x)    `(macro-slot 2 ,d ,x))
(##define-macro (macro-denv-exception-handler d)        `(macro-slot 3 ,d))
(##define-macro (macro-denv-exception-handler-set! d x) `(macro-slot 3 ,d ,x))
(##define-macro (macro-denv-debugging-settings d)       `(macro-slot 4 ,d))
(##define-macro (macro-denv-debugging-settings-set! d x)`(macro-slot 4 ,d ,x))
(##define-macro (macro-denv-input-port d)               `(macro-slot 5 ,d))
(##define-macro (macro-denv-input-port-set! d x)        `(macro-slot 5 ,d ,x))
(##define-macro (macro-denv-output-port d)              `(macro-slot 6 ,d))
(##define-macro (macro-denv-output-port-set! d x)       `(macro-slot 6 ,d ,x))

; Environment tree nodes.

(##define-macro (macro-make-env nv l r)       `(##vector ,nv ,l ,r))
(##define-macro (macro-env-name-val e)        `(macro-slot 0 ,e))
(##define-macro (macro-env-name-val-set! e x) `(macro-slot 0 ,e ,x))
(##define-macro (macro-env-left e)            `(macro-slot 1 ,e))
(##define-macro (macro-env-left-set! e x)     `(macro-slot 1 ,e ,x))
(##define-macro (macro-env-right e)           `(macro-slot 2 ,e))
(##define-macro (macro-env-right-set! e x)    `(macro-slot 2 ,e ,x))

; Binding of special dynamic variables.

(##define-macro (macro-dynamic-bind var val thunk)
  `(let ((val ,val) (thunk ,thunk))

     (##declare (not interrupts-enabled))

     (let ((current-denv (macro-thread-denv (macro-current-thread))))
       (##dynamic-env-bind
        (macro-make-denv
         (macro-denv-global current-denv)
         (macro-denv-local current-denv)
         ,(if (eq? var 'interrupt-mask)
            `val
            `(macro-denv-interrupt-mask current-denv))
         ,(if (eq? var 'exception-handler)
            `val
            `(macro-denv-exception-handler current-denv))
         ,(if (eq? var 'debugging-settings)
            `val
            `(macro-denv-debugging-settings current-denv))
         ,(if (eq? var 'input-port)
            `val
            `(macro-denv-input-port current-denv))
         ,(if (eq? var 'output-port)
            `val
            `(macro-denv-output-port current-denv)))
        thunk))))

(##define-macro (macro-current-interrupt-mask)
  `(macro-denv-interrupt-mask (macro-thread-denv (macro-current-thread))))

(##define-macro (macro-current-exception-handler)
  `(macro-denv-exception-handler (macro-thread-denv (macro-current-thread))))

(##define-macro (macro-debugging-settings-port)
  `(macro-denv-debugging-settings (macro-thread-denv (macro-current-thread))))

(##define-macro (macro-current-input-port)
  `(macro-denv-input-port (macro-thread-denv (macro-current-thread))))

(##define-macro (macro-current-output-port)
  `(macro-denv-output-port (macro-thread-denv (macro-current-thread))))

; Exception raising.

(##define-macro (macro-raise obj)
  `(let ((obj ,obj))
     (##declare (not safe)) ; avoid procedure check on the call to the handler
     ((macro-current-exception-handler) obj)))

(##define-macro (macro-abort obj)
  `(let ((obj ,obj))
     (##declare (not safe)) ; avoid procedure check on the call to the handler
     ((macro-current-exception-handler) obj)
     ((macro-current-exception-handler)
      (macro-make-noncontinuable-exception obj))))

;------------------------------------------------------------------------------

; Representation of time objects.

; A time object is represented by an object vector of length 1
; slot 0 = number of clock ticks (flonum) since the reference point

(##define-macro (macro-time? obj)
  `(macro-structure-object? ,obj ##time-tag))

(##define-macro (macro-check-time var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-time? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-time ',(car form) ,@(cdr form))
          `(##trap-check-time* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-check-absrel-time var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (or (##real? ,var)
             (macro-time? ,var))
       ,expr
       ,(if (list? form)
          `(##trap-check-time ',(car form) ,@(cdr form))
          `(##trap-check-time* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-check-absrel-time-or-false var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (or (##not ,var)
             (##real? ,var)
             (macro-time? ,var))
       ,expr
       ,(if (list? form)
          `(##trap-check-time ',(car form) ,@(cdr form))
          `(##trap-check-time* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-make-time seconds)
  `(let ((seconds ,seconds))
     (macro-make-structure-object ##time-tag seconds)))

(##define-macro (macro-time-seconds t)        `(macro-slot 1 ,t))
(##define-macro (macro-time-seconds-set! t x) `(macro-slot 1 ,t ,x))

;------------------------------------------------------------------------------

; Representation of thread related exceptions.

(##define-macro (macro-make-deadlock-exception)
  `##deadlock-exception)

(##define-macro (macro-deadlock-exception? obj)
  `(##eq? ,obj ##deadlock-exception))

(##define-macro (macro-make-join-timeout-exception)
  `##join-timeout-exception)

(##define-macro (macro-join-timeout-exception? obj)
  `(##eq? ,obj ##join-timeout-exception))

(##define-macro (macro-make-abandoned-mutex-exception)
  `##abandoned-mutex-exception)

(##define-macro (macro-abandoned-mutex-exception? obj)
  `(##eq? ,obj ##abandoned-mutex-exception))

(##define-macro (macro-make-started-thread-exception)
  `##started-thread-exception)

(##define-macro (macro-started-thread-exception? obj)
  `(##eq? ,obj ##started-thread-exception))

(##define-macro (macro-make-terminated-thread-exception)
  `##terminated-thread-exception)

(##define-macro (macro-terminated-thread-exception? obj)
  `(##eq? ,obj ##terminated-thread-exception))

(##define-macro (macro-uncaught-exception? obj)
  `(macro-structure-object? ,obj ##uncaught-exception-tag))

(##define-macro (macro-check-uncaught-exception var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-uncaught-exception? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-uncaught-exception ',(car form) ,@(cdr form))
          `(##trap-check-uncaught-exception* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-make-uncaught-exception obj)
  `(let ((obj ,obj))
     (macro-make-structure-object ##uncaught-exception-tag obj)))

(##define-macro (macro-uncaught-exception-reason exc)
  `(macro-slot 1 ,exc))
(##define-macro (macro-uncaught-exception-reason-set! exc obj)
  `(macro-slot 1 ,exc ,obj))

(##define-macro (macro-scheduler-exception? obj)
  `(macro-structure-object? ,obj ##scheduler-exception-tag))

(##define-macro (macro-check-scheduler-exception var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-scheduler-exception? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-scheduler-exception ',(car form) ,@(cdr form))
          `(##trap-check-scheduler-exception* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-make-scheduler-exception obj)
  `(let ((obj ,obj))
     (macro-make-structure-object ##scheduler-exception-tag obj)))

(##define-macro (macro-scheduler-exception-reason exc)
  `(macro-slot 1 ,exc))
(##define-macro (macro-scheduler-exception-reason-set! exc obj)
  `(macro-slot 1 ,exc ,obj))

(##define-macro (macro-noncontinuable-exception? obj)
  `(macro-structure-object? ,obj ##noncontinuable-exception-tag))

(##define-macro (macro-check-noncontinuable-exception var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-noncontinuable-exception? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-noncontinuable-exception ',(car form) ,@(cdr form))
          `(##trap-check-noncontinuable-exception* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-make-noncontinuable-exception obj)
  `(let ((obj ,obj))
     (macro-make-structure-object ##noncontinuable-exception-tag obj)))

(##define-macro (macro-noncontinuable-exception-reason exc)
  `(macro-slot 1 ,exc))
(##define-macro (macro-noncontinuable-exception-reason-set! exc obj)
  `(macro-slot 1 ,exc ,obj))

(##define-macro (macro-unbound-dynamic-variable-exception? obj)
  `(macro-structure-object? ,obj ##unbound-dynamic-variable-exception-tag))

(##define-macro (macro-check-unbound-dynamic-variable-exception var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-unbound-dynamic-variable-exception? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-unbound-dynamic-variable-exception ',(car form) ,@(cdr form))
          `(##trap-check-unbound-dynamic-variable-exception* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-make-unbound-dynamic-variable-exception name)
  `(let ((name ,name))
     (macro-make-structure-object ##unbound-dynamic-variable-exception-tag name)))

(##define-macro (macro-unbound-dynamic-variable-exception-name exc)
  `(macro-slot 1 ,exc))
(##define-macro (macro-unbound-dynamic-variable-exception-name-set! exc obj)
  `(macro-slot 1 ,exc ,obj))

(##define-macro (macro-low-level-exception? obj)
  `(macro-structure-object? ,obj ##low-level-exception-tag))

(##define-macro (macro-check-low-level-exception var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-low-level-exception? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-low-level-exception ',(car form) ,@(cdr form))
          `(##trap-check-low-level-exception* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-make-low-level-exception err-code)
  `(let ((err-code ,err-code))
     (macro-make-structure-object ##low-level-exception-tag err-code)))

(##define-macro (macro-low-level-exception-err-code exc)
  `(macro-slot 1 ,exc))
(##define-macro (macro-low-level-exception-err-code-set! exc obj)
  `(macro-slot 1 ,exc ,obj))

;------------------------------------------------------------------------------

; Primordial thread settings.

(##define-macro (macro-primordial-thread-base-exception-handler)
  `##thread-base-exception-handler);;;;;;;;;;;;;;;;

(##define-macro (macro-primordial-thread-base-priority)
  (exact->inexact (- (expt 10 10))))

(##define-macro (macro-primordial-thread-quantum)
  (exact->inexact 4/100))

(##define-macro (macro-primordial-thread-priority-boost)
  (exact->inexact (expt 10 -6)))

;------------------------------------------------------------------------------

; Representation of thread groups, threads, mutexes and condition variables.

; A thread group contains a double-ended-queue of the threads that are
; in the group and a double-ended-queue of the thread groups that are
; in the group.  A thread contains a double-ended-queue of the mutexes
; that are owned by the thread.  Conversely, a mutex contains a queue
; of the threads blocked on the mutex.  A condition variable also
; contains a queue of the threads blocked on the condition variable.

; Representation of blocked thread queues.

(##define-macro (macro-btq-deq-next node)        `(macro-slot 1 ,node))
(##define-macro (macro-btq-deq-next-set! node x) `(macro-slot 1 ,node ,x))
(##define-macro (macro-btq-deq-prev node)        `(macro-slot 2 ,node))
(##define-macro (macro-btq-deq-prev-set! node x) `(macro-slot 2 ,node ,x))
(##define-macro (macro-btq-color node)           `(macro-slot 3 ,node))
(##define-macro (macro-btq-color-set! node x)    `(macro-slot 3 ,node ,x))
(##define-macro (macro-btq-parent node)          `(macro-slot 4 ,node))
(##define-macro (macro-btq-parent-set! node x)   `(macro-slot 4 ,node ,x))
(##define-macro (macro-btq-left node)            `(macro-slot 5 ,node))
(##define-macro (macro-btq-left-set! node x)     `(macro-slot 5 ,node ,x))
(##define-macro (macro-btq-right node)           `(macro-slot 6 ,node))
(##define-macro (macro-btq-right-set! node x)    `(macro-slot 6 ,node ,x))
(##define-macro (macro-btq-leftmost node)        `(macro-slot 6 ,node))
(##define-macro (macro-btq-leftmost-set! node x) `(macro-slot 6 ,node ,x))
(##define-macro (macro-btq-owner node)           `(macro-slot 7 ,node))
(##define-macro (macro-btq-owner-set! node x)    `(macro-slot 7 ,node ,x))

(##define-macro (macro-btq-deq-init! deq)
  `(let ((deq ,deq))

     (##declare (not interrupts-enabled))

     (macro-btq-deq-next-set! deq deq)
     (macro-btq-deq-prev-set! deq deq)))

(##define-macro (macro-btq-deq-remove! item)
  `(let ((item ,item))

     (##declare (not interrupts-enabled))

     (let ((item-next (macro-btq-deq-next item))
           (item-prev (macro-btq-deq-prev item)))
       (macro-btq-deq-next-set! item-prev item-next)
       (macro-btq-deq-prev-set! item-next item-prev))))

(##define-macro (macro-btq-deq-insert! deq item)
  `(let ((deq ,deq) (item ,item))

     (##declare (not interrupts-enabled))

     ; add item to tail of deq

     (let ((deq-last (macro-btq-deq-prev deq)))
       (macro-btq-deq-next-set! deq-last item)
       (macro-btq-deq-prev-set! deq item)
       (macro-btq-deq-next-set! item deq)
       (macro-btq-deq-prev-set! item deq-last))))

(##define-macro (macro-btq-link! mutex thread)
  `(let ((mutex ,mutex) (thread ,thread))

     (##declare (not interrupts-enabled))

     (macro-btq-deq-insert! thread mutex)
     (macro-btq-owner-set! mutex thread)))

(##define-macro (macro-btq-unlink! mutex state)
  `(let ((mutex ,mutex) (state ,state))

     (##declare (not interrupts-enabled))

     (macro-btq-deq-init! mutex)
     (macro-btq-owner-set! mutex state)))

; Representation of timeout queues.

(##define-macro (macro-toq-color node)           `(macro-slot 8 ,node))
(##define-macro (macro-toq-color-set! node x)    `(macro-slot 8 ,node ,x))
(##define-macro (macro-toq-parent node)          `(macro-slot 9 ,node))
(##define-macro (macro-toq-parent-set! node x)   `(macro-slot 9 ,node ,x))
(##define-macro (macro-toq-left node)            `(macro-slot 10 ,node))
(##define-macro (macro-toq-left-set! node x)     `(macro-slot 10 ,node ,x))
(##define-macro (macro-toq-right node)           `(macro-slot 11 ,node))
(##define-macro (macro-toq-right-set! node x)    `(macro-slot 11 ,node ,x))
(##define-macro (macro-toq-leftmost node)        `(macro-slot 11 ,node))
(##define-macro (macro-toq-leftmost-set! node x) `(macro-slot 11 ,node ,x))

; Representation of threads.

; The state of a thread is determined by the content of the
; "end-condvar", "exception?" and "result" fields:
;
;  thread state               "end-condvar" "exception?" "result"
;  not yet started            condvar       thunk        #f
;  started, never run         condvar       thunk        ##thread-start-action!
;  started, has run           condvar       #f           action procedure/#f/#t
;  terminated with result     #f            #f           result object
;  terminated with exception  #f            #t           exception object

(##define-macro (macro-thread-frame t)               `(macro-slot 0 ,t))
(##define-macro (macro-thread-frame-set! t x)        `(macro-slot 0 ,t ,x))
; fields 1 and 2 are the deq links of blocked thread queues owned by the thread
; fields 3 to 6 are for keeping this thread in a blocked thread queue
(##define-macro (macro-thread-tgroup t)              `(macro-slot 7 ,t))
(##define-macro (macro-thread-tgroup-set! t x)       `(macro-slot 7 ,t ,x))
; fields 8 to 11 are for keeping this thread in a timeout queue
; fields 12 and 13 are the deq links of threads in this thread group
(##define-macro (macro-thread-floats t)              `(macro-slot 14 ,t))
(##define-macro (macro-thread-floats-set! t x)       `(macro-slot 14 ,t ,x))
(##define-macro (macro-thread-name t)                `(macro-slot 15 ,t))
(##define-macro (macro-thread-name-set! t x)         `(macro-slot 15 ,t ,x))
(##define-macro (macro-thread-end-condvar t)         `(macro-slot 16 ,t))
(##define-macro (macro-thread-end-condvar-set! t x)  `(macro-slot 16 ,t ,x))
(##define-macro (macro-thread-exception? t)          `(macro-slot 17 ,t))
(##define-macro (macro-thread-exception?-set! t x)   `(macro-slot 17 ,t ,x))
(##define-macro (macro-thread-result t)              `(macro-slot 18 ,t))
(##define-macro (macro-thread-result-set! t x)       `(macro-slot 18 ,t ,x))
(##define-macro (macro-thread-specific t)            `(macro-slot 19 ,t))
(##define-macro (macro-thread-specific-set! t x)     `(macro-slot 19 ,t ,x))
(##define-macro (macro-thread-denv t)                `(macro-slot 20 ,t))
(##define-macro (macro-thread-denv-set! t x)         `(macro-slot 20 ,t ,x))
(##define-macro (macro-thread-denv-cache1 t)         `(macro-slot 21 ,t))
(##define-macro (macro-thread-denv-cache1-set! t x)  `(macro-slot 21 ,t ,x))
(##define-macro (macro-thread-denv-cache2 t)         `(macro-slot 22 ,t))
(##define-macro (macro-thread-denv-cache2-set! t x)  `(macro-slot 22 ,t ,x))
(##define-macro (macro-thread-denv-cache3 t)         `(macro-slot 23 ,t))
(##define-macro (macro-thread-denv-cache3-set! t x)  `(macro-slot 23 ,t ,x))

; Access to floating point fields.

(##define-macro (macro-timeout f)                  `(##f64vector-ref ,f 0))
(##define-macro (macro-timeout-set! f x)           `(##f64vector-set! ,f 0 ,x))
(##define-macro (macro-base-priority f)            `(##f64vector-ref ,f 1))
(##define-macro (macro-base-priority-set! f x)     `(##f64vector-set! ,f 1 ,x))
(##define-macro (macro-quantum f)                  `(##f64vector-ref ,f 2))
(##define-macro (macro-quantum-set! f x)           `(##f64vector-set! ,f 2 ,x))
(##define-macro (macro-quantum-used f)             `(##f64vector-ref ,f 3))
(##define-macro (macro-quantum-used-set! f x)      `(##f64vector-set! ,f 3 ,x))
(##define-macro (macro-priority-boost f)           `(##f64vector-ref ,f 4))
(##define-macro (macro-priority-boost-set! f x)    `(##f64vector-set! ,f 4 ,x))
(##define-macro (macro-boosted-priority f)         `(##f64vector-ref ,f 5))
(##define-macro (macro-boosted-priority-set! f x)  `(##f64vector-set! ,f 5 ,x))
(##define-macro (macro-effective-priority f)       `(##f64vector-ref ,f 6))
(##define-macro (macro-effective-priority-set! f x)`(##f64vector-set! ,f 6 ,x))

(##define-macro (macro-thread-timeout t)
  `(macro-timeout (macro-thread-floats ,t)))

(##define-macro (macro-thread-timeout-set! t x)
  `(macro-timeout-set! (macro-thread-floats ,t) ,x))

(##define-macro (macro-thread-base-priority t)
  `(macro-base-priority (macro-thread-floats ,t)))

(##define-macro (macro-thread-base-priority-set! t x)
  `(macro-base-priority-set! (macro-thread-floats ,t) ,x))

(##define-macro (macro-thread-quantum t)
  `(macro-quantum (macro-thread-floats ,t)))

(##define-macro (macro-thread-quantum-set! t x)
  `(macro-quantum-set! (macro-thread-floats ,t) ,x))

(##define-macro (macro-thread-quantum-used t)
  `(macro-quantum-used (macro-thread-floats ,t)))

(##define-macro (macro-thread-quantum-used-set! t x)
  `(macro-quantum-used-set! (macro-thread-floats ,t) ,x))

(##define-macro (macro-thread-priority-boost t)
  `(macro-priority-boost (macro-thread-floats ,t)))

(##define-macro (macro-thread-priority-boost-set! t x)
  `(macro-priority-boost-set! (macro-thread-floats ,t) ,x))

(##define-macro (macro-thread-boosted-priority t)
  `(macro-boosted-priority (macro-thread-floats ,t)))

(##define-macro (macro-thread-boosted-priority-set! t x)
  `(macro-boosted-priority-set! (macro-thread-floats ,t) ,x))

(##define-macro (macro-thread-effective-priority t)
  `(macro-effective-priority (macro-thread-floats ,t)))

(##define-macro (macro-thread-effective-priority-set! t x)
  `(macro-effective-priority-set! (macro-thread-floats ,t) ,x))

(##define-macro (macro-thread-object? obj)
  `(##thread? ,obj))

(##define-macro (macro-make-thread-object . slots)
  `(##subtype-set!
    (##vector 0 ,@slots)
    (macro-subtype-continuation)))

(##define-macro (macro-thread? obj)
  `(macro-thread-object? ,obj))

(##define-macro (macro-check-thread var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-thread? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-thread ',(car form) ,@(cdr form))
          `(##trap-check-thread* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-make-thread thunk name tgroup)
  `(let ((thunk ,thunk) (name ,name) (tgroup ,tgroup))

     (##declare (not interrupts-enabled))

     (let* ((current-thread
             (macro-current-thread))
            (thread
             (macro-make-thread-object
              #f
              #f
              #f
              #f
              #f
              #f
              tgroup
              #f
              #f
              #f
              #f
              #f
              #f
              (let* ((floats
                      (macro-thread-floats current-thread))
                     (base-priority
                      (macro-base-priority floats))
                     (priority-boost
                      (macro-priority-boost floats)))
                (##f64vector
                 (macro-inexact-+0)
                 base-priority
                 (macro-quantum floats)
                 (macro-inexact-+0)
                 priority-boost
                 base-priority
                 base-priority))
              name
              (macro-make-condvar #f)
              thunk
              #f
              (##void)
              (let ((denv (macro-thread-denv current-thread)))
                (macro-make-denv
                 (macro-denv-global denv)
                 (macro-denv-local denv)
                 (macro-denv-interrupt-mask denv)
                 (macro-thread-base-exception-handler)
                 (macro-denv-debugging-settings denv)
                 (macro-denv-input-port denv)
                 (macro-denv-output-port denv)))
              (macro-thread-denv-cache1 current-thread)
              (macro-thread-denv-cache2 current-thread)
              (macro-thread-denv-cache3 current-thread))))
       (macro-btq-deq-init! thread)
;;;;;;       (macro-tgroup-threads-deq-insert! tgroup thread)
       thread)))

(##define-macro (macro-thread-btq-remove-if-in-btq! thread)
  `(let ((thread ,thread))

     (##declare (not interrupts-enabled))

     (if (macro-btq-parent thread)
       (##thread-btq-remove! thread))))

(##define-macro (macro-thread-toq-remove-if-in-toq! thread)
  `(let ((thread ,thread))

     (##declare (not interrupts-enabled))

     (if (macro-toq-parent thread)
       (##thread-toq-remove! thread))))

(##define-macro (macro-thread-reschedule-if-needed!)
  `(let ()

     (##declare (not interrupts-enabled))

     (let ((leftmost (macro-btq-leftmost (macro-run-queue))))
       (if (##not (##eq? leftmost (macro-current-thread)))
         (##thread-reschedule!)
         (##void)))))

(##define-macro (macro-thread-save! proc arg1 arg2 arg3)
  `(##thread-save! ,proc ,arg1 ,arg2 ,arg3))

(##define-macro (macro-thread-restore! thread thunk)
  `(##thread-restore! ,thread ,thunk))

(##define-macro (macro-run-queue)
;  `(##scheduler-state);;;;;;;
  `(##run-queue);;;;;;;;;
)

(##define-macro (macro-primordial-thread)
  `(macro-run-queue-primordial-thread (macro-run-queue)))

(##define-macro (macro-current-thread)
  `(##current-thread))

(##define-macro (macro-thread-base-exception-handler)
  `##thread-base-exception-handler)

(##define-macro (macro-thread-higher-prio? t1 t2)
  `(let ((t1 ,t1) (t2 ,t2))
     (let ((floats1 (macro-thread-floats t1))
           (floats2 (macro-thread-floats t2)))
       (##flonum.< (macro-effective-priority floats2) ; high priority first
                   (macro-effective-priority floats1)))))

(##define-macro (macro-thread-sooner? t1 t2)
  `(let ((t1 ,t1) (t2 ,t2))
     (let ((floats1 (macro-thread-floats t1))
           (floats2 (macro-thread-floats t2)))
       (##flonum.< (macro-timeout floats1)
                   (macro-timeout floats2)))))

(##define-macro (macro-thread-sooner-or-simultaneous-and-higher-prio? t1 t2)
  `(let ((t1 ,t1) (t2 ,t2))
     (let ((floats1 (macro-thread-floats t1))
           (floats2 (macro-thread-floats t2)))
       (if (##not (##flonum.= (macro-timeout floats1)
                              (macro-timeout floats2)))
         (##flonum.< (macro-timeout floats1)
                     (macro-timeout floats2))
         (##flonum.< (macro-effective-priority floats2) ; high priority first
                     (macro-effective-priority floats1))))))

(##define-macro (macro-thread-boost-and-clear-quantum-used! thread)
  `(let ((thread ,thread))

     (##declare (not interrupts-enabled))

     (let ((floats (macro-thread-floats thread)))
       (macro-quantum-used-set! floats (macro-inexact-+0))
       (if (##not (##flonum.= (##flonum.+ (macro-base-priority floats)
                                          (macro-priority-boost floats))
                              (macro-boosted-priority floats)))
         (begin

           ; save old boosted priority for ##thread-boosted-priority-changed!

           (macro-temp-set!
            (macro-thread-floats (macro-run-queue))
            (macro-boosted-priority floats))

           (macro-boosted-priority-set!
            floats
            (##flonum.+ (macro-base-priority floats)
                        (macro-priority-boost floats)))

            (##thread-boosted-priority-changed! thread))))))

(##define-macro (macro-thread-unboost-and-clear-quantum-used! thread)
  `(let ((thread ,thread))

     (##declare (not interrupts-enabled))

     (let ((floats (macro-thread-floats thread)))
       (macro-quantum-used-set! floats (macro-inexact-+0))
       (if (##not (##flonum.= (macro-base-priority floats)
                              (macro-boosted-priority floats)))
         (begin

           ; save old boosted priority for ##thread-boosted-priority-changed!

           (macro-temp-set!
            (macro-thread-floats (macro-run-queue))
            (macro-boosted-priority floats))

           (macro-boosted-priority-set!
            floats
            (macro-base-priority floats))

            (##thread-boosted-priority-changed! thread))))))

(##define-macro (macro-thread-inherit-priority! thread parent);;;;;;;;;;;;;;;
  `(let ((thread ,thread) (parent ,parent))

     (##declare (not interrupts-enabled))

     (let ((thread-floats (macro-thread-floats thread))
           (parent-floats (macro-thread-floats parent)))
       (if (##flonum.< (macro-effective-priority thread-floats)
                       (macro-effective-priority parent-floats))
         (begin
(set! pcase1 (##fixnum.+ pcase1 1));;;;;;;;;;;;;;;;;;
           (macro-effective-priority-set!
            thread-floats
            (macro-effective-priority parent-floats))
           (##thread-effective-priority-changed! thread #t))
(set! pcase2 (##fixnum.+ pcase2 1));;;;;;;;;;;;;;;;;;
))))

; Representation of mutexes.

; fields 1 and 2 are for maintaining this mutex in a deq of btqs
; fields 3 to 5 are for maintaining a queue of blocked threads
; field 6 is the leftmost thread in the queue of blocked threads
; field 7 is the owner of the mutex (or 'not-owned or 'abandoned or #f)
(##define-macro (macro-mutex-name m)        `(macro-slot 8 ,m))
(##define-macro (macro-mutex-name-set! m x) `(macro-slot 8 ,m ,x))

(##define-macro (macro-mutex? obj)
  `(macro-structure-object? ,obj ##mutex-tag))

(##define-macro (macro-check-mutex var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-mutex? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-mutex ',(car form) ,@(cdr form))
          `(##trap-check-mutex* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-make-mutex name)
  `(let ((name ,name))
     (let ((mutex
            (macro-make-structure-object
             ##mutex-tag
             #f
             #f
             #f
             #f
             #f
             #f
             #f
             name)))
       (macro-btq-deq-init! mutex)
       (macro-btq-init! mutex)
       mutex)))

(##define-macro (macro-mutex-unlocked-not-abandoned-and-not-multiprocessor? mutex)
  `(##not (macro-btq-owner ,mutex)))

(##define-macro (macro-mutex-lock! mutex absrel-timeout owner)
  `(let ((mutex ,mutex) (absrel-timeout ,absrel-timeout) (owner ,owner))

     (##declare (not interrupts-enabled))

     (let ((state (macro-btq-owner mutex)))
       (if state
         (##mutex-lock-out-of-line! mutex absrel-timeout owner)
         (begin
           (macro-btq-link! mutex owner)
           #t)))))

(##define-macro (macro-mutex-lock-anonymously! mutex absrel-timeout)
  `(let ((mutex ,mutex) (absrel-timeout ,absrel-timeout))

     (##declare (not interrupts-enabled))

     (let ((state (macro-btq-owner mutex)))
       (if state
         (##mutex-lock-out-of-line! mutex absrel-timeout #f)
         (begin
           (macro-btq-owner-set! mutex 'not-owned)
           #t)))))

(##define-macro (macro-mutex-unlock! mutex)
  `(let ((mutex ,mutex))

     (##declare (not interrupts-enabled))

     (macro-btq-deq-remove! mutex)
     (let ((leftmost (macro-btq-leftmost mutex)))
       (if (##eq? leftmost mutex)
         (begin
           (macro-btq-unlink! mutex #f)
           (##void))
         (##mutex-signal! mutex leftmost #f)))))

(##define-macro (macro-mutex-unlock-no-reschedule! mutex)
  `(let ((mutex ,mutex))

     (##declare (not interrupts-enabled))

     (macro-btq-deq-remove! mutex)
     (let ((leftmost (macro-btq-leftmost mutex)))
       (if (##eq? leftmost mutex)
         (begin
           (macro-btq-unlink! mutex #f)
           (##void))
         (##mutex-signal-no-reschedule! mutex leftmost #f)))))

; Representation of condition variables.

; fields 1 and 2 are for maintaining this condition variable in a deq of btqs
; fields 3 to 5 are for maintaining a queue of blocked threads
; field 6 is the leftmost thread in the queue of blocked threads
; field 7 is the owner of the condition variable
(##define-macro (macro-condvar-name c)            `(macro-slot 8 ,c))
(##define-macro (macro-condvar-name-set! c x)     `(macro-slot 8 ,c ,x))

(##define-macro (macro-condvar? obj)
  `(macro-structure-object? ,obj ##condvar-tag))

(##define-macro (macro-check-condvar var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-condvar? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-condvar ',(car form) ,@(cdr form))
          `(##trap-check-condvar* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-make-condvar name)
  `(let ((name ,name))
     (let ((condvar
            (macro-make-structure-object
             ##condvar-tag
             #f
             #f
             #f
             #f
             #f
             #f
             #f
             name)))
       (macro-btq-deq-init! condvar)
       (macro-btq-init! condvar)
       condvar)))

; Representation of thread groups.

(##define-macro (macro-tgroup-tgroups-deq-next g)       `(macro-slot 1 ,g))
(##define-macro (macro-tgroup-tgroups-deq-next-set! g x)`(macro-slot 1 ,g ,x))
(##define-macro (macro-tgroup-tgroups-deq-prev g)       `(macro-slot 2 ,g))
(##define-macro (macro-tgroup-tgroups-deq-prev-set! g x)`(macro-slot 2 ,g ,x))
(##define-macro (macro-tgroup-tgroups g)                `(macro-slot 3 ,g))
(##define-macro (macro-tgroup-tgroups-set! g x)         `(macro-slot 3 ,g ,x))
(##define-macro (macro-tgroup-parent g)                 `(macro-slot 4 ,g))
(##define-macro (macro-tgroup-parent-set! g x)          `(macro-slot 4 ,g ,x))
(##define-macro (macro-tgroup-name g)                   `(macro-slot 5 ,g))
(##define-macro (macro-tgroup-name-set! g x)            `(macro-slot 5 ,g ,x))
(##define-macro (macro-tgroup-suspend-condvar g)        `(macro-slot 6 ,g))
(##define-macro (macro-tgroup-suspend-condvar-set! g x) `(macro-slot 6 ,g ,x))
(##define-macro (macro-tgroup-unused1 g)                `(macro-slot 7 ,g))
(##define-macro (macro-tgroup-unused1-set! g x)         `(macro-slot 7 ,g ,x))
(##define-macro (macro-tgroup-unused2 g)                `(macro-slot 8 ,g))
(##define-macro (macro-tgroup-unused2-set! g x)         `(macro-slot 8 ,g ,x))
(##define-macro (macro-tgroup-unused3 g)                `(macro-slot 9 ,g))
(##define-macro (macro-tgroup-unused3-set! g x)         `(macro-slot 9 ,g ,x))
(##define-macro (macro-tgroup-unused4 g)                `(macro-slot 10 ,g))
(##define-macro (macro-tgroup-unused4-set! g x)         `(macro-slot 10 ,g ,x))
(##define-macro (macro-tgroup-unused5 g)                `(macro-slot 11 ,g))
(##define-macro (macro-tgroup-unused5-set! g x)         `(macro-slot 11 ,g ,x))
(##define-macro (macro-tgroup-threads-deq-next g)       `(macro-slot 12 ,g))
(##define-macro (macro-tgroup-threads-deq-next-set! g x)`(macro-slot 12 ,g ,x))
(##define-macro (macro-tgroup-threads-deq-prev g)       `(macro-slot 13 ,g))
(##define-macro (macro-tgroup-threads-deq-prev-set! g x)`(macro-slot 13 ,g ,x))

(##define-macro (macro-tgroup? obj)
  `(macro-structure-object? ,obj ##tgroup-tag))

(##define-macro (macro-check-tgroup var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (macro-tgroup? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-tgroup ',(car form) ,@(cdr form))
          `(##trap-check-tgroup* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-make-tgroup name parent)
  `(let ((name ,name) (parent ,parent))
     (let* ((tgroups
             (##vector #f #f #f))
            (tgroup
             (macro-make-structure-object
              ##tgroup-tag
              #f
              #f
              tgroups
              parent
              name
              #f
              #f
              #f
              #f
              #f
              #f
              #f
              #f)))
       (macro-tgroup-tgroups-deq-init! tgroups)
       (macro-tgroup-threads-deq-init! tgroup)
       (if parent
         (macro-tgroup-tgroups-deq-insert!
          (macro-tgroup-tgroups parent)
          tgroup))
       tgroup)))

(##define-macro (macro-tgroup-tgroups-deq-init! deq)
  `(let ((deq ,deq))

     (##declare (not interrupts-enabled))

     (macro-tgroup-tgroups-deq-next-set! deq deq)
     (macro-tgroup-tgroups-deq-prev-set! deq deq)))

(##define-macro (macro-tgroup-tgroups-deq-remove! item)
  `(let ((item ,item))

     (##declare (not interrupts-enabled))

     (let ((item-next (macro-tgroup-tgroups-deq-next item))
           (item-prev (macro-tgroup-tgroups-deq-prev item)))
       (macro-tgroup-tgroups-deq-next-set! item-prev item-next)
       (macro-tgroup-tgroups-deq-prev-set! item-next item-prev))))

(##define-macro (macro-tgroup-tgroups-deq-insert! deq item)
  `(let ((deq ,deq) (item ,item))

     (##declare (not interrupts-enabled))

     ; add item to tail of deq

     (let ((deq-last (macro-tgroup-tgroups-deq-prev deq)))
       (macro-tgroup-tgroups-deq-next-set! deq-last item)
       (macro-tgroup-tgroups-deq-prev-set! deq item)
       (macro-tgroup-tgroups-deq-next-set! item deq)
       (macro-tgroup-tgroups-deq-prev-set! item deq-last))))

(##define-macro (macro-tgroup-threads-deq-init! deq)
  `(let ((deq ,deq))

     (##declare (not interrupts-enabled))

     (macro-tgroup-threads-deq-next-set! deq deq)
     (macro-tgroup-threads-deq-prev-set! deq deq)))

(##define-macro (macro-tgroup-threads-deq-remove! item)
  `(let ((item ,item))

     (##declare (not interrupts-enabled))

     (let ((item-next (macro-tgroup-threads-deq-next item))
           (item-prev (macro-tgroup-threads-deq-prev item)))
       (macro-tgroup-threads-deq-next-set! item-prev item-next)
       (macro-tgroup-threads-deq-prev-set! item-next item-prev))))

(##define-macro (macro-tgroup-threads-deq-insert! deq item)
  `(let ((deq ,deq) (item ,item))

     (##declare (not interrupts-enabled))

     ; add item to tail of deq

     (let ((deq-last (macro-tgroup-threads-deq-prev deq)))
       (macro-tgroup-threads-deq-next-set! deq-last item)
       (macro-tgroup-threads-deq-prev-set! deq item)
       (macro-tgroup-threads-deq-next-set! item deq)
       (macro-tgroup-threads-deq-prev-set! item deq-last))))

; Representation of the run queue.

; fields 1 and 2 are the deq links of blocking device condvars
; fields 3 to 5 are for maintaining a queue of runnable threads
; field 6 is the leftmost thread in the queue of runnable threads
; field 7 must be #f (the queue of runnable threads has no owner)
; fields 8 to 10 are for maintaining a timeout queue of threads
; field 11 is the leftmost thread in the timeout queue of threads
(##define-macro (macro-run-queue-primordial-thread rq)
  `(macro-slot 12 ,rq))
(##define-macro (macro-run-queue-primordial-thread-set! rq x)
  `(macro-slot 12 ,rq ,x))
(##define-macro (macro-run-queue-unused rq)
  `(macro-slot 13 ,rq))
(##define-macro (macro-run-queue-unused-set! rq x)
  `(macro-slot 13 ,rq ,x))
; field 14 is for storing the current time, heartbeat interval and a
; temporary float

(##define-macro (macro-current-time f)             `(##f64vector-ref ,f 0))
(##define-macro (macro-current-time-set! f x)      `(##f64vector-set! ,f 0 ,x))
(##define-macro (macro-heartbeat-interval f)       `(##f64vector-ref ,f 1))
(##define-macro (macro-heartbeat-interval-set! f x)`(##f64vector-set! ,f 1 ,x))
(##define-macro (macro-temp f)                     `(##f64vector-ref ,f 2))
(##define-macro (macro-temp-set! f x)              `(##f64vector-set! ,f 2 ,x))

(##define-macro (macro-update-current-time!)
  `(##get-current-time! (macro-thread-floats (macro-run-queue))))

(##define-macro (macro-make-run-queue)
  `(let ((run-queue
          (macro-make-structure-object
           '#(run-queue -1)
           #f
           #f
           #f
           #f
           #f
           #f
           #f
           #f
           #f
           #f
           #f
           #f
           #f
           (##f64vector (macro-inexact-+0)
                        (macro-inexact-+0)
                        (macro-inexact-+0)))))
     (macro-btq-deq-init! run-queue)
     (macro-btq-init! run-queue)
     (macro-toq-init! run-queue)
     run-queue))

;------------------------------------------------------------------------------
