/* file: "os.c" */

/* Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved. */

/*
 * This module implements the operating system specific routines
 * including:
 *
 *  - OS specific initialization/finalization
 *  - process termination
 *  - fatal error handling
 *  - conversion of error codes to error messages
 *  - low-level memory allocation
 *  - time management
 *  - process times (user time, system time and real time).
 *  - heartbeat interrupt handling
 *  - user interrupt handling
 *  - access to environment variables
 *  - shell command
 *  - dynamic loading
 *  - dynamic C compilation
 *  - floating point environment setup
 *  - virtual memory statistics
 *  - filesystem path expansion
 *  - formatting of source code position
 *  - operations on I/O devices
 */

#define ___INCLUDED_FROM_OS
#define ___VERSION_MAJOR 21
#include "gambit.h"

#include "os.h"
#include "setup.h"
#include "mem.h"
#include "c_intf.h"

#define USE_LINEEDIT
#define LINEEDIT_WITH_NONBLOCKING_IO
#define LINEEDIT_SUPPORT_F5_TO_F12
#define LINEEDIT_WITH_LOCAL_CLIPBOARD
#define LINEEDIT_REPORT_CHAR_ENCODING_ERRORS


/*---------------------------------------------------------------------------*/

/* Conversion of OS error codes to ___ERR_CODE. */

#ifdef USE_errno

___HIDDEN ___ERR_CODE err_code_from_errno ___PVOID
{
  int e = errno;

  if (e == 0)
    return ___UNKNOWN_ERR;

  return ___ERR_CODE_FROM_ERRNO(e);
}

#endif


#ifdef USE_GetLastError

___HIDDEN ___ERR_CODE err_code_from_GetLastError ___PVOID
{
  DWORD e = GetLastError ();

  if (e == ERROR_SUCCESS)
    return ___UNKNOWN_ERR;

  return ___ERR_CODE_FROM_HRESULT(e);
}

#endif


#ifdef USE_WSAGetLastError

___HIDDEN ___ERR_CODE err_code_from_WSAGetLastError ___PVOID
{
  DWORD e = WSAGetLastError ();

  if (e == ERROR_SUCCESS)
    return ___UNKNOWN_ERR;

  return ___ERR_CODE_FROM_WSAERROR(e);
}

#endif


#ifdef USE_OSErr

___HIDDEN ___ERR_CODE err_code_from_OSErr
   ___P((OSErr e),
        (e)
OSErr e;)
{
  if (e == noErr)
    return ___UNKNOWN_ERR;

  return ___ERR_CODE_FROM_OSERR(e);
}

#endif


/*---------------------------------------------------------------------------*/

/* OS specific initialization/finalization. */

#ifndef USE_MACOS

___HIDDEN void setup_os_specific ___PVOID
{
}


___HIDDEN void cleanup_os_specific ___PVOID
{
}


#endif


#ifdef USE_MACOS

/* Which features are available. */

___HIDDEN Boolean has_GetUTCDateTime = 0;
___HIDDEN Boolean has_GetDateTime = 0;
___HIDDEN Boolean has_ReadLocation = 0;
___HIDDEN Boolean has_Delay = 0;
___HIDDEN Boolean has_IdleUpdate = 0;
___HIDDEN Boolean has_WaitNextEvent = 0;
___HIDDEN Boolean has_OSDispatch = 0;
___HIDDEN Boolean has_FindFolder = 0;
___HIDDEN Boolean has_AliasMgr = 0;
___HIDDEN Boolean has_AppleEvents = 0;


#define test_bit(n,i) (n&(1<<i))


___HIDDEN TrapType get_trap_type
   ___P((short trap_num),
        (trap_num)
short trap_num;)
{
  /* OS traps start with A0, Tool traps with A8 or AA. */

  if (trap_num & 0x0800)
    return ToolTrap;
  else
    return OSTrap;
}


___HIDDEN short nb_toolbox_traps ___PVOID
{
  /* InitGraf (trap $A86E) is always implemented. */

  if (NGetTrapAddress (0xA86E, ToolTrap) == NGetTrapAddress (0xAA6E, ToolTrap))
    return (0x200);
  else
    return (0x400);
}


___HIDDEN Boolean trap_exists
  ___P((short trap_num),
       (trap_num)
short trap_num;)
{
  TrapType typ = get_trap_type (trap_num);
  if ((typ == ToolTrap) && ((trap_num &= 0x07FF) >= nb_toolbox_traps ()))
    return false;
  return (NGetTrapAddress (_Unimplemented, ToolTrap) !=
          NGetTrapAddress (trap_num, typ));
}


/* String conversion utilities. */

___HIDDEN Boolean c2pascal
   ___P((char *cstr,
         StringPtr pstr,
         int max_length),
        (cstr,
         pstr,
         max_length)
char *cstr;
StringPtr pstr;
int max_length;)
{
  StringPtr p1 = pstr+1;
  char *p2 = cstr;
  while (max_length > 0 && *p2 != '\0')
    {
       *p1++ = *p2++;
       max_length--;
    }
  if (*p2 != '\0')
    return 0;
  else
    {
      pstr[0] = p2-cstr;
      return 1;
    }
}


___HIDDEN Boolean pascal2c
   ___P((StringPtr pstr,
         char *cstr,
         int max_length),
        (pstr,
         cstr,
         max_length)
StringPtr pstr;
char *cstr;
int max_length;)
{
  char *p1 = cstr;
  StringPtr p2 = pstr+1;
  int len = pstr[0];
  if (len > max_length)
    return 0;
  else
    {
      while (len-- > 0)
        *p1++ = *p2++;
      *p1++ = '\0';
      return 1;
    }
}


#define DIR_SEPARATOR1 ':'
#define PARENT_HOP ":"

#define DIR_SEPARATOR(c)((c) == DIR_SEPARATOR1)
#define SEPARATOR(c)DIR_SEPARATOR(c)

___HIDDEN OSErr make_ResolvedFSSpec
   ___P((short vol,
         long dir,
         ConstStr255Param path,
         FSSpec *spec),
        (vol,
         dir,
         path,
         spec)
short vol;
long dir;
ConstStr255Param path;
FSSpec *spec;)
{
  OSErr err;
  Str255 name;
  StringPtr start = ___CAST(StringPtr,path+1);
  StringPtr end = start + path[0];
  StringPtr p1 = start;
  StringPtr p2 = name+1;
  CInfoPBRec pb;
  Boolean is_folder;
  Boolean is_aliased;

  if (!has_AliasMgr)
    return unimpErr;

  spec->vRefNum = vol;
  spec->parID = dir;

  do
    {
      p2 = name+1;
      while (p1 < end && DIR_SEPARATOR(*p1))  /* copy leading ':'s */
        *p2++ = *p1++;
      while (p1 < end && !DIR_SEPARATOR(*p1)) /* copy name that follows */
        *p2++ = *p1++;
      if (p1 < end && DIR_SEPARATOR(*p1))     /* end with a ':' if folder */
        *p2++ = DIR_SEPARATOR1;
      name[0] = p2 - (name+1);

      err = FSMakeFSSpec (spec->vRefNum, spec->parID, name, spec);
      if (err == fnfErr && p1 == end)
        return noErr;
      if (err != noErr)
        return err;

      if ((err = ResolveAliasFile (spec, 1, &is_folder, &is_aliased)) != noErr)
        return err;
      if (is_folder)
        {
          pb.dirInfo.ioNamePtr = spec->name;
          pb.dirInfo.ioVRefNum = spec->vRefNum;
          pb.dirInfo.ioDrDirID = spec->parID;
          pb.dirInfo.ioFDirIndex = 0;
          if ((err = PBGetCatInfoSync (&pb)) != noErr)
            return err;
          spec->parID = pb.hFileInfo.ioDirID;
          spec->name[0] = 0;
        }
      else if (p1 < end)
        return dirNFErr;
    } while (p1 < end);

  return noErr;
}


___HIDDEN OSErr ResolvedFSSpec_to_fullpath
   ___P((FSSpec *spec,
         StringPtr fullpath),
        (spec,
         fullpath)
FSSpec *spec;
StringPtr fullpath;)
{
  OSErr err;
  int i;
  Str255 result;
  StringPtr p = result + sizeof(result);
  CInfoPBRec pb;
  Str31 name;

  for (i = spec->name[0]; i > 0; i--)
    *--p = spec->name[i];

  pb.dirInfo.ioNamePtr = name;
  pb.dirInfo.ioVRefNum = spec->vRefNum;
  pb.dirInfo.ioDrParID = spec->parID;
  pb.dirInfo.ioFDirIndex = -1;

  do
    {
      pb.dirInfo.ioDrDirID = pb.dirInfo.ioDrParID;
      if ((err = PBGetCatInfoSync (&pb)) != noErr)
        return err;
      if (p-name[0]-1 < result)
        return bdNamErr; /* file name is too long */
      *--p = DIR_SEPARATOR1;
      for (i = name[0]; i > 0; i--)
        *--p = name[i];
    } while (pb.dirInfo.ioDrDirID != fsRtDirID);

  i = result + sizeof(result) - p;
  *fullpath++ = i;
  while (i > 0)
    {
      *fullpath++ = *p++;
      i--;
    }

  return noErr;
}


___HIDDEN ___BOOL path_expand_to_absolute
   ___P((char *path,
         char *new_path,
         long max_length),
        (path,
         new_path,
         max_length)
char *path;
char *new_path;
long max_length)
{
  ___BOOL result = 0;
  FSSpec spec;
  short vol;
  long dir;
  char tmp[___PATH_MAX_LENGTH+1];
  Str255 ppath;

  if (path[0] == '~')
    {
      if (path[1] == '~')
        {
          /* "~~" or "~~:xxx..." */

          int i = 0;
          int j = 0;
          int sep = 0;
          char *gambc_dir;

          if (!has_FindFolder)
            goto ret;

          if (path[2]!='\0' && !DIR_SEPARATOR(path[2]))
            goto ret;

#ifdef ___GAMBCDIR
          gambc_dir = ___GAMBCDIR;
#else
          gambc_dir = ":Gambit-C:gambc";
#endif

          i += 2;

          while (*gambc_dir != '\0')
            if (j < ___PATH_MAX_LENGTH)
              {
                tmp[j] = *gambc_dir++;
                j++;
              }
            else
              goto ret;

          while (path[i] != '\0')
            if (j < ___PATH_MAX_LENGTH)
              {
                if (DIR_SEPARATOR(path[i]))
                  sep = 1;
                tmp[j++] = path[i++];
              }
            else
              goto ret;

          if (!sep)
            if (j < ___PATH_MAX_LENGTH)
              tmp[j++] = DIR_SEPARATOR1;
            else
              goto ret;

          tmp[j] = '\0';
          path = tmp;

          if (FindFolder (kOnSystemDisk,
                          kPreferencesFolderType,
                          0,
                          &vol,
                          &dir)
              != noErr)
            goto ret;
        }
      else if (path[1]!='\0' && !DIR_SEPARATOR(path[1]))
        {
          /* "~user" or "~user:xxx..." */

          goto ret; /* no equivalent on Macintosh */
        }
      else
        {
          /* "~" or "~:xxx..." */

          path++;
          vol = 0; /* use default volume and directory
                      (folder containing application) */
          dir = 0;
        }
    }
  else
    {
      vol = 0; /* use default volume and directory
                  (folder containing application) */
      dir = 0;
    }

  if (!c2pascal (path, ppath, 255) ||
      make_ResolvedFSSpec (vol, dir, ppath, &spec) != noErr ||
      ResolvedFSSpec_to_fullpath (&spec, ppath) != noErr ||
      !pascal2c (ppath, new_path, max_length))
    goto ret;

  result = 1;

 ret:

  return result;
}


___HIDDEN OSErr copy_file_sectors
   ___P((short src_refnum,
         short dst_refnum),
        (src_refnum,
         dst_refnum)
short src_refnum;
short dst_refnum;)
{
  OSErr err1, err2;
  char buf[2048];
  long count1, count2;

  do
    {
      count1 = sizeof (buf);
      err1 = FSRead (src_refnum, &count1, buf);
      if (err1 != noErr && err1 != eofErr)
        return err1;
      count2 = count1;
      err2 = FSWrite (dst_refnum, &count2, buf);
      if (err2 != noErr || count1 != count2)
        return err2;
    } while (err1 != eofErr);

  return noErr;
}


___HIDDEN OSErr copy_file
   ___P((FSSpec src_spec,
         FSSpec dst_spec),
        (src_spec,
         dst_spec)
FSSpec src_spec;
FSSpec dst_spec;)
{
  OSErr err, err2;
  short src_refnum, dst_refnum;
  FInfo src_info;

  if (((err = FSpDelete (&dst_spec)) == noErr || err == fnfErr) &&
      (err = FSpGetFInfo (&src_spec, &src_info)) == noErr &&
      (err = FSpCreate (&dst_spec, '????', '????', 0)) == noErr)
    {
      src_info.fdFlags = src_info.fdFlags & ~kHasBeenInited;
      if ((err = FSpSetFInfo (&dst_spec, &src_info) == noErr) &&
          (err = FSpOpenRF (&src_spec, fsRdPerm, &src_refnum) == noErr))
        {
          if ((err = FSpOpenRF (&dst_spec, fsWrPerm, &dst_refnum)) == noErr)
            {
              err = copy_file_sectors (src_refnum, dst_refnum);
              err2 = FSClose (dst_refnum);
              if (err == noErr)
                err = err2;
            }
          err2 = FSClose (src_refnum);
          if (err == noErr)
            err = err2;
          if (err == noErr &&
              (err = FSpOpenDF (&src_spec, fsRdPerm, &src_refnum) == noErr))
            {
              if ((err = FSpOpenDF (&dst_spec, fsWrPerm, &dst_refnum)) == noErr)
                {
                  err = copy_file_sectors (src_refnum, dst_refnum);
                  err2 = FSClose (dst_refnum);
                  if (err == noErr)
                    err = err2;
                }
              err2 = FSClose (src_refnum);
              if (err == noErr)
                err = err2;
            }
        }
      if (err != noErr)
        FSpDelete (&dst_spec);
    }

  return err;
}


___HIDDEN void setup_os_specific ___PVOID
{
  has_GetUTCDateTime = trap_exists (_UTCDateTime); /*********** correct? */
  has_GetDateTime = trap_exists (_GetDateTime);
  has_ReadLocation = trap_exists (_ReadLocation);
  has_Delay = trap_exists (_Delay);

  has_IdleUpdate = trap_exists (_IdleUpdate);
  has_WaitNextEvent = trap_exists (_WaitNextEvent);
  has_OSDispatch = trap_exists (_OSDispatch);

  {
    long response;

    has_FindFolder = (Gestalt (gestaltFindFolderAttr, &response) == noErr &&
                      test_bit (response, gestaltFindFolderPresent));
    has_AliasMgr = (Gestalt (gestaltAliasMgrAttr, &response) == noErr &&
                    test_bit (response, gestaltAliasMgrPresent));
    has_AppleEvents = (Gestalt (gestaltAppleEventsAttr, &response) == noErr &&
                       test_bit (response, gestaltAppleEventsPresent));
  }
}


___HIDDEN void cleanup_os_specific ___PVOID
{
}


#endif

/*---------------------------------------------------------------------------*/

/* Process termination. */

void ___exit_process
   ___P((int status),
        (status)
int status;)
{
  exit (status);
}


/*---------------------------------------------------------------------------*/

/* Fatal error handling. */

void ___fatal_error
   ___P((char *msg),
        (msg)
char *msg;)
{
  if (___setup_params.fatal_error != 0)
    ___setup_params.fatal_error (msg);
  else
    fprintf (stderr, "*** FATAL ERROR -- %s\n", msg);

  ___exit_process (1);
}


/*---------------------------------------------------------------------------*/

/* Conversion of error codes to error messages. */

___HIDDEN char *c_type_name_table[] =
{
  "int8",
  "unsigned-int8",
  "int16",
  "unsigned-int16",
  "int32",
  "unsigned-int32",
  "int64",
  "unsigned-int64",
  "float32",
  "float64",
  "char",
  "signed-char",
  "unsigned-char",
  "latin1",
  "ucs2",
  "ucs4",
  "short",
  "unsigned-short",
  "int",
  "unsigned-int",
  "long",
  "unsigned-long",
  "float",
  "double",
  "pointer",
  "nonnull-pointer",
  "function",
  "bool",
  "char-string",
  "latin1-string",
  "utf8-string",
  "ucs2-string",
  "ucs4-string",
  "VARIANT",
  "(heap overflow)"
};


void ___err_code_to_string
   ___P((___ERR_CODE err,
         char *str),
        (err,
         str)
___ERR_CODE err;
char *str;)
{
  int facility = ___ERR_CODE_FACILITY(err);

  if (facility >= ___ERR_CODE_FACILITY_SYSTEM)
    {
      /* System specific error code */

      if (err == ___CDEF_HEAP_OVERFLOW_ERR)
        {
          sprintf (str, "Heap overflow while allocating stack marker");
        }
      else if (err == ___IMPL_LIMIT_ERR)
        {
          sprintf (str, "Implementation limit exceeded");
        }
      else if (err == ___HEAP_OVERFLOW_ERR)
        {
          sprintf (str, "Heap overflow");
        }
      else if (err == ___CLOSED_DEVICE_ERR)
        {
          sprintf (str, "Device is closed");
        }
      else if (err == ___INVALID_OP_ERR)
        {
          sprintf (str, "Invalid operation");
        }
      else if ((err >= ___STOC_BASE && err <= ___STOC_MAX) ||
               (err >= ___CTOS_BASE && err <= ___CTOS_MAX))
        {
          int arg_num, c_type_index;
          char *dir;
          if (err <= ___STOC_MAX)
            {
              arg_num = (err-___STOC_BASE) & ((1<<7)-1);
              c_type_index = (err-___STOC_BASE) >> 7;
              dir = "to C";
            }
          else
            {
              arg_num = (err-___CTOS_BASE) & ((1<<7)-1);
              c_type_index = (err-___CTOS_BASE) >> 7;
              dir = "from C";
            }
          if (arg_num == ___RETURN_POS)
            sprintf (str,
                     "Can't convert result %s %s",
                     dir,
                     c_type_name_table[c_type_index]);
          else if (arg_num == 0)
            sprintf (str,
                     "Can't convert %s %s",
                     dir,
                     c_type_name_table[c_type_index]);
          else
            sprintf (str,
                     "Can't convert argument %d %s %s",
                     arg_num,
                     dir,
                     c_type_name_table[c_type_index]);
        }
      else
        sprintf (str, "Unknown error");
    }
  else if (facility >= ___ERR_CODE_FACILITY_ERRNO)
    {
      /* ANSI-C errno error code */

      char *msg;
      int i;

      msg = strerror (___ERR_CODE_TO_ERRNO(err));

      if (msg == NULL)
        msg = "Unknown error";

      i = 0;
      while (i < ___ERR_MAX_LENGTH)
        if ((str[i++] = *msg++) == '\0')
          break;
      str[i] = '\0';
    }
  else if (facility >= ___ERR_CODE_FACILITY_MACOS)
    {
      /* MACOS error code */

#ifndef USE_XXXX
      str[0] = '\0';
#endif

#ifdef USE_XXXX
      str[0] = '\0';
#endif
    }
  else
    {
      /* Windows HRESULT error code */

#ifndef USE_FormatMessage
      str[0] = '\0';
#endif

#ifdef USE_FormatMessage
      DWORD len = FormatMessage 
                    (FORMAT_MESSAGE_FROM_SYSTEM+FORMAT_MESSAGE_MAX_WIDTH_MASK,
                     NULL,
                     ___ERR_CODE_TO_HRESULT(err),
                     MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
                     str,
                     ___ERR_MAX_LENGTH,
                     NULL);

      if (len == 0)
        str[0] = '\0';
#endif
    }
}


/*---------------------------------------------------------------------------*/

/* Low-level memory allocation. */

#ifdef ___DEBUG_ALLOC_MEM

___HIDDEN long alloc_mem_calls;
___HIDDEN long free_mem_calls;

#endif


___HIDDEN void setup_alloc_mem ___PVOID
{
#ifdef ___DEBUG_ALLOC_MEM
  alloc_mem_calls = 0;
  free_mem_calls = 0;
#endif
}


___HIDDEN void cleanup_alloc_mem ___PVOID
{
#ifdef ___DEBUG_ALLOC_MEM
  if (alloc_mem_calls != free_mem_calls)
    fprintf (stderr,
             "*** MEMORY LEAK: alloc_mem_calls = %ld  free_mem_calls = %ld\n",
             alloc_mem_calls,
             free_mem_calls);
#endif
}


void *___alloc_mem
   ___P((unsigned long bytes),
        (bytes)
unsigned long bytes;)
{
#ifdef ___DEBUG_ALLOC_MEM
#ifdef USE_WIN32
  InterlockedIncrement (&alloc_mem_calls);
#else
  alloc_mem_calls++;
#endif
#endif

#ifdef USE_TempNewHandle
  if (has_OSDispatch)
    {
      OSErr e;
      Ptr p;
      Handle h = TempNewHandle (sizeof (Handle) + bytes, &err);
      if (e != noErr || h == 0)
        return 0;
      HLock (h);
      p = *h;
      *___CAST(Handle*,p) = h;
      return p + sizeof (Handle);
    }
  else
    return malloc (bytes);
#else
  return malloc (bytes);
#endif
}


void ___free_mem
   ___P((void *ptr),
        (ptr)
void *ptr;)
{
#ifdef ___DEBUG_ALLOC_MEM
#ifdef USE_WIN32
  InterlockedIncrement (&free_mem_calls);
#else
  free_mem_calls++;
#endif
#endif

#ifdef USE_TempNewHandle
  if (has_OSDispatch)
    {
      OSErr e;
      Handle h = *___CAST(Handle*,___CAST(Ptr,ptr) - sizeof (Handle));
      HUnlock (h);
      TempDisposeHandle (h, &e);
    }
  else
    free (ptr);
#else
  free (ptr);
#endif
}


/*---------------------------------------------------------------------------*/

/* Time management. */

#define TIME_REFERENCE_POINT 1970

#define DAYS_UNTIL_JAN_1_2000(year) \
(730119 - ((year)-1)*365 - ((year)-1)/4 + ((year)-1)/100 - ((year)-1)/400)

#define DAYS_UNTIL_TIME_REFERENCE_POINT(year) \
(DAYS_UNTIL_JAN_1_2000(year) - DAYS_UNTIL_JAN_1_2000(TIME_REFERENCE_POINT))

#define JAN_1(year)(DAYS_UNTIL_TIME_REFERENCE_POINT(year) * 24 * 60 * 60)

#ifdef ___FLOAT_TIME_REPRESENTATION

#define TIME_POS_INFINITY (1.0/0.0)  /* positive infinity */
#define TIME_NEG_INFINITY (-1.0/0.0) /* negative infinity */
#define TIME_ZERO 0.0

#endif

#ifdef ___INT_TIME_REPRESENTATION

#define TIME_POS_INFINITY { 2147483647, 999999999 }
#define TIME_NEG_INFINITY { -2147483648, 0 }
#define TIME_ZERO { 0, 0 }

#endif


___time ___time_pos_infinity = TIME_POS_INFINITY;
___time ___time_neg_infinity = TIME_NEG_INFINITY;
___time ___time_zero = TIME_ZERO;


void ___time_from_nsecs
   ___P((___time *tim,
         ___SM32 secs,
         ___SM32 nsecs),
        (tim,
         secs,
         nsecs)
___time *tim;
___SM32 secs;
___SM32 nsecs;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  *tim = secs + nsecs / 1000000000.0;
#endif

#ifdef ___INT_TIME_REPRESENTATION
  tim->secs = secs + nsecs / 1000000000;
  tim->nsecs = nsecs % 1000000000;
#endif
}


void ___time_add
   ___P((___time *tim1,
         ___time tim2),
        (tim1,
         tim2)
___time *tim1;
___time tim2;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  *tim1 += tim2;
#endif

#ifdef ___INT_TIME_REPRESENTATION
  ___SM32 total_secs = tim1->secs + tim2.secs;
  ___SM32 total_nsecs = tim1->nsecs + tim2.nsecs;
  tim1->secs += tim2.secs + total_nsecs / 1000000000;
  tim1->nsecs = total_nsecs % 1000000000;
#endif
}


void ___time_subtract
   ___P((___time *tim1,
         ___time tim2),
        (tim1,
         tim2)
___time *tim1;
___time tim2;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  *tim1 -= tim2;
#endif

#ifdef ___INT_TIME_REPRESENTATION
  ___SM32 total_nsecs;

  if (tim2.nsecs > tim1->nsecs)
    {
      tim1->secs -= tim2.secs + 1;
      tim1->nsecs += 1000000000 - tim2.nsecs;
    }
  else
    {
      tim1->secs -= tim2.secs;
      tim1->nsecs -= tim2.nsecs;
    }
#endif
}


___BOOL ___time_equal
   ___P((___time tim1,
         ___time tim2),
        (tim1,
         tim2)
___time tim1;
___time tim2;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  return (tim1 == tim2);
#endif

#ifdef ___INT_TIME_REPRESENTATION
  return (tim1.secs == tim2.secs && tim1.nsecs == tim2.nsecs);
#endif
}


___BOOL ___time_less
   ___P((___time tim1,
         ___time tim2),
        (tim1,
         tim2)
___time tim1;
___time tim2;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  return (tim1 < tim2);
#endif

#ifdef ___INT_TIME_REPRESENTATION
  return (tim1.secs < tim2.secs ||
          (tim1.secs == tim2.secs && tim1.nsecs < tim2.nsecs));
#endif
}


___BOOL ___time_positive
   ___P((___time tim),
        (tim)
___time tim;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  return (tim > 0.0);
#endif

#ifdef ___INT_TIME_REPRESENTATION
  return (tim.secs > 0 ||
          (tim.secs == 0 && tim.nsecs > 0));
#endif
}


void ___time_get_current_time
   ___P((___time *tim),
        (tim)
___time *tim;)
{
#ifndef USE_time
#ifndef USE_ftime
#ifndef USE_gettimeofday
#ifndef USE_clock_gettime
#ifndef USE_getclock
#ifndef USE__ftime
#ifndef USE_GetSystemTime
#ifndef USE_MACOS

  *tim = ___time_neg_infinity;

#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif

#ifdef USE_time

  time_t t;
  time (&t);
  ___time_from_nsecs (tim, t-JAN_1(1970), 0);

#endif

#ifdef USE_ftime

  struct timeb tb;
  ftime (&tb);
  ___time_from_nsecs (tim, tb.time-JAN_1(1970), tb.millitm*1000000);

#endif

#ifdef USE_gettimeofday

  struct timeval tv;
  if (gettimeofday (&tv, NULL) == 0)
    ___time_from_nsecs (tim, tv.tv_sec-JAN_1(1970), tv.tv_usec*1000);
  else
    *tim = ___time_neg_infinity;

#endif

#ifdef USE_clock_gettime

  /* The clock_gettime function is in POSIX.1b (realtime programming). */

  struct timespec ts;
  if (clock_gettime (CLOCK_REALTIME, &ts) == 0)
    ___time_from_nsecs (tim, ts.tv_sec-JAN_1(1970), ts.tv_nsec);
  else
    *tim = ___time_neg_infinity;

#endif

#ifdef USE_getclock

  /* The getclock function is in the Tru64 UNIX standard C library. */

  struct timespec ts;
  if (getclock (TIMEOFDAY, &ts) == 0)
    ___time_from_nsecs (tim, ts.tv_sec-JAN_1(1970), ts.tv_nsec);
  else
    *tim = ___time_neg_infinity;

#endif

#ifdef USE__ftime

  struct _timeb tb;
  _ftime (&tb);
  ___time_from_nsecs (tim, tb.time-JAN_1(1970), tb.millitm*1000000);

#endif

#ifdef USE_GetSystemTime

  SYSTEMTIME st;
  FILETIME ft;
  GetSystemTime (&st);
  if (SystemTimeToFileTime (&st, &ft))
    {
      LONGLONG x = *___CAST(LONGLONG*,&ft);
      ___SM32 secs = x / 10000000 - JAN_1(1601);
      ___SM32 nsecs = x % 10000000 * 100;
      ___time_from_nsecs (tim, secs, nsecs);
    }
  else
    *tim = ___time_neg_infinity;

#endif

#ifdef USE_MACOS

  if (has_GetUTCDateTime)
    {
      UTCDateTime dt;

      if (GetUTCDateTime (&dt, kUTCDefaultOptions) == noErr)
        ___time_from_nsecs (tim,
                            dt.lowSeconds-JAN_1(1904),
                            ___CAST(___SM32,dt.fraction)*15258 +
                            ((___CAST(___SM32,dt.fraction)*101)>>7));
      else
        *tim = ___time_neg_infinity;
    }
  else if (has_GetDateTime)
    {
      unsigned long t;

      GetDateTime (&t);  /* Get seconds since 01-01-1904 in local time. */

      if (has_ReadLocation)
        {
          /* Adjust to UTC based on machine location. */

          MachineLocation here;
          long offset;

          ReadLocation (&here);
          offset = here.u.gmtDelta & 0x00ffffff;
          if (offset & 0x00800000)
            offset |= 0xff000000;
          t -= offset;
        }

      ___time_from_nsecs (tim, t-JAN_1(1904), 0);
    }
  else
    *tim = ___time_neg_infinity;

#endif
}


void ___time_to_seconds
   ___P((___time tim,
         ___F64 *seconds),
        (tim,
         seconds)
___time tim;
___F64 *seconds;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  *seconds = tim;
#endif

#ifdef ___INT_TIME_REPRESENTATION
  *seconds = tim.secs + tim.nsecs / 1000000000.0;
#endif
}


void ___time_from_seconds
   ___P((___time *tim,
         ___F64 seconds),
        (tim,
         seconds)
___time *tim;
___F64 seconds;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  *tim = seconds;
#endif

#ifdef ___INT_TIME_REPRESENTATION
  if (seconds >= 2147483648.0) /* more than 68 years is pos infinity! */
    *tim = ___time_pos_infinity;
  else if (seconds < -2147483648.0) /* less than -68 years is neg infinity! */
    *tim = ___time_neg_infinity;
  else
    {
      ___SM32 secs = seconds;
      ___SM32 nsecs = (seconds - secs) * 1000000000.0;
      if (nsecs < 0)
        {
          secs--;
          nsecs += 1000000000;
        }
      tim->secs = secs;
      tim->nsecs = nsecs;
    }
#endif
}


___HIDDEN void absolute_time_to_relative_time
   ___P((___time tim,
         ___time *rtim),
        (tim,
         rtim)
___time tim;
___time *rtim;)
{
  if (___time_equal (tim, ___time_pos_infinity))
    *rtim = ___time_pos_infinity;
  else
    {
      ___time now;
      *rtim = tim;
      ___time_get_current_time (&now);
      ___time_subtract (rtim, now);
    }
}


#ifndef LINEEDIT_WITH_NONBLOCKING_IO

___HIDDEN void absolute_time_sleep
   ___P((___time tim),
        (tim)
___time tim;)
{
  absolute_time_to_relative_time (tim, &tim);

  if (___time_positive (tim))
    {
      ___SM32 nsecs;

#ifdef ___FLOAT_TIME_REPRESENTATION
      if (tim > 2.0)
        nsecs = 2000000000;
      else
        nsecs = tim * 1000000000.0;
#endif

#ifdef ___INT_TIME_REPRESENTATION
      if (tim.secs > 2)
        nsecs = 2000000000;
      else
        nsecs = tim.secs * 1000000000 + tim.nsecs;
#endif

#ifndef USE_sleep
#ifndef USE_nanosleep
#ifndef USE_Sleep
#ifndef USE_MACOS
      /* Can't sleep... so stay awake! */
#endif
#endif
#endif
#endif

#ifdef USE_sleep
      sleep (nsecs / 1000000000);
#endif

#ifdef USE_nanosleep
      {
        /* The nanosleep function is in POSIX.1b (realtime programming). */
        struct timespec ts;
        ts.tv_sec = nsecs / 1000000000;
        ts.tv_nsec = nsecs;
        nanosleep (&ts, NULL);
      }
#endif

#ifdef USE_Sleep
      Sleep (nsecs / 1000000);
#endif

#ifdef USE_MACOS
      {
        ___U32 ticks = nsecs / 16666666;
        if (has_IdleUpdate)
          IdleUpdate ();
        if (has_WaitNextEvent)
          {
            EventRecord er;
            WaitNextEvent (0, &er, ticks, NULL);
          }
        else if (has_Delay)
          Delay (ticks, &ticks);
      }
#endif
    }
}

#endif


#ifdef USE_select

___HIDDEN void absolute_time_to_nonnegative_timeval
   ___P((___time tim,
         struct timeval **tv),
        (tim,
         tv)
___time tim;
struct timeval **tv;)
{
  if (___time_equal (tim, ___time_pos_infinity))
    *tv = NULL;
  else
    {
      struct timeval *t = *tv;

      if (___time_positive (tim))
        {
#ifdef ___FLOAT_TIME_REPRESENTATION
          if (tim >= 2147483648.0) /* upper bound is 68 years! */
            {
              t->tv_sec = 2147483647;
              t->tv_usec = 999999;
            }
          else
            {
              t->tv_sec = tim;
              t->tv_usec = (tim - t->tv_sec) * 1000000.0;
            }
#endif

#ifdef ___INT_TIME_REPRESENTATION
          t->tv_sec = tim.secs;
          t->tv_usec = tim.nsecs / 1000;
#endif
        }
      else
        {
          t->tv_sec = 0;
          t->tv_usec = 0;
        }
    }
}

#endif


#ifdef USE_MsgWaitForMultipleObjects

/*************** obsoleted by set_heartbeat_interval */

___HIDDEN void absolute_time_to_nonnegative_msecs
   ___P((___time tim,
         DWORD *msecs),
        (tim,
         msecs)
___time tim;
DWORD *msecs;)
{
  if (___time_equal (tim, ___time_pos_infinity))
    *msecs = INFINITE;
  else
    {
      if (___time_positive (tim))
        {

*********** make sure INFINITE != 2147483647

#ifdef ___FLOAT_TIME_REPRESENTATION
          if (tim >= 2147483.647)
            *msecs = 2147483647; /* upper bound is 24.8 days! */
          else
            *msecs = ___CAST(int,tim*1000.0);
#endif

#ifdef ___INT_TIME_REPRESENTATION
          if (tim.secs > 2147483 ||
              (tim.secs == 2147483 && tim.nsecs >= 647000000))
            *msecs = 2147483647; /* upper bound is 24.8 days! */
          else
            *msecs = tim.secs * 1000 + tim.nsecs / 1000000;
#endif
        }
      else
        *msecs = 0;
    }
}

#endif


/*---------------------------------------------------------------------------*/

/* Process times (user time, system time and real time). */

___HIDDEN ___F64 process_start_seconds; /* process start in real time */

void ___process_times
   ___P((___F64 *user,
         ___F64 *sys,
         ___F64 *real),
        (user,
         sys,
         real)
___F64 *user;
___F64 *sys;
___F64 *real;)
{

#ifndef USE_times
#ifndef USE_clock
#ifndef USE_getrusage
#ifndef USE_GetProcessTime
#ifndef USE_TickCount
#ifndef USE_DosQuerySysInfo

  *user = 0.0; /* can't get time... result is 0 */
  *sys = 0.0;

#endif
#endif
#endif
#endif
#endif
#endif

#ifdef USE_times

#ifdef CLK_TCK
#define CLOCK_TICKS_PER_SEC CLK_TCK
#else
#ifdef _SC_CLOCKS_PER_SEC
#define CLOCK_TICKS_PER_SEC sysconf (_SC_CLOCKS_PER_SEC)
#else
#ifdef _SC_CLK_TCK
#define CLOCK_TICKS_PER_SEC sysconf (_SC_CLK_TCK)
#else
@error Cannot find a definition for CLOCK_TICKS_PER_SEC
#endif
#endif
#endif

  struct tms t;

  times (&t);

  *user = ___CAST(___F64,t.tms_utime) / CLOCK_TICKS_PER_SEC;
  *sys = ___CAST(___F64,t.tms_stime) / CLOCK_TICKS_PER_SEC;

#endif

#ifdef USE_clock

  *user = ___CAST(___F64,clock ()) / CLOCKS_PER_SEC;
  *sys = 0.0; /* fake system time */

#endif

#ifdef USE_getrusage

  struct rusage ru;

  if (getrusage (RUSAGE_SELF, &ru) == 0)
    {
      *user = ru.ru_utime.tv_sec + ___CAST(___F64,ru.ru_utime.tv_usec) / 1.0e6;
      *sys = ru.ru_stime.tv_sec + ___CAST(___F64,ru.ru_stime.tv_usec) / 1.0e6;
    }
  else
    {
      *user = 0.0; /* can't get time... result is 0 */
      *sys = 0.0;
    }

#endif

#ifdef USE_GetProcessTime

  HANDLE p;
  FILETIME creation_time, exit_time, sys_time, user_time;

  p = GetCurrentProcess ();

  if (GetProcessTimes (p, &creation_time, &exit_time, &sys_time, &user_time))
    {
      *user = (user_time.dwHighDateTime * 4294967296.0 +
               user_time.dwLowDateTime) / 1.0e7;
      *sys = (sys_time.dwHighDateTime * 4294967296.0 +
              sys_time.dwLowDateTime) / 1.0e7;
    }
  else
    {
      *user = 0.0; /* can't get time... result is 0 */
      *sys = 0.0;
    }

#endif

#ifdef USE_TickCount

  static ___U32 origin = 0;
  ___U32 now;

  while (origin == 0)
    origin = TickCount ();

  now = TickCount ();

  *user = ___CAST(___F64,(now-origin)) / 60;
  *sys = 0.0; /* fake system time */

#endif

#ifdef USE_DosQuerySysInfo

  static ___U32 origin = 0;
  ___U32 now;

  while (origin == 0)
    DosQuerySysInfo (QSV_MS_COUNT, QSV_MS_COUNT, &origin, sizeof (origin));

  DosQuerySysInfo (QSV_MS_COUNT, QSV_MS_COUNT, &now, sizeof (now));

  *user = ___CAST(___F64,(now-origin)) / 1.0e3; /* are the units right? */
  *sys = 0.0; /* fake system time */

#endif

  {
    ___time now;
    ___F64 seconds;

    ___time_get_current_time (&now);
    ___time_to_seconds (now, &seconds);

    *real = seconds - process_start_seconds;
  }
}


___HIDDEN void setup_process_times ___PVOID
{
  ___F64 user, sys;

  process_start_seconds = 0.0;
  ___process_times (&user, &sys, &process_start_seconds);

}


___HIDDEN void cleanup_process_times ___PVOID
{
}


/*---------------------------------------------------------------------------*/

/* Heartbeat interrupt handling. */

/*
 * Use the process-time timer unless only the real-time timer is
 * available (e.g. DJGPP).  Note that on some systems (e.g. MkLinux)
 * ITIMER_VIRTUAL is an enum type, not a macro.
 */

#ifdef ITIMER_VIRTUAL
#define HEARTBEAT_ITIMER ITIMER_VIRTUAL
#define HEARTBEAT_SIG SIGVTALRM
#else
#ifdef ITIMER_REAL
#define HEARTBEAT_ITIMER ITIMER_REAL
#define HEARTBEAT_SIG SIGALRM
#else
#define HEARTBEAT_ITIMER ITIMER_VIRTUAL
#define HEARTBEAT_SIG SIGVTALRM
#endif
#endif


#ifndef USE_setitimer
#ifndef USE_dos_setvect_1Ch
#ifndef USE_DosStartTimer
#ifndef USE_VInstall
#ifndef USE_CreateThread

___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        (seconds)
___F64 seconds;)
{
  return 0.0;
}


___HIDDEN void disable_heartbeat_interrupt ___PVOID
{
}


___HIDDEN void enable_heartbeat_interrupt ___PVOID
{
}


___HIDDEN void setup_heartbeat_interrupt_handling ___PVOID
{
}


___HIDDEN void cleanup_heartbeat_interrupt_handling ___PVOID
{
}


#endif
#endif
#endif
#endif
#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_setitimer

___HIDDEN void heartbeat_interrupt_handler
   ___P((int sig),
        (sig)
int sig;)
{
#ifdef USE_signal
  signal (HEARTBEAT_SIG, heartbeat_interrupt_handler);
#endif
  ___raise_interrupt (___INTR_HEARTBEAT);
}


___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        (seconds)
___F64 seconds;)
{
  struct itimerval tv;

  if (seconds < 0.0) /* turn heartbeat off */
    {
      tv.it_interval.tv_sec  = 0;
      tv.it_interval.tv_usec = 0;
      tv.it_value.tv_sec     = 0;
      tv.it_value.tv_usec    = 0;
      setitimer (HEARTBEAT_ITIMER, &tv, 0);
      return 0.0;
    }
  else
    {
      int secs;
      int usecs;

      if (seconds >= 2147483648.0) /* upper bound is 68 years! */
        {
          secs = 2147483647;
          usecs = 999999;
        }
      else
        {
          secs = seconds;
          usecs = (seconds - secs) * 1000000.0;
          if (secs <= 0 && usecs == 0) /* use smallest interval */
            {
              secs = 0;
              usecs = 1;
            }
        }

      tv.it_interval.tv_sec  = secs;
      tv.it_interval.tv_usec = usecs;
      tv.it_value.tv_sec     = secs;
      tv.it_value.tv_usec    = usecs;
      setitimer (HEARTBEAT_ITIMER, &tv, 0);
      getitimer (HEARTBEAT_ITIMER, &tv);
      return tv.it_interval.tv_sec + tv.it_interval.tv_usec / 1000000.0;
    }
}


___HIDDEN void disable_heartbeat_interrupt ___PVOID
{
#ifdef USE_sigaction
  struct sigaction act;
  act.sa_handler = SIG_IGN;
  act.sa_flags = 0;
  sigemptyset (&act.sa_mask);
  sigaction (HEARTBEAT_SIG, &act, 0);
#endif

#ifdef USE_signal
  signal (HEARTBEAT_SIG, SIG_IGN);
#endif
}


___HIDDEN void enable_heartbeat_interrupt ___PVOID
{
#ifdef USE_sigaction
  struct sigaction act;
  act.sa_handler = heartbeat_interrupt_handler;
  act.sa_flags = 0;
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;
#endif
  sigemptyset (&act.sa_mask);
  sigaction (HEARTBEAT_SIG, &act, 0);
#endif

#ifdef USE_signal
  signal (HEARTBEAT_SIG, heartbeat_interrupt_handler);
#endif
}


___HIDDEN void setup_heartbeat_interrupt_handling ___PVOID
{
  ___set_heartbeat_interval (-1.0);
  enable_heartbeat_interrupt ();
}


___HIDDEN void cleanup_heartbeat_interrupt_handling ___PVOID
{
  ___set_heartbeat_interval (-1.0);
  disable_heartbeat_interrupt ();
}


#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_dos_setvect_1Ch

/* 1Ch = timer tick vector */
___HIDDEN ___BOOL heartbeat_enabled;
___HIDDEN void (__interrupt __far *prev_vector_1Ch) ___PVOID;
___HIDDEN long heartbeat_interval;
___HIDDEN long heartbeat_countdown;


___HIDDEN void __interrupt __far heartbeat_interrupt_handler ___PVOID
{
  if (heartbeat_countdown > 1)
    heartbeat_countdown--;
  else if (heartbeat_countdown == 1 && heartbeat_enabled)
    {
      heartbeat_countdown = heartbeat_interval;
      ___raise_interrupt (___INTR_HEARTBEAT);
    }
  _chain_intr (prev_vector_1Ch);
}


___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        (seconds)
___F64 seconds;)
{
  if (seconds < 0.0) /* turn heartbeat off */
    {
      _dos_setvect (0x1C, prev_vector_1Ch);
      return 0.0;
    }
  else
    {
      ___F64 ticks = seconds * 18.2; /* 18.2 ticks/sec */
      int t;

      if (ticks >= 2147483648.0) /* upper bound is 3.7 years! */
        t = 2147483647;
      else
        {
          t = ticks;
          if (t <= 0) /* use smallest interval */
            t = 1;
        }

      heartbeat_interval = t;
      heartbeat_countdown = heartbeat_interval;
      _dos_setvect (0x1C, heartbeat_interrupt_handler);
      return t / 18.2;
    }
}


___HIDDEN void disable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 0;
}


___HIDDEN void enable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 1;
}


___HIDDEN void setup_heartbeat_interrupt_handling ___PVOID
{
  prev_vector_1Ch = _dos_getvect (0x1C);
  heartbeat_enabled = 1;
  ___set_heartbeat_interval (-1.0);
}


___HIDDEN void cleanup_heartbeat_interrupt_handling ___PVOID
{
  ___set_heartbeat_interval (-1.0);
  heartbeat_enabled = 0;
}


#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_DosStartTimer

___HIDDEN ___BOOL heartbeat_enabled;
___HIDDEN ___BOOL heartbeat_hev;
___HIDDEN ___BOOL heartbeat_tid;
___HIDDEN ___BOOL heartbeat_htimer;
___HIDDEN HEV    hev;
___HIDDEN TID    tid;
___HIDDEN HTIMER htimer;


___HIDDEN void APIENTRY heartbeat_thread_code
   ___P((ULONG param),
        (param)
ULONG param;)
{
  ULONG pc = 0;
  while (1)
    {
      DosWaitEventSem (hev, SEM_INDEFINITE_WAIT);
      DosResetEventSem (hev, &pc);
      if (heartbeat_enabled)
        {
          DosEnterCritSec ();
          ___raise_interrupt (___INTR_HEARTBEAT);
          DosExitCritSec ();
        }
    }
}


___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        (seconds)
___F64 seconds;)
{
  if (!heartbeat_hev || !heartbeat_tid)
    return 0.0;
  if (heartbeat_htimer)
    {
      DosStopTimer (htimer);
      heartbeat_htimer = 0;
    }
  if (seconds < 0.0) /* turn heartbeat off */
    return 0.0;
  else
    {
      ___F64 msecs = seconds * 1000.0;
      ULONG m;

      if (msecs >= 4294967295.0) /* upper bound is 49.7 days! */
        m = 4294967295;
      else
        {
          m = msecs;
          if (m == 0) /* use smallest interval */
            m = 1;
        }

      if (DosStartTimer (m, ___CAST(HSEM,hev), &htimer) == NO_ERROR)
        {
          heartbeat_htimer = 1;
          return m / 1000.0;
        }
      else
        return 0.0;
    }
}


___HIDDEN void disable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 0;
}


___HIDDEN void enable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 1;
}


___HIDDEN void setup_heartbeat_interrupt_handling ___PVOID
{
  heartbeat_enabled = 0;
  heartbeat_hev     = 0;
  heartbeat_tid     = 0;
  heartbeat_htimer  = 0;
  if (DosCreateEventSem (0, &hev, DC_SEM_SHARED, FALSE) != NO_ERROR)
    return;
  heartbeat_hev = 1;
  if (DosCreateThread (&tid,
                       heartbeat_thread_code,
                       0,
                       CREATE_READY|STACK_COMMITTED,
                       4096)
      != NO_ERROR)
    return;
  heartbeat_tid = 1;
  DosSetPriority (PRTYS_THREAD, PRTYC_NOCHANGE, PRTYD_MAXIMUM, tid);
  heartbeat_enabled = 1;
}


___HIDDEN void cleanup_heartbeat_interrupt_handling ___PVOID
{
  heartbeat_enabled = 0;
  if (heartbeat_htimer)
    {
      DosStopTimer (htimer);
      heartbeat_htimer = 0;
    }
  if (heartbeat_tid)
    {
      DosKillThread (tid);
      heartbeat_tid = 0;
    }
  if (heartbeat_hev)
    {
      DosCloseEventSem (hev);
      heartbeat_hev = 0;
    }
}


#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_VInstall

___HIDDEN ___BOOL heartbeat_task_installed;
___HIDDEN short heartbeat_task_ticks;
___HIDDEN VBLTask heartbeat_task;
___HIDDEN ___BOOL heartbeat_enabled;


___HIDDEN void heartbeat_task_code ___PVOID
{
  if (heartbeat_enabled)
    ___raise_interrupt (___INTR_HEARTBEAT);
  heartbeat_task.vblCount = heartbeat_task_ticks;
}


#ifdef __POWERPC__

#define heartbeat_task_proc heartbeat_task_code

#else

___HIDDEN asm void heartbeat_task_proc ___PVOID
{
    move.l a5,-(sp)
    move.l 0x904,a5
    jsr    heartbeat_task_code
    move.l (sp)+,a5
    rts
}

#endif


___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        (seconds)
___F64 seconds;)
{
  if (heartbeat_task_installed)
    if (VRemove (___CAST(QElemPtr,&heartbeat_task)) != noErr)
      {
        heartbeat_task_installed = 0;
        return 0.0;
      }

  if (seconds < 0.0) /* turn heartbeat off */
    return 0.0;
  else
    {
      ___F64 ticks = seconds * 60.0; /* 60 ticks/sec */
      short t;

      if (ticks >= 32767.0) /* upper bound is 9 minutes! */
        t = 32767;
      else
        {
          t = ticks;
          if (t <= 0) /* use smallest interval */
            t = 1;
        }

      heartbeat_task_ticks    = t;
      heartbeat_task.qType    = vType;
      heartbeat_task.vblAddr  = NewVBLProc (heartbeat_task_proc);
      heartbeat_task.vblCount = t;
      heartbeat_task.vblPhase = 0;
      if (VInstall (___CAST(QElemPtr,&heartbeat_task)) == noErr)
        {
          heartbeat_task_installed = 1;
          return t / 60.0;
        }
      else
        return 0.0;
    }
}


___HIDDEN void disable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 0;
}


___HIDDEN void enable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 1;
}


___HIDDEN void setup_heartbeat_interrupt_handling ___PVOID
{
  heartbeat_enabled = 1;
  heartbeat_task_installed = 0;
}


___HIDDEN void cleanup_heartbeat_interrupt_handling ___PVOID
{
  ___set_heartbeat_interval (-1.0);
  heartbeat_enabled = 0;
}


#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_CreateThread

/*
 * The HEARTBEAT_PRIORITY should be high so that heartbeats will occur
 * as close as possible to the correct time.
 */

#define HEARTBEAT_PRIORITY THREAD_PRIORITY_HIGHEST

___HIDDEN HANDLE heartbeat_thread;
___HIDDEN HANDLE heartbeat_update; /* heartbeat needs updating when signaled */
___HIDDEN DWORD heartbeat_interval;
___HIDDEN ___BOOL heartbeat_enabled;


___HIDDEN DWORD WINAPI heartbeat_thread_proc
   ___P((LPVOID param),
        (param)
LPVOID param;)
{
  DWORD interval = heartbeat_interval;

  while (1)
    {
      DWORD n;

      if (interval < 0)
        n = WaitForSingleObject (heartbeat_update, INFINITE);
      else
        n = WaitForSingleObject (heartbeat_update, interval);

      interval = heartbeat_interval;

      if (interval == 0)
        return 0;

      switch (n)
        {
        case WAIT_ABANDONED:
        case WAIT_OBJECT_0:
          break;

        case WAIT_TIMEOUT:
          if (heartbeat_enabled)
            ___raise_interrupt (___INTR_HEARTBEAT);
          break;

        default: /* WAIT_FAILED */
          return 0;
        }
    }

  return 0;
}


___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        (seconds)
___F64 seconds;)
{
  if (seconds < 0.0) /* turn heartbeat off */
    {
      heartbeat_interval = -1;
      SetEvent (heartbeat_update); /* ignore error */
      return 0.0;
    }
  else
    {
      ___F64 msecs = seconds * 1000.0;
      DWORD m;

      if (msecs >= 2147483648.0) /* upper bound is 24.8 days! */
        m = 2147483647; /************* make sure INFINITE != 2147483647 */
      else
        {
          m = msecs;
          if (m == 0) /* use smallest interval */
            m = 1;
        }

      heartbeat_interval = m;
      SetEvent (heartbeat_update); /* ignore error */
      return m / 1000.0;
    }
}


___HIDDEN void disable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 0;
}


___HIDDEN void enable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 1;
}


___HIDDEN void setup_heartbeat_interrupt_handling ___PVOID
{
  DWORD thread_id;

  heartbeat_enabled = 1;
  heartbeat_interval = -1;

  heartbeat_update =
    CreateEvent (NULL,  /* can't inherit */
                 FALSE, /* auto reset */
                 FALSE, /* not signaled */
                 NULL); /* no name */

  if (heartbeat_update == NULL)
    {
      heartbeat_thread = NULL;
      return;
    }

  heartbeat_thread =
    CreateThread (NULL,  /* no security attributes        */
                  65536, /* committed stack size          */
                  heartbeat_thread_proc,
                  0,     /* argument to thread function   */
                  0,     /* use default creation flags    */
                  &thread_id);

  if (heartbeat_thread == NULL)
    {
      CloseHandle (heartbeat_update); /* ignore error */
      heartbeat_update = NULL;
      heartbeat_thread = NULL;
      return;
    }

  SetThreadPriority (heartbeat_thread, HEARTBEAT_PRIORITY); /* ignore error */

  return;
}


___HIDDEN void cleanup_heartbeat_interrupt_handling ___PVOID
{
  if (heartbeat_thread != NULL)
    {
      heartbeat_enabled = 0;
      heartbeat_interval = 0;
      SetEvent (heartbeat_update); /* ignore error */
      WaitForSingleObject (heartbeat_thread, INFINITE); /* ignore error */
      CloseHandle (heartbeat_update); /* ignore error */
      heartbeat_update = NULL;
      heartbeat_thread = NULL;
    }
}


#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* User interrupt handling. */

/***************** osoleted by console device? */

#ifndef USE_sigaction_SIGINT
#ifndef USE_signal_SIGINT
#ifndef USE_dos_setvect_23h

___HIDDEN void setup_user_interrupt_handling ___PVOID
{
}

___HIDDEN void cleanup_user_interrupt_handling ___PVOID
{
}

___HIDDEN void disable_user_interrupt ___PVOID
{
}

___HIDDEN void enable_user_interrupt ___PVOID
{
}

#endif
#endif
#endif

/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_sigaction_SIGINT

___HIDDEN void user_interrupt_handler
   ___P((int sig),
        (sig)
int sig;)
{
  ___raise_interrupt (___INTR_USER);
}

___HIDDEN void setup_user_interrupt_handling ___PVOID
{
  struct sigaction act;

  act.sa_handler = user_interrupt_handler;
  act.sa_flags = 0;
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;
#endif
  sigemptyset (&act.sa_mask);
  sigaction (SIGINT, &act, 0);
}

___HIDDEN void cleanup_user_interrupt_handling ___PVOID
{
  struct sigaction act;

  act.sa_handler = SIG_DFL;
  act.sa_flags = 0;
  sigemptyset (&act.sa_mask);
  sigaction (SIGINT, &act, 0);
}

___HIDDEN void disable_user_interrupt ___PVOID
{
  sigset_t mask;

  sigemptyset (&mask);
  sigaddset (&mask, SIGINT);
  sigprocmask (SIG_BLOCK, &mask, 0);
}

___HIDDEN void enable_user_interrupt ___PVOID
{
  sigset_t mask;

  sigemptyset (&mask);
  sigaddset (&mask, SIGINT);
  sigprocmask (SIG_UNBLOCK, &mask, 0);
}

#endif

/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_signal_SIGINT

___HIDDEN void user_interrupt_handler
   ___P((int sig),
        (sig)
int sig;)
{
  signal (SIGINT, user_interrupt_handler);
  ___raise_interrupt (___INTR_USER);
}

___HIDDEN void setup_user_interrupt_handling ___PVOID
{
  signal (SIGINT, user_interrupt_handler);
}

___HIDDEN void cleanup_user_interrupt_handling ___PVOID
{
  signal (SIGINT, SIG_DFL);
}

___HIDDEN void disable_user_interrupt ___PVOID
{
  signal (SIGINT, SIG_IGN);
}

___HIDDEN void enable_user_interrupt ___PVOID
{
  signal (SIGINT, user_interrupt_handler);
}

#endif

/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_dos_setvect_23h

/* 23h = ctrl-c vector */
___HIDDEN void (__interrupt __far *prev_vector_23h) ___PVOID;

/* 1Bh = ctrl-break vector */
___HIDDEN void (__interrupt __far *prev_vector_1Bh) ___PVOID;

___HIDDEN void __interrupt __far user_interrupt_handler ___PVOID
{
  ___raise_interrupt (___INTR_USER);
}

___HIDDEN void setup_user_interrupt_handling ___PVOID
{
  prev_vector_23h = _dos_getvect (0x23);
  prev_vector_1Bh = _dos_getvect (0x1B);
  _dos_setvect (0x23, user_interrupt_handler);
  _dos_setvect (0x1B, user_interrupt_handler);
}

___HIDDEN void cleanup_user_interrupt_handling ___PVOID
{
  _dos_setvect (0x23, prev_vector_23h);
  _dos_setvect (0x1B, prev_vector_1Bh);
}

___HIDDEN void disable_user_interrupt ___PVOID
{
}

___HIDDEN void enable_user_interrupt ___PVOID
{
}

#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

___HIDDEN void setup_interrupt_handling ___PVOID
{
  setup_user_interrupt_handling ();
  setup_heartbeat_interrupt_handling ();
}

___HIDDEN void cleanup_interrupt_handling ___PVOID
{
  cleanup_heartbeat_interrupt_handling ();
  cleanup_user_interrupt_handling ();
}

___HIDDEN void disable_interrupts ___PVOID
{
  disable_heartbeat_interrupt ();
  disable_user_interrupt ();
}

___HIDDEN void enable_interrupts ___PVOID
{
  enable_user_interrupt ();
  enable_heartbeat_interrupt ();
}


/*---------------------------------------------------------------------------*/

#if 0
/* OS event handling. */

/******************obsolete?*/

___WORD ___os_event_get ___PVOID
{
  ___WORD event;

  if (___setup_params.event_get != 0)
    return ___setup_params.event_get ();

  event = ___FAL;

#ifdef BUFFERED_STDOUT

  fflush (stdout);

#endif

#ifdef USE_MACOS

#ifdef USE_mac_gui

  {
    EventRecord er;

    if (mac_gui_get_next_event (&er))
      {
        event = ___alloc_scmobj (___sU8VECTOR, sizeof(er), ___STILL);
        if (event != ___FAL)
          {
            *(EventRecord*)___BODY_AS(event,___tSUBTYPED) = er;
            ___still_obj_refcount_dec (event);
          }
      }
  }

#else

  {
    EventRecord er;
    RgnHandle cursor_rgn;

    cursor_rgn = NewRgn ();
    if (has_IdleUpdate)
      IdleUpdate ();
    if (has_WaitNextEvent ?
        WaitNextEvent (everyEvent, &er, 0, cursor_rgn) :
        GetNextEvent (everyEvent, &er))
      {
        event = ___alloc_scmobj (___sU8VECTOR, sizeof(er), ___STILL);
        if (event != ___FAL)
          {
            *(EventRecord*)___BODY_AS(event,___tSUBTYPED) = er;
            ___still_obj_refcount_dec (event);
          }
      }
    DisposeRgn (cursor_rgn);
  }

#endif

#endif

#ifdef __WATCOMC__
#ifdef __SW_BW

  _dwYield ();

#endif
#endif

#ifdef USE_autocad_gui

#if 0
  /***************************************/
  if (acedCommand (0) == RTCAN)
    ___raise_interrupt (___INTR_USER);

  acedUpdateDisplay ();
#endif

#if 1
  if (acedUsrBrk ())
    ___raise_interrupt (___INTR_USER);
#endif

#endif

  return event;
}

___WORD ___os_event_handler ___P((___WORD event),(event)
___WORD event;)
{
  if (___setup_params.event_handler != 0)
    return ___setup_params.event_handler (event);

#ifdef USE_MACOS

#ifdef USE_mac_gui

  if (mac_gui_handle_event ((EventRecord*)___BODY_AS(event,___tSUBTYPED), NULL))
    if (mac_gui_quit ())
      ___exit_process (1);

#else

  SIOUXHandleOneEvent ((EventRecord*)___BODY_AS(event,___tSUBTYPED));

#endif

#endif

  return ___TRU;
}

#endif


/*---------------------------------------------------------------------------*/

/* Access to environment variables. */

char *___getenv
   ___P((char *name),
        (name)
char *name;)
{
  return getenv (name);
}


/*---------------------------------------------------------------------------*/

/* Shell command. */

int ___shell_command
   ___P((char *cmd),
        (cmd)
char *cmd;)
{
  int result;

  disable_interrupts ();
  result = system (cmd);
  enable_interrupts ();

  return result;
}


/*---------------------------------------------------------------------------*/

/* Dynamic loading. */

#ifdef USE_dlopen
#define DL_DESCR void *
#endif

#ifdef USE_shl_load
#define DL_DESCR shl_t
#endif

#ifdef USE_LoadLibrary
#define DL_DESCR HANDLE
#define DL_ERROR_BUF_SIZE 256
___HIDDEN char dl_error_buffer[DL_ERROR_BUF_SIZE];
#endif

#ifdef USE_DosLoadModule
#define DL_DESCR HMODULE
#define DL_ERROR_BUF_SIZE 256
___HIDDEN char dl_error_buffer[DL_ERROR_BUF_SIZE];
#endif

#ifdef USE_dxe_load
#define DL_DESCR void *
#endif

#ifdef USE_GetDiskFragment
#define DL_DESCR CFragConnectionID
#define DL_ERROR_BUF_SIZE 256
___HIDDEN char dl_error_buffer[DL_ERROR_BUF_SIZE];
#endif


#ifdef DL_DESCR

typedef struct dl_entry
  {
    struct dl_entry *next;
    DL_DESCR descr;
  } dl_entry;

___HIDDEN dl_entry *dl_list;

#endif


___HIDDEN void setup_dynamic_load ___PVOID
{
#ifdef DL_DESCR

  dl_list = 0;

#endif
}


void *___dynamic_load
   ___P((char *path,
         char *name,
         char **errmsg),
        (path,
         name,
         errmsg)
char *path;
char *name;
char **errmsg;)
{
#ifndef DL_DESCR

  *errmsg = "Dynamic loading is not available on this platform";
  return 0;

#else

  void *result = 0;
  dl_entry *p;

  p = ___CAST(dl_entry*,___alloc_mem (sizeof (dl_entry)));
  if (p == 0)
    {
      *errmsg = "Can't allocate dynamic loader entry";
      return 0;
    }

#ifdef USE_dlopen

#ifdef RTLD_NOW
  p->descr = dlopen (path, RTLD_NOW);
#else
  p->descr = dlopen (path, 1);
#endif

  if (p->descr != 0)
    if ((result = dlsym (p->descr, name)) == 0)
      dlclose (p->descr);

  if (result == 0)
    *errmsg = ___CAST(char*,dlerror ());

#endif

#ifdef USE_shl_load

  p->descr = shl_load (path, BIND_IMMEDIATE, 0);
  if (p->descr == 0)
    *errmsg = strerror (errno);
  else
    if (shl_findsym (&p->descr, name, TYPE_PROCEDURE, &result))
      {
        shl_unload (p->descr);
        result = 0;
        *errmsg = "Can't find symbol";
      }

#endif

#ifdef USE_LoadLibrary

  p->descr = LoadLibrary (path);
  if (p->descr != 0)
    if ((result = GetProcAddress (p->descr, name)) == 0)
      FreeLibrary (p->descr);

  if (result == 0)
    {
      FormatMessage 
        (FORMAT_MESSAGE_FROM_SYSTEM+FORMAT_MESSAGE_MAX_WIDTH_MASK,
         0,
         GetLastError (),
         MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
         dl_error_buffer,
         DL_ERROR_BUF_SIZE,
         0);
      *errmsg = dl_error_buffer;
    }

#endif

#ifdef USE_DosLoadModule

  {
    HMODULE hmodule;
    APIRET rc;
    char real_path[280];
    int i;

    i = 0;
    while (path[i] != '\0')
      i++;

    if (i <= 4 ||
        path[i-1] != 'l' && path[i-1] != 'L' ||
        path[i-2] != 'l' && path[i-2] != 'L' ||
        path[i-3] != 'd' && path[i-3] != 'D' ||
        path[i-4] != '.')
      {
        /* if path doesn't end in ".dll" we must find the real ".dll" path */

        FILE *f = fopen (path, "r");
        if (f == 0)
          {
            *errmsg = "Invalid Name";
            return 0;
          }
        if (!fgets (real_path, 280, f))
          {
            *errmsg = "Invalid Redirection";
            fclose (f);
            return 0;
          }
        fclose (f);
        path = real_path;
      }

    if (DosLoadModule (dl_error_buffer, DL_ERROR_BUF_SIZE, path, &hmodule)
        != NO_ERROR)
      *errmsg = dl_error_buffer;
    else
      {
        p->descr = hmodule;
        rc = DosQueryProcAddr (hmodule, 0L, name, &result);
        if (rc != NO_ERROR)
          {
            switch (rc)
              {
              case ERROR_INVALID_HANDLE:
                *errmsg = "Invalid Handle";
                break;
              case ERROR_INVALID_NAME:
                *errmsg = "Invalid Name";
                break;
              case ERROR_INVALID_ORDINAL:
                *errmsg = "Invalid Ordinal";
                break;
              case ERROR_ENTRY_IS_CALLGATE:
                *errmsg = "Entry is callgate";
                break;
              default:
                *errmsg = "Unknown error";
                break;
              }
            DosFreeModule (hmodule);
          }
      }
  }

#endif

#ifdef USE_dxe_load

  result = _dxe_load (path);
  p->descr = result;

  if (result == 0)
    *errmsg = "Unknown error";

#endif

#ifdef USE_GetDiskFragment

  {
    OSErr err;
    Ptr mainadr, procadr;
    Str63 ppath;
    Str255 pname;
    Str255 pmsg;
    char msg[256];
    FSSpec spec;

    if (!c2pascal (path, ppath, sizeof(ppath)-1))
      *errmsg = "Path is too long";
    else if (!c2pascal (name, pname, sizeof(pname)-1))
      *errmsg = "Module name is too long";
    else if (make_ResolvedFSSpec (0, 0, ppath, &spec) != noErr)
      *errmsg = "Invalid path";
    else
      {
        err = GetDiskFragment (&spec, 0, kCFragGoesToEOF, ppath,
                               kPrivateCFragCopy, &p->descr, &mainadr, pmsg);
        if (err != noErr)
          {
            pascal2c (pmsg, msg, 255);
            sprintf (dl_error_buffer,
                     "GetDiskFragment failed with error code %d on \"%s\"",
                     err,
                     msg);
            *errmsg = dl_error_buffer;
          }
        else
          {
            if (FindSymbol (p->descr, pname, &procadr, kCodeCFragSymbol)
                != noErr)
              {
                *errmsg = "FindSymbol failed";
                CloseConnection (&p->descr);
              }
            else
              result = ___CAST(void*,procadr);
          }
      }
  }

#endif

  if (result == 0)
    ___free_mem (p);
  else
    {
      p->next = dl_list;
      dl_list = p;
    }

  return result;

#endif
}


___HIDDEN void cleanup_dynamic_load ___PVOID
{
#ifdef DL_DESCR

  dl_entry *p = dl_list;
  while (p != 0)
    {
      dl_entry *next = p->next;

#ifdef USE_dlopen
      dlclose (p->descr);
#endif

#ifdef USE_shl_load
      shl_unload (p->descr);
#endif

#ifdef USE_LoadLibrary
      FreeLibrary (p->descr);
#endif

#ifdef USE_DosLoadModule
      DosFreeModule (p->descr);
#endif

#ifdef USE_dxe_load
#endif

#ifdef USE_GetDiskFragment
      CloseConnection (&p->descr);
#endif

      ___free_mem (p);
      p = next;
    }
  dl_list = 0;

#endif  
}


/*---------------------------------------------------------------------------*/

/* Dynamic C compilation. */

#ifdef USE_CWIDE

/*
 * On the Macintosh, a C module can be dynamically compiled by
 * sending AppleEvents to the CodeWarrior IDE.  This code has been
 * tested with CWPro 1 on a PowerPC based Macintosh.
 */

#define CWIDE_EVENT_CLASS 'MMPR'
#define CWIDE_SIGNATURE 'CWIE'

#define CWIDE_no_result     0
#define CWIDE_list_of_short 1
#define CWIDE_list_of_ErrM  2

enum
  {
    errShell_ActionFailed  = 1,
    errShell_FileNotFound  = 2,
    errShell_DuplicateFile = 3,
    errShell_CompileError  = 4,
    errShell_MakeFailed    = 5,
    errShell_NoOpenProject = 6,
    errShell_WindowNotOpen = 7
  };


___HIDDEN AEAddressDesc CWIDE_address;


___HIDDEN OSErr setup_CWIDE ___PVOID
{
  long signature = CWIDE_SIGNATURE;

  CWIDE_address.dataHandle = nil;

  return AECreateDesc (typeApplSignature,
                       ___CAST(Ptr,&signature),
                       sizeof (long),
                       &CWIDE_address);
}


___HIDDEN OSErr cleanup_CWIDE ___PVOID
{
  if (CWIDE_address.dataHandle != nil)
    return AEDisposeDesc (&CWIDE_address);
  else
    return noErr;
}


___HIDDEN OSErr CWIDE_send_ae
   ___P((AppleEvent *ae,
         int res_kind,
         long *CWIDE_err),
        (ae,
         res_kind,
         CWIDE_err)
AppleEvent ae;
int res_kind;
long *CWIDE_err;)
{
  OSErr err;
  AppleEvent reply_ae;

  reply_ae.dataHandle = nil;

  err = AESend (ae,
                &reply_ae,
                kAEWaitReply,
                kAENormalPriority,
                kNoTimeOut,
                nil,
                nil);

  if (err == noErr && res_kind != CWIDE_no_result)
    switch (res_kind)
      {
        case CWIDE_list_of_short:
        case CWIDE_list_of_ErrM:
          {
            AEDescList list;
            list.dataHandle = nil;
            err = AEGetParamDesc (&reply_ae, keyAEResult, typeAEList, &list);
            if (err == noErr)
              if (res_kind == CWIDE_list_of_short)
                {
                  short error_code;
                  AEKeyword key;
                  DescType typ;
                  Size siz;
                  err = AEGetNthPtr (&list,
                                     1,
                                     typeShortInteger,
                                     &key,
                                     &typ,
                                     ___CAST(Ptr,&error_code),
                                     sizeof (short),
                                     &siz);
                  if (err == noErr)
                    *CWIDE_err = error_code;
                }
              else
                {
                  long nb_errors;
                  err = AECountItems (&list, &nb_errors);
                  if (err == noErr)
                    *CWIDE_err = nb_errors;
                }
            if (list.dataHandle != nil)
              AEDisposeDesc (&list);
            break;
          }
      }

  if (reply_ae.dataHandle != nil)
    AEDisposeDesc (&reply_ae);

  return err;
}


___HIDDEN OSErr CWIDE_make_ae
   ___P((AEEventClass clas,
         AEEventID id,
         AppleEvent *ae),
        (clas,
         id,
         ae)
AEEventClass clas;
AEEventID id;
AppleEvent *ae;)
{
  return AECreateAppleEvent (clas,
                             id,
                             &CWIDE_address,
                             kAutoGenerateReturnID,
                             kAnyTransactionID,
                             ae);
}


___HIDDEN OSErr CWIDE_file_param
   ___P((FSSpec *file,
         AEKeyword key,
         AppleEvent *ae),
        (file,
         key,
         ae)
FSSpec *file;
AEKeyword key;
AppleEvent *ae;)
{
  OSErr err;
  AEDescList list;

  list.dataHandle = nil;

  err = AECreateList (nil, 0, false, &list);
  if (err == noErr)
    {
      err = AEPutPtr (&list,
                      1,
                      typeFSS,
                      ___CAST(Ptr,file),
                      sizeof (FSSpec));
      if (err == noErr)
        err = AEPutParamDesc (ae, key, &list);
    }

  if (list.dataHandle != nil)
    AEDisposeDesc (&list);

  return err;
}


___HIDDEN OSErr CWIDE_no_param
   ___P((AEEventID id,
         int res_kind,
         long *CWIDE_err),
        (id,
         res_kind,
         CWIDE_err)
AEEventID id;
int res_kind;
long *CWIDE_err;)
{
  OSErr err;
  AppleEvent ae;

  ae.dataHandle = nil;

  err = CWIDE_make_ae (CWIDE_EVENT_CLASS, id, &ae);
  if (err == noErr)
    err = CWIDE_send_ae (&ae, res_kind, CWIDE_err);

  if (ae.dataHandle != nil)
    AEDisposeDesc (&ae);

  return err;
}


___HIDDEN OSErr CWIDE_open_project
   ___P((FSSpec *file),
        (file)
FSSpec *file;)
{
  long CWIDE_err;
  OSErr err;
  AppleEvent ae;

  ae.dataHandle = nil;

  err = CWIDE_make_ae ('aevt', 'odoc', &ae);
  if (err == noErr)
    {
      err = CWIDE_file_param (file, keyDirectObject, &ae);
      if (err == noErr)
        err = CWIDE_send_ae (&ae, CWIDE_no_result, &CWIDE_err);
    }

  if (ae.dataHandle != nil)
    AEDisposeDesc (&ae);

  return err;
}


___HIDDEN OSErr CWIDE_close_project ___PVOID
{
  long CWIDE_err;
  return CWIDE_no_param ('ClsP', CWIDE_no_result, &CWIDE_err);
}


___HIDDEN OSErr CWIDE_make_project
   ___P((long *CWIDE_err),
        (CWIDE_err)
long *CWIDE_err;)
{
  OSErr err;
  AppleEvent ae;

  ae.dataHandle = nil;

  err = CWIDE_make_ae (CWIDE_EVENT_CLASS, 'Make', &ae);
  if (err == noErr)
    {
      Boolean extern_editor = true;
      err = AEPutParamPtr (&ae,
                           'Errs',
                           typeBoolean,
                           ___CAST(Ptr,&extern_editor),
                           sizeof (Boolean));
      if (err == noErr)
        err = CWIDE_send_ae (&ae, CWIDE_list_of_ErrM, CWIDE_err);
    }

  if (ae.dataHandle != nil)
    AEDisposeDesc (&ae);

  return err;
}


___HIDDEN OSErr CWIDE_add_file
   ___P((FSSpec *file,
         short segment_num,
         long *CWIDE_err),
        (file,
         segment_num,
         CWIDE_err)
FSSpec *file;
short segment_num;
long *CWIDE_err;)
{
  OSErr err;
  AppleEvent ae;

  ae.dataHandle = nil;

  err = CWIDE_make_ae (CWIDE_EVENT_CLASS, 'AddF', &ae);
  if (err == noErr)
    {
      err = CWIDE_file_param (file, keyDirectObject, &ae);
      if (err == noErr)
        {
          err = AEPutParamPtr (&ae,
                               'Segm',
                               typeShortInteger,
                               ___CAST(Ptr,&segment_num),
                               sizeof (short));
          if (err == noErr)
            err = CWIDE_send_ae (&ae, CWIDE_list_of_short, CWIDE_err);
        }
    }

  if (ae.dataHandle != nil)
    AEDisposeDesc (&ae);

  return err;
}


___HIDDEN OSErr CWIDE_set_prefs
   ___P((Str255 name,
         AERecord *rec),
        (name,
         rec)
Str255 name;
AERecord *rec;)
{
  long CWIDE_err;
  OSErr err;
  AppleEvent ae;

  ae.dataHandle = nil;

  err = CWIDE_make_ae (CWIDE_EVENT_CLASS, 'Pref', &ae);
  if (err == noErr)
    {
      err = AEPutParamPtr (&ae,
                           'PNam',
                           typeChar,
                           ___CAST(Ptr,(name+1)),
                           name[0]);
      if (err == noErr)
        {
          err = AEPutParamDesc (&ae, 'PRec', rec);
          if (err == noErr)
            err = CWIDE_send_ae (&ae, CWIDE_no_result, &CWIDE_err);
        }
    }

  if (ae.dataHandle != nil)
    AEDisposeDesc (&ae);

  return err;
}


___HIDDEN OSErr CWIDE_set_output
   ___P((Str255 path),
        (path)
Str255 path;)
{
  OSErr err;
  AERecord rec;
  err = AECreateList (nil, 0, true, &rec);
  if (err == noErr)
    {
      err = AEPutKeyPtr (&rec,
                         'PR02',
                         typeChar,
                         ___CAST(Ptr,(path+1)),
                         path[0]);
      if (err == noErr)
        err = CWIDE_set_prefs ("\pPPC Project", &rec);
    }

  if (rec.dataHandle != nil)
    AEDisposeDesc (&rec);

  return err;
}


___HIDDEN int CWIDE_is_a_regular_file
   ___P((char *path),
        (path)
char *path;)
{
  struct stat sbuf;
  return (stat (path, &sbuf) == 0 && S_ISREG(sbuf.st_mode));
}


___BOOL ___dynamic_cc
   ___P((char *cc_cmd,
         char *ld_cmd,
         char *root_path,
         char *output_path,
         int ld_flip,
         char **errmsg),
        (cc_cmd,
         ld_cmd,
         root_path,
         output_path,
         ld_flip,
         errmsg)
char *cc_cmd;
char *ld_cmd;
char *root_path;
char *output_path;
int ld_flip;
char **errmsg;)
{
  long CWIDE_err;
  OSErr err;
  FSSpec source_spec, project_spec, model_spec;
  char source_path[___PATH_MAX_LENGTH+1];
  char project_path[___PATH_MAX_LENGTH+1];
  char model_path[___PATH_MAX_LENGTH+1];
  Str255 source_ppath;
  Str255 project_ppath;
  Str255 output_ppath;
  Str255 model_ppath;
  char *p;

  /******* we should check that we will not overflow the strings ******/

  sprintf (source_path, "%s.c", root_path);
  sprintf (project_path, "%s.prj", output_path);

  p = output_path; /* strip directory */
  while (*p != '\0')
    {
      if (DIR_SEPARATOR(*p))
        output_path = p+1;
      p++;
    }

  if (!path_expand_to_absolute (cc_cmd, model_path, ___PATH_MAX_LENGTH) ||
      !CWIDE_is_a_regular_file (model_path))
    {
      *errmsg = "The model project file is missing";
      return 1;
    }

  if (!c2pascal (source_path, source_ppath, sizeof (source_ppath)-1) ||
      !c2pascal (project_path, project_ppath, sizeof (project_ppath)-1) ||
      !c2pascal (output_path, output_ppath, sizeof (output_ppath)-1) ||
      !c2pascal (model_path, model_ppath, sizeof (model_ppath)-1) ||
      (err = make_ResolvedFSSpec (0, 0, source_ppath, &source_spec)) != noErr ||
      (err = make_ResolvedFSSpec (0, 0, project_ppath, &project_spec)) != noErr ||
      (err = make_ResolvedFSSpec (0, 0, model_ppath, &model_spec)) != noErr ||
      (err = copy_file (model_spec, project_spec)) != noErr ||
      (err = setup_CWIDE ()) != noErr)
    {
      *errmsg = "The setting up of the CodeWarrior IDE failed";
      return 1;
    }

  if ((err = CWIDE_close_project ()) != noErr ||
      (err = CWIDE_open_project (&project_spec)) != noErr)
    {
      if (err == connectionInvalid)
        *errmsg = "The CodeWarrior IDE is not running";
      else
        *errmsg = "The CodeWarrior IDE failed to open the project file";
    }
  else if ((err = CWIDE_add_file (&source_spec, 0, &CWIDE_err)) != noErr)
    *errmsg = "The CodeWarrior IDE failed to add the source file to the project";
  else if (CWIDE_err == errShell_FileNotFound)
    {
      err = fnfErr;
      *errmsg = "The source file could not be found";
    }
  else if ((err = CWIDE_set_output (output_ppath)) != noErr)
    *errmsg = "The CodeWarrior IDE failed to set the output file";
  else if ((err = CWIDE_make_project (&CWIDE_err)) != noErr)
    *errmsg = "The CodeWarrior IDE failed to make the project";
  else if (CWIDE_err > 0)
    {
      err = fnfErr; /* anything can do */
      *errmsg = "Compilation failed at C level";
    }
  else if ((err = CWIDE_close_project ()) != noErr)
    *errmsg = "The CodeWarrior IDE could not close the project file";
  else
    remove (project_path);

  cleanup_CWIDE ();

  return err != noErr;
}


#else

___BOOL ___dynamic_cc
   ___P((char *cc_cmd,
         char *ld_cmd,
         char *root_path,
         char *output_path,
         int ld_flip,
         char **errmsg),
        (cc_cmd,
         ld_cmd,
         root_path,
         output_path,
         ld_flip,
         errmsg)
char *cc_cmd;
char *ld_cmd;
char *root_path;
char *output_path;
int ld_flip;
char **errmsg;)
{
  char cmd[___CMD_MAX_LENGTH+1];

  /******* we should check that we will not overflow the string ******/

  sprintf (cmd, cc_cmd, root_path, root_path);

  if (___shell_command (cmd) != 0)
    {
      *errmsg = "Compilation failed at C level";
      return 1;
    }

  if (ld_flip)
    sprintf (cmd, ld_cmd, output_path, root_path);
  else
    sprintf (cmd, ld_cmd, root_path, output_path);

  if (___shell_command (cmd) != 0)
    {
      *errmsg = "Linking failed at C level";
      return 1;
    }

  return 0;
}


#endif


/*---------------------------------------------------------------------------*/

/* Floating point environment setup. */

___HIDDEN void setup_fp ___PVOID
{
#ifdef USE_control87

#ifdef __LCC__

#define FP_EXC_MASK _MCW_EM
#define FP_EXC_CW \
(_EM_INVALID+_EM_ZERODIVIDE+_EM_OVERFLOW+_EM_UNDERFLOW+_EM_INEXACT+_EM_DENORMAL)

  _control87 (FP_EXC_CW, FP_EXC_MASK);

#else

#define FP_EXC_MASK MCW_EM
#define FP_EXC_CW \
(EM_INVALID+EM_ZERODIVIDE+EM_OVERFLOW+EM_UNDERFLOW+EM_INEXACT+EM_DENORMAL)

  _control87 (FP_EXC_CW, FP_EXC_MASK);

#endif

#endif

#ifdef USE__FPU_SETCW

#define FPU_CW \
(_FPU_MASK_IM+_FPU_MASK_ZM+_FPU_MASK_OM+_FPU_MASK_UM+_FPU_MASK_PM+ \
_FPU_MASK_DM+_FPU_DOUBLE+_FPU_RC_NEAREST)

  fpu_control_t cw = FPU_CW;
  _FPU_SETCW (cw);

#endif
}


___HIDDEN void cleanup_fp ___PVOID
{
}


/*---------------------------------------------------------------------------*/

/* Virtual memory statistics. */

void ___vm_stats
   ___P((long *minflt,
         long *majflt),
        (minflt,
         majflt)
long *minflt;
long *majflt;)
{
#ifndef USE_getrusage

  *minflt = 0; /* can't get statistics... result is 0 */
  *majflt = 0;

#endif

#ifdef USE_getrusage

  struct rusage ru;

  if (getrusage (RUSAGE_SELF, &ru) == 0)
    {
      *minflt = ru.ru_minflt;
      *majflt = ru.ru_majflt;
    }
  else
    {
      *minflt = 0; /* can't get statistics... result is 0 */
      *majflt = 0;
    }

#endif
}


/*---------------------------------------------------------------------------*/

/* Filesystem path expansion. */

#ifndef USE_MACOS

#ifdef USE_DOS_PATHS

#ifdef USE_DOS
#define DIR_SEPARATOR1 '\\'
#define DIR_SEPARATOR2 '/'
#define PARENT_HOP "..\\"
#else
#define DIR_SEPARATOR1 '/'
#define DIR_SEPARATOR2 '\\'
#define PARENT_HOP "../"
#endif
#define DRIVE_SEPARATOR ':'

#define DIR_SEPARATOR(c)((c) == DIR_SEPARATOR1 || (c) == DIR_SEPARATOR2)
#define SEPARATOR(c)(DIR_SEPARATOR(c) || (c) == DRIVE_SEPARATOR)

#else

#define DIR_SEPARATOR1 '/'
#define PARENT_HOP "../"

#define DIR_SEPARATOR(c)((c) == DIR_SEPARATOR1)
#define SEPARATOR(c)DIR_SEPARATOR(c)

#endif


#ifdef USE_DOS

___HIDDEN ___BOOL normalize_dir_path
   ___P((char *dir_path,
         char *norm_dir_path,
         long max_length),
        (dir_path,
         norm_dir_path,
         max_length)
char *dir_path;
char *norm_dir_path;
long max_length;)
{
  char cwd1[___PATH_MAX_LENGTH+1];
  ___BOOL result = 0;
  if (getcwd (cwd1, ___PATH_MAX_LENGTH+1) != 0)
    {
      long i = 0;
      while (dir_path[i] != '\0' && !SEPARATOR(dir_path[i]))
        i++;
      if (dir_path[i] == DRIVE_SEPARATOR && dir_path[i+1] != '\0')
        {
          char cwd2[___PATH_MAX_LENGTH+1];
          char tmp = dir_path[i+1];
          dir_path[i+1] = '\0';
          if (chdir (dir_path) == 0)
            {
              if (getcwd (cwd2, ___PATH_MAX_LENGTH+1) != 0)
                {
                  dir_path[i+1] = tmp;
                  if (chdir (dir_path) == 0)
                    {
                      result = getcwd (norm_dir_path, max_length) != 0;
                      chdir (cwd2);
                    }
                }
              chdir (cwd1);
            }
        }
      else if (chdir (dir_path) == 0)
        {
          result = getcwd (norm_dir_path, max_length) != 0;
          chdir (cwd1);
        }
    }
  return result;
}


#else

___HIDDEN ___BOOL normalize_dir_path
   ___P((char *dir_path,
         char *norm_dir_path,
         long max_length),
        (dir_path,
         norm_dir_path,
         max_length)
char *dir_path;
char *norm_dir_path;
long max_length;)
{
  char cwd[___PATH_MAX_LENGTH+1];
  ___BOOL result = 0;
  if (getcwd (cwd, ___PATH_MAX_LENGTH+1) != 0 && chdir (dir_path) == 0)
    {
      result = getcwd (norm_dir_path, max_length) != 0;
      chdir (cwd);
    }
  return result;
}


#endif


___HIDDEN ___BOOL path_expand_to_absolute
   ___P((char *path,
         char *new_path,
         long max_length),
        (path,
         new_path,
         max_length)
char *path;
char *new_path;
long max_length;)
{
  ___BOOL result = 0;
  char tmp[___PATH_MAX_LENGTH+1];
  long i = 0;
  long j = 0;
  long k;

  disable_interrupts ();

  if (path[0] == '\0')
    {
      /* Empty path expands to current working directory. */

      if (getcwd (new_path, max_length) == 0)
        goto ret;

      goto terminal_dir_separator;
    }

  if (path[0] == '~')
    {
      /*
       * Path starting with a '~' refers to a user's home directory or
       * the Gambit-C installation directory.  So get into "tilde_dir"
       * the path of the appropriate directory.
       */

      char *tilde_dir;

      if (path[1] == '~')
        {
          /* "~~" or "~~/xxx..." */

          if (path[2]!='\0' && !DIR_SEPARATOR(path[2]))
            goto ret;
          tilde_dir = getenv ("GAMBCDIR");
          if (tilde_dir == 0)
#ifdef ___GAMBCDIR
            tilde_dir = ___GAMBCDIR;
#else
#ifdef USE_DOS_PATHS
            tilde_dir = "c:/gambc/";
#else
            tilde_dir = "/usr/local/share/gambc/";
#endif
#endif
          i += 2;
        }
      else
        {
          /* "~" or "~/xxx..." or "~user" or "~user/xxx..." */

          i++;
          while (path[i]!='\0' && !DIR_SEPARATOR(path[i]))
            {
              if (i-1 > ___PATH_MAX_LENGTH)
                goto ret;
              tmp[i-1] = path[i];
              i++;
            }

          if (i == 1)
            {
              tilde_dir = getenv ("HOME");
              if (tilde_dir == 0)
#ifdef ___HOME
                tilde_dir = ___HOME;
#else
                goto ret;
#endif
            }
          else
#ifdef USE_UNIX
            {
              struct passwd *p;
              tmp[i-1] = '\0';
              p = getpwnam (tmp);
              if (p == 0)
                goto ret;
              tilde_dir = p->pw_dir;
              if (tilde_dir == 0)
                goto ret;
            }
#else
            goto ret;
#endif
        }

      /* Copy "tilde_dir" to head of "tmp". */

      while (*tilde_dir != '\0')
        {
          if (j > ___PATH_MAX_LENGTH)
            goto ret;
          tmp[j++] = *tilde_dir++;
        }
    }

  k = i;

  /*
   * At this point the first "k" characters of the path have
   * been processed.  Either k==0 or path[k]=='\0' or path[k] is
   * a path separator.
   */

  /* Copy the "non-tilde" part of the path to "tmp". */

  while (path[i] != '\0')
    {
      if (j > ___PATH_MAX_LENGTH)
        goto ret;
      tmp[j++] = path[i++];
    }

  if (j > ___PATH_MAX_LENGTH)
    goto ret;

  tmp[j] = '\0';

#ifdef USE_fullpath

  if (_fullpath (new_path, tmp, max_length+1) == 0)
    goto ret;

  {
    struct stat sbuf;
    if (stat (new_path, &sbuf) == 0 && S_ISDIR(sbuf.st_mode))
      {
        /* Make sure the directory path ends in a path separator. */

        terminal_dir_separator:
        j = strlen (new_path);
        if (j==0 || !DIR_SEPARATOR(new_path[j-1]))
          {
            new_path[j++] = DIR_SEPARATOR1;
            if (j > max_length)
              goto ret;
            new_path[j] = '\0';
          }
      }
  }

#else

  /* Check if this is the path of an existing directory. */

  if (normalize_dir_path (tmp, new_path, max_length))
    {
      /* Make sure the directory path ends in a path separator. */

      terminal_dir_separator:
      j = strlen (new_path);
      if (j==0 || !DIR_SEPARATOR(new_path[j-1]))
        {
          new_path[j++] = DIR_SEPARATOR1;
          if (j > max_length)
            goto ret;
          new_path[j] = '\0';
        }
    }
  else
    {
      /*
       * The path is not that of an existing directory.  Get the path
       * of the enclosing directory by removing the part that follows
       * the last path separator.
       */

      i = strlen (path);
      while (i > k && !SEPARATOR(path[i]))
        {
          i--;
          j--;
        }

      if (SEPARATOR(path[i]))
        {
          /* Check if the enclosing directory exists. */

          i++;
          j++;
          if (j > ___PATH_MAX_LENGTH)
            goto ret;
          tmp[j] = '\0';
          if (!normalize_dir_path (tmp, new_path, max_length))
            goto ret;
        }
      else if (i == 0)
        {
          /*
           * No enclosing directory specified so use current working
           * directory instead.
           */

          if (getcwd (new_path, max_length) == 0)
            goto ret;
        }
      else
        goto ret;

      /* Append the "non-tilde" part of the path. */

      j = strlen (new_path);
      if (!DIR_SEPARATOR(new_path[j-1]))
        new_path[j++] = DIR_SEPARATOR1;

      while (path[i] != '\0')
        {
          if (j > max_length)
            goto ret;
          new_path[j++] = path[i++];
        }

      if (j > max_length)
        goto ret;

      new_path[j] = '\0';
    }

#endif

  result = 1;

  ret:

  enable_interrupts ();

  return result;
}


#endif


___BOOL ___path_absolute
   ___P((char *path),
        (path)
char *path;)
{
#ifdef USE_MACOS

  char *p = path;
  if (*p == '~')
    return 1;
  while (*p != '\0' && !SEPARATOR(*p))
    p++;
  return p != path && DIR_SEPARATOR(*p);

#else

#ifdef USE_DOS_PATHS

  char *p = path;
  if (*p == '~' || DIR_SEPARATOR(*p))
    return 1;
  while (*p != '\0' && !SEPARATOR(*p))
    p++;
  return *p == DRIVE_SEPARATOR && DIR_SEPARATOR(p[1]);

#else

  return *path == '~' || DIR_SEPARATOR(*path);
  
#endif

#endif
}


___BOOL ___path_extension
   ___P((char *path,
         char *ext,
         long max_length),
        (path,
         ext,
         max_length)
char *path;
char *ext;
long max_length;)
{
  char *p, *end = path;
  while (*end != '\0')
    end++;
  p = end;
  while (p > path && *p != '.' && !SEPARATOR(*p))
    p--;
  if (*p != '.')
    p = end;
  if (end - p > max_length)
    return 0;
  while (p <= end)
    *ext++ = *p++;
  return 1;
}


___BOOL ___path_strip_extension
   ___P((char *path,
         char *new_path,
         long max_length),
        (path,
         new_path,
         max_length)
char *path;
char *new_path;
long max_length;)
{
  char *p, *end = path;
  while (*end != '\0')
    end++;
  p = end;
  while (p > path && *p != '.' && !SEPARATOR(*p))
    p--;
  if (*p != '.')
    p = end;
  if (p - path > max_length)
    return 0;
  while (path < p)
    *new_path++ = *path++;
  *new_path++ = '\0';
  return 1;
}


___BOOL ___path_directory
   ___P((char *path,
         char *dir,
         long max_length),
        (path,
         dir,
         max_length)
char *path;
char *dir;
long max_length;)
{
  char *p = path, *end = path;
  while (*end != '\0')
    {
      end++;
      if (SEPARATOR(end[-1]))
        p = end;
    }
  if (p - path > max_length)
    return 0;
  while (path < p)
    *dir++ = *path++;
  *dir++ = '\0';
  return 1;
}


___BOOL ___path_strip_directory
   ___P((char *path,
         char *new_path,
         long max_length),
        (path,
         new_path,
         max_length)
char *path;
char *new_path;
long max_length;)
{
  char *p = path, *end = path;
  while (*end != '\0')
    {
      end++;
      if (SEPARATOR(end[-1]))
        p = end;
    }
  if (end - p > max_length)
    return 0;
  while (p <= end)
    *new_path++ = *p++;
  return 1;
}


___BOOL ___path_drive
   ___P((char *path,
         char *drive,
         long max_length),
        (path,
         drive,
         max_length)
char *path;
char *drive;
long max_length;)
{
#ifdef USE_DOS_PATHS

  char *p = path;
  while (*p != '\0' && !SEPARATOR(*p))
    p++;
  if (*p != DRIVE_SEPARATOR)
    {
      drive[0] = '\0';
      return 1;
    }
  p++;
  if (p - path > max_length)
    return 0;
  while (path < p)
    *drive++ = *path++;
  *drive++ = '\0';
  return 1;

#else

  drive[0] = '\0';
  return 1;

#endif
}


___BOOL ___path_strip_drive
   ___P((char *path,
         char *new_path,
         long max_length),
        (path,
         new_path,
         max_length)
char *path;
char *new_path;
long max_length;)
{
#ifdef USE_DOS_PATHS

  char *p = path, *end = path;
  while (*end != '\0')
    end++;
  while (*p != '\0' && !SEPARATOR(*p))
    p++;
  if (*p != DRIVE_SEPARATOR)
    p = path;
  else
    p++;
  if (end - p > max_length)
    return 0;
  while (p <= end)
    *new_path++ = *p++;
  return 1;

#else

  char *end = path;
  while (*end != '\0')
    end++;
  if (end - path > max_length)
    return 0;
  while (path <= end)
    *new_path++ = *path++;
  return 1;

#endif
}


___BOOL ___path_expand
   ___P((char *path,
         char *new_path,
         long max_length,
         int format),
        (path,
         new_path,
         max_length,
         format)
char *path;
char *new_path;
long max_length;
int format;)
{
  int expansion, subdir, nb_parent_hops;
  char *p1, *p2, *q1, *q2;
  char path_cwd[___PATH_MAX_LENGTH+1];

  if (!path_expand_to_absolute (path, new_path, max_length))
    return 0;

  if (format == ___PATH_ABSOLUTE)
    return 1;

  if (!path_expand_to_absolute ("", path_cwd, max_length))
    return 0;

#ifdef USE_DOS_PATHS

  /* return absolute path if drives are different */

  if (path_cwd[0] == '\0' ||
      path_cwd[1] != DRIVE_SEPARATOR ||
      path_cwd[2] != DIR_SEPARATOR1 ||
      new_path[0] != path_cwd[0] ||
      new_path[1] != DRIVE_SEPARATOR ||
      new_path[2] != DIR_SEPARATOR1)
    return 1;

#endif

  p1 = path_cwd;
  p2 = new_path;
  q1 = new_path;

  while (*p1 != '\0' && *p1 == *p2)
    {
      if (DIR_SEPARATOR(*p2))
        q1 = p2+1;
      p1++;
      p2++;
    }

#ifdef USE_MACOS

  /* if volume names are different an absolute path is needed */

  if (q1 == new_path)
    return 1;

#endif

  subdir = 0;
  q2 = q1;
  while (*q2 != '\0')
    {
      if (DIR_SEPARATOR(*q2))
        subdir = 1;
      q2++;
    }

  nb_parent_hops = 0;
  while (*p1 != '\0')
    {
      if (DIR_SEPARATOR(*p1))
        nb_parent_hops++;
      p1++;
    }

#ifdef USE_MACOS

  if (nb_parent_hops > 0 || subdir)
    nb_parent_hops++;

#endif

  expansion = nb_parent_hops*(sizeof (PARENT_HOP)-1) - (q1-new_path);

  if (expansion >= 0)
    {
      if (format == ___PATH_SHORTEST)
        return 1;
      if ((q2-new_path)+expansion > max_length)
        return 0;
      *(q2+expansion) = '\0';
      while (q1 < q2)
        {
          q2--;
          *(q2+expansion) = *q2;
        }
    }
  else
    {
      while (q1 < q2)
        {
          *(q1+expansion) = *q1;
          q1++;
        }
      *(q1+expansion) = '\0';
    }

  while (nb_parent_hops > 0)
    {
      char *p = PARENT_HOP;
      while (*p != '\0')
        *new_path++ = *p++;
      nb_parent_hops--;
    }

  return 1;
}


/*---------------------------------------------------------------------------*/

/* Formatting of source code position. */

char *___format_filepos
   ___P((char *path,
         long filepos,
         ___BOOL pinpoint),
        (path,
         filepos,
         pinpoint)
char *path;
long filepos;
___BOOL pinpoint;)
{
#ifdef USE_MACOS

#ifdef USE_mac_gui

  if (pinpoint)
    mac_gui_highlight (path, filepos);

#endif

#endif

  return 0; /* Use default format for displaying location */
}


/*---------------------------------------------------------------------------*/

/* Operations on I/O devices. */

/* User module setup. */

typedef struct user_module_struct
  {
    int refcount; /* number of threads using this structure */

#ifndef USE_WIN32

#define USER_MODULE_INIT_OS_SPECIFIC

#endif

#ifdef USE_WIN32

    HANDLE always_signaled;  /* this event is always signaled */
    HANDLE abort_select;     /* ___device_select exits when this is signaled */

#define USER_MODULE_INIT_OS_SPECIFIC , NULL, NULL

#endif
  } user_module;

___HIDDEN user_module um =
{
  0
  USER_MODULE_INIT_OS_SPECIFIC
};


#ifdef USE_WIN32

___HIDDEN BOOL WINAPI console_event_handler
   ___P((DWORD dwCtrlType),
        (dwCtrlType)
DWORD dwCtrlType;)
{
#if 0
  /*****************************/
  switch (dwCtrlType)
    {
    case CTRL_C_EVENT:
      um.got_event = 1;
      break;
    case CTRL_BREAK_EVENT:
      um.got_event = 2;
      break;
    case CTRL_CLOSE_EVENT:
      um.got_event = 3;
      break;
    case CTRL_LOGOFF_EVENT:
      um.got_event = 4;
      break;
    case CTRL_SHUTDOWN_EVENT:
      um.got_event = 5;
      break;
    default:
      um.got_event = 999;
    }

  SetEvent (um.abort_select); /* ignore error */

  return TRUE;
#else
  return FALSE;
#endif
}

#endif


___HIDDEN ___WORD user_setup ___PVOID
{
  if (um.refcount == 0)
    {
#ifdef USE_WIN32

      ___WORD e = ___FIX(___NO_ERR);
      DWORD thread_id;

      um.always_signaled = NULL;
      um.abort_select = NULL;

      um.always_signaled =
        CreateEvent (NULL,  /* can't inherit */
                     TRUE,  /* manual reset */
                     TRUE,  /* signaled */
                     NULL); /* no name */

      if (um.always_signaled == NULL)
        e = ___FIX(err_code_from_GetLastError ());
      else
        {
          um.abort_select =
            CreateEvent (NULL,  /* can't inherit */
                         TRUE,  /* manual reset */
                         FALSE, /* not signaled */
                         NULL); /* no name */

          if (um.abort_select == NULL)
            e = ___FIX(err_code_from_GetLastError ());
          else
            {
              if (!SetConsoleCtrlHandler (console_event_handler, TRUE))/****************zzzzzzzzzzzzzzz*/
                e = ___FIX(err_code_from_GetLastError ());
            }
        }

      if (e != ___FIX(___NO_ERR))
        {
          if (um.abort_select != NULL)
            CloseHandle (um.abort_select); /* ignore error */

          if (um.always_signaled != NULL)
            CloseHandle (um.always_signaled); /* ignore error */

          return e;
        }

#endif
    }

  um.refcount++;

  return ___FIX(___NO_ERR);
}


___HIDDEN ___WORD user_cleanup ___PVOID
{
  if (--um.refcount == 0)
    {
#ifdef USE_WIN32

      SetConsoleCtrlHandler (console_event_handler, FALSE); /* ignore error *//**************zzzzzzzzzzzzz*/

      CloseHandle (um.abort_select); /* ignore error */

      CloseHandle (um.always_signaled); /* ignore error */

#endif
    }

  return ___FIX(___NO_ERR);
}


/*---------------------------------------------------------------------------*/

#ifdef USE_PUMPS

/* Nonblocking pipes */

/*
 * A pipe is a unidirectional FIFO data structure that buffers data
 * between a "writer" that writes data to the pipe and a "reader" that
 * reads this data.  Independently from the FIFO buffer, a pipe has a
 * "reader error indicator" and a "writer error indicator".  When
 * created the error indicators are set to "no error" and the FIFO
 * contains no data.

 * Typically the writer writes bytes and out-of-band messages at one
 * end with a "write" operation and the reader reads these bytes and
 * out-of-band messages in the same order at the other end with a
 * "read" operation.  Out-of-band messages are used to request some
 * special action from the reader and are expected to be infrequent.
 * When a read operation extracts an out-of-band message from the pipe
 * it reports that zero bytes can be read from the pipe and sets the
 * out-of-band structure (one of the parameters of the read
 * operation).  The writer can set the "writer error indicator" when
 * it needs to transmit an error code to the reader.  After all
 * previously written bytes and out-of-band messages are read by the
 * reader, the next read operation will return this error code and the
 * error indicator is reset to "no error".  Similarly, the reader can
 * set the "reader error indicator" when it needs to transmit an error
 * code to the writer.  The next write operation will return this
 * error code and the error indicator is reset to "no error".

 * The nonblocking pipes implemented here have read and write
 * operations that don't block the calling thread.  When the data
 * transfer requested can't be performed immediately these operations
 * will return with a special error code (EAGAIN).  This allows the
 * calling thread to do other things before attempting the operation
 * again.  We say a pipe is "ready for an operation" if calling that
 * operation will not return EAGAIN (i.e. the operations will succeed
 * with no error or will fail with some error different from EAGAIN).

 * Specifically,
 *
 *   a pipe is ready for reading when
 *     - the "writer error indicator" is not "no error"
 *     - or
 *       - the "reader error indicator" is "no error"
 *       - and
 *         - an out-of-band message is in the pipe
 *         - or at least one byte is in the FIFO buffer
 *
 *   a pipe is ready for writing when
 *     - the "reader error indicator" is not "no error"
 *     - or
 *       - the "writer error indicator" is "no error"
 *       - and no out-of-band message is in the pipe
 *       - and at least one byte is free in the FIFO buffer
 *       - and the previous read operation did not return an error or
 *         an out-of-band message
 *         (at creation a pipe is not ready for writing and the first
 *         "read" operation will make the pipe ready for writing, in other
 *         words, after it reports an error or out-of-band message, the
 *         writer blocks until the reader shows some interest in more data)

 * This table summarizes the states of a pipe (an R means ready for
 * reading and a W means ready for writing and a w means ready for
 * writing iff the most recent read operation did not return an
 * error or out-of-band message):

 *        no writer :  writer     reader      bytes in FIFO buffer
 *          error   :  error      error?      (N is size of buffer)
 *        +----+----+----+----+
 *        |  w | R  | R  | R  |     no           0 bytes
 *        +----+----+----+----+
 *        | Rw | R  | R  | R  |     no           > 0 bytes but < N
 *        +----+----+----+----+
 *        | R  | R  | R  | R  |     no           = N
 *        +----+----+----+----+
 *        |  W |  W | RW | RW |     yes          0 bytes
 *        +----+----+----+----+
 *        |  W |  W | RW | RW |     yes          > 0 bytes but < N
 *        +----+----+----+----+
 *        |  W |  W | RW | RW |     yes          = N
 *        +----+----+----+----+
 *        :!oob:oob :!oob:oob :

 */

typedef struct nonblocking_pipe_oob_msg_struct
  {
    int op;         /* operation */
    DWORD param[2]; /* parameters */
  } nonblocking_pipe_oob_msg;

typedef struct nonblocking_pipe_struct
  {
    HANDLE mutex;     /* mutex to serialize access to the pipe */
    HANDLE revent;    /* event, signaled iff ready for reading */
    HANDLE wevent;    /* event, signaled iff ready for writing */
    ___WORD rerr;     /* the reader error indicator */
    ___WORD werr;     /* the writer error indicator */
    ___BOOL oob;      /* the out-of-band indicator */
    nonblocking_pipe_oob_msg oob_msg; /* out-of-band message */
    int rd;           /* the read pointer of the circular buffer */
    int wr;           /* the write pointer of the circular buffer */
    int size;         /* size of the circular buffer */
    ___U8 *buffer;    /* the circular buffer */
  } nonblocking_pipe;

___HIDDEN ___WORD nonblocking_pipe_setup
   ___P((nonblocking_pipe *pipe,
         int size),
        (pipe,
         size)
nonblocking_pipe *pipe;
int size;)
{
  ___U8 *buffer;
  HANDLE mutex;
  HANDLE revent;
  HANDLE wevent;

  buffer = ___CAST(___U8*,___alloc_mem (size));

  if (buffer == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  mutex = CreateMutex (NULL,  /* can't inherit */
                       FALSE, /* unlocked */
                       NULL); /* no name */

  if (mutex == NULL)
    {
      ___WORD e = ___FIX(err_code_from_GetLastError ());
      ___free_mem (buffer);
      return e;
    }

  revent = CreateEvent (NULL,  /* can't inherit */
                        TRUE,  /* manual reset */
                        FALSE, /* not signaled */
                        NULL); /* no name */

  if (revent == NULL)
    {
      ___WORD e = ___FIX(err_code_from_GetLastError ());
      CloseHandle (mutex); /* ignore error */
      ___free_mem (buffer);
      return e;
    }

  wevent = CreateEvent (NULL,  /* can't inherit */
                        TRUE,  /* manual reset */
                        FALSE, /* not signaled */
                        NULL); /* no name */

  if (wevent == NULL)
    {
      ___WORD e = ___FIX(err_code_from_GetLastError ());
      CloseHandle (revent); /* ignore error */
      CloseHandle (mutex); /* ignore error */
      ___free_mem (buffer);
      return e;
    }

  pipe->mutex = mutex;
  pipe->revent = revent;
  pipe->wevent = wevent;
  pipe->rerr = ___FIX(___NO_ERR);
  pipe->werr = ___FIX(___NO_ERR);
  pipe->oob = 0;
  pipe->rd = 0;
  pipe->wr = 0;
  pipe->size = size;
  pipe->buffer = buffer;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD nonblocking_pipe_cleanup
   ___P((nonblocking_pipe *pipe),
        (pipe)
nonblocking_pipe *pipe;)
{
  CloseHandle (pipe->wevent); /* ignore error */
  CloseHandle (pipe->revent); /* ignore error */
  CloseHandle (pipe->mutex); /* ignore error */
  ___free_mem (pipe->buffer);

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD nonblocking_pipe_set_reader_err
   ___P((nonblocking_pipe *pipe,
         ___WORD err),
        (pipe,
         err)
nonblocking_pipe *pipe;
___WORD err;)
{
  if (WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return ___FIX(err_code_from_GetLastError ());

  /* note: the reader error indicator may get overwritten */

  pipe->rerr = err;

  SetEvent (pipe->wevent); /* ignore error */

  if (pipe->werr == ___FIX(___NO_ERR))
    ResetEvent (pipe->revent); /* ignore error */

  ReleaseMutex (pipe->mutex); /* ignore error */

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD nonblocking_pipe_set_writer_err
   ___P((nonblocking_pipe *pipe,
         ___WORD err),
        (pipe,
         err)
nonblocking_pipe *pipe;
___WORD err;)
{
  if (WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return ___FIX(err_code_from_GetLastError ());

  /* note: the writer error indicator may get overwritten */

  pipe->werr = err;

  SetEvent (pipe->revent); /* ignore error */

  if (pipe->rerr == ___FIX(___NO_ERR))
    ResetEvent (pipe->wevent); /* ignore error */

  ReleaseMutex (pipe->mutex); /* ignore error */

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD nonblocking_pipe_write_oob
   ___P((nonblocking_pipe *pipe,
         nonblocking_pipe_oob_msg *oob_msg),
        (pipe,
         oob_msg)
nonblocking_pipe *pipe;
nonblocking_pipe_oob_msg *oob_msg;)
{
  if (WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return ___FIX(err_code_from_GetLastError ());

  if (pipe->werr != ___FIX(___NO_ERR) || pipe->oob)
    {
      ReleaseMutex (pipe->mutex); /* ignore error */
      return ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));
    }

  pipe->oob = 1;
  pipe->oob_msg = *oob_msg;

  if (pipe->rerr == ___FIX(___NO_ERR))
    {
      SetEvent (pipe->revent); /* ignore error */
      ResetEvent (pipe->wevent); /* ignore error */
    }

  ReleaseMutex (pipe->mutex); /* ignore error */

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD nonblocking_pipe_read_ready_wait
   ___P((nonblocking_pipe *pipe),
        (pipe)
nonblocking_pipe *pipe;)
{
  ___WORD werr;

  if (WaitForSingleObject (pipe->revent, INFINITE) == WAIT_FAILED ||
      WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return ___FIX(err_code_from_GetLastError ());

  werr = pipe->werr;

  if (werr != ___FIX(___NO_ERR))
    {
      pipe->werr = ___FIX(___NO_ERR);

      if (pipe->rerr != ___FIX(___NO_ERR) ||
          (pipe->rd == pipe->wr && !pipe->oob))
        ResetEvent (pipe->revent); /* ignore error */
    }

  ReleaseMutex (pipe->mutex); /* ignore error */

  return werr;
}

___HIDDEN ___WORD nonblocking_pipe_write_ready_wait
   ___P((nonblocking_pipe *pipe),
        (pipe)
nonblocking_pipe *pipe;)
{
  ___WORD rerr;

  if (WaitForSingleObject (pipe->wevent, INFINITE) == WAIT_FAILED ||
      WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return ___FIX(err_code_from_GetLastError ());

  rerr = pipe->rerr;

  if (rerr != ___FIX(___NO_ERR))
    {
      pipe->rerr = ___FIX(___NO_ERR);

      ResetEvent (pipe->wevent); /* ignore error */
    }

  ReleaseMutex (pipe->mutex); /* ignore error */

  return rerr;
}

___HIDDEN ___WORD nonblocking_pipe_read
   ___P((nonblocking_pipe *pipe,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done,
         nonblocking_pipe_oob_msg *oob_msg),
        (pipe,
         buf,
         len,
         len_done,
         oob_msg)
nonblocking_pipe *pipe;
___U8 *buf;
___stream_index len;
___stream_index *len_done;
nonblocking_pipe_oob_msg *oob_msg;)
{
  ___WORD rerr;
  ___WORD werr;
  DWORD rd;
  DWORD wr;
  ___U8 *p;
  DWORD end;
  DWORD n;
  DWORD i;

  if (len <= 0)
    return ___FIX(___UNKNOWN_ERR);

  if (WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return ___FIX(err_code_from_GetLastError ());

  rerr = pipe->rerr;
  werr = pipe->werr;
  rd = pipe->rd;
  wr = pipe->wr;

  if (rerr != ___FIX(___NO_ERR))
    {
      /* there is a reader error */

      if (werr == ___FIX(___NO_ERR))
        werr = ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));
      else
        {
          pipe->werr = ___FIX(___NO_ERR);
          ResetEvent (pipe->revent); /* ignore error */
        }
      ReleaseMutex (pipe->mutex); /* ignore error */
      return werr;
    }

  /* there is no reader error */

  if (rd == wr)
    {
      /* no bytes in FIFO buffer */

      if (pipe->oob)
        {
          /* out-of-band present */

          *oob_msg = pipe->oob_msg;
          pipe->oob = 0;
          if (werr == ___FIX(___NO_ERR))
            {
              ResetEvent (pipe->revent); /* ignore error */
#if 0
              /******************zzzzzzzzzzzzzz****/
              SetEvent (pipe->wevent); /* ignore error */
#endif
            }
          ReleaseMutex (pipe->mutex); /* ignore error */
          *len_done = 0;
          return ___FIX(___NO_ERR);
        }

      /* out-of-band not present */

      if (werr != ___FIX(___NO_ERR))
        {
          /* there is a writer error */

          pipe->werr = ___FIX(___NO_ERR);

          ResetEvent (pipe->revent); /* ignore error */
          ReleaseMutex (pipe->mutex); /* ignore error */
          return werr;
        }

      /* there is no writer error */

      SetEvent (pipe->wevent); /* ignore error */
      ReleaseMutex (pipe->mutex); /* ignore error */
      return ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));
    }

  /* at least one byte in FIFO buffer */

  end = pipe->size - rd; /* number of bytes from rd to end */

  n = wr + end; /* number of bytes available */
  if (n >= pipe->size)
    n -= pipe->size;

  if (n > len) /* don't transfer more than len */
    n = len;

  *len_done = n;

  p = pipe->buffer + rd; /* prepare transfer source */

  rd = rd + n;
  if (rd >= pipe->size)
    rd -= pipe->size;

  pipe->rd = rd;

  if (werr == ___FIX(___NO_ERR) && !pipe->oob)
    {
      /* there is no writer error and out-of-band not present */

      if (rd == wr) /* the FIFO will be empty? */
        ResetEvent (pipe->revent); /* ignore error */

      /* the FIFO will not be full */

      SetEvent (pipe->wevent); /* ignore error */
    }

  if (n <= end)
    {
      /* only need to transfer n bytes starting from original rd */

      for (i=n; i>0; i--)
        *buf++ = *p++;
    }
  else
    {
      /* need to transfer end bytes starting from original rd */

      for (i=end; i>0; i--)
        *buf++ = *p++;

      /* and to transfer n-end bytes starting from 0 */

      p = pipe->buffer;

      for (i=n-end; i>0; i--)
        *buf++ = *p++;
    }

  ReleaseMutex (pipe->mutex); /* ignore error */

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD nonblocking_pipe_write
   ___P((nonblocking_pipe *pipe,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (pipe,
         buf,
         len,
         len_done)
nonblocking_pipe *pipe;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___WORD rerr;
  ___WORD werr;
  DWORD rd;
  DWORD wr;
  ___U8 *p;
  DWORD end;
  DWORD n;
  DWORD i;

  if (len <= 0)
    return ___FIX(___UNKNOWN_ERR);

  if (WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return ___FIX(err_code_from_GetLastError ());

  rerr = pipe->rerr;
  werr = pipe->werr;
  rd = pipe->rd;
  wr = pipe->wr;

  if (werr != ___FIX(___NO_ERR))
    {
      /* there is a writer error */

      if (rerr == ___FIX(___NO_ERR))
        rerr = ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));
      else
        {
          pipe->rerr = ___FIX(___NO_ERR);
          ResetEvent (pipe->wevent); /* ignore error */
        }
      ReleaseMutex (pipe->mutex); /* ignore error */
      return rerr;
    }

  /* there is no writer error */

  if (rerr != ___FIX(___NO_ERR))
    {
      /* there is a reader error */

      pipe->rerr = ___FIX(___NO_ERR);

      if (rd != wr || pipe->oob) /* FIFO is not empty */
        SetEvent (pipe->revent); /* ignore error */

      ResetEvent (pipe->wevent); /* ignore error */
      ReleaseMutex (pipe->mutex); /* ignore error */
      return rerr;
    }

  /* there is no reader error */

  if (wr + 1 - rd == 0 || wr + 1 - rd == pipe->size || pipe->oob)
    {
      /* FIFO buffer is full or out-of-band present */

      ReleaseMutex (pipe->mutex); /* ignore error */
      return ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));
    }

  /* FIFO buffer is not full and out-of-band not present */

  end = pipe->size - wr; /* number of bytes from wr to end */

  n = rd + end - 1; /* number of bytes available */
  if (n >= pipe->size)
    n -= pipe->size;

  if (n > len) /* don't transfer more than len */
    n = len;

  *len_done = n;

  p = pipe->buffer + wr; /* prepare transfer source */

  wr = wr + n;
  if (wr >= pipe->size)
    wr -= pipe->size;

  pipe->wr = wr;

  if (wr + 1 - rd == 0 || wr + 1 - rd == pipe->size) /* FIFO will be full? */
    ResetEvent (pipe->wevent); /* ignore error */

  /* FIFO will not be empty */

  SetEvent (pipe->revent); /* ignore error */

  if (n <= end)
    {
      /* only need to transfer n bytes starting from original wr */

      for (i=n; i>0; i--)
        *p++ = *buf++;
    }
  else
    {
      /* need to transfer end bytes starting from original wr */

      for (i=end; i>0; i--)
        *p++ = *buf++;

      /* and to transfer n-end bytes starting from 0 */

      p = pipe->buffer;

      for (i=n-end; i>0; i--)
        *p++ = *buf++;
    }

  ReleaseMutex (pipe->mutex); /* ignore error */

  return ___FIX(___NO_ERR);
}

#endif


/*---------------------------------------------------------------------------*/

/* Generic device operations. */

___WORD ___device_select
   ___P((___device **devs,
         int nb_read_devs,
         int nb_write_devs,
         ___time timeout),
        (devs,
         nb_read_devs,
         nb_write_devs,
         timeout)
___device **devs;
int nb_read_devs;
int nb_write_devs;
___time timeout;)
{
#define SELECT_DEVS_MAX 8192

  int nb_devs;
  ___device_select_state state;
  int pass;
  ___BOOL all_select_setups_done;
  ___device *select_devs[SELECT_DEVS_MAX];
  int i;

  nb_devs = nb_read_devs + nb_write_devs;
  if (nb_devs > SELECT_DEVS_MAX) /* ignore last devices if too many */
    nb_devs = SELECT_DEVS_MAX;
  
  state.devs = devs;

  state.timeout = timeout;

#ifdef USE_select

  state.highest_fd_plus_1 = 0;

  FD_ZERO(&state.readfds);
  FD_ZERO(&state.writefds);
  FD_ZERO(&state.exceptfds);

#endif

#ifdef USE_MsgWaitForMultipleObjects

#if 0
  state.timeout = 5000;/*******************/
#else
  state.timeout = INFINITE;
#endif

  state.timeout_dev_pos = -1;

  state.message_queue_mask = 0;
  state.message_queue_dev_pos = -1;

  state.wait_objs_buffer[0] = um.abort_select;
  state.wait_objs_buffer[1] = um.heartbeat_thread;

  state.nb_wait_objs = 2;

#endif

  for (i=nb_devs-1; i>=0; i--)
    {
      ___device *d = devs[i];
      select_devs[i] = d;
      if (d != NULL)
        d->select_step = 0;
    }

  pass = ___SELECT_PASS_1;
  all_select_setups_done = 0;

  while (!all_select_setups_done)
    {
      all_select_setups_done = 1;

      for (i=nb_devs-1; i>=0; i--)
        {
          ___WORD e;
          ___device *d = select_devs[i];
          if (d != NULL)
            {
              if ((e = ___device_select_virt
                         (d,
                          i>=nb_read_devs,
                          i,
                          pass,
                          &state))
                  != ___FIX(___NO_ERR))
                {
                  if (e != ___FIX(___SELECT_SETUP_DONE))
                    return e;
                  select_devs[i] = NULL;
                }
              else
                all_select_setups_done = 0;
            }
        }

      pass++;
    }

#ifdef USE_select

  {
    ___WORD e;/***************unused*/
    ___time delta;
    struct timeval delta_tv_struct;
    struct timeval *delta_tv = &delta_tv_struct;

    absolute_time_to_relative_time (state.timeout, &delta);

    absolute_time_to_nonnegative_timeval (delta, &delta_tv);

#if 0
    /****************************/
    printf ("timeout = %20.10g\n",state.timeout);
    printf ("delta = %20.10g\n",delta);
    printf ("delta_tv = %p secs=%d  usecs=%d\n",delta_tv,delta_tv_struct.tv_sec,delta_tv_struct.tv_usec);
    /*      fcntl (0, F_SETFL, 0); /* set nonblocking mode */
#endif

#if 0
    /****************************/
    {
      static FILE* logfile = NULL;
      if (logfile == NULL) logfile = fopen ("logfile", "w");
      if (logfile != NULL)
        fprintf(logfile,"select  secs=%d  usecs=%d  highest_fd_plus_1 = %d\n",delta_tv->tv_sec,delta_tv->tv_usec,state.highest_fd_plus_1);
      else
        fprintf(logfile,"select  NULL  highest_fd_plus_1 = %d\n",state.highest_fd_plus_1);
#endif

    state.select_result =
      select (state.highest_fd_plus_1,
              &state.readfds,
              &state.writefds,
              &state.exceptfds,
              delta_tv);

#if 0
    /****************************/

    if (FD_ISSET(0,&state.readfds)) fprintf(logfile,"0 ready for reading\n");
    if (FD_ISSET(1,&state.readfds)) fprintf(logfile,"1 ready for reading\n");
    if (FD_ISSET(3,&state.readfds)) fprintf(logfile,"3 ready for reading\n");
    if (FD_ISSET(4,&state.readfds)) fprintf(logfile,"4 ready for reading\n");
    if (FD_ISSET(0,&state.writefds)) fprintf(logfile,"0 ready for writing\n");
    if (FD_ISSET(1,&state.writefds)) fprintf(logfile,"1 ready for writing\n");
    if (FD_ISSET(3,&state.writefds)) fprintf(logfile,"3 ready for writing\n");
    if (FD_ISSET(4,&state.writefds)) fprintf(logfile,"4 ready for writing\n");
    if (FD_ISSET(0,&state.exceptfds)) fprintf(logfile,"0 ready for exceptions\n");
    if (FD_ISSET(1,&state.exceptfds)) fprintf(logfile,"1 ready for exceptions\n");
    if (FD_ISSET(3,&state.exceptfds)) fprintf(logfile,"3 ready for exceptions\n");
    if (FD_ISSET(4,&state.exceptfds)) fprintf(logfile,"4 ready for exceptions\n");

    /*      fcntl (0, F_SETFL, O_NONBLOCK); /* set nonblocking mode */
    fprintf(logfile,"select result = %d        errno = %d\n",state.select_result, errno);/*****************/

    if (state.select_result<0)
      fprintf(logfile,"secs=%d  usecs=%d  state.highest_fd_plus_1 = %d    select result = %d\n",delta_tv_struct.tv_sec,delta_tv_struct.tv_usec,state.highest_fd_plus_1,  errno);/*****************/
    fflush(logfile);
    }
#endif

    if (state.select_result < 0)
      return ___FIX(err_code_from_errno ());
  }

#endif

#ifdef USE_MsgWaitForMultipleObjects

  if (state.timeout == 0)
    goto timed_out; /* don't even call MsgWaitForMultipleObjects */

  while (state.nb_wait_objs > 0 || state.message_queue_mask != 0)
    {
      DWORD n;

      n = MsgWaitForMultipleObjects
            (state.nb_wait_objs,
             state.wait_objs_buffer,
             FALSE,
             state.timeout,
             state.message_queue_mask);

      if (n == WAIT_FAILED)
        return ___FIX(err_code_from_GetLastError ());

      if (n >= WAIT_OBJECT_0 &&
          n <= WAIT_OBJECT_0+state.nb_wait_objs)
        n -= WAIT_OBJECT_0;
      else if (n >= WAIT_ABANDONED_0 &&
               n <= WAIT_ABANDONED_0+state.nb_wait_objs-1)
        n -= WAIT_ABANDONED_0;
      else
        {
          /* n == WAIT_TIMEOUT */

          /*
           * The call to MsgWaitForMultipleObjects timed out.  Mark
           * the appropriate device "ready".
           */

          if (state.timeout != 0)
            {
              /* first call to MsgWaitForMultipleObjects */

              timed_out:

              i = state.timeout_dev_pos;
              if (i >= 0)
                devs[i] = NULL;
            }

          break;
        }

      if (n == state.nb_wait_objs)
        {
          /*
           * The message queue contains a message that is of interest.
           * Mark the appropriate device "ready".
           */

          i = state.message_queue_dev_pos;
          if (i >= 0)
            devs[i] = NULL;

          /*
           * Don't check if other devices are ready because this might
           * cause an infinite loop.
           */

          break;
        }
      else if (n == 0)
        {
          /*
           * The call to device_select must be aborted because the
           * abort_select event is set.  This occurs when an interrupt
           * (such as a CTRL-C user interrupt) needs to be serviced
           * promptly by the main program.
           */

          /**********************/
          ResetEvent (um.abort_select); /* ignore error */

          break;
        }
      else if (n == 1)
        {
          /*
           * The heartbeat thread has died.  This is normally due to
           * the program being terminated abruptly by the user (for
           * example by using the thread manager or the "shutdown"
           * item in the start menu).  When this happens we must
           * initiate a clean termination of the program.
           */

          fprintf (stderr, "heartbeat thread died\n");
          fflush (stderr);/***********/

          return ___FIX(___UNKNOWN_ERR);
        }
      else
        {
          /* Mark the appropriate device "ready". */

          i = state.wait_obj_to_dev_pos[n];
          if (i >= 0)
            devs[i] = NULL;

          /* Prepare to check remaining devices. */

          state.wait_obj_to_dev_pos[n] =
            state.wait_obj_to_dev_pos[--state.nb_wait_objs];
        }

      state.timeout = 0; /* next MsgWaitForMultipleObjects will only poll */
    }

#endif

  for (i=nb_devs-1; i>=0; i--)
    {
      ___WORD e;
      ___device *d = devs[i];
      if (d != NULL)
        if ((e = ___device_select_virt
                   (d,
                    i>=nb_read_devs,
                    i,
                    ___SELECT_PASS_CHECK,
                    &state))
            != ___FIX(___NO_ERR))
          return e;
    }

  return ___FIX(___NO_ERR);
}

#ifdef USE_select

___HIDDEN void device_select_add_fd
   ___P((___device_select_state *state,
         int fd,
         ___BOOL for_writing),
        (state,
         fd,
         for_writing)
___device_select_state *state;
int fd;
___BOOL for_writing;)
{
  if (for_writing)
    FD_SET(fd, &state->writefds);
  else
    FD_SET(fd, &state->readfds);

  if (fd >= state->highest_fd_plus_1)
    state->highest_fd_plus_1 = fd+1;
}

#endif

#ifdef USE_MsgWaitForMultipleObjects

___HIDDEN void device_select_add_wait_obj
   ___P((___device_select_state *state,
         int i,
         HANDLE wait_obj),
        (state,
         i,
         wait_obj)
___device_select_state *state;
int i;
HANDLE wait_obj;)
{
  int j = state->nb_wait_objs;

  if (j < MAXIMUM_WAIT_OBJECTS)
    {
      state->wait_objs_buffer[j] = wait_obj;
      state->wait_obj_to_dev_pos[j] = i;
      state->nb_wait_objs = j+1;
    }
}

#endif

___WORD ___device_flush_write
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___device_flush_write_virt (self);
}

___WORD ___device_close
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  return ___device_close_virt (self, direction);
}

___HIDDEN void device_add_to_group
   ___P((___device_group *dgroup,
         ___device *dev),
        (dgroup,
         dev)
___device_group *dgroup;
___device *dev;)
{
  ___device *head = dgroup->list;

  dev->group = dgroup;

  if (head == NULL)
    {
      dev->prev = dev;
      dev->next = dev;
      dgroup->list = dev;
    }
  else
    {
      ___device *tail = head->prev;
      dev->prev = tail;
      dev->next = head;
      head->prev = dev;
      tail->next = dev;
    }
}

___HIDDEN void device_remove_from_group
   ___P((___device *dev),
        (dev)
___device *dev;)
{
  ___device_group *dgroup = dev->group;
  ___device *prev = dev->prev;
  ___device *next = dev->next;

  if (prev == dev)
    dgroup->list = NULL;
  else
    {
      next->prev = prev;
      prev->next = next;
      dev->prev = dev;
      dev->next = dev;
      if (dgroup->list == dev)
        dgroup->list = next;
    }

  dev->group = NULL;
}

___HIDDEN void device_add_ref
   ___P((___device *self),
        (self)
___device *self;)
{
#ifdef USE_WIN32
  InterlockedIncrement (&self->refcount);
#else
  self->refcount++;
#endif
}

___WORD ___device_release
   ___P((___device *self),
        (self)
___device *self;)
{
  ___WORD e = ___FIX(___NO_ERR);

  if (
#ifdef USE_WIN32
      !InterlockedDecrement (&self->refcount)
#else
      --self->refcount == 0
#endif
      )
    {
      e = ___device_release_virt (self);
      ___free_mem (self);
    }

  return e;
}

___WORD ___device_cleanup
   ___P((___device *self),
        (self)
___device *self;)
{
  ___WORD e;
  ___device *devs[1];

  if (self->group == NULL)
    return ___FIX(___UNKNOWN_ERR);

  device_remove_from_group (self);

  while (1)
    {
      e = ___device_close (self, ___DIRECTION_RD);
      if (e == ___FIX(___NO_ERR))
        break;
      if (e != ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)))
        return e;

      devs[0] = self;
      e = ___device_select (devs, 1, 0, ___time_pos_infinity);
      if (e != ___FIX(___NO_ERR))
        return e;
    }

  while (1)
    {
      e = ___device_close (self, ___DIRECTION_WR);
      if (e == ___FIX(___NO_ERR))
        break;
      if (e != ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)))
        return e;

      devs[0] = self;
      e = ___device_select (devs, 0, 1, ___time_pos_infinity);
      if (e != ___FIX(___NO_ERR))
        return e;
    }

  return ___device_release (self);
}

___WORD ___device_group_setup
   ___P((___device_group **dgroup),
        (dgroup)
___device_group **dgroup;)
{
  ___WORD e;
  ___device_group *g;

  g = ___CAST(___device_group*,___alloc_mem (sizeof (___device_group)));

  if (g == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  e = user_setup ();

  if (e != ___FIX(___NO_ERR))
    {
      ___free_mem (g);
      return e;
    }

  g->list = NULL;

  *dgroup = g;

  return ___FIX(___NO_ERR);
}

___WORD ___device_group_cleanup
   ___P((___device_group *dgroup),
        (dgroup)
___device_group *dgroup;)
{
  while (dgroup->list != NULL)
    if (___device_cleanup (dgroup->list) != ___FIX(___NO_ERR))
      break;

  user_cleanup (); /* ignore error */

  ___free_mem (dgroup);

  return ___FIX(___NO_ERR);
}

/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Timer device. */

/*
 * Timer devices are not particularly useful, given that the scheduler
 * implements timeouts.  Use this as an example.
 */

___HIDDEN int device_timer_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___TIMER_KIND;
}

___HIDDEN ___WORD device_timer_select_virt
   ___P((___device *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_timer *d = ___CAST(___device_timer*,self);

  if (pass == ___SELECT_PASS_1)
    {
      if (___time_less (d->expiry, state->timeout))
        state->timeout = d->expiry;
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  if (state->select_result == 0)
    {
      if (___time_equal (d->expiry, state->timeout))
        state->devs[i] = NULL;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD device_timer_release_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD device_timer_flush_write_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD device_timer_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___device_timer_vtbl ___device_timer_table =
{
  {
    device_timer_kind,
    device_timer_select_virt,
    device_timer_release_virt,
    device_timer_flush_write_virt,
    device_timer_close_virt
  }
};

___WORD ___device_timer_setup
   ___P((___device_timer **dev,
         ___device_group *dgroup),
        (dev,
         dgroup)
___device_timer **dev;
___device_group *dgroup;)
{
  ___device_timer *d;

  d = ___CAST(___device_timer*,___alloc_mem (sizeof (___device_timer)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  d->base.vtbl = &___device_timer_table;
  d->base.refcount = 1;
  d->base.direction = ___DIRECTION_RD | ___DIRECTION_WR;
  d->base.read_stage = ___STAGE_OPEN;
  d->base.write_stage = ___STAGE_OPEN;

  d->expiry = ___time_pos_infinity;

  *dev = d;

  device_add_to_group (dgroup, &d->base);

  return ___FIX(___NO_ERR);
}

void ___device_timer_set_expiry
   ___P((___device_timer *dev,
         ___time expiry),
        (dev,
         expiry)
___device_timer *dev;
___time expiry;)
{
  dev->expiry = expiry;
}

/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Byte stream devices. */

#ifdef USE_PUMPS

#define OOB_EOS         0
#define OOB_SEEK        1
#define OOB_FLUSH_WRITE 2

typedef struct ___device_stream_pump_struct
  {
    HANDLE thread;
    nonblocking_pipe pipe;
  } ___device_stream_pump;

#define PIPE_BUFFER_SIZE 16384

/*
 * It is important that the PUMP_PRIORITY be above that of the main
 * program (which is THREAD_PRIORITY_NORMAL) so that I/O operations,
 * such as flushing output buffers, will be done as soon as the pump
 * reads it from the nonblocking pipe.  Thus the operation will appear
 * to be synchronous to the main program.
 */

#define PUMP_PRIORITY THREAD_PRIORITY_ABOVE_NORMAL

___HIDDEN ___WORD ___device_stream_pump_setup
   ___P((___device_stream_pump **pump,
         DWORD stack_size,
         LPTHREAD_START_ROUTINE proc,
         LPVOID arg),
        (pump,
         stack_size,
         proc,
         arg)
___device_stream_pump **pump;
DWORD stack_size;
LPTHREAD_START_ROUTINE proc;
LPVOID arg;)
{
  ___WORD e;
  ___device_stream_pump *p;
  HANDLE thread_handle;
  DWORD thread_id;

  p = ___CAST(___device_stream_pump*,
              ___alloc_mem (sizeof (___device_stream_pump)));

  if (p == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  e = nonblocking_pipe_setup (&p->pipe, PIPE_BUFFER_SIZE+1);

  if (e != ___FIX(___NO_ERR))
    {
      ___free_mem (p);
      return e;
    }

  *pump = p; /* set before thread created to avoid race condition */

  thread_handle =
    CreateThread (NULL,         /* no security attributes       */
                  stack_size,   /* stack size                   */
                  proc,         /* thread procedure             */
                  arg,          /* argument to thread procedure */
                  0,            /* use default creation flags   */
                  &thread_id);

  if (thread_handle == NULL ||
      !SetThreadPriority (thread_handle, PUMP_PRIORITY))
    {
      e = ___FIX(err_code_from_GetLastError ());
      nonblocking_pipe_cleanup (&p->pipe);
      ___free_mem (p);
      *pump = NULL; /* make sure caller does not think a pump was created */
      return e;
    }

  p->thread = thread_handle;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_stream_pump_reader_kill
   ___P((___device_stream_pump *pump),
        (pump)
___device_stream_pump *pump;)
{
  return nonblocking_pipe_set_reader_err
           (&pump->pipe,
            ___FIX(___KILL_PUMP));
}

___HIDDEN ___WORD ___device_stream_pump_writer_kill
   ___P((___device_stream_pump *pump),
        (pump)
___device_stream_pump *pump;)
{
  return nonblocking_pipe_set_writer_err
           (&pump->pipe,
            ___FIX(___KILL_PUMP));
}

___HIDDEN ___WORD ___device_stream_pump_wait
   ___P((___device_stream_pump *pump),
        (pump)
___device_stream_pump *pump;)
{
  DWORD code;

  code = WaitForSingleObject (pump->thread, 0);

  if (code == WAIT_FAILED)
    return ___FIX(err_code_from_GetLastError ());

  if (code == WAIT_TIMEOUT)
    return ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_stream_pump_cleanup
   ___P((___device_stream_pump *pump),
        (pump)
___device_stream_pump *pump;)
{
  CloseHandle (pump->thread); /* ignore error */
  nonblocking_pipe_cleanup (&pump->pipe); /* ignore error */
  ___free_mem (pump);

  return ___FIX(___NO_ERR);
}

#endif

___HIDDEN ___WORD ___device_stream_select_virt
   ___P((___device *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_stream *d = ___CAST(___device_stream*,self);

#ifdef USE_PUMPS

  if (for_writing)
    {
      ___device_stream_pump *p = d->write_pump;
      if (p != NULL)
        {
          HANDLE wait_obj;

          if (d->base.write_stage == ___STAGE_OPEN)
            wait_obj = p->pipe.wevent;
          else
            wait_obj = p->thread;

          device_select_add_wait_obj (state, i, wait_obj);

          return ___FIX(___SELECT_SETUP_DONE);
        }
    }
  else
    {
      ___device_stream_pump *p = d->read_pump;
      if (p != NULL)
        {
          HANDLE wait_obj;

          if (d->base.read_stage == ___STAGE_OPEN)
            wait_obj = p->pipe.revent;
          else
            wait_obj = p->thread;

          device_select_add_wait_obj (state, i, wait_obj);

          return ___FIX(___SELECT_SETUP_DONE);
        }
    }

#endif

  return ___device_stream_select_raw_virt
           (d,
            for_writing,
            i,
            pass,
            state);
}

___HIDDEN ___WORD ___device_stream_release_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  ___WORD e;
  ___device_stream *d = ___CAST(___device_stream*,self);

  e = ___device_stream_release_raw_virt (d);

#ifdef USE_PUMPS

  {
    ___device_stream_pump *p;

    p = d->read_pump;
    if (p != NULL)
      ___device_stream_pump_cleanup (p); /* ignore error */

    p = d->write_pump;
    if (p != NULL)
      ___device_stream_pump_cleanup (p); /* ignore error */
  }

#endif

  return e;
}

___HIDDEN ___WORD ___device_stream_flush_write_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  ___device_stream *d = ___CAST(___device_stream*,self);

#ifdef USE_PUMPS

  {
    ___device_stream_pump *p = d->write_pump;

    if (p != NULL)
      {
        nonblocking_pipe_oob_msg oob_msg;

        oob_msg.op = OOB_FLUSH_WRITE;

        return nonblocking_pipe_write_oob (&p->pipe, &oob_msg);
      }
  }

#endif

  return ___device_stream_flush_write_raw_virt (d);
}

___WORD ___device_stream_seek
   ___P((___device_stream *self,
         ___stream_index pos),
        (self,
         pos)
___device_stream *self;
___stream_index pos;)
{
#ifdef USE_PUMPS

  {
    ___device_stream_pump *p = self->write_pump;

    if (p != NULL)
      {
        nonblocking_pipe_oob_msg oob_msg;

        oob_msg.op = OOB_SEEK;
        oob_msg.param[0] = pos;

        return nonblocking_pipe_write_oob (&p->pipe, &oob_msg);
      }
  }

#endif

  return ___device_stream_seek_raw_virt (self, pos);
}

___WORD ___device_stream_read
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
#ifdef USE_PUMPS

  {
    ___device_stream_pump *p = self->read_pump;

    if (p != NULL)
      {
        nonblocking_pipe_oob_msg oob_msg;
        return nonblocking_pipe_read (&p->pipe, buf, len, len_done, &oob_msg);
      }
  }

#endif

  return ___device_stream_read_raw_virt (self, buf, len, len_done);
}

___WORD ___device_stream_write
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
#ifdef USE_PUMPS

  {
    ___device_stream_pump *p = self->write_pump;

    if (p != NULL)
      return nonblocking_pipe_write (&p->pipe, buf, len, len_done);
  }

#endif

  return ___device_stream_write_raw_virt (self, buf, len, len_done);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef USE_PUMPS

___HIDDEN DWORD WINAPI ___device_stream_read_pump_proc
   ___P((LPVOID param),
        (param)
LPVOID param;)
{
  ___device_stream *dev = (___device_stream*)param;
  nonblocking_pipe *p = &dev->read_pump->pipe;
  ___WORD e;
  DWORD len;
  DWORD n;
  DWORD i;
  ___U8 buf[PIPE_BUFFER_SIZE];
  nonblocking_pipe_oob_msg oob_msg;

  while (1)
    {
      /* wait until the reader needs some data */

      e = nonblocking_pipe_write_ready_wait (p);

      /* read some characters from device */

      if (e == ___FIX(___NO_ERR))
        e = ___device_stream_read_raw_virt (dev, buf, PIPE_BUFFER_SIZE, &len);

      if (e == ___FIX(___NO_ERR))
        {
          if (len == 0)
            {
              /* we reached the end-of-stream */

              oob_msg.op = OOB_EOS;

              while ((e = nonblocking_pipe_write_oob (p, &oob_msg))
                     == ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)))
                {
                  /* suspend thread until operation can be performed */

                  e = nonblocking_pipe_write_ready_wait (p);
                  if (e != ___FIX(___NO_ERR))
                    break;
                }
              if (e != ___FIX(___NO_ERR))
                break;
            }
          else
            {
              /* write to the pipe the bytes that were read */

              i = 0;

              while (i < len)
                {
                  while ((e = nonblocking_pipe_write (p, buf+i, len-i, &n))
                         == ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)))
                    {
                      /* suspend thread until operation can be performed */

                      e = nonblocking_pipe_write_ready_wait (p);
                      if (e != ___FIX(___NO_ERR))
                        break;
                    }
                  if (e != ___FIX(___NO_ERR))
                    break;
                  i += n;
                }
            }
        }

      if (e != ___FIX(___NO_ERR))
        {
          if (e == ___FIX(___KILL_PUMP)) /* terminate? */
            break;

          if (e == ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)))
            continue;

          /* report the failure back through the pipe */

          e = nonblocking_pipe_set_writer_err (p, e);

          if (e != ___FIX(___NO_ERR))
            {
              /*
               * The failure could not be reported. To avoid an
               * infinite loop the thread is terminated.
               */

              ExitThread (0);
            }
        }
    }

  device_release (&dev->base); /* ignore error */

  return 0;
}

___HIDDEN DWORD WINAPI ___device_stream_write_pump_proc
   ___P((LPVOID param),
        (param)
LPVOID param;)
{
  ___device_stream *dev = (___device_stream*)param;
  nonblocking_pipe *p = &dev->write_pump->pipe;
  ___WORD e;
  DWORD len;
  DWORD n;
  DWORD i;
  ___U8 buf[PIPE_BUFFER_SIZE];
  nonblocking_pipe_oob_msg oob_msg;

  while (1)
    {
      /* get from the pipe some bytes to write to the device */

      while ((e = nonblocking_pipe_read (p, buf, PIPE_BUFFER_SIZE, &len, &oob_msg))
             == ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)))
        {
          /* suspend thread until operation can be performed */

          e = nonblocking_pipe_read_ready_wait (p);
          if (e != ___FIX(___NO_ERR))
            break;
        }

      if (e == ___FIX(___NO_ERR))
        {
          if (len > 0)
            {
              /* write to the device the bytes that were read from the pipe */

              i = 0;

              while (i < len)
                {
                  e = ___device_stream_write_raw_virt (dev, buf+i, len-i, &n);
                  if (e != ___FIX(___NO_ERR))
                    break;
                  i += n;
                }
            }
          else
            {
              switch (oob_msg.op)
                {
                case OOB_FLUSH_WRITE:
                  e = ___device_stream_flush_write_raw_virt (dev);
                  break;
                case OOB_SEEK:
                  fprintf (stderr,"***** got OOB_SEEK %d\n", oob_msg.param[0]);fflush (stderr);
                  e = ___device_stream_seek_raw_virt (dev, oob_msg.param[0]);
                  break;
                case OOB_EOS:
                  fprintf (stderr,"***** got OOB_EOS\n");fflush (stderr);
                  break;
                }
            }
        }

      if (e != ___FIX(___NO_ERR))
        {
          if (err == ___FIX(___KILL_PUMP)) /* terminate? */
            break;

          /* report the failure back through the pipe */

          e = nonblocking_pipe_set_reader_err (p, e);

          if (e != ___FIX(___NO_ERR))
            {
              /*
               * The failure could not be reported. To avoid an
               * infinite loop the thread is terminated.
               */

              ExitThread (0);
            }
        }
    }

  device_release (&dev->base); /* ignore error */

  return 0;
}

#endif

___WORD ___device_stream_setup
   ___P((___device_stream *dev,
         ___device_group *dgroup,
         int direction,
         int pumps_on),
        (dev,
         dgroup,
         direction,
         pumps_on)/*********************/
___device_stream *dev;
___device_group *dgroup;
int direction;
int pumps_on;)
{
  ___WORD e;/*********unused*/

  dev->base.refcount = 1;
  dev->base.direction = direction;
  dev->base.read_stage = ___STAGE_CLOSED;
  dev->base.write_stage = ___STAGE_CLOSED;

#ifdef USE_PUMPS
  dev->read_pump = NULL;
  dev->write_pump = NULL;
#endif

  device_add_to_group (dgroup, &dev->base);

  if (direction & ___DIRECTION_RD)
    {
      dev->base.read_stage = ___STAGE_OPEN;

#ifdef USE_PUMPS

      if (pumps_on & ___DIRECTION_RD)
        {
          device_add_ref (&dev->base);

          e = ___device_stream_pump_setup
                (&dev->read_pump,
                 65536,/********************/
                 ___device_stream_read_pump_proc,
                 dev);

          if (e != ___FIX(___NO_ERR))
            {
              device_release (&dev->base); /* ignore error */
              device_cleanup (&dev->base); /* ignore error */
              return e;
            }
        }

#endif
    }

  if (direction & ___DIRECTION_WR)
    {
      dev->base.write_stage = ___STAGE_OPEN;

#ifdef USE_PUMPS

      if (pumps_on & ___DIRECTION_WR)
        {
          device_add_ref (&dev->base);

          e = ___device_stream_pump_setup
                (&dev->write_pump,
                 65536,/********************/
                 ___device_stream_write_pump_proc,
                 dev);

          if (e != ___FIX(___NO_ERR))
            {
              device_release (&dev->base); /* ignore error */
              device_cleanup (&dev->base); /* ignore error */
              return e;
            }
        }

#endif
    }

  return ___FIX(___NO_ERR);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef USE_WIN32

/* Console stream device. */

typedef struct ___device_console_struct
  {
    ___device_stream base;
    HANDLE h;
  } ___device_console;

typedef struct ___device_console_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_console_vtbl;

___HIDDEN int ___device_console_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___CONSOLE_DEVICE_KIND;
}

___HIDDEN ___WORD ___device_console_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___WORD e;
  ___device_console *d = (___device_console*)self;
  ___device_stream_pump *p = d->base.read_pump;

  if (p != NULL)
    {
      INPUT_RECORD ir;
      DWORD len;

      d->base.base.read_stage = ___STAGE_CLOSING1;

      /* Terminate the reader pump cleanly. */

      e = ___device_stream_pump_reader_kill (p);

      if (e != ___FIX(___NO_ERR))
        return e;

      while (1)
        {
          /*
           * The reader pump might be blocked trying to read from the
           * console.  To unblock the reader pump we simulate the user
           * typing a "return" at the console.
           */

          ir.EventType = KEY_EVENT;
          ir.Event.KeyEvent.bKeyDown          = 1;
          ir.Event.KeyEvent.wRepeatCount      = 1;
          ir.Event.KeyEvent.wVirtualKeyCode   = 0;
          ir.Event.KeyEvent.wVirtualScanCode  = 0;
#ifdef UNICODE
          ir.Event.KeyEvent.uChar.UnicodeChar = 13; /* "return" */
#else
          ir.Event.KeyEvent.uChar.AsciiChar   = 13; /* "return" */
#endif
          ir.Event.KeyEvent.dwControlKeyState = 0;

          if (!WriteConsoleInput (d->h, &ir, 1, &len))
            return err_code_from_GetLastError ();

          if (len == 1)
            break;

          /*
           * It is possible that len != 1 if the input buffer is full
           * when WriteConsoleInput is called, so we need to empty the
           * console input buffer and try again.
           */

          if (!FlushConsoleInputBuffer (d->h))
            return ___FIX(err_code_from_GetLastError ());
        }

      /* Wait for the pump to terminate. */

      e = ___device_stream_pump_wait (p);

      if (e != ___FIX(___NO_ERR))
        return e;

      /* Don't leave anything in the console input buffer. */

      FlushConsoleInputBuffer (d->h); /* ignore error */
    }

  d->base.base.read_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_console_close_write_virt.......
   ___P((___device *self),
        (self)
___device *self;)
{
  ___WORD e;
  ___device_console *d = (___device_console*)self;
  ___device_stream_pump *p = d->base.write_pump;

  if (p != NULL)
    {
      d->base.base.write_stage = ___STAGE_CLOSING1;

      /* Terminate the writer pump cleanly. */

      e = ___device_stream_pump_writer_kill (p);

      if (e != ___FIX(___NO_ERR))
        return e;

      /*
       * The writer pump can't be blocked trying to write to the
       * console because writes to the console are synchronous.
       */

      /* Wait for the pump to terminate. */

      e = ___device_stream_pump_wait (p);

      if (e != ___FIX(___NO_ERR))
        return e;
    }

  d->base.base.write_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_console_select_raw_virt
   ___P((___device_stream *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device_stream *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  if ((for_writing ? d->base.write_stage : d->base.read_stage)
      != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (pass == ___SELECT_PASS_1)
    {
      device_select_add_wait_obj (state, i, um.always_signaled);
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  check!

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_console_release_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_console_flush_write_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_console_seek_raw_virt
   ___P((___device_stream *self,
         ___stream_index pos),
        (self,
         pos)
___device_stream *self;
___stream_index pos;)
{
  return ___FIX(___UNKNOWN_ERR);
}

___HIDDEN ___WORD ___device_console_read_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_console *d = (___device_console*)self;
  DWORD chars_len;      /* number of characters to transfer */
  DWORD chars_len_done; /* number of characters transfered */

  if (d->base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  chars_len = len / sizeof (TCHAR);

  if (chars_len < 1)
    return ___FIX(___UNKNOWN_ERR);

  if (!ReadConsole (d->h, buf, chars_len, &chars_len_done, NULL))
    return ___FIX(err_code_from_GetLastError ());

  *len_done = chars_len_done * sizeof (TCHAR);

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_console_write_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_console *d = (___device_console*)self;
  DWORD chars_len;      /* number of characters to transfer */
  DWORD chars_len_done; /* number of characters transfered */

  if (d->base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  chars_len = len / sizeof (TCHAR);

  if (chars_len < 1)
    return ___UNKNOWN_ERR;

  if (!WriteConsole (d->h, buf, chars_len, &chars_len_done, NULL))
    return ___FIX(err_code_from_GetLastError ());

  *len_done = chars_len_done * sizeof (TCHAR);

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_console_width_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(80);
}

___HIDDEN ___WORD ___device_console_default_options_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___STREAM_DEFAULT_OPTIONS(___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___NO_BUFFERING,
                                   ___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___NO_BUFFERING);
}

___HIDDEN ___device_console_vtbl ___device_console_table =
{
  {
    {
      ___device_console_kind,
      ___device_stream_select_virt,
      ___device_stream_release_virt,
      ___device_stream_flush_write_virt,
      ___device_console_close_virt
    },
    ___device_console_select_raw_virt,
    ___device_console_release_raw_virt,
    ___device_console_flush_write_raw_virt,
    ___device_console_seek_raw_virt,
    ___device_console_read_raw_virt,
    ___device_console_write_raw_virt,
    ___device_console_width_virt,
    ___device_console_default_options_virt
  }
};

___WORD ___device_console_setup_from_handle
   ___P((___device_console **dev,
         ___device_group *dgroup,
         HANDLE h,
         int direction),
        (dev,
         dgroup,
         h,
         direction)
___device_console **dev;
___device_group *dgroup;
HANDLE h;
int direction;)
{
  ___device_console *d;

  d = ___CAST(___device_console*,___alloc_mem (sizeof (___device_console)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  d->base.base.vtbl = &___device_console_table;
  d->h = h;

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            ___DIRECTION_RD);
}

#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef USE_WIN32

/* Serial stream device */

typedef struct ___device_serial_struct
  {
    ___device_stream base;
    HANDLE h;
  } ___device_serial;

typedef struct ___device_serial_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_serial_vtbl;

___HIDDEN int ___device_serial_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___SERIAL_DEVICE_KIND;
}

___HIDDEN ___WORD ___device_serial_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___WORD e;
  ___device_serial *d = (___device_serial*)self;
  ___device_stream_pump *p = d->base.read_pump;

  if (p != NULL)
    {
      d->base.base.read_stage = ___STAGE_CLOSING1;

      /* Terminate the reader pump cleanly. */

      e = ___device_stream_pump_reader_kill (p);

      if (e != ___FIX(___NO_ERR))
        return e;

      /*
       * The reader pump might be blocked trying to read from the
       * serial device.  The call to PurgeComm will resume the reader
       * pump.
       */

      if (!PurgeComm (d->h, PURGE_RXABORT))
        return ___FIX(err_code_from_GetLastError ());

      /* Wait for the pump to terminate. */

      e = ___device_stream_pump_wait (p);

      if (e != ___FIX(___NO_ERR))
        return e;
    }

  /* Close handle when other side is also closed. */

  if (d->base.base.write_stage == ___STAGE_CLOSED)
    if (!CloseHandle (d->h))
      return ___FIX(err_code_from_GetLastError ());

  d->base.base.read_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_serial_close_write_virt...............
   ___P((___device *self),
        (self)
___device *self;)
{
  ___WORD e;
  ___device_serial *d = (___device_serial*)self;
  ___device_stream_pump *p = d->base.write_pump;

  if (p != NULL)
    {
      d->base.base.write_stage = ___STAGE_CLOSING1;

      /* Terminate the writer pump cleanly. */

      e = ___device_stream_pump_writer_kill (p);

      if (e != ___FIX(___NO_ERR))
        return e;

      /*
       * The writer pump might be blocked trying to write to the
       * serial device but eventually it will resume when the serial
       * buffer is sufficiently empty.
       */

#if 0
/***********************/
      /* Speed things up by calling PurgeComm. */

      if (!PurgeComm (d->h, PURGE_TXABORT))
        return ___FIX(err_code_from_GetLastError ());
#endif

      /* Wait for the pump to terminate. */

      e = ___device_stream_pump_wait (p);

      if (e != ___FIX(___NO_ERR))
        return e;
    }

  /* Close handle when other side is also closed. */

  if (d->base.base.read_stage == ___STAGE_CLOSED)
    if (!CloseHandle (d->h))
      return ___FIX(err_code_from_GetLastError ());

  d->base.base.write_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_serial_select_raw_virt
   ___P((___device_stream *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device_stream *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  if ((for_writing ? d->base.write_stage : d->base.read_stage)
      != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (pass == ___SELECT_PASS_1)
    {
      device_select_add_wait_obj (state, i, um.always_signaled);
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  check!

    return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_serial_release_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_serial_flush_write_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  ___device_serial *d = (___device_serial*)self;

  if (d->base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (!FlushFileBuffers (d->h))
    return ___FIX(err_code_from_GetLastError ());

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_serial_seek_raw_virt
   ___P((___device_stream *self,
         ___stream_index pos),
        (self,
         pos)
___device_stream *self;
___stream_index pos;)
{
  return ___FIX(___UNKNOWN_ERR);
}

___HIDDEN ___WORD ___device_serial_read_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_serial *d = (___device_serial*)self;

  if (d->base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (!ReadFile (d->h, buf, len, len_done, NULL))
    return ___FIX(err_code_from_GetLastError ());

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_serial_write_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_serial *d = (___device_serial*)self;

  if (d->base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (!WriteFile (d->h, buf, len, len_done, NULL))
    {
      DWORD err = GetLastError ();

      /*
       * Even though WriteFile has reported a failure, the operation
       * was executed correctly (i.e. len_done contains the number of
       * bytes written) if GetLastError returns ERROR_SUCCESS.
       */

      if (err != ERROR_SUCCESS)
        return ___FIX(___ERR_CODE_FROM_HRESULT(err));
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_serial_width_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(80);
}

___HIDDEN ___WORD ___device_serial_default_options_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___STREAM_DEFAULT_OPTIONS(___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING,
                                   ___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING);
}

___HIDDEN ___device_serial_vtbl ___device_serial_table =
{
  {
    {
      ___device_serial_kind,
      ___device_stream_select_virt,
      ___device_stream_release_virt,
      ___device_stream_flush_write_virt,
      ___device_serial_close_virt
    },
    ___device_serial_select_raw_virt,
    ___device_serial_release_raw_virt,
    ___device_serial_flush_write_raw_virt,
    ___device_serial_seek_raw_virt,
    ___device_serial_read_raw_virt,
    ___device_serial_write_raw_virt,
    ___device_serial_width_virt,
    ___device_serial_default_options_virt
  }
};

___HIDDEN ___WORD device_set_comm_state
   ___P((___device *dev,
         LPCTSTR def),
        (dev,
         def)
___device *dev;
LPCTSTR def;)
{
  ___device_serial *d;
  DCB dcb;

  if (dev->vtbl != (void*)&___device_serial_table)
    return ___FIX(___UNKNOWN_ERR);

  d = (___device_serial*)dev;

  FillMemory (&dcb, sizeof (dcb), 0);

  if (!GetCommState (d->h, &dcb) ||
      !BuildCommDCB (def, &dcb) ||
      !SetCommState (d->h, &dcb))
    return ___FIX(err_code_from_GetLastError ());

  return ___FIX(___NO_ERR);
}

___WORD ___device_serial_setup_from_handle
   ___P((___device_serial **dev,
         ___device_group *dgroup,
         HANDLE h,
         int direction),
        (dev,
         dgroup,
         h,
         direction)
___device_serial **dev;
___device_group *dgroup;
HANDLE h;
int direction;)
{
  ___WORD e;
  ___device_serial *d;
  COMMTIMEOUTS cto;

  d = ___CAST(___device_serial*,___alloc_mem (sizeof (___device_serial)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  /*
   * Setup serial device so that ReadFile will return as soon as a
   * character is available.
   */

  cto.ReadIntervalTimeout         = MAXDWORD;
  cto.ReadTotalTimeoutMultiplier  = MAXDWORD;
  cto.ReadTotalTimeoutConstant    = 1;
  cto.WriteTotalTimeoutMultiplier = 0;
  cto.WriteTotalTimeoutConstant   = 0;

  if (!SetCommTimeouts (h, &cto)
#ifdef BIG_SERIAL_BUFFERS
      || !SetupComm (h, 65536, 65536))
#endif
     )
    {
      e = ___FIX(err_code_from_GetLastError ());
      ___free_mem (&d->base.base);
      return e;
    }

  d->base.base.vtbl = &___device_serial_table;
  d->h = h;

  e = device_set_comm_state (&d->base.base, _T("")); /* will set default */

  if (e != ___FIX(___NO_ERR))
    {
      ___free_mem (&d->base.base);
      return e;
    }

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            ___DIRECTION_RD|___DIRECTION_WR);
}

#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef USE_WIN32

/* Pipe stream device */

typedef struct ___device_pipe_struct
  {
    ___device_stream base;
    HANDLE h;
  } ___device_pipe;

typedef struct ___device_pipe_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_pipe_vtbl;

___HIDDEN int ___device_pipe_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___PIPE_DEVICE_KIND;
}

___HIDDEN ___WORD ___device_pipe_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___WORD e;
  ___device_pipe *d = (___device_pipe*)self;
  ___device_stream_pump *p = d->base.read_pump;

  if (p != NULL)
    {
      d->base.base.read_stage = ___STAGE_CLOSING1;

      /* Terminate the reader pump cleanly. */

      /*
       * The reader pump might be blocked trying to read from the
       * pipe.  Unfortunately there is no way to force the read
       * operation to return.  We can't kill the pump with
       * TerminateThread because this might leave some system
       * resources in an inconsistent state.  We simply leave the pump
       * as is so that when the read operation completes (because some
       * data is written to the other end of the pipe or the pipe is
       * closed) the pump will cleanup the resources associated with
       * the pipe and terminate.
       */

      e = ___device_stream_pump_reader_kill (p);

      if (e != ___FIX(___NO_ERR))
        return e;
    }

  d->base.base.read_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_pipe_close_write_virt...............
   ___P((___device *self),
        (self)
___device *self;)
{
  ___WORD e;
  ___device_pipe *d = (___device_pipe*)self;
  ___device_stream_pump *p = d->base.write_pump;

  if (p != NULL)
    {
      d->base.base.write_stage = ___STAGE_CLOSING1;

      /* Terminate the writer pump cleanly. */

      /*
       * The writer pump might be blocked trying to write to the pipe.
       * Unfortunately there is no way to force the write operation to
       * return.  We can't kill the pump with TerminateThread because
       * this might leave some system resources in an inconsistent
       * state.  We simply leave the pump as is so that when the write
       * operation completes (because some data is read from the other
       * end of the pipe or the pipe is closed) the pump will cleanup
       * the resources associated with the pipe and terminate.
       */

      e = ___device_stream_pump_writer_kill (p);

      if (e != ___FIX(___NO_ERR))
        return e;
    }

  d->base.base.write_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_pipe_select_raw_virt
   ___P((___device_stream *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device_stream *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  if ((for_writing ? d->base.write_stage : d->base.read_stage)
      != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (pass == ___SELECT_PASS_1)
    {
      device_select_add_wait_obj (state, i, um.always_signaled);
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  check!

    return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_pipe_release_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  ___device_pipe *d = (___device_pipe*)self;

  /* Close handle. */

#if 0
  if (!CloseHandle (d->h))
    return ___FIX(err_code_from_GetLastError ());
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_pipe_flush_write_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_pipe_seek_raw_virt
   ___P((___device_stream *self,
         ___stream_index pos),
        (self,
         pos)
___device_stream *self;
___stream_index pos;)
{
  return ___FIX(___UNKNOWN_ERR);
}

___HIDDEN ___WORD ___device_pipe_read_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_pipe *d = (___device_pipe*)self;

  if (d->base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (!ReadFile (d->h, buf, len, len_done, NULL))
    {
      ___WORD = ___FIX(err_code_from_GetLastError ());

      if (e == ___FIX(___ERR_CODE_FROM_HRESULT(ERROR_BROKEN_PIPE)))
        {
          *len_done = 0;
          return ___FIX(___NO_ERR);
        }

      return e;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_pipe_write_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_pipe *d = (___device_pipe*)self;

  if (d->base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (!WriteFile (d->h, buf, len, len_done, NULL))
    return ___FIX(err_code_from_GetLastError ());

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_pipe_width_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(80);
}

___HIDDEN ___WORD ___device_pipe_default_options_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___STREAM_DEFAULT_OPTIONS(___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING,
                                   ___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING);
}

___HIDDEN ___device_pipe_vtbl ___device_pipe_table =
{
  {
    {
      ___device_pipe_kind,
      ___device_stream_select_virt,
      ___device_stream_release_virt,
      ___device_stream_flush_write_virt,
      ___device_pipe_close_virt
    },
    ___device_pipe_select_raw_virt,
    ___device_pipe_release_raw_virt,
    ___device_pipe_flush_write_raw_virt,
    ___device_pipe_seek_raw_virt,
    ___device_pipe_read_raw_virt,
    ___device_pipe_write_raw_virt,
    ___device_pipe_width_virt,
    ___device_pipe_default_options_virt
  }
};

___WORD ___device_pipe_setup_from_handle
   ___P((___device_pipe **dev,
         ___device_group *dgroup,
         HANDLE h,
         int direction),
        (dev,
         dgroup,
         h,
         direction)
___device_pipe **dev;
___device_group *dgroup;
HANDLE h;
int direction;)
{
  ___device_pipe *d;

  d = ___CAST(___device_pipe*,___alloc_mem (sizeof (___device_pipe)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  d->base.base.vtbl = &___device_pipe_table;
  d->h = h;

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            ___DIRECTION_RD|___DIRECTION_WR);
}

#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef USE_WIN32

/* File stream device */

typedef struct ___device_file_struct
  {
    ___device_stream base;
    HANDLE h;
  } ___device_file;

typedef struct ___device_file_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_file_vtbl;

___HIDDEN int ___device_file_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FILE_DEVICE_KIND;
}

___HIDDEN ___WORD ___device_file_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___WORD e;
  ___device_file *d = (___device_file*)self;
  ___device_stream_pump *p = d->base.read_pump;

  if (p != NULL)
    {
      d->base.base.read_stage = ___STAGE_CLOSING1;

      /* Terminate the reader pump cleanly. */

      e = ___device_stream_pump_reader_kill (p);

      if (e != ___FIX(___NO_ERR))
        return e;

      /*
       * The reader pump might be blocked trying to read from the file
       * (because the disk is seeking to the right track or is
       * spinning up or the user has removed the floppy containing the
       * file, etc) but we can expect the operation to return
       * eventually, and the pump will terminate.
       */

      /* Wait for the pump to terminate. */

      e = ___device_stream_pump_wait (p);

      if (e != ___FIX(___NO_ERR))
        return e;
    }

  /* Close handle when other side is also closed. */

  if (d->base.base.write_stage == ___STAGE_CLOSED)
    if (!CloseHandle (d->h))
      return ___FIX(err_code_from_GetLastError ());

  d->base.base.read_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_close_write_virt.................
   ___P((___device *self),
        (self)
___device *self;)
{
  ___WORD e;
  ___device_file *d = (___device_file*)self;
  ___device_stream_pump *p = d->base.write_pump;

  if (p != NULL)
    {
      d->base.base.write_stage = ___STAGE_CLOSING1;

      /* Terminate the writer pump cleanly. */

      e = ___device_stream_pump_writer_kill (p);

      if (e != ___FIX(___NO_ERR))
        return e;

      /*
       * The writer pump might be blocked trying to write to the file
       * (because the disk is seeking to the right track or is
       * spinning up or the user has removed the floppy containing the
       * file, etc) but we can expect the operation to return
       * eventually, and the pump will terminate.
       */

      /* Wait for the pump to terminate. */

      e = ___device_stream_pump_wait (p);

      if (e != ___FIX(___NO_ERR))
        return e;
    }

  /* Close handle when other side is also closed. */

  if (d->base.base.read_stage == ___STAGE_CLOSED)
    if (!CloseHandle (d->h))
      return ___FIX(err_code_from_GetLastError ());

  d->base.base.write_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_select_raw_virt
   ___P((___device_stream *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device_stream *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  if ((for_writing ? d->base.write_stage : d->base.read_stage)
      != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (pass == ___SELECT_PASS_1)
    {
      device_select_add_wait_obj (state, i, um.always_signaled);
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  check!

    return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_release_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_flush_write_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  ___device_file *d = (___device_file*)self;

  if (d->base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (!FlushFileBuffers (d->h))
    return ___FIX(err_code_from_GetLastError ());

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_seek_raw_virt
   ___P((___device_stream *self,
         ___stream_index pos),
        (self,
         pos)
___device_stream *self;
___stream_index pos;)
{
  return ___FIX(___UNKNOWN_ERR);
}

___HIDDEN ___WORD ___device_file_read_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_file *d = (___device_file*)self;

  if (d->base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (!ReadFile (d->h, buf, len, len_done, NULL))
    return ___FIX(err_code_from_GetLastError ());

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_write_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_file *d = (___device_file*)self;

  if (d->base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (!WriteFile (d->h, buf, len, len_done, NULL))
    return ___FIX(err_code_from_GetLastError ());

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_width_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(80);
}

___HIDDEN ___WORD ___device_file_default_options_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___STREAM_DEFAULT_OPTIONS(___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING,
                                   ___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING);
}

___HIDDEN ___device_file_vtbl ___device_file_table =
{
  {
    {
      ___device_file_kind,
      ___device_stream_select_virt,
      ___device_stream_release_virt,
      ___device_stream_flush_write_virt,
      ___device_file_close_virt
    },
    ___device_file_select_raw_virt,
    ___device_file_release_raw_virt,
    ___device_file_flush_write_raw_virt,
    ___device_file_seek_raw_virt,
    ___device_file_read_raw_virt,
    ___device_file_write_raw_virt,
    ___device_file_width_virt,
    ___device_file_default_options_virt
  }
};

___HIDDEN ___WORD ___device_file_setup_from_handle
   ___P((___device_file **dev,
         ___device_group *dgroup,
         HANDLE h,
         int direction),
        (dev,
         dgroup,
         h,
         direction)
___device_file **dev;
___device_group *dgroup;
HANDLE h;
int direction;)
{
  ___device_file *d;

  d = ___CAST(___device_file*,___alloc_mem (sizeof (___device_file)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  d->base.base.vtbl = &___device_file_table;
  d->h = h;

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            ___DIRECTION_RD|___DIRECTION_WR);
}

#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef USE_WIN32

/* TCP client stream device */

/*
  int err;
  int len;
  int val;

  len = sizeof (val);
  err = getsockopt (___CAST(SOCKET,h), SOL_SOCKET, SO_TYPE, &val, &len);
  if (err == 0)
    return ___DEVICE_STREAM_SOCKET;
*/

typedef struct ___device_tcp_client_struct
  {
    ___device_stream base;
    SOCKET s;
    WSAEVENT event;   /* for implementing "___device_select" */
    long notify;      /* for implementing "___device_select" */
  } ___device_tcp_client;

typedef struct ___device_tcp_client_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_tcp_client_vtbl;

___HIDDEN int ___device_tcp_client_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___TCP_CLIENT_DEVICE_KIND;
}

___HIDDEN ___WORD ___device_tcp_client_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___device_tcp_client *d = (___device_tcp_client*)self;

  d->base.base.read_stage = ___STAGE_CLOSING1;

  /* Shutdown receiving side. */

  if (shutdown (d->s, SD_RECEIVE) == SOCKET_ERROR)
    return ___FIX(err_code_from_WSAGetLastError ());

  /* Close socket when other side is also closed. */

  if (d->base.base.write_stage == ___STAGE_CLOSED)
    if (closesocket (d->s) == SOCKET_ERROR)
      {
        ___WORD e = ___FIX(err_code_from_WSAGetLastError ());

        if (e == ___FIX(___ERR_CODE_FROM_WSAERROR(WSAEWOULDBLOCK)))
          return ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));

        return e;
      }

  d->base.base.read_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_close_write_virt.............
   ___P((___device *self),
        (self)
___device *self;)
{
  ___device_tcp_client *d = (___device_tcp_client*)self;

  d->base.base.write_stage = ___STAGE_CLOSING1;

  /* Shutdown sending side. */

  if (shutdown (d->s, SD_SEND) == SOCKET_ERROR)
    return ___FIX(err_code_from_WSAGetLastError ());

  /* Close socket when other side is also closed. */

  if (d->base.base.read_stage == ___STAGE_CLOSED)
    if (closesocket (d->s) == SOCKET_ERROR)
      {
        ___WORD e = ___FIX(err_code_from_WSAGetLastError ());

        if (e == ___FIX(___ERR_CODE_FROM_WSAERROR(WSAEWOULDBLOCK)))
          return ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));

        return e;
      }

  d->base.base.write_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_select_raw_virt
   ___P((___device_stream *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device_stream *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_tcp_client *d = (___device_tcp_client*)self;

  if ((for_writing ? d->base.write_stage : d->base.read_stage)
      != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  switch (pass)
    {
    case ___SELECT_PASS_1:
      if (d->base.base.select_step == 0)
        {
          if (!WSAResetEvent (d->event))
            return ___FIX(err_code_from_WSAGetLastError ());
          d->notify = 0;
          d->base.base.select_step = 1;
        }
      if (for_writing)
        {
          if (d->base.base.write_stage == ___STAGE_OPEN)
            d->notify |= FD_WRITE;
        }
      else
        {
          if (d->base.base.read_stage == ___STAGE_OPEN)
            d->notify |= FD_READ | FD_CLOSE;
        }
      return ___FIX(___NO_ERR);

    case ___SELECT_PASS_2:
      if (WSAEventSelect (d->s, d->event, d->notify) == SOCKET_ERROR)
        return ___FIX(err_code_from_WSAGetLastError ());
      return ___FIX(___NO_ERR);

    case ___SELECT_PASS_3:
      device_select_add_wait_obj (state, i, d->event);
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  check!

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_release_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  ___device_tcp_client *d = (___device_tcp_client*)self;

  /* Close event. */

  if (!CloseHandle (d->event))
    return ___FIX(err_code_from_GetLastError ());

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_flush_write_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_seek_raw_virt
   ___P((___device_stream *self,
         ___stream_index pos),
        (self,
         pos)
___device_stream *self;
___stream_index pos;)
{
  return ___FIX(___UNKNOWN_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_read_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_tcp_client *d = (___device_tcp_client*)self;
  int n;

  if (d->base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if ((n = recv (d->s, buf, len, 0)) == SOCKET_ERROR)
    {
      ___WORD e = ___FIX(err_code_from_WSAGetLastError ());

      if (e == ___FIX(___ERR_CODE_FROM_WSAERROR(WSAEWOULDBLOCK)))
        return ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));

      return e;
    }

  *len_done = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_write_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_tcp_client *d = (___device_tcp_client*)self;
  int n;

  if (d->base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if ((n = send (d->s, buf, len, 0)) == SOCKET_ERROR)
    {
      ___WORD e = ___FIX(err_code_from_WSAGetLastError ());

      if (e == ___FIX(___ERR_CODE_FROM_WSAERROR(WSAEWOULDBLOCK)))
        return ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));

      return e;
    }

  *len_done = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_width_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(80);
}

___HIDDEN ___WORD ___device_tcp_client_default_options_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___STREAM_DEFAULT_OPTIONS(___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING,
                                   ___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING);
}

___HIDDEN ___device_tcp_client_vtbl ___device_tcp_client_table =
{
  {
    {
      ___device_tcp_client_kind,
      ___device_stream_select_virt,
      ___device_stream_release_virt,
      ___device_stream_flush_write_virt,
      ___device_tcp_client_close_virt
    },
    ___device_tcp_client_select_raw_virt,
    ___device_tcp_client_release_raw_virt,
    ___device_tcp_client_flush_write_raw_virt,
    ___device_tcp_client_seek_raw_virt,
    ___device_tcp_client_read_raw_virt,
    ___device_tcp_client_write_raw_virt,
    ___device_tcp_client_width_virt,
    ___device_tcp_client_default_options_virt
  }
};

___WORD ___device_tcp_client_setup_from_socket
   ___P((___device_tcp_client **dev,
         ___device_group *dgroup,
         SOCKET s,
         int direction),
        (dev,
         dgroup,
         s,
         direction)
___device_tcp_client **dev;
___device_group *dgroup;
SOCKET s;
int direction;)
{
  ___device_tcp_client *d;
  WSAEVENT event;
  unsigned long flag;

  d = ___CAST(___device_tcp_client*,___alloc_mem (sizeof (___device_tcp_client)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  /*
   * Setup socket to perform nonblocking I/O.
   */

  flag = 1; /* request nonblocking I/O */

  if (ioctlsocket (s, FIONBIO, &flag) == SOCKET_ERROR ||
      (event = WSACreateEvent ()) == WSA_INVALID_EVENT)
    {
      ___WORD e = ___FIX(err_code_from_WSAGetLastError ());
      ___free_mem (&d->base.base);
      return e;
    }

  d->base.base.vtbl = &___device_tcp_client_table;
  d->s = s;
  d->event = event;

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            0);
}

#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef USE_WIN32

___HIDDEN int ___device_stream_kind_from_handle
   ___P((HANDLE h),
        (h)
HANDLE h;)
{
  DWORD n;
  CONSOLE_CURSOR_INFO cinfo;
  DCB dcb;
  BY_HANDLE_FILE_INFORMATION finfo;

  if (GetNumberOfConsoleInputEvents (h, &n))
    return ___CONSOLE_DEVICE_KIND;

  if (GetConsoleCursorInfo (h, &cinfo))
    return ___CONSOLE_DEVICE_KIND;

  if (GetCommState (h, &dcb))
    return ___SERIAL_DEVICE_KIND;

  if (GetFileType (h) == FILE_TYPE_PIPE)
    return ___PIPE_DEVICE_KIND;

  if (GetFileInformationByHandle (h, &finfo))
    return ___FILE_DEVICE_KIND;

  return ___NONE_KIND;
}

___HIDDEN int ___device_stream_direction_from_handle
   ___P((HANDLE h),
        (h)
HANDLE h;)
{
  DWORD n;
  int direction = 0;

  if (ReadFile (h, NULL, 0, &n, NULL))
    direction |= ___DIRECTION_RD;

  if (WriteFile (h, NULL, 0, &n, NULL))
    direction |= ___DIRECTION_WR;

  return direction;
}

___WORD ___device_stream_setup_from_handle
   ___P((___device_stream **dev,
         ___device_group *dgroup,
         HANDLE h,
         int kind,
         int direction),
        (dev,
         dgroup,
         h,
         kind,
         direction)
___device_stream **dev;
___device_group *dgroup;
HANDLE h;
int kind;
int direction;)
{
  ___WORD e = ___FIX(___UNKNOWN_ERR);

  if (kind == ___NONE_KIND)
    kind = ___device_stream_kind_from_handle (h);

  if (direction == 0)
    direction = ___device_stream_direction_from_handle (h);

  switch (kind)
    {
    case ___CONSOLE_DEVICE_KIND:
      {
        ___device_console *d;
        e = ___device_console_setup_from_handle (&d, dgroup, h, direction);
        *dev = ___CAST(___device_stream*,d);
        break;
      }

    case ___SERIAL_DEVICE_KIND:
      {
        ___device_serial *d;
        e = ___device_serial_setup_from_handle (&d, dgroup, h, direction);
        *dev = ___CAST(___device_stream*,d);
        break;
      }

    case ___PIPE_DEVICE_KIND:
      {
        ___device_pipe *d;
        e = ___device_pipe_setup_from_handle (&d, dgroup, h, direction);
        *dev = ___CAST(___device_stream*,d);
        break;
      }

    case ___FILE_DEVICE_KIND:
      {
        ___device_file *d;
        e = ___device_file_setup_from_handle (&d, dgroup, h, direction);
        *dev = ___CAST(___device_stream*,d);
        break;
      }
    }

  return e;
}

___WORD ___device_stream_setup_from_path
   ___P((___device_stream **dev,
         ___device_group *dgroup,
         LPCTSTR path,
         int flags,
         int mode),
        (dev,
         dgroup,
         path,
         flags,
         mode)
___device_stream **dev;
___device_group *dgroup;
LPCTSTR path;
int flags;
int mode;)
{
  HANDLE h;
  ??? fl;
  int direction;

  switch ((flags >> 4) & 3)
    {
    default:
    case 1:
      fl = GENERIC_READ;
      direction = ___DIRECTION_RD;
      break;
    case 2:
      fl = GENERIC_WRITE;
      direction = ___DIRECTION_WR;
      break;
    case 3:
      fl = GENERIC_READ|GENERIC_WRITE;
      direction = ___DIRECTION_RD|___DIRECTION_WR;
      break;
    }

  if (flags & 16)
    fl |= O_APPEND;

  switch ((flags >> 1) & 3)
    {
    default:
    case 0: fl |= OPEN_EXISTING; break;
    case 1: fl |= OPEN_ALWAYS; break;
    case 2: fl |= ???; break;
    }

  if (flags & 1)
    fl |= O_TRUNC???;

  fl |= O_NONBLOCK???;


if ((dfile = CreateFile(DataFile, GENERIC_READ
                        | GENERIC_WRITE, FILE_SHARE_READ, NULL,
                        OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL)) ==
    INVALID_HANDLE_VALUE)

  h = CreateFile (path,
                  fl,
                  0,             /* exclusive access mode */
                  NULL,          /* no security attributes */
                  OPEN_EXISTING, /* comm devices must use OPEN_EXISTING */
                  0,             /* no overlapped I/O */
                  NULL);

  if (h == INVALID_HANDLE_VALUE)
    return ___FIX(err_code_from_GetLastError ());

  return ___device_stream_setup_from_handle
           (dev,
            dgroup,
            h,
            ___NONE_KIND,
            direction);
}

#endif


#ifdef USE_TERMCAP
#define USE_TERMCAP_OR_TERMINFO
#endif

#ifdef USE_TERMINFO
#define USE_TERMCAP_OR_TERMINFO
#endif

#ifdef USE_TERMCAP_OR_TERMINFO

/*================================================================*/
#define UNICODE_NUL         0
#define UNICODE_BELL        7
#define UNICODE_BACKSPACE   8
#define UNICODE_TAB         9
#define UNICODE_LINEFEED    10
#define UNICODE_RETURN      13
#define UNICODE_ESCAPE      27
#define UNICODE_SPACE       32
#define UNICODE_DOUBLEQUOTE 34
#define UNICODE_SHARP       35
#define UNICODE_QUOTE       39
#define UNICODE_COMMA       44
#define UNICODE_1           49
#define UNICODE_2           50
#define UNICODE_3           51
#define UNICODE_4           52
#define UNICODE_6           54
#define UNICODE_8           56
#define UNICODE_SEMICOLON   59
#define UNICODE_AT          64
#define UNICODE_UPPER_F     70
#define UNICODE_LOWER_F     102
#define UNICODE_UPPER_S     83
#define UNICODE_LOWER_S     115
#define UNICODE_UPPER_U     85
#define UNICODE_LOWER_U     117
#define UNICODE_LPAREN      40
#define UNICODE_RPAREN      41
#define UNICODE_LBRACKET    91
#define UNICODE_BACKSLASH   92
#define UNICODE_RBRACKET    93
#define UNICODE_LBRACE      123
#define UNICODE_VBAR        124
#define UNICODE_RBRACE      125
#define UNICODE_RUBOUT      127
#define UNICODE_REPLACEMENT 0xfffd

/*---------------------------------------------------------------------------*/

/* Data types.  */

/* Type of characters that can be read from and written to a tty */

typedef ___CS_SELECT(___LATIN1,___UCS2,___UCS4) tty_char;

#define TTY_CHAR_MAX ___MAX_CHR

/* Tty management data structures */

#define TTY_TAB_STOP 8

#define TTY_RBUF_SIZE 100
#define TTY_WBUF_SIZE 100

#define DEFAULT_EMACS_BINDINGS 1
#define DEFAULT_PAREN_BALANCE_DURATION_NSECS 500000000 /* 1/2 second */

typedef int tty_input_mode;

#define RAW_INPUT_MODE      1
#define LINEEDIT_INPUT_MODE 256

typedef int tty_term_mode;

#define INITIAL_TERM_MODE 0
#define CBREAK_TERM_MODE  1
#define RAW_TERM_MODE     2

typedef int tty_encoding;

#define ASCII_ENCODING  0
#define LATIN1_ENCODING 1
#define UTF8_ENCODING   2
#define UCS2LE_ENCODING 3
#define UCS2BE_ENCODING 4
#define UCS4LE_ENCODING 5
#define UCS4BE_ENCODING 6
#define NATIVE_ENCODING 7

typedef int tty_text_attrs;

#define NORMAL_STYLE  0
#define BOLD_STYLE    1
#define REVERSE_STYLE 2

#define DEFAULT_COLOR 8

#define GET_STYLE(x)(((x)>>8)&3)
#define GET_FOREGROUND_COLOR(x)((x)&15)
#define GET_BACKGROUND_COLOR(x)(((x)>>4)&15)

#define MAKE_TEXT_ATTRS(s,f,b)(((s)<<8)+(f)+((b)<<4))

#define DEFAULT_TEXT_ATTRS \
MAKE_TEXT_ATTRS(NORMAL_STYLE,DEFAULT_COLOR,DEFAULT_COLOR)

typedef int tty_signal;

#define USER_INTERRUPT_SIGNAL 0

#ifndef USE_POSIX
#ifndef USE_WIN32
#ifndef USE_MACOS

#define NB_TTY_SIGNALS 1

#endif
#endif
#endif

#ifdef USE_POSIX

#define WINDOW_SIZE_SIGNAL 1
#define CONTINUE_SIGNAL    2
#define NB_TTY_SIGNALS     3

#define CONSOLE_TERMINAL_TYPE_ENV_VAR "TERM"

#endif

#ifdef USE_WIN32

#define NB_TTY_SIGNALS 1

#endif

#ifdef USE_MACOS

/********************/

#endif

typedef struct tty_state_struct
  {
    struct tty_state_struct *prev; /* circular list links */
    struct tty_state_struct *next;

    int refcount; /* nb of "tty" structures that point to this "tty_state" */

    tty_input_mode input_mode; /* input mode (RAW or not, LINEEDIT or not) */
    tty_term_mode term_mode;   /* terminal mode (INITIAL, CBREAK, RAW) */
    tty_encoding encoding;     /* external char encoding (ASCII, UTF8, ...) */

    /* read buffer, needed for decoding characters */
    int rlo;
    int rhi;
    ___U8 rbuf[TTY_RBUF_SIZE];

    /* write buffer, needed for encoding characters */
    int wlo;
    int whi;
    ___U8 wbuf[TTY_WBUF_SIZE];
    ___WORD werr; /* write error flag for latest write operation */

    ___BOOL received_signal[NB_TTY_SIGNALS]; /* signals that were received */

#ifdef USE_POSIX
    int fd;                         /* file descriptor of the terminal */
    struct termios initial_termios; /* initial terminal driver settings */
#endif

#ifdef USE_WIN32
    HANDLE input;
    HANDLE output;
#define INPUT_RBUF_SIZE 10
    INPUT_RECORD input_rbuf[INPUT_RBUF_SIZE];
    int input_rlo;
    int input_rhi;
#endif

#ifdef USE_MACOS
    /*****************/
#endif

#ifdef USE_LINEEDIT
    struct lineedit_state_struct *les; /* line edit state */
#endif
  } tty_state;

typedef struct tty_struct
  {
    tty_state *state; /* NULL means not yet open console */
  } tty;

typedef struct tty_module_struct
  {
    tty_state *list; /* circular list of tty_state's */
    tty console;     /* the console is a special tty */
    void (*user_interrupt_handler) ___PVOID; /* called on a user interrupt */
  } tty_module;

___HIDDEN tty_module tm; /* tty module's global state */

/*---------------------------------------------------------------------------*/

/*********************/

___HIDDEN ___WORD tty_set_term_mode
   ___P((tty_state *state,
         tty_term_mode mode),
        ());

___HIDDEN ___WORD tty_set_terminal_type
   ___P((tty *t,
         char *terminal_type,
         ___BOOL emacs_bindings),
        ());

___HIDDEN ___WORD tty_set_encoding
   ___P((tty *t,
         tty_encoding encoding),
        ());

___HIDDEN ___WORD tty_cleanup
   ___P((tty *t),
        ());

/*---------------------------------------------------------------------------*/

#ifdef USE_LINEEDIT

/* Extensible string data structure */

typedef tty_char extensible_string_char;

typedef struct extensible_string_struct
  {
    extensible_string_char *buffer;
    int length;
    int max_length;
  } extensible_string;

/* Tty line editing data structures */

#define LINEEDIT_EVENT_NONE           0
#define LINEEDIT_EVENT_KEY            1
#define LINEEDIT_EVENT_RETURN         2
#define LINEEDIT_EVENT_BACK           3
#define LINEEDIT_EVENT_BACK_WORD      4
#define LINEEDIT_EVENT_TAB            5
#define LINEEDIT_EVENT_MARK           6
#define LINEEDIT_EVENT_PASTE          7
#define LINEEDIT_EVENT_CUT            8
#define LINEEDIT_EVENT_CUT_RIGHT      9
#define LINEEDIT_EVENT_CUT_LEFT       10
#define LINEEDIT_EVENT_REFRESH        11
#define LINEEDIT_EVENT_TRANSPOSE      12
#define LINEEDIT_EVENT_TRANSPOSE_WORD 13
#define LINEEDIT_EVENT_UP             14
#define LINEEDIT_EVENT_DOWN           15
#define LINEEDIT_EVENT_RIGHT          16
#define LINEEDIT_EVENT_RIGHT_WORD     17
#define LINEEDIT_EVENT_LEFT           18
#define LINEEDIT_EVENT_LEFT_WORD      19
#define LINEEDIT_EVENT_HOME           20
#define LINEEDIT_EVENT_HOME_DOC       21
#define LINEEDIT_EVENT_INSERT         22
#define LINEEDIT_EVENT_DELETE         23
#define LINEEDIT_EVENT_DELETE_WORD    24
#define LINEEDIT_EVENT_END            25
#define LINEEDIT_EVENT_END_DOC        26
#define LINEEDIT_EVENT_F1             27
#define LINEEDIT_EVENT_META_F1        28
#define LINEEDIT_EVENT_F2             29
#define LINEEDIT_EVENT_META_F2        30
#define LINEEDIT_EVENT_F3             31
#define LINEEDIT_EVENT_META_F3        32
#define LINEEDIT_EVENT_F4             33
#define LINEEDIT_EVENT_META_F4        34

#ifdef LINEEDIT_SUPPORT_F5_TO_F12
#define LINEEDIT_EVENT_F5             35
#define LINEEDIT_EVENT_META_F5        36
#define LINEEDIT_EVENT_F6             37
#define LINEEDIT_EVENT_META_F6        38
#define LINEEDIT_EVENT_F7             39
#define LINEEDIT_EVENT_META_F7        40
#define LINEEDIT_EVENT_F8             41
#define LINEEDIT_EVENT_META_F8        42
#define LINEEDIT_EVENT_F9             43
#define LINEEDIT_EVENT_META_F9        44
#define LINEEDIT_EVENT_F10            45
#define LINEEDIT_EVENT_META_F10       46
#define LINEEDIT_EVENT_F11            47
#define LINEEDIT_EVENT_META_F11       48
#define LINEEDIT_EVENT_F12            49
#define LINEEDIT_EVENT_META_F12       50
#define LINEEDIT_EVENT_LAST           LINEEDIT_EVENT_META_F12
#else
#define LINEEDIT_EVENT_LAST           LINEEDIT_EVENT_META_F4
#endif

#define LINEEDIT_CAP_CLEAR 0
#define LINEEDIT_CAP_ED    1
#define LINEEDIT_CAP_CUU1  2
#define LINEEDIT_CAP_CUD1  3
#define LINEEDIT_CAP_CUU   4
#define LINEEDIT_CAP_CUD   5
#define LINEEDIT_CAP_CUF   6
#define LINEEDIT_CAP_CUB   7
#define LINEEDIT_CAP_SGR0  8
#define LINEEDIT_CAP_BOLD  9
#define LINEEDIT_CAP_REV   10
#define LINEEDIT_CAP_SETAF 11
#define LINEEDIT_CAP_SETAB 12
#define LINEEDIT_CAP_LAST  12

typedef struct lineedit_event_struct
  {
    int event_kind;
    tty_char event_char;
  } lineedit_event;

typedef struct lineedit_history_struct
  {
    struct lineedit_history_struct *prev; /* circular list links */
    struct lineedit_history_struct *next;
    extensible_string actual;
    extensible_string edited; /* buffer != NULL when being edited */
  } lineedit_history;

#define WITH_ESC_PREFIX ((___U8)(1<<7))

/*
 * LARGEST_DECODER is the size of the largest input decoder over all
 * terminals defined in the termcap database.  The largest input
 * decoder is for the "ti926-8" terminal.  Note that this value must
 * be updated when the keyboard bindings are changed.
 */

#ifdef LINEEDIT_SUPPORT_F5_TO_F12
#define LARGEST_DECODER 177
#else
#define LARGEST_DECODER 128
#endif

/*
 * The type lineedit_input_decoder_state must at least contain the
 * range 0 .. LINEEDIT_EVENT_LAST + LARGEST_DECODER + 1.
 */

#if LINEEDIT_EVENT_LAST + LARGEST_DECODER + 1 <= 255

typedef ___U8 lineedit_input_decoder_state;
#define LINEEDIT_INPUT_DECODER_STATE_MAX 255

#else

typedef ___U16 lineedit_input_decoder_state;
#define LINEEDIT_INPUT_DECODER_STATE_MAX 65535

#endif

#define LINEEDIT_INPUT_DECODER_MAX_LENGTH \
(LINEEDIT_INPUT_DECODER_STATE_MAX-LINEEDIT_EVENT_LAST-1)

typedef struct lineedit_input_test_struct
  {
    ___U8 trigger; /* when this byte is read, action is performed */
    lineedit_input_decoder_state action; /* event type or new state */
    lineedit_input_decoder_state next;   /* next test when byte != trigger */
  } lineedit_input_test;

typedef struct lineedit_input_decoder_struct
  {
    lineedit_input_test *buffer;
    int length;
    int max_length;
  } lineedit_input_decoder;

#define LINEEDIT_MODE_SCHEME 1

typedef struct lineedit_state_struct
  {
    struct tty_state_struct *ts;

    int edit_mode;

    ___BOOL paste_cancel;
    int paste_index;
    tty_char *paste_text;

    int input_rlo;           /* index of next character to read */
    extensible_string input; /* last input line that was entered */

    int output_wlo;
    extensible_string output;

    int max_history_length;
    int history_length; /* number of history lines, not counting last */
    lineedit_history *hist_last;

    struct lineedit_state_undo_struct
    {
      lineedit_history *hist;
      int edit_point;     /* position in buffer where edit ops take place */
      int mark_point;     /* position in buffer of mark */
      int screen_cursor;  /* position in screen of cursor */
      int cursor_to_point;/* offset to add to cursor to get buffer position */
      ___BOOL delayed_wrap;
      ___BOOL paren_balance_trigger;
      ___BOOL paren_balance_in_progress;
      ___time paren_balance_end;
      tty_text_attrs attrs;
    } current;

    int paren_balance_duration_nsecs;

    tty_text_attrs input_attrs;
    tty_text_attrs output_attrs;

    int screen_rows;
    int screen_cols;
    int screen_size;
    ___BOOL has_auto_left_margin;
    ___BOOL has_auto_right_margin;
    ___BOOL has_eat_newline_glitch;
    ___BOOL linefeed_moves_to_left_margin;
    int tab_stop;

    lineedit_input_decoder input_decoder;
    lineedit_input_decoder_state input_decoder_state;

#ifdef LINEEDIT_WITH_LOCAL_CLIPBOARD
    extensible_string clipboard;
#endif

#ifdef USE_TERMCAP_OR_TERMINFO
    char *capability[LINEEDIT_CAP_LAST+1];
#endif

#ifdef USE_WIN32
    /***********************/
#endif

#ifdef USE_MACOS
    /***********************/
#endif
  } lineedit_state;

#define LINEEDIT_MAX_TAB_STOP 8

typedef struct lineedit_module_struct
  {
    tty_char tab_spaces[LINEEDIT_MAX_TAB_STOP];
#ifdef USE_TERMCAP_OR_TERMINFO
    lineedit_state *current_state; /* parameter of lineedit_output_putc */
    ___WORD last_err;              /* parameter of lineedit_output_putc */
#endif
  } lineedit_module;

___HIDDEN lineedit_module lem;

/*************************/

/* Forward declarations */

___HIDDEN ___WORD tty_read_noncanonical
   ___P((tty_state *state,
         tty_char *buf,
         int len),
        ());

___HIDDEN void lineedit_track_cursor_movement
   ___P((lineedit_state *state,
         tty_char *buf,
         int len),
        ());

___HIDDEN ___WORD tty_write_to_wbuf
   ___P((tty_state *state,
         tty_char *buf,
         int len),
        ());

___HIDDEN void tty_wbuf_drain
   ___P((tty_state *state),
        ());

/*---------------------------------------------------------------------------*/

/* Extensible string routines */

___HIDDEN ___WORD extensible_string_setup
   ___P((extensible_string *str,
         int n),
        (str,
         n)
extensible_string *str;
int n;)
{
#define EXTENSIBLE_STRING_INITIAL_BUFFER_SIZE 1

  if (n < EXTENSIBLE_STRING_INITIAL_BUFFER_SIZE)
    n = EXTENSIBLE_STRING_INITIAL_BUFFER_SIZE;

  str->buffer = ___CAST(extensible_string_char*,
                        ___alloc_mem (n * sizeof (extensible_string_char)));

  if (str->buffer == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  str->length = 0;
  str->max_length = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD extensible_string_cleanup
   ___P((extensible_string *str),
        (str)
extensible_string *str;)
{
  ___free_mem (str->buffer);
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD extensible_string_copy
   ___P((extensible_string_char *src,
         int len,
         extensible_string *dst,
         int fudge),
        (src,
         len,
         dst,
         fudge)
extensible_string_char *src;
int len;
extensible_string *dst;
int fudge;)
{
  extensible_string_char *buf;

  buf = ___CAST(extensible_string_char*,
                ___alloc_mem ((len+fudge) * sizeof (extensible_string_char)));

  if (buf == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  dst->buffer = buf;
  dst->length = len;
  dst->max_length = len+fudge;

  while (len-- > 0)
    buf[len] = src[len];

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD extensible_string_set_length
   ___P((extensible_string *str,
         int len,
         int fudge),
        (str,
         len,
         fudge)
extensible_string *str;
int len;
int fudge;)
{
  if (len > str->max_length || 2*len+1 < str->max_length)
    {
      int i;
      int new_max_length = (fudge<0) ? 3*len/2+1 : len+fudge;
      extensible_string_char *old_buffer = str->buffer;
      extensible_string_char *new_buffer = (new_max_length>10000000)?NULL: ___CAST(extensible_string_char*,___alloc_mem (new_max_length * sizeof (extensible_string_char)));/*********************/

      if (new_buffer == NULL)
        return ___FIX(___HEAP_OVERFLOW_ERR);

      i = str->length;
      if (i > len)
        i = len;

      while (i-- > 0)
        new_buffer[i] = old_buffer[i];

      ___free_mem (old_buffer);
      str->buffer = new_buffer;
      str->max_length = new_max_length;
    }

  str->length = len;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD extensible_string_delete
   ___P((extensible_string *str,
         int pos,
         int len),
        (str,
         pos,
         len)
extensible_string *str;
int pos;
int len;)
{
  int i;

  if (pos < 0)
    pos = 0;
  else if (pos > str->length)
    pos = str->length;

  i = str->length - pos;
  if (len < 0)
    len = 0;
  else if (len > i)
    len = i;

  for (i=pos; i<str->length-len; i++)
    str->buffer[i] = str->buffer[i+len];

  str->length -= len;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD extensible_string_insert
   ___P((extensible_string *str,
         int pos,
         int len,
         extensible_string_char *chars),
        (str,
         pos,
         len,
         chars)
extensible_string *str;
int pos;
int len;
extensible_string_char *chars;)
{
  ___WORD e;
  int i;

  if (pos < 0)
    pos = 0;
  else if (pos > str->length)
    pos = str->length;

  if (len < 0)
    len = 0;

  if ((e = extensible_string_set_length (str, str->length+len, -1))
      == ___FIX(___NO_ERR))
    {
      for (i=str->length-len-1; i>=pos; i--)
        str->buffer[i+len] = str->buffer[i];

      for (i=len-1; i>=0; i--)
        str->buffer[i+pos] = chars[i];
    }

  return e;
}

/*---------------------------------------------------------------------------*/

/* Tty line editing input decoder routines */

___HIDDEN ___WORD lineedit_input_decoder_setup
   ___P((lineedit_input_decoder *decoder),
        (decoder)
lineedit_input_decoder *decoder;)
{
#define LINEEDIT_INPUT_DECODER_INITIAL_BUFFER_SIZE 1 /******************/

  int n = LINEEDIT_INPUT_DECODER_INITIAL_BUFFER_SIZE;

  decoder->buffer = ___CAST(lineedit_input_test*,
                            ___alloc_mem (n * sizeof (lineedit_input_test)));

  if (decoder->buffer == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  decoder->length = 0;
  decoder->max_length = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_input_decoder_cleanup
   ___P((lineedit_input_decoder *decoder),
        (decoder)
lineedit_input_decoder *decoder;)
{
  ___free_mem (decoder->buffer);
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_input_decoder_set_length
   ___P((lineedit_input_decoder *decoder,
         int len,
         int fudge),
        (decoder,
         len,
         fudge)
lineedit_input_decoder *decoder;
int len;
int fudge;)
{
  if (len > LINEEDIT_INPUT_DECODER_MAX_LENGTH)
    return ___FIX(___UNKNOWN_ERR);

  if (len > decoder->max_length || 2*len+1 < decoder->max_length)
    {
      int i;
      int new_max_length = (fudge<0) ? 3*len/2+1 : len+fudge;
      lineedit_input_test *old_buffer = decoder->buffer;
      lineedit_input_test *new_buffer =
        ___CAST(lineedit_input_test*,
                ___alloc_mem (new_max_length * sizeof (lineedit_input_test)));

      if (new_buffer == NULL)
        return ___FIX(___HEAP_OVERFLOW_ERR);

      i = decoder->length;
      if (i > len)
        i = len;

      while (i-- > 0)
        new_buffer[i] = old_buffer[i];

      ___free_mem (old_buffer);
      decoder->buffer = new_buffer;
      decoder->max_length = new_max_length;
    }

  decoder->length = len;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_input_decoder_add
   ___P((lineedit_input_decoder *decoder,
         char *seq,
         ___U8 event),
        (decoder,
         seq,
         event)
lineedit_input_decoder *decoder;
char *seq;
___U8 event;)
{
  ___WORD e;
  ___U8 b;
  int i;
  char *p = seq;

  if (event & WITH_ESC_PREFIX)
    b = UNICODE_ESCAPE;
  else
    b = (___U8)*p++;

if (0)
  {/****************************/
    char *p = seq;
    if (b < 32)
      printf ("(%d)", b);
    else
      printf ("%c", b);
    while (*p != 0)
      if (*p < 32)
        printf ("(%d)", *p++);
      else
        printf ("%c", *p++);
    printf ("\n");
    fflush (stdout);
  }

  if (decoder->length > 0)
    {
      i = 0;
      while (b != (___U8)'\0' || p == seq+1)
        {
          if (decoder->buffer[i].trigger == b)
            {
              int a = decoder->buffer[i].action;
              if (a < decoder->length)
                {
                  i = a;
                  b = (___U8)*p++;
                }
              else
                return ___FIX(___NO_ERR); /* prefix of sequence exists */
            }
          else
            {
              int n = decoder->buffer[i].next;
              if (n < decoder->length)
                i = n;
              else
                {
                  decoder->buffer[i].next = decoder->length;
                  break;
                }
            }
        }
    }

  if (b != (___U8)'\0' || p == seq+1)
    {
      /* sequence is not a prefix of a preexisting sequence */

      while (b != (___U8)'\0' || p == seq+1)
        {
          i = decoder->length;

          if ((e = lineedit_input_decoder_set_length (decoder, i+1, -1))
              != ___FIX(___NO_ERR))
            return e;

          decoder->buffer[i].trigger = b;
          decoder->buffer[i].action = i+1;
          decoder->buffer[i].next = LINEEDIT_INPUT_DECODER_STATE_MAX;

          b = (___U8)*p++;
        }
      decoder->buffer[i].action =
        LINEEDIT_INPUT_DECODER_STATE_MAX-(event&~WITH_ESC_PREFIX);
    }

  return ___FIX(___NO_ERR);
}

typedef struct lineedit_defseq_struct
  {
    char *seq;
    ___U8 event;
  } lineedit_defseq;

___HIDDEN lineedit_defseq lineedit_defseq_common[] =
{
  /* sequences that all terminals support */

   { "\012",     LINEEDIT_EVENT_RETURN                           }
  ,{ "\015",     LINEEDIT_EVENT_RETURN                           }
  ,{ "\010",     LINEEDIT_EVENT_BACK                             }
  ,{ "\010",     LINEEDIT_EVENT_BACK_WORD      | WITH_ESC_PREFIX }
  ,{ "\011",     LINEEDIT_EVENT_TAB                              }
};

___HIDDEN lineedit_defseq lineedit_defseq_map_rubout_to_backspace[] =
{
  /* sequences that map the rubout key to backspace */

   { "\177",     LINEEDIT_EVENT_BACK                             }
  ,{ "\177",     LINEEDIT_EVENT_BACK_WORD      | WITH_ESC_PREFIX }
};

___HIDDEN lineedit_defseq lineedit_defseq_emacs[] =
{
  /* sequences specific to emacs mode */

   { "\0",       LINEEDIT_EVENT_MARK                             }
  ,{ "\031",     LINEEDIT_EVENT_PASTE                            }
  ,{ "\027",     LINEEDIT_EVENT_CUT                              }
  ,{ "\013",     LINEEDIT_EVENT_CUT_RIGHT                        }
  ,{ "\025",     LINEEDIT_EVENT_CUT_LEFT                         }
  ,{ "\014",     LINEEDIT_EVENT_REFRESH                          }
  ,{ "\024",     LINEEDIT_EVENT_TRANSPOSE                        }
  ,{ "\024",     LINEEDIT_EVENT_TRANSPOSE_WORD | WITH_ESC_PREFIX }
  ,{ "t",        LINEEDIT_EVENT_TRANSPOSE_WORD | WITH_ESC_PREFIX }
  ,{ "T",        LINEEDIT_EVENT_TRANSPOSE_WORD | WITH_ESC_PREFIX }

  ,{ "\020",     LINEEDIT_EVENT_UP                               }
  ,{ "\016",     LINEEDIT_EVENT_DOWN                             }
  ,{ "\006",     LINEEDIT_EVENT_RIGHT                            }
  ,{ "\002",     LINEEDIT_EVENT_LEFT                             }
  ,{ "\006",     LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "f",        LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "F",        LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "\002",     LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }
  ,{ "b",        LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }
  ,{ "B",        LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }

  ,{ "\001",     LINEEDIT_EVENT_HOME                             }
  ,{ "\004",     LINEEDIT_EVENT_DELETE                           }
  ,{ "\005",     LINEEDIT_EVENT_END                              }
  ,{ "\004",     LINEEDIT_EVENT_DELETE_WORD    | WITH_ESC_PREFIX }
};

___HIDDEN lineedit_defseq lineedit_defseq_widespread[] =
{
  /* sequences that many types of terminals support */

   { "[A",       LINEEDIT_EVENT_UP             | WITH_ESC_PREFIX }
  ,{ "[B",       LINEEDIT_EVENT_DOWN           | WITH_ESC_PREFIX }
  ,{ "[C",       LINEEDIT_EVENT_RIGHT          | WITH_ESC_PREFIX }
  ,{ "[D",       LINEEDIT_EVENT_LEFT           | WITH_ESC_PREFIX }

  ,{ "\033[A",   LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033[B",   LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033[C",   LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "\033[D",   LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }

  ,{ "OA",       LINEEDIT_EVENT_UP             | WITH_ESC_PREFIX }
  ,{ "OB",       LINEEDIT_EVENT_DOWN           | WITH_ESC_PREFIX }
  ,{ "OC",       LINEEDIT_EVENT_RIGHT          | WITH_ESC_PREFIX }
  ,{ "OD",       LINEEDIT_EVENT_LEFT           | WITH_ESC_PREFIX }

  ,{ "\033OA",   LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033OB",   LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033OC",   LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "\033OD",   LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }

  ,{ "A",        LINEEDIT_EVENT_UP             | WITH_ESC_PREFIX }
  ,{ "B",        LINEEDIT_EVENT_DOWN           | WITH_ESC_PREFIX }
  ,{ "C",        LINEEDIT_EVENT_RIGHT          | WITH_ESC_PREFIX }
  ,{ "D",        LINEEDIT_EVENT_LEFT           | WITH_ESC_PREFIX }

  ,{ "\033A",    LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033B",    LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033C",    LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "\033D",    LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "[5A",      LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "[5B",      LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "[5C",      LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "[5D",      LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }
#endif

  ,{ "[1~",      LINEEDIT_EVENT_HOME           | WITH_ESC_PREFIX }
  ,{ "[2~",      LINEEDIT_EVENT_INSERT         | WITH_ESC_PREFIX }
  ,{ "[3~",      LINEEDIT_EVENT_DELETE         | WITH_ESC_PREFIX }
  ,{ "[4~",      LINEEDIT_EVENT_END            | WITH_ESC_PREFIX }
  ,{ "[5~",      LINEEDIT_EVENT_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "[6~",      LINEEDIT_EVENT_END_DOC        | WITH_ESC_PREFIX }

  ,{ "\033[1~",  LINEEDIT_EVENT_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "\033[2~",  LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033[3~",  LINEEDIT_EVENT_DELETE_WORD    | WITH_ESC_PREFIX }
  ,{ "\033[4~",  LINEEDIT_EVENT_END_DOC        | WITH_ESC_PREFIX }
  ,{ "\033[5~",  LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033[6~",  LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }

  ,{ "[H",       LINEEDIT_EVENT_HOME           | WITH_ESC_PREFIX }
  ,{ "[L",       LINEEDIT_EVENT_INSERT         | WITH_ESC_PREFIX }
  ,{ "[F",       LINEEDIT_EVENT_END            | WITH_ESC_PREFIX }
  ,{ "\177",     LINEEDIT_EVENT_DELETE                           }
  ,{ "\033[H",   LINEEDIT_EVENT_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "\177",     LINEEDIT_EVENT_DELETE_WORD    | WITH_ESC_PREFIX }
  ,{ "\033[F",   LINEEDIT_EVENT_END_DOC        | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "[1;5~",    LINEEDIT_EVENT_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "[2;5~",    LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "[3;5~",    LINEEDIT_EVENT_DELETE_WORD    | WITH_ESC_PREFIX }
  ,{ "[4;5~",    LINEEDIT_EVENT_END_DOC        | WITH_ESC_PREFIX }
  ,{ "[5;5~",    LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "[6;5~",    LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }

  ,{ "[5H",      LINEEDIT_EVENT_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "[5L",      LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "[5F",      LINEEDIT_EVENT_END_DOC        | WITH_ESC_PREFIX }
#endif

  ,{ "OP",       LINEEDIT_EVENT_F1             | WITH_ESC_PREFIX }
  ,{ "OQ",       LINEEDIT_EVENT_F2             | WITH_ESC_PREFIX }
  ,{ "OR",       LINEEDIT_EVENT_F3             | WITH_ESC_PREFIX }
  ,{ "OS",       LINEEDIT_EVENT_F4             | WITH_ESC_PREFIX }

  ,{ "\033OP",   LINEEDIT_EVENT_META_F1        | WITH_ESC_PREFIX }
  ,{ "\033OQ",   LINEEDIT_EVENT_META_F2        | WITH_ESC_PREFIX }
  ,{ "\033OR",   LINEEDIT_EVENT_META_F3        | WITH_ESC_PREFIX }
  ,{ "\033OS",   LINEEDIT_EVENT_META_F4        | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "O5P",      LINEEDIT_EVENT_META_F1        | WITH_ESC_PREFIX }
  ,{ "O5Q",      LINEEDIT_EVENT_META_F2        | WITH_ESC_PREFIX }
  ,{ "O5R",      LINEEDIT_EVENT_META_F3        | WITH_ESC_PREFIX }
  ,{ "O5S",      LINEEDIT_EVENT_META_F4        | WITH_ESC_PREFIX }
#endif

  ,{ "[11~",     LINEEDIT_EVENT_F1             | WITH_ESC_PREFIX }
  ,{ "[12~",     LINEEDIT_EVENT_F2             | WITH_ESC_PREFIX }
  ,{ "[13~",     LINEEDIT_EVENT_F3             | WITH_ESC_PREFIX }
  ,{ "[14~",     LINEEDIT_EVENT_F4             | WITH_ESC_PREFIX }

  ,{ "\033[11~", LINEEDIT_EVENT_META_F1        | WITH_ESC_PREFIX }
  ,{ "\033[12~", LINEEDIT_EVENT_META_F2        | WITH_ESC_PREFIX }
  ,{ "\033[13~", LINEEDIT_EVENT_META_F3        | WITH_ESC_PREFIX }
  ,{ "\033[14~", LINEEDIT_EVENT_META_F4        | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "[11;5~",   LINEEDIT_EVENT_META_F1        | WITH_ESC_PREFIX }
  ,{ "[12;5~",   LINEEDIT_EVENT_META_F2        | WITH_ESC_PREFIX }
  ,{ "[13;5~",   LINEEDIT_EVENT_META_F3        | WITH_ESC_PREFIX }
  ,{ "[14;5~",   LINEEDIT_EVENT_META_F4        | WITH_ESC_PREFIX }
#endif

#ifdef LINEEDIT_SUPPORT_F5_TO_F12

  ,{ "[15~",     LINEEDIT_EVENT_F5             | WITH_ESC_PREFIX }
  ,{ "[17~",     LINEEDIT_EVENT_F6             | WITH_ESC_PREFIX }
  ,{ "[18~",     LINEEDIT_EVENT_F7             | WITH_ESC_PREFIX }
  ,{ "[19~",     LINEEDIT_EVENT_F8             | WITH_ESC_PREFIX }
  ,{ "[20~",     LINEEDIT_EVENT_F9             | WITH_ESC_PREFIX }
  ,{ "[21~",     LINEEDIT_EVENT_F10            | WITH_ESC_PREFIX }
  ,{ "[23~",     LINEEDIT_EVENT_F11            | WITH_ESC_PREFIX }
  ,{ "[24~",     LINEEDIT_EVENT_F12            | WITH_ESC_PREFIX }

  ,{ "\033[15~", LINEEDIT_EVENT_META_F5        | WITH_ESC_PREFIX }
  ,{ "\033[17~", LINEEDIT_EVENT_META_F6        | WITH_ESC_PREFIX }
  ,{ "\033[18~", LINEEDIT_EVENT_META_F7        | WITH_ESC_PREFIX }
  ,{ "\033[19~", LINEEDIT_EVENT_META_F8        | WITH_ESC_PREFIX }
  ,{ "\033[20~", LINEEDIT_EVENT_META_F9        | WITH_ESC_PREFIX }
  ,{ "\033[21~", LINEEDIT_EVENT_META_F10       | WITH_ESC_PREFIX }
  ,{ "\033[23~", LINEEDIT_EVENT_META_F11       | WITH_ESC_PREFIX }
  ,{ "\033[24~", LINEEDIT_EVENT_META_F12       | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "[15;5~",   LINEEDIT_EVENT_META_F5        | WITH_ESC_PREFIX }
  ,{ "[17;5~",   LINEEDIT_EVENT_META_F6        | WITH_ESC_PREFIX }
  ,{ "[18;5~",   LINEEDIT_EVENT_META_F7        | WITH_ESC_PREFIX }
  ,{ "[19;5~",   LINEEDIT_EVENT_META_F8        | WITH_ESC_PREFIX }
  ,{ "[20;5~",   LINEEDIT_EVENT_META_F9        | WITH_ESC_PREFIX }
  ,{ "[21;5~",   LINEEDIT_EVENT_META_F10       | WITH_ESC_PREFIX }
  ,{ "[23;5~",   LINEEDIT_EVENT_META_F11       | WITH_ESC_PREFIX }
  ,{ "[24;5~",   LINEEDIT_EVENT_META_F12       | WITH_ESC_PREFIX }
#endif

#endif
};

___HIDDEN ___WORD lineedit_defseq_install_table
   ___P((lineedit_input_decoder *decoder,
         lineedit_defseq *table,
         int len),
        (decoder,
         table,
         len)
lineedit_input_decoder *decoder;
lineedit_defseq *table;
int len;)
{
  ___WORD e;
  int i;

  for (i=0; i<len; i++)
    if ((e = lineedit_input_decoder_add
               (decoder,
                table[i].seq,
                table[i].event))
        != ___FIX(___NO_ERR))
      return e;

  return ___FIX(___NO_ERR);
}

#ifdef USE_TERMCAP_OR_TERMINFO

typedef struct lineedit_defkey_struct
  {
    char *cap;
    ___U8 event_no_escape;
    ___U8 event_with_escape;
  } lineedit_defkey;

#ifdef USE_TERMCAP
#define DEFKEY(termcap_name,terminfo_name,event_no_escape,event_with_escape) \
{ termcap_name, event_no_escape, event_with_escape | WITH_ESC_PREFIX }
#else
#define DEFKEY(termcap_name,terminfo_name,event_no_escape,event_with_escape) \
{ terminfo_name, event_no_escape, event_with_escape | WITH_ESC_PREFIX }
#endif

typedef struct lineedit_defcap_struct
  {
    char *cap;
  } lineedit_defcap;

#ifdef USE_TERMCAP
#define DEFCAP(termcap_name,terminfo_name) { termcap_name }
#else
#define DEFCAP(termcap_name,terminfo_name) { terminfo_name }
#endif

___HIDDEN lineedit_defcap lineedit_defcap_table[] =
{
  /* order must match LINEEDIT_CAP_XXX */

   DEFCAP( "cl", "clear" ) /* erase screen and home cursor */
  ,DEFCAP( "cd", "ed"    ) /* erase to end of screen       */
  ,DEFCAP( "up", "cuu1"  ) /* move up one line             */
  ,DEFCAP( "do", "cud1"  ) /* move down one line           */
  ,DEFCAP( "UP", "cuu"   ) /* move up N lines              */
  ,DEFCAP( "DO", "cud"   ) /* move down N lines            */
  ,DEFCAP( "RI", "cuf"   ) /* move right N columns         */
  ,DEFCAP( "LE", "cub"   ) /* move left N columns          */
  ,DEFCAP( "me", "sgr0"  ) /* turn off all attributes      */
  ,DEFCAP( "md", "bold"  ) /* turn on bold attribute       */
  ,DEFCAP( "mr", "rev"   ) /* turn on reverse attribute    */
  ,DEFCAP( "AF", "setaf" ) /* ANSI set foreground color    */
  ,DEFCAP( "AB", "setab" ) /* ANSI set background color    */
};

___HIDDEN lineedit_defkey lineedit_defkey_table[] =
{
   DEFKEY( "ku", "kcuu1", LINEEDIT_EVENT_UP,       LINEEDIT_EVENT_NONE        )
  ,DEFKEY( "kd", "kcud1", LINEEDIT_EVENT_DOWN,     LINEEDIT_EVENT_NONE        )
  ,DEFKEY( "kr", "kcuf1", LINEEDIT_EVENT_RIGHT,    LINEEDIT_EVENT_RIGHT_WORD  )
  ,DEFKEY( "kl", "kcub1", LINEEDIT_EVENT_LEFT,     LINEEDIT_EVENT_LEFT_WORD   )
  ,DEFKEY( "kh", "khome", LINEEDIT_EVENT_HOME,     LINEEDIT_EVENT_HOME_DOC    )
  ,DEFKEY( "kI", "kich1", LINEEDIT_EVENT_INSERT,   LINEEDIT_EVENT_NONE        )
  ,DEFKEY( "kD", "kdch1", LINEEDIT_EVENT_DELETE,   LINEEDIT_EVENT_DELETE_WORD )
  ,DEFKEY( "@7", "kend",  LINEEDIT_EVENT_END,      LINEEDIT_EVENT_END_DOC     )
  ,DEFKEY( "kP", "kpp",   LINEEDIT_EVENT_HOME_DOC, LINEEDIT_EVENT_NONE        )
  ,DEFKEY( "kN", "knp",   LINEEDIT_EVENT_END_DOC,  LINEEDIT_EVENT_NONE        )
  ,DEFKEY( "k1", "kf1",   LINEEDIT_EVENT_F1,       LINEEDIT_EVENT_META_F1     )
  ,DEFKEY( "k2", "kf2",   LINEEDIT_EVENT_F2,       LINEEDIT_EVENT_META_F2     )
  ,DEFKEY( "k3", "kf3",   LINEEDIT_EVENT_F3,       LINEEDIT_EVENT_META_F3     )
  ,DEFKEY( "k4", "kf4",   LINEEDIT_EVENT_F4,       LINEEDIT_EVENT_META_F4     )
#ifdef LINEEDIT_SUPPORT_F5_TO_F12
  ,DEFKEY( "k5", "kf5",   LINEEDIT_EVENT_F5,       LINEEDIT_EVENT_META_F5     )
  ,DEFKEY( "k6", "kf6",   LINEEDIT_EVENT_F6,       LINEEDIT_EVENT_META_F6     )
  ,DEFKEY( "k7", "kf7",   LINEEDIT_EVENT_F7,       LINEEDIT_EVENT_META_F7     )
  ,DEFKEY( "k8", "kf8",   LINEEDIT_EVENT_F8,       LINEEDIT_EVENT_META_F8     )
  ,DEFKEY( "k9", "kf9",   LINEEDIT_EVENT_F9,       LINEEDIT_EVENT_META_F9     )
  ,DEFKEY( "k;", "kf10",  LINEEDIT_EVENT_F10,      LINEEDIT_EVENT_META_F10    )
  ,DEFKEY( "F1", "kf11",  LINEEDIT_EVENT_F11,      LINEEDIT_EVENT_META_F11    )
  ,DEFKEY( "F2", "kf12",  LINEEDIT_EVENT_F12,      LINEEDIT_EVENT_META_F12    )
#endif
};

___HIDDEN ___WORD lineedit_defkey_install
   ___P((lineedit_input_decoder *decoder,
         char *cap,
         ___U8 event_no_escape,
         ___U8 event_with_escape),
        (decoder,
         cap,
         event_no_escape,
         event_with_escape)
lineedit_input_decoder *decoder;
char *cap;
___U8 event_no_escape;
___U8 event_with_escape;)
{
  ___WORD e = ___FIX(___NO_ERR);
  char *seq;

#ifdef USE_TERMCAP
  seq = tgetstr (cap, NULL);
#else
  seq = tigetstr (cap);
#endif

  if (seq != (char*)-1 && seq != NULL)
    if ((e = lineedit_input_decoder_add (decoder, seq, event_no_escape))
        != ___FIX(___NO_ERR) &&
        event_with_escape != (LINEEDIT_EVENT_NONE|WITH_ESC_PREFIX))
      e = lineedit_input_decoder_add (decoder, seq, event_with_escape);

  return e;
}

___HIDDEN ___WORD lineedit_defkey_install_table
   ___P((lineedit_input_decoder *decoder,
         lineedit_defkey *table,
         int len),
        (decoder,
         table,
         len)
lineedit_input_decoder *decoder;
lineedit_defkey *table;
int len;)
{
  ___WORD e;
  int i;

  for (i=0; i<len; i++)
    if ((e = lineedit_defkey_install
               (decoder,
                table[i].cap,
                table[i].event_no_escape,
                table[i].event_with_escape))
        != ___FIX(___NO_ERR))
      return e;

  return ___FIX(___NO_ERR);
}

#endif

#define NBELEMS(table)(sizeof (table) / sizeof (table[0]))

___HIDDEN ___WORD lineedit_input_decoder_init
   ___P((lineedit_input_decoder *decoder,
         ___BOOL map_rubout_to_backspace,
         ___BOOL emacs_bindings,
         ___BOOL termcap_or_terminfo_keys),
        (decoder,
         map_rubout_to_backspace,
         emacs_bindings,
         termcap_or_terminfo_keys)
lineedit_input_decoder *decoder;
___BOOL map_rubout_to_backspace;
___BOOL emacs_bindings;
___BOOL termcap_or_terminfo_keys;)
{
  ___WORD e;

  decoder->length = 0;

  if ((e = lineedit_defseq_install_table
             (decoder,
              lineedit_defseq_common,
              NBELEMS(lineedit_defseq_common)))
      != ___FIX(___NO_ERR))
    return e;

  if (map_rubout_to_backspace)
    if ((e = lineedit_defseq_install_table
               (decoder,
                lineedit_defseq_map_rubout_to_backspace,
                NBELEMS(lineedit_defseq_map_rubout_to_backspace)))
        != ___FIX(___NO_ERR))
      return e;

  if (emacs_bindings)
    if ((e = lineedit_defseq_install_table
               (decoder,
                lineedit_defseq_emacs,
                NBELEMS(lineedit_defseq_emacs)))
        != ___FIX(___NO_ERR))
      return e;

#ifdef USE_TERMCAP_OR_TERMINFO
  if (termcap_or_terminfo_keys)
    if ((e = lineedit_defkey_install_table
               (decoder,
                lineedit_defkey_table,
                NBELEMS(lineedit_defkey_table)))
        != ___FIX(___NO_ERR))
      return e;
#endif

  if ((e = lineedit_defseq_install_table
             (decoder,
              lineedit_defseq_widespread,
              NBELEMS(lineedit_defseq_widespread)))
      != ___FIX(___NO_ERR))
    return e;

#if 0
#if 0
  {int i;
  /***************************/
  i = 0;
  while (i<decoder->length)
    {
      if (decoder->buffer[i].trigger < UNICODE_SPACE)
      printf ("%d trigger=%d action=%d next=%d\n",
              i,
              decoder->buffer[i].trigger,
              decoder->buffer[i].action,
              decoder->buffer[i].next);
      else
      printf ("%d trigger='%c' action=%d next=%d\n",
              i,
              decoder->buffer[i].trigger,
              decoder->buffer[i].action,
              decoder->buffer[i].next);
      i++;
    }
  }
#else
  printf ("%d\n",decoder->length);
#endif
  fflush (stdout);
#endif

  return ___FIX(___NO_ERR);
}

/*---------------------------------------------------------------------------*/

/* Tty line editing routines */

___HIDDEN ___WORD lineedit_history_begin_edit
   ___P((lineedit_state *state,
         lineedit_history *h),
        (state,
         h)
lineedit_state *state;
lineedit_history *h;)
{
#define LINEEDIT_FUDGE 80

  ___WORD e;

  if (h->edited.buffer == NULL)
    {
      if ((e = extensible_string_copy
                 (h->actual.buffer,
                  h->actual.length,
                  &h->edited,
                  LINEEDIT_FUDGE))
          != ___FIX(___NO_ERR))
        return e;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN void lineedit_history_end_edit
   ___P((lineedit_state *state,
         lineedit_history *h),
        (state,
         h)
lineedit_state *state;
lineedit_history *h;)
{
  if (h->edited.buffer != NULL)
    {
      extensible_string_cleanup (&h->edited); /* ignore error */
      h->edited.buffer = NULL;
    }
}

___HIDDEN ___WORD lineedit_history_setup
   ___P((lineedit_state *state,
         lineedit_history **hist),
        (state,
         hist)
lineedit_state *state;
lineedit_history **hist;)
{
  ___WORD e;
  lineedit_history *h;

  h = ___CAST(lineedit_history*,___alloc_mem (sizeof (lineedit_history)));

  if (h == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  if ((e = extensible_string_setup (&h->actual, 0)) != ___FIX(___NO_ERR))
    {
      ___free_mem (h);
      return e;
    }

  h->edited.buffer = NULL;

  h->prev = h; /* create a circular list */
  h->next = h;

  *hist = h;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_history_cleanup
   ___P((lineedit_state *state,
         lineedit_history *h),
        (state,
         h)
lineedit_state *state;
lineedit_history *h;)
{
  lineedit_history_end_edit (state, h);

  extensible_string_cleanup (&h->actual); /* ignore error */

  ___free_mem (h);

  return ___FIX(___NO_ERR);
}

___HIDDEN void lineedit_history_remove
   ___P((lineedit_state *state,
         lineedit_history *item),
        (state,
         item)
lineedit_state *state;
lineedit_history *item;)
{
  lineedit_history *prev = item->prev;
  lineedit_history *next = item->next;

  if (prev == item)
    state->hist_last = NULL;
  else
    {
      next->prev = prev;
      prev->next = next;
      item->prev = item;
      item->next = item;
      if (state->hist_last == item)
        state->hist_last = prev;
    }
}

___HIDDEN void lineedit_history_trim
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  while (state->history_length > state->max_history_length)
    {
      lineedit_history *first = state->hist_last->next;
      lineedit_history_remove (state, first);
      lineedit_history_cleanup (state, first); /* ignore error */
      state->history_length--;
    }
}

___HIDDEN void lineedit_set_max_history_length
   ___P((lineedit_state *state,
         int max_history_length),
        (state,
         max_history_length)
lineedit_state *state;
int max_history_length;)
{
  if (max_history_length >= 0)
    {
      state->max_history_length = max_history_length;
      lineedit_history_trim (state);
    }
}

___HIDDEN void lineedit_history_add
   ___P((lineedit_state *state,
         lineedit_history *item),
        (state,
         item)
lineedit_state *state;
lineedit_history *item;)
{
  lineedit_history *dest = state->hist_last;

  if (dest == NULL)
    {
      item->prev = item;
      item->next = item;
    }
  else
    {
      lineedit_history *after_dest = dest->next;
      item->next = after_dest;
      item->prev = dest;
      dest->next = item;
      after_dest->prev = item;
    }

  state->hist_last = item;
  state->history_length++;
}

___HIDDEN ___WORD lineedit_output_drain
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  /*
   * Drain the output generated by the previous line editing
   * operation.
   */

  ___WORD e;
  ___WORD n;
  int wlo;
  int length;
  tty_state *ts = state->ts;

  if ((e = ts->werr) != ___FIX(___NO_ERR))
    {
      ts->werr = ___FIX(___NO_ERR);
      return e;
    }

  tty_wbuf_drain (ts); /* error code returned in ts->werr */

  wlo = state->output_wlo;
  length = state->output.length;

  while (wlo < length)
    {
      if ((e = ts->werr) != ___FIX(___NO_ERR))
        {
          ts->werr = ___FIX(___NO_ERR);
          state->output_wlo = wlo;
          return e;
        }

      n = tty_write_to_wbuf (ts, &state->output.buffer[wlo], length - wlo);

      if (n <= ___FIX(0))
        {
          if (n == ___FIX(0))
            n = ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));
#ifdef LINEEDIT_REPORT_CHAR_ENCODING_ERRORS
          else if (n == ___FIX(___STOC_LATIN1STRING_ERR))
            wlo++;
#endif
          state->output_wlo = wlo;
          return n;
        }

      wlo += ___INT(n);

      tty_wbuf_drain (ts); /* error code returned in ts->werr */
    }

  if (extensible_string_set_length (&state->output, 0, 1) == ___FIX(___NO_ERR))
    wlo = 0;

  state->output_wlo = wlo;

  return ___FIX(___NO_ERR);
}

___HIDDEN void lineedit_track_cursor_movement
   ___P((lineedit_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
lineedit_state *state;
tty_char *buf;
int len;)
{
  /*
   * Even with the termcap/terminfo database it is impossible to tell
   * exactly how the cursor will move when an arbitrary sequence of
   * characters is sent to the terminal.  So we make general
   * assumptions about the terminal behavior with a bias towards
   * "xterm", which is probably the terminal that will most often be
   * used under UNIX.
   */

  /*
   * The fields state->current.screen_cursor,
   * state->current.cursor_to_point, and state->current.delayed_wrap
   * are updated accordingly.
   */

  int screen_cols = state->screen_cols;
  int screen_cursor = state->current.screen_cursor;
  int cursor_to_point = state->current.cursor_to_point;
  ___BOOL delayed_wrap = state->current.delayed_wrap;
  tty_char c;
  int col;

  while (len-- > 0)
    {
      c = *buf++;

      switch (c)
        {
        case UNICODE_BELL:
          break;

        case UNICODE_BACKSPACE:
          if (screen_cursor % screen_cols > 0 ||
              (screen_cursor > 0 && state->has_auto_left_margin))
            screen_cursor--;
          delayed_wrap = 0;
          break;

        case UNICODE_TAB:
          col = ((screen_cursor % screen_cols) / state->tab_stop + 1)
                * state->tab_stop;
          if (col >= screen_cols)
            {
              if (state->has_auto_right_margin &&
                  !state->has_eat_newline_glitch)
                {
                  screen_cursor = (screen_cursor / screen_cols + 1)
                                  * screen_cols;
                  if (screen_cursor >= state->screen_size)
                    {
                      screen_cursor -= screen_cols;
                      cursor_to_point += screen_cols;
                    }
                }
              else
                {
                  col = screen_cols-1; /* don't wrap */
                  screen_cursor = (screen_cursor / screen_cols) * screen_cols
                                  + col;
                }
            }
          else
            screen_cursor = (screen_cursor / screen_cols) * screen_cols + col;
          break;

        case UNICODE_LINEFEED:
          if (state->linefeed_moves_to_left_margin)
            screen_cursor = (screen_cursor / screen_cols + 1) * screen_cols;
          else
            screen_cursor = screen_cursor + screen_cols;
          if (screen_cursor >= state->screen_size)
            {
              screen_cursor -= screen_cols;
              cursor_to_point += screen_cols;
            }
          delayed_wrap = 0;
          break;

        case UNICODE_RETURN:
          screen_cursor = (screen_cursor / screen_cols) * screen_cols;
          delayed_wrap = 0;
          break;

        default:
          if (delayed_wrap)
            {
              screen_cursor++;
              if (screen_cursor >= state->screen_size)
                {
                  screen_cursor -= screen_cols;
                  cursor_to_point += screen_cols;
                }
            }

          screen_cursor++;

          if (screen_cursor % screen_cols != 0)
            delayed_wrap = 0;
          else if (state->has_auto_right_margin)
            {
              if (state->has_eat_newline_glitch)
                {
                  screen_cursor--;
                  delayed_wrap = 1; /* wrap later */
                }
              else
                {
                  if (screen_cursor >= state->screen_size)
                    {
                      screen_cursor -= screen_cols;
                      cursor_to_point += screen_cols;
                    }
                  delayed_wrap = 0;
                }
            }
          else
            {
              screen_cursor--;
              delayed_wrap = 0;
            }
        }
    }

  state->current.screen_cursor = screen_cursor;
  state->current.cursor_to_point = cursor_to_point;
  state->current.delayed_wrap = delayed_wrap;
}

___HIDDEN ___WORD lineedit_output_string
   ___P((lineedit_state *state,
         char *str,
         int rep),
        (state,
         str,
         rep)
lineedit_state *state;
char *str;
int rep;)
{
  /*
   * Output the C string "str" multiple times ("rep" is the repeat
   * factor).
   */

  while (rep-- > 0)
    {
      char *p = str;
      while (*p != '\0')
        {
          ___WORD e;
          tty_char c = *p++;
          if ((e = extensible_string_insert
                     (&state->output,
                      state->output.length,
                      1,
                      &c))
              != ___FIX(___NO_ERR))
            return e;
#if 0
printf ("%c", c); fflush(stdout);/******************/
#endif
        }
    }

  return ___FIX(___NO_ERR);
}

#ifdef USE_TERMCAP_OR_TERMINFO

___HIDDEN int lineedit_output_putc
   ___P((int c),
        (c)
int c;)
{
  /*
   * This routine outputs a C character.  It is needed for the call to
   * the termcap/terminfo "tputs" routine.  The parameters it uses are
   * stored in the lineedit module structure "lem".
   */

  if (lem.last_err == ___FIX(___NO_ERR))
    {
      tty_char ch = c;
      lem.last_err = extensible_string_insert
                       (&lem.current_state->output,
                        lem.current_state->output.length,
                        1,
                        &ch);
    }

  return c;
}

___HIDDEN ___WORD lineedit_output_cap
   ___P((lineedit_state *state,
         int cap,
         int arg,
         int rep),
        (state,
         cap,
         arg,
         rep)
lineedit_state *state;
int cap;
int arg;
int rep;)
{
  /*
   * Output the sequence for the terminal capability "cap" with
   * integer parameter "arg" (-1 means no parameter).  "rep" is the
   * repeat factor.
   */

  char *str = state->capability[cap];

  if (arg >= 0)
    {
#ifdef USE_TERMCAP
      str = tgoto (str, arg, arg);
#else
      str = tparm (str, arg);
#endif
    }

  if (cap < LINEEDIT_CAP_SGR0)
    state->current.delayed_wrap = 0; /* cursor commands cancel delayed wrap */

  lem.current_state = state;
  lem.last_err = ___FIX(___NO_ERR);

  if (tputs (str, 1, lineedit_output_putc) == ERR &&
      lem.last_err == ___FIX(___NO_ERR))
    lem.last_err = ___FIX(___UNKNOWN_ERR);

  return lem.last_err;
}

#endif

___HIDDEN ___WORD lineedit_output_set_text_attrs
   ___P((lineedit_state *state,
         tty_text_attrs attrs),
        (state,
         attrs)
lineedit_state *state;
tty_text_attrs attrs;)
{
  /*
   * Output the control sequence to set the text attributes of the
   * next characters sent to the terminal.
   */

#ifdef USE_TERMCAP_OR_TERMINFO

  ___WORD e;
  int turn_on;
  tty_text_attrs current_attrs = state->current.attrs;

  if (current_attrs == attrs)
    return ___FIX(___NO_ERR);

  turn_on = GET_STYLE(attrs);

  if ((GET_STYLE(current_attrs) & ~turn_on) != 0 ||
      (GET_FOREGROUND_COLOR(attrs) >= DEFAULT_COLOR &&
       GET_FOREGROUND_COLOR(current_attrs) < DEFAULT_COLOR) ||
      (GET_BACKGROUND_COLOR(attrs) >= DEFAULT_COLOR &&
       GET_BACKGROUND_COLOR(current_attrs) < DEFAULT_COLOR))
    {
      if (state->capability[LINEEDIT_CAP_SGR0] != NULL)
        if ((e = lineedit_output_cap (state, LINEEDIT_CAP_SGR0, -1, 1))
            != ___FIX(___NO_ERR))
          return e;
      current_attrs = MAKE_TEXT_ATTRS(0,DEFAULT_COLOR,DEFAULT_COLOR);
    }
  else
    turn_on = (~GET_STYLE(current_attrs) & turn_on);

  if ((turn_on & BOLD_STYLE) &&
      state->capability[LINEEDIT_CAP_BOLD] != NULL)
    if ((e = lineedit_output_cap (state, LINEEDIT_CAP_BOLD, -1, 1))
        != ___FIX(___NO_ERR))
      return e;

  if ((turn_on & REVERSE_STYLE) &&
      state->capability[LINEEDIT_CAP_REV] != NULL)
    if ((e = lineedit_output_cap (state, LINEEDIT_CAP_REV, -1, 1))
        != ___FIX(___NO_ERR))
      return e;

  if (GET_FOREGROUND_COLOR(attrs) < DEFAULT_COLOR &&
      GET_FOREGROUND_COLOR(attrs) != GET_FOREGROUND_COLOR(current_attrs) &&
      state->capability[LINEEDIT_CAP_SETAF] != NULL)
    if ((e = lineedit_output_cap
               (state,
                LINEEDIT_CAP_SETAF,
                GET_FOREGROUND_COLOR(attrs),
                1))
        != ___FIX(___NO_ERR))
      return e;

  if (GET_BACKGROUND_COLOR(attrs) < DEFAULT_COLOR &&
      GET_BACKGROUND_COLOR(attrs) != GET_BACKGROUND_COLOR(current_attrs) &&
      state->capability[LINEEDIT_CAP_SETAB] != NULL)
    if ((e = lineedit_output_cap
               (state,
                LINEEDIT_CAP_SETAB,
                GET_BACKGROUND_COLOR(attrs),
                1))
        != ___FIX(___NO_ERR))
      return e;

  state->current.attrs = attrs;

#endif

#ifdef USE_WIN32

  {
    /*********************/

    WORD attributes = 0;

    if (GET_STYLE(attrs) & BOLD_STYLE)
      attributes |= FOREGROUND_INTENSITY;

    if (GET_FOREGROUND_COLOR(attrs) == DEFAULT_COLOR)
      attributes |= (FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE);
    else
      {
        if (GET_FOREGROUND_COLOR(attrs) & 1)
          attributes |= FOREGROUND_RED;
        if (GET_FOREGROUND_COLOR(attrs) & 2)
          attributes |= FOREGROUND_GREEN;
        if (GET_FOREGROUND_COLOR(attrs) & 4)
          attributes |= FOREGROUND_BLUE;
      }

    if (GET_BACKGROUND_COLOR(attrs) == DEFAULT_COLOR)
      attributes |= 0;
    else
      {
        if (GET_BACKGROUND_COLOR(attrs) & 1)
          attributes |= BACKGROUND_RED;
        if (GET_BACKGROUND_COLOR(attrs) & 2)
          attributes |= BACKGROUND_GREEN;
        if (GET_BACKGROUND_COLOR(attrs) & 4)
          attributes |= BACKGROUND_BLUE;
      }

    lineedit_output_drain (state); /* ignore error */

    if (!SetConsoleTextAttribute
           (state->ts->output,
            attributes))
      return ___FIX(err_code_from_GetLastError ());
  }
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_output_chars
   ___P((lineedit_state *state,
         tty_char *buf,
         int len,
         tty_text_attrs attrs),
        (state,
         buf,
         len,
         attrs)
lineedit_state *state;
tty_char *buf;
int len;
tty_text_attrs attrs;)
{
  /*
   * Output "len" characters from the buffer "buf" and keep track of
   * cursor motion.
   */

  ___WORD e;

  if ((e = lineedit_output_set_text_attrs (state, attrs))
      != ___FIX(___NO_ERR))
    return e;

  e = extensible_string_insert
        (&state->output,
         state->output.length,
         len,
         buf);
  if (0) {int i; for (i=0;i<len;i++) printf ("%c", buf[i]); fflush(stdout);/******************/}

  lineedit_track_cursor_movement (state, buf, len);

  return e;
}

___HIDDEN ___WORD lineedit_output_char_repetition
   ___P((lineedit_state *state,
         tty_char c,
         int rep,
         tty_text_attrs attrs),
        (state,
         c,
         rep,
         attrs)
lineedit_state *state;
tty_char c;
int rep;
tty_text_attrs attrs;)
{
  /*
   * Output a character "rep" times and keep track of cursor motion.
   */

#define CHAR_BUFFER_SIZE (80*50)

  ___WORD e;
  tty_char char_buffer[CHAR_BUFFER_SIZE];
  int n;

  n = rep;
  if (n > CHAR_BUFFER_SIZE)
    n = CHAR_BUFFER_SIZE;

  while (n > 0)
    char_buffer[--n] = c;

  while (rep > 0)
    {
      n = rep;
      if (n > CHAR_BUFFER_SIZE)
        n = CHAR_BUFFER_SIZE;
      if ((e = lineedit_output_chars (state, char_buffer, n, attrs))
          != ___FIX(___NO_ERR))
        return e;
      rep -= n;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN tty_text_attrs lineedit_erase_attrs
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  tty_text_attrs output_attrs = state->output_attrs;
  int output_style = GET_STYLE(output_attrs);
  int current_style = GET_STYLE(state->current.attrs);

  return MAKE_TEXT_ATTRS((current_style & BOLD_STYLE) |
                         (output_style & REVERSE_STYLE),
                         GET_FOREGROUND_COLOR(output_attrs),
                         GET_BACKGROUND_COLOR(output_attrs));
}

___HIDDEN ___WORD lineedit_output_spaces
   ___P((lineedit_state *state,
         int rep),
        (state,
         rep)
lineedit_state *state;
int rep;)
{
  return lineedit_output_char_repetition
           (state,
            UNICODE_SPACE,
            rep,
            lineedit_erase_attrs (state));
}

___HIDDEN ___WORD lineedit_erase_to_end_of_screen
   ___P((lineedit_state *state,
         int pos,
         int chars_to_erase),
        (state,
         pos,
         chars_to_erase)
lineedit_state *state;
int pos;
int chars_to_erase;)
{
  /*
   * Erases from the current position up to the end of the screen.
   * "pos" is the start position and "chars_to_erase" is the number of
   * characters that are erased.
   */

/***************************strange interface**/

  ___WORD e;

#ifdef USE_TERMCAP_OR_TERMINFO
  if (chars_to_erase > 3 &&
      state->capability[LINEEDIT_CAP_ED] != NULL)
    {
      if ((e = lineedit_output_set_text_attrs
                 (state,
                  lineedit_erase_attrs (state)))
          == ___FIX(___NO_ERR))
        e = lineedit_output_cap (state, LINEEDIT_CAP_ED, -1, 1);
    }
  else
#endif
    e = lineedit_output_spaces (state, chars_to_erase);

  return e;
}

___HIDDEN ___WORD lineedit_output_chars_at
   ___P((lineedit_state *state,
         tty_char *str,
         int chars,
         int chars_to_erase,
         int pos),
        (state,
         str,
         chars,
         chars_to_erase,
         pos)
lineedit_state *state;
tty_char *str;
int chars;
int chars_to_erase;
int pos;)
{
  ___WORD e;/**************unused*/

  /******************not used!

  if ((e = lineedit_prepare_to_write_at (state, pos))
      != ___FIX(___NO_ERR))
    return e;

  if ((e = lineedit_output_chars (state, str, chars, state->input_attrs))
       != ___FIX(___NO_ERR))
    return e;
  */

  return lineedit_erase_to_end_of_screen (state, pos+chars, chars_to_erase);
}

___HIDDEN ___WORD lineedit_output_current_hist
   ___P((lineedit_state *state,
         int start,
         int len),
        (state,
         start,
         len)
lineedit_state *state;
int start;
int len;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  int to_end;
  int n;

  if (start < 0)
    {
      n = -start;
      if (n > len)
        n = len;
      if ((e = lineedit_output_spaces (state, n)) != ___FIX(___NO_ERR))
        return e;
      start += n;
      len -= n;
    }

  to_end = edited->length - start;

  if (to_end > 0)
    {
      if (to_end > len)
        n = len;
      else
        n = to_end;

      if ((e = lineedit_output_chars
                 (state,
                  &edited->buffer[start],
                  n,
                  state->input_attrs))
          != ___FIX(___NO_ERR))
        return e;

      len -= n;

      if (n == to_end &&
          state->current.delayed_wrap &&
          state->current.screen_cursor + 1 < state->screen_size &&
          len < 1)
        len = 1;
    }

  return lineedit_erase_to_end_of_screen (state, 999999, len);/*******/
}

___HIDDEN ___WORD lineedit_output_force_delayed_wrap
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  /*
   * Forces the cursor to wrap to the next line if the wrap was
   * delayed.
   */

  if (state->current.delayed_wrap)
    return lineedit_output_current_hist
             (state,
              state->current.screen_cursor+1+state->current.cursor_to_point,
              1);

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_move_screen_cursor
   ___P((lineedit_state *state,
         int pos),
        (state,
         pos)
lineedit_state *state;
int pos;)
{
  /*
   * This routine sends the appropriate commands to the terminal to
   * move the cursor to position "pos".
   */

  /*
   * The fields state->current.screen_cursor and
   * state->current.delayed_wrap are updated accordingly.
   */

  ___WORD e;
  int screen_cursor = state->current.screen_cursor;
  int screen_cols;
  int pos_col;
  int col_dist;
  int row_dist;
  int dist;

  if (pos == screen_cursor)
    return ___FIX(___NO_ERR);

  screen_cols = state->screen_cols;
  pos_col = pos % screen_cols;
  col_dist = pos_col - screen_cursor % screen_cols;
  row_dist = pos / screen_cols - screen_cursor / screen_cols;

#ifdef USE_TERMCAP_OR_TERMINFO

  while (1)
    {
      /*
       * Check if the cursor can be moved backward by writing no more
       * than 4 backspace characters (if on the same row or column) or
       * 8 backspace characters (otherwise).  This is probably no more
       * than is required by cursor movement commands.
       */

      if (pos < screen_cursor &&
          (row_dist == 0 || state->has_auto_left_margin) &&
          screen_cursor - pos <= ((row_dist == 0 || col_dist == 0) ? 4 : 8))
        break;

      /*
       * Check if the cursor can be moved forward by writing no more
       * than 4 plain characters (if on the same row or column) or 8
       * plain characters (otherwise).  This is probably no more than
       * is required by cursor movement commands.
       */

      if (pos > screen_cursor &&
          screen_cursor + state->current.cursor_to_point >= 0 &&
          (row_dist == 0 || state->has_auto_right_margin) &&
          pos - (screen_cursor + state->current.delayed_wrap)
              + ((pos_col == 0 && state->has_eat_newline_glitch) ? 2 : 0) <=
              ((row_dist == 0 || col_dist == 0) ? 4 : 8))
        break;

      /*
       * If the cursor is on the target row, move the cursor to the
       * target column using cursor movement commands, if the terminal
       * supports this.
       */

      if (row_dist == 0)
        {
          if (col_dist == 0 ||
              ((col_dist > 0)
               ? (state->capability[LINEEDIT_CAP_CUF] == NULL)
               : (state->capability[LINEEDIT_CAP_CUB] == NULL)))
            break;

          if ((e = lineedit_output_cap
                     (state,
                      (col_dist > 0) ? LINEEDIT_CAP_CUF : LINEEDIT_CAP_CUB,
                      (col_dist > 0) ? col_dist : -col_dist,
                      1))
              != ___FIX(___NO_ERR))
            return e;

          screen_cursor += col_dist;
          state->current.screen_cursor = screen_cursor;
          break;
        }

      /*
       * Move the cursor to the target row using cursor movement
       * commands, if the terminal supports this.
       */

      if ((row_dist > 0)
          ? (state->capability[LINEEDIT_CAP_CUD] == NULL &&
             state->capability[LINEEDIT_CAP_CUD1] == NULL)
          : (state->capability[LINEEDIT_CAP_CUU] == NULL &&
             state->capability[LINEEDIT_CAP_CUU1] == NULL))
        break;

      if ((row_dist > 0)
          ? (state->capability[LINEEDIT_CAP_CUD] == NULL ||
             (state->capability[LINEEDIT_CAP_CUD1] != NULL && row_dist == 1))
          : (state->capability[LINEEDIT_CAP_CUU] == NULL ||
             (state->capability[LINEEDIT_CAP_CUU1] != NULL && row_dist == -1)))
        e = lineedit_output_cap
              (state,
               (row_dist > 0) ? LINEEDIT_CAP_CUD1 : LINEEDIT_CAP_CUU1,
               -1,
               (row_dist > 0) ? row_dist : -row_dist);
      else
        e = lineedit_output_cap
              (state,
               (row_dist > 0) ? LINEEDIT_CAP_CUD : LINEEDIT_CAP_CUU,
               (row_dist > 0) ? row_dist : -row_dist,
               1);

      if (e != ___FIX(___NO_ERR))
        return e;

      /* Cursor commands to change row were successful. */

      screen_cursor += row_dist * screen_cols;
      state->current.screen_cursor = screen_cursor;
      row_dist = 0;
    }

#endif

  /*
   * If the cursor is still not at the target position, move the
   * cursor by writing plain characters or backspace characters.  We
   * assume that the terminal has auto left/right margins if the
   * cursor needs to change row.
   */

  dist = pos - screen_cursor;

  if (dist != 0)
    {
      if (dist < 0)
        {
          if ((e = lineedit_output_char_repetition
                     (state,
                      UNICODE_BACKSPACE,
                      -dist,
                      state->current.attrs))
              != ___FIX(___NO_ERR))
            return e;
        }
      else
        {
          int start = screen_cursor + state->current.cursor_to_point +
                      state->current.delayed_wrap;
          int len = dist - state->current.delayed_wrap;
          ___BOOL need_backspace =
            (pos_col == 0 && state->has_eat_newline_glitch);

          if ((e = lineedit_output_current_hist
                     (state,
                      start,
                      len+need_backspace))
              != ___FIX(___NO_ERR))
            return e;

          if (need_backspace)
            {
              if ((e = lineedit_output_char_repetition
                         (state,
                          UNICODE_BACKSPACE,
                          1,
                          state->current.attrs))
                  != ___FIX(___NO_ERR))
                return e;
            }
        }

      screen_cursor += dist;
      state->current.screen_cursor = screen_cursor;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_left_margin_of_next_row
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  ___WORD e;

  if (!state->linefeed_moves_to_left_margin)
    if ((e = lineedit_output_char_repetition
               (state,
                UNICODE_RETURN,
                1,
                state->output_attrs))
        != ___FIX(___NO_ERR))
      return e;

  return lineedit_output_char_repetition
           (state,
            UNICODE_LINEFEED,
            1,
            state->output_attrs);
}

___HIDDEN ___WORD lineedit_prepare_to_write_at
   ___P((lineedit_state *state,
         int pos),
        (state,
         pos)
lineedit_state *state;
int pos;)
{
  /*
   * This routine sends commands to the terminal such that the next
   * character sent to the terminal will show up at position "pos".
   * The position can be equal to state->screen_size, in which case
   * either the screen is scrolled right away or the next character
   * sent to the terminal will cause the screen to scroll.
   */

  ___WORD e;

  if (state->current.screen_cursor + state->current.delayed_wrap == pos)
    return ___FIX(___NO_ERR);

  if (pos >= state->screen_size)
    {
      if ((e = lineedit_move_screen_cursor (state, state->screen_size-1))
          != ___FIX(___NO_ERR))
        return e;
    }
  else
    {
      if ((e = lineedit_move_screen_cursor (state, pos))
          != ___FIX(___NO_ERR))
        return e;

      if (!state->current.delayed_wrap)
        return ___FIX(___NO_ERR);

      /*
       * Output a backspace and a character.  Note that this assumes
       * the screen is at least 2 columns wide.
       */

      if ((e = lineedit_output_char_repetition
                 (state,
                  UNICODE_BACKSPACE,
                  1,
                  state->current.attrs))
          != ___FIX(___NO_ERR))
        return e;
    }

  return lineedit_output_current_hist
           (state,
            state->current.screen_cursor + state->current.cursor_to_point,
            1);
}

___HIDDEN ___WORD lineedit_newline
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  int pos = edited->length - state->current.cursor_to_point;

  if (pos < 0)
    pos = 0;
  else if (pos >= state->screen_size)
    pos = state->screen_size-1;

  if ((e = lineedit_prepare_to_write_at (state, pos))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_left_margin_of_next_row (state);
}

___HIDDEN ___WORD lineedit_erase_screen
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  ___WORD e;

#ifdef USE_TERMCAP_OR_TERMINFO
  if (state->capability[LINEEDIT_CAP_CLEAR] != NULL)
    {
      if ((e = lineedit_output_set_text_attrs
                 (state,
                  lineedit_erase_attrs (state)))
          == ___FIX(___NO_ERR))
        e = lineedit_output_cap (state, LINEEDIT_CAP_CLEAR, -1, 1);
    }
  else
#endif
    e = lineedit_newline (state);

  if (e == ___FIX(___NO_ERR))
    {
      state->current.screen_cursor = 0; /****************/
    }

  return e;
}

___HIDDEN ___WORD lineedit_copy_to_clipboard
   ___P((lineedit_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
lineedit_state *state;
tty_char *buf;
int len;)
{
  /*
   * Copies "len" characters from the buffer "buf" to the clipboard.
   */

#ifdef LINEEDIT_WITH_LOCAL_CLIPBOARD

  ___WORD e;
  extensible_string str;

  if ((e = extensible_string_copy (buf, len, &str, 0)) != ___FIX(___NO_ERR))
    return e;

  extensible_string_cleanup (&state->clipboard); /* ignore error */

  state->clipboard = str;

#else

  /***********************/

#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_paste_from_clipboard
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  /*
   * Paste the content of the clipboard to the input stream (the next
   * characters read from the terminal will in fact come from the
   * pasted text).
   */

#ifdef LINEEDIT_WITH_LOCAL_CLIPBOARD

  int len = state->clipboard.length;
  tty_char *str;

  str = ___CAST(tty_char*,___alloc_mem ((len+1) * sizeof (tty_char)));

  if (str == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  str[len] = 0;
  while (len-- > 0)
    str[len] = state->clipboard.buffer[len];

  if (state->paste_text != NULL)
    ___free_mem (state->paste_text);

  state->paste_index = 0;
  state->paste_text = str;

#else

  /*************************/

#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_get_event
   ___P((lineedit_state *state,
         lineedit_event *ev),
        (state,
         ev)
lineedit_state *state;
lineedit_event *ev;)
{
  /*
   * This routine obtains the next line editing event from the
   * terminal.  An event can be a key press or some editing operation
   * (like "delete next word").
   */

  lineedit_input_decoder_state s = state->input_decoder_state;
  tty_char c;
  ___BOOL first_char = (s == 0);
  ___WORD n;

  next_char:

  n = tty_read_noncanonical (state->ts, &c, 1);

  if (n == ___FIX(1))
    {
      while (s < state->input_decoder.length)
        {
          if (state->input_decoder.buffer[s].trigger == c)
            {
              int a = state->input_decoder.buffer[s].action;
              if (a < LINEEDIT_INPUT_DECODER_MAX_LENGTH)
                {
                  s = a;
                  first_char = 0;
                  goto next_char;
                }
              state->input_decoder_state = 0;
              ev->event_kind = LINEEDIT_INPUT_DECODER_STATE_MAX-a;
              return ___FIX(___NO_ERR);
            }
          else
            s = state->input_decoder.buffer[s].next;
        }
      if (first_char)
        {
          ev->event_kind = LINEEDIT_EVENT_KEY;
          ev->event_char = c;
          return ___FIX(___NO_ERR);
        }
      s = 0; /* ignore the sequence including last character read */
    }
  else if (n != ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)))
    return n;

  state->input_decoder_state = s;
  ev->event_kind = LINEEDIT_EVENT_NONE;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_set_terminal_type
   ___P((lineedit_state *state,
         char *terminal_type,
         ___BOOL emacs_bindings),
        (state,
         terminal_type,
         emacs_bindings)
lineedit_state *state;
char *terminal_type;
___BOOL emacs_bindings;)
{
  ___WORD e;
  int screen_rows = 24;
  int screen_cols = 80;
  ___BOOL has_auto_left_margin = 1;
  ___BOOL has_auto_right_margin = 1;
  ___BOOL has_eat_newline_glitch = 0;

#ifdef USE_TERMCAP_OR_TERMINFO

  int i;

  if (terminal_type != NULL)
    {
#ifdef USE_TERMCAP

      char termcap_buffer[4080]; /* the value 4080 is used by GNU's readline */

      if (tgetent (termcap_buffer, terminal_type) != 1)
        return ___FIX(___UNKNOWN_ERR);

      screen_rows = tgetnum ("li");
      screen_cols = tgetnum ("co");
      has_auto_left_margin = (tgetflag ("bw") != 0);
      has_auto_right_margin = (tgetflag ("am") != 0);
      has_eat_newline_glitch = (tgetflag ("xn") != 0);

#else

      int errret;

      if (setupterm (terminal_type, 1, &errret) == ERR)
        return ___FIX(___UNKNOWN_ERR);

      screen_rows = tigetnum ("lines");
      screen_cols = tigetnum ("cols");
      has_auto_left_margin = (tigetflag ("bw") != 0);
      has_auto_right_margin = (tigetflag ("am") != 0);
      has_eat_newline_glitch = (tigetflag ("xenl") != 0);

#endif
    }

  /*
   * Setup the capability table.
   */

  for (i=0; i<LINEEDIT_CAP_LAST+1; i++)
    {
      char *seq = state->capability[i];

      if (seq != NULL)
        {
          ___free_mem (seq);
          state->capability[i] = NULL;
        }
    }

  if (terminal_type != NULL)
    for (i=0; i<LINEEDIT_CAP_LAST+1; i++)
      {
        char *seq;

#ifdef USE_TERMCAP
        seq = tgetstr (lineedit_defcap_table[i].cap, NULL);
#else
        seq = tigetstr (lineedit_defcap_table[i].cap);
#endif

        if (seq != (char*)-1 && seq != NULL)
          {
            char *p;

            /*
             * Reject any sequence that includes a linefeed if the
             * terminal driver automatically adds a carriage return.
             */

            if (state->linefeed_moves_to_left_margin)
              {
                p = seq;
                while (*p != '\0')
                  if (*p != UNICODE_LINEFEED)
                    p++;
                  else
                    break;
                if (*p != '\0')
                  seq = NULL;
              }

            /*
             * Keep a copy of the sequence.
             */

            if (seq != NULL)
              {
                int len = 0;
                while (seq[len] != '\0')
                  len++;
                p = ___CAST(char*,___alloc_mem (len));
                if (p != NULL)
                  {
                    p[len] = '\0';
                    while (len-- > 0)
                      p[len] = seq[len];
                  }
                seq = p;
              }
          }
        else
          seq = NULL;

        state->capability[i] = seq;
      }

#endif

  /*
   * Initialize input decoder so that it recognizes special keys.  Map
   * rubout key to backspace if terminal driver is setup to erase when
   * a rubout is typed.
   */

  state->input_decoder_state = 0;

  if ((e = lineedit_input_decoder_init
             (&state->input_decoder,
#ifdef USE_POSIX
              (state->ts->initial_termios.c_cc[VERASE] == UNICODE_RUBOUT),
#else
              0,
#endif
              emacs_bindings,
              terminal_type != NULL))
      != ___FIX(___NO_ERR))
    return e;

  if (screen_rows <= 0)
    screen_rows = 24;
  if (screen_cols <= 0)
    screen_cols = 80;

  state->screen_rows = screen_rows;
  state->screen_cols = screen_cols;
  state->screen_size = screen_rows * screen_cols;
  state->has_auto_left_margin = has_auto_left_margin;
  state->has_auto_right_margin = has_auto_right_margin;
  state->has_eat_newline_glitch = has_eat_newline_glitch;

#if 0
  printf ("%s screen_rows=%d screen_cols=%d alm=%d arm=%d eng=%d\n", terminal_type,screen_rows, screen_cols,
          has_auto_left_margin,
          has_auto_right_margin,
          has_eat_newline_glitch);fflush (stdout);
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_set_paren_balance_duration
   ___P((lineedit_state *state,
         int nsecs),
        (state,
         nsecs)
lineedit_state *state;
int nsecs;)
{
  state->paren_balance_duration_nsecs = nsecs;
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_setup
   ___P((lineedit_state **les,
         tty_state *ts),
        (les,
         ts)
lineedit_state **les;
tty_state *ts;)
{
  ___WORD e;
  lineedit_history *h;
  lineedit_state *state;

  *les = NULL;

  /* allocate a new lineedit_state */

  state = ___CAST(lineedit_state*,___alloc_mem (sizeof (lineedit_state)));

  if (state == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  *les = state;

  state->ts = ts;

  state->edit_mode = LINEEDIT_MODE_SCHEME;

  state->paste_cancel = 0;
  state->paste_index = 0;
  state->paste_text = NULL;

  state->input_rlo = 0;
  state->input.buffer = NULL;

  state->output_wlo = 0;

  state->max_history_length = 1000000; /* set a very high limit */
  state->history_length = -1;
  state->hist_last = NULL;

  state->current.edit_point = 0;
  state->current.mark_point = 0;
  state->current.screen_cursor = 0;
  state->current.cursor_to_point = 0;
  state->current.delayed_wrap = 0;
  state->current.paren_balance_in_progress = 0;
  state->current.attrs = DEFAULT_TEXT_ATTRS;

  state->paren_balance_duration_nsecs = DEFAULT_PAREN_BALANCE_DURATION_NSECS;

  state->input_attrs = DEFAULT_TEXT_ATTRS;
  state->output_attrs = DEFAULT_TEXT_ATTRS;

  state->screen_rows = 24;
  state->screen_cols = 80;
  state->screen_size = 24*80;

  state->has_auto_left_margin = 1;
  state->has_auto_right_margin = 1;
  state->has_eat_newline_glitch = 0;
  state->linefeed_moves_to_left_margin = 1;

  state->tab_stop = TTY_TAB_STOP;

  if ((e = extensible_string_setup (&state->output, 0)) == ___FIX(___NO_ERR))
    {
      if ((e = lineedit_history_setup (state, &h)) == ___FIX(___NO_ERR))
        {
          lineedit_history_add (state, h);
          state->current.hist = h;
          if ((e = lineedit_history_begin_edit (state, h))
              == ___FIX(___NO_ERR))
            {
              if ((e = lineedit_input_decoder_setup (&state->input_decoder))
                  == ___FIX(___NO_ERR))
                {
                  if ((e = lineedit_set_terminal_type
                             (state,
                              NULL,
                              DEFAULT_EMACS_BINDINGS))
                      == ___FIX(___NO_ERR))
                    {
#ifdef LINEEDIT_WITH_LOCAL_CLIPBOARD
                      if ((e = extensible_string_setup (&state->clipboard, 0))
                          == ___FIX(___NO_ERR))
                        {
#endif
#ifdef USE_TERMCAP_OR_TERMINFO
                          int i;
                          for (i=0; i<LINEEDIT_CAP_LAST+1; i++)
                            state->capability[i] = NULL;
#endif
#ifdef USE_WIN32
                          /***********************/
#endif
#ifdef USE_MACOS
                          /***********************/
#endif
#ifdef LINEEDIT_WITH_LOCAL_CLIPBOARD
                          if (e != ___FIX(___NO_ERR))
                            extensible_string_cleanup
                              (&state->clipboard); /* ignore error */
                        }
#endif
                    }
                  if (e != ___FIX(___NO_ERR))
                    lineedit_input_decoder_cleanup
                      (&state->input_decoder); /* ignore error */
                }
            }
          if (e != ___FIX(___NO_ERR))
            lineedit_history_cleanup (state, h); /* ignore error */
        }
      if (e != ___FIX(___NO_ERR))
        extensible_string_cleanup (&state->output); /* ignore error */
    }

  if (e != ___FIX(___NO_ERR))
    ___free_mem (state);

  return e;
}

___HIDDEN ___WORD lineedit_cleanup
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  lineedit_output_set_text_attrs
    (state,
     DEFAULT_TEXT_ATTRS); /* ignore error */

  lineedit_output_drain (state); /* ignore error */

  if (state->paste_text != NULL)
    ___free_mem (state->paste_text); /* discard paste text */

  if (state->input.buffer != NULL)
    extensible_string_cleanup (&state->input); /* ignore error */

  extensible_string_cleanup (&state->output); /* ignore error */

  state->max_history_length = -1;
  lineedit_history_trim (state);

#ifdef LINEEDIT_WITH_LOCAL_CLIPBOARD
  extensible_string_cleanup (&state->clipboard); /* ignore error */
#endif

#ifdef USE_TERMCAP_OR_TERMINFO
  {
    int i;

    lineedit_input_decoder_cleanup (&state->input_decoder); /* ignore error */

    for (i=0; i<LINEEDIT_CAP_LAST+1; i++)
      {
        char *seq = state->capability[i];
        if (seq != NULL)
          ___free_mem (seq);
      }
  }
#endif

#ifdef USE_WIN32
  /*********************/
#endif

#ifdef USE_MACOS
  /*********************/
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_move_edit_point
   ___P((lineedit_state *state,
         int pos),
        (state,
         pos)
lineedit_state *state;
int pos;)
{
  /*
   * Moves the edit point of the current edited line.  The screen may
   * get scrolled in order to keep the edit point visible.
   */

  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  int screen_cursor;
  int screen_cols;
  int screen_rows;
  int screen_size;
  int cursor_to_point;
  int cursor_row;
  int first_row;
  int last_row;
  int nb_rows;
  int all_fits;
  int n;
  int start;

  if (pos < 0 ||
      pos > edited->length)
    return ___FIX(___INVALID_OP_ERR);

  screen_cols = state->screen_cols;
  screen_rows = state->screen_rows;
  screen_size = state->screen_size;
  cursor_to_point = state->current.cursor_to_point;
  screen_cursor = pos - cursor_to_point; /* new cursor position in screen */

  /*
   * Check if the text from the cursor to the position just after end
   * of the edited line can all fit on the screen (while maintaining
   * column alignment).
   */

  cursor_row = screen_cursor;
  if (cursor_row < 0)
    cursor_row -= screen_cols - 1;
  cursor_row /= screen_cols;

  first_row = -cursor_to_point;
  if (first_row < 0)
    first_row -= screen_cols - 1;
  first_row /= screen_cols;

  last_row = edited->length - cursor_to_point;
  if (last_row < 0)
    last_row -= screen_cols - 1;
  last_row /= screen_cols;

  all_fits = (last_row - cursor_row < screen_rows);

  /*
   * Determine how many rows to scroll and in which direction (nb_rows
   * is positive when scrolling forward and negative when scrolling
   * backward).
   */

  if (all_fits)
    {
      if (cursor_row >= 0 &&
          last_row < screen_rows)
        /* Don't scroll if text from cursor to end is visible right now. */
        nb_rows = 0;
      else
        /* Scroll so that last row of edited line is at bottom of screen. */
        nb_rows = last_row - (screen_rows - 1);
    }
  else
    {
      if (cursor_row >= 0 &&
          cursor_row < screen_rows)
        /* Don't scroll if cursor is visible right now. */
        nb_rows = 0;
      else
        {
          /*
           * Scroll cursor row to bottom of screen unless this makes
           * the beginning of the edited line visible, in which case
           * align the beginning of the edited line to the top of the
           * screen.
           */

          if (cursor_row < first_row + (screen_rows - 1))
            nb_rows = first_row;
          else
            nb_rows = cursor_row - (screen_rows - 1);
        }
    }

  /*
   * Perform scrolling of screen.
   */

  if (nb_rows != 0)
    {
      screen_cursor -= screen_cols * nb_rows;

      if (nb_rows < 0)
        {
          /* scroll screen backward by whole rows */

          if (pos < screen_cursor)
            screen_cursor = pos; /* align start of buffer with screen home */

          start = pos - screen_cursor;

          n = screen_size; /* write to the whole screen */

          if ((e = lineedit_prepare_to_write_at (state, 0))
              != ___FIX(___NO_ERR))
            return e;
        }
      else
        {
          /* scroll screen forward by whole rows */

          start = screen_size + cursor_to_point;

          n = screen_cols * nb_rows; /* write characters that scroll in */

          if ((e = lineedit_prepare_to_write_at (state, screen_size))
              != ___FIX(___NO_ERR))
            return e;
        }

      if (!state->has_eat_newline_glitch)/*****************/
        n--; /* to avoid wraparound at end of screen */

      if (n > edited->length - start)
        n = edited->length - start;

      if ((e = lineedit_output_current_hist (state, start, n))
          != ___FIX(___NO_ERR))
        return e;

      if (all_fits)
        if ((e = lineedit_output_force_delayed_wrap (state))
            != ___FIX(___NO_ERR))
          return e;
    }

  if ((e = lineedit_move_screen_cursor (state, screen_cursor))
      != ___FIX(___NO_ERR))
    return e;

  state->current.edit_point = pos;
  state->current.cursor_to_point = pos - screen_cursor;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_update_region
   ___P((lineedit_state *state,
         int start,
         int end,
         ___BOOL allow_scrolling),
        (state,
         start,
         end,
         allow_scrolling)
lineedit_state *state;
int start;
int end;
___BOOL allow_scrolling;)
{
  ___WORD e;
  int screen_size = state->screen_size;
  int cursor_to_point = state->current.cursor_to_point;
  int pos = start - cursor_to_point;

  if (!state->has_eat_newline_glitch)/*****************/
    screen_size--; /* to avoid wraparound at end of screen */

  if (pos >= screen_size ||
      end - cursor_to_point <= 0)
    return ___FIX(___NO_ERR);

  if (pos < 0)
    pos = 0;

  start = pos + cursor_to_point;

  if ((e = lineedit_prepare_to_write_at (state, pos))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_output_current_hist (state, start, end - start);
}

___HIDDEN ___WORD lineedit_refresh
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;

  if ((e = lineedit_erase_screen (state))
      != ___FIX(___NO_ERR))
    return e;

  state->current.cursor_to_point = -state->current.screen_cursor;

  if ((e = lineedit_update_region (state, 0, edited->length, 1))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_move_edit_point (state, state->current.edit_point);
}

___HIDDEN ___BOOL lineedit_word_boundary
   ___P((lineedit_state *state,
         int back,
         int pos,
         int *final_pos),
        (state,
         back,
         pos,
         final_pos)
lineedit_state *state;
int back;
int pos;
int *final_pos;)
{
  extensible_string *edited = &state->current.hist->edited;
  tty_char *buf = edited->buffer;
  int len = edited->length;
  int depth = 0;
  tty_char in_string = UNICODE_NUL; /* assume not inside a string */
  tty_char c;

#define OPEN_PAREN(x) \
((x) == UNICODE_LPAREN || (x) == UNICODE_LBRACKET || (x) == UNICODE_LBRACE)

#define CLOSE_PAREN(x) \
((x) == UNICODE_RPAREN || (x) == UNICODE_RBRACKET || (x) == UNICODE_RBRACE)

#define STRING_DELIM(x) \
((x) == UNICODE_DOUBLEQUOTE || (x) == UNICODE_VBAR)

#define INT_VECTOR(x) \
((x) == UNICODE_LOWER_S || (x) == UNICODE_UPPER_S || \
 (x) == UNICODE_LOWER_U || (x) == UNICODE_UPPER_U)

#define FLOAT_VECTOR(x) \
((x) == UNICODE_LOWER_F || (x) == UNICODE_UPPER_F)

  if (back == 0)
    {
      /* skip whitespace */

      while (pos < len)
        {
          c = buf[pos];
          if (c <= UNICODE_SPACE)
            pos++;
          else
            break;
        }

      /* skip standard read-macros */

      while (pos < len)
        {
          c = buf[pos];
          if (c == UNICODE_QUOTE)
            pos++;
          else if (c == UNICODE_COMMA)
            {
              pos++;
              if (pos < len && buf[pos] == UNICODE_AT)
                pos++;
            }
          else
            break;
        }

      if (pos == len)
        {
          depth = -1; /* force error */
          goto done_forward;
        }

      /* handle head specially */

      c = buf[pos++];

      if (pos < len && c == UNICODE_SHARP)
        {
          /* handle #\xxx, #(...), #f32(...), etc */

          c = buf[pos++];
          if (c == UNICODE_BACKSLASH)
            {
              if (pos < len)
                pos++;
            }
          else if (OPEN_PAREN(c))
            depth++;
          else
            {
              ___BOOL f = FLOAT_VECTOR(c);
              if (pos < len && (f || INT_VECTOR(c)))
                {
                  c = buf[pos++];
                  if ((!f && pos < len && c == UNICODE_8) ||
                      (pos+1 < len &&
                       ((!f && c == UNICODE_1 && buf[pos] == UNICODE_6) ||
                        (c == UNICODE_3 && buf[pos] == UNICODE_2) ||
                        (c == UNICODE_6 && buf[pos] == UNICODE_4)) &&
                       (pos++, 1)))
                    {
                      c = buf[pos++];
                      if (OPEN_PAREN(c))
                        depth++;
                      else
                        pos--;
                    }
                }
            }
        }
      else if (OPEN_PAREN(c))
        depth++;
      else if (CLOSE_PAREN(c))
        {
          depth = -1; /* force error */
          goto done_forward;
        }
      else if (STRING_DELIM(c))
        in_string = c;

      while (pos < len)
        if (in_string == UNICODE_NUL)
          {
            c = buf[pos];

            if (c == UNICODE_SHARP)
              {
                if (pos+2 < len && buf[pos+1] == UNICODE_BACKSLASH)
                  pos += 2;
              }
            else if (c <= UNICODE_SPACE ||
                     c == UNICODE_SEMICOLON ||
                     c == UNICODE_QUOTE ||
                     c == UNICODE_COMMA)
              {
                if (depth == 0)
                  break;
              }
            else if (OPEN_PAREN(c))
              {
                if (depth == 0)
                  break;
                depth++;
              }
            else if (CLOSE_PAREN(c))
              {
                if (depth == 0)
                  break;
                depth--;
                if (depth == 0)
                  {
                    pos++;
                    break;
                  }
              }
            else if (STRING_DELIM(c))
              {
                if (depth == 0)
                  break;
                in_string = c;
              }

            pos++;
          }
        else
          {
            c = buf[pos++];

            if (c == UNICODE_BACKSLASH) /* character is escaped */
              {
                if (pos < len)
                  pos++;
              }
            else if (c == in_string) /* end of string reached */
              {
                if (depth == 0)
                  break;
                in_string = UNICODE_NUL;
              }
          }

      done_forward:;
    }
  else
    {
      /* skip whitespace */

      while (pos > 0)
        {
          c = buf[pos-1];
          if (c <= UNICODE_SPACE)
            pos--;
          else
            break;
        }

      if (pos == 0)
        {
          depth = -1; /* force error */
          goto done_backward;
        }

      /* handle tail specially */

      c = buf[--pos];

      if (pos >= 2 &&
          buf[pos-2] == UNICODE_SHARP &&
          buf[pos-1] == UNICODE_BACKSLASH)
        pos -= 2;
      else if (OPEN_PAREN(c))
        {
          depth = -1; /* force error */
          goto done_backward;
        }
      else if (CLOSE_PAREN(c))
        depth++;
      else if (STRING_DELIM(c))
        in_string = c;
      else if (c == UNICODE_QUOTE ||
               c == UNICODE_COMMA ||
               (pos >= 1 && c == UNICODE_AT && buf[pos-1] == UNICODE_COMMA))
        {
          pos++;
          goto skip_read_macros;
        }

      while (pos > 0)
        if (in_string == UNICODE_NUL)
          {
            c = buf[pos-1];

            if (pos >= 3 &&
                buf[pos-3] == UNICODE_SHARP &&
                buf[pos-2] == UNICODE_BACKSLASH)
              pos -= 2;
            else if (c <= UNICODE_SPACE ||
                     c == UNICODE_SEMICOLON ||
                     c == UNICODE_QUOTE ||
                     c == UNICODE_COMMA)
              {
                if (depth == 0)
                  break;
              }
            else if (pos >= 2 &&
                     c == UNICODE_AT &&
                     buf[pos-2] == UNICODE_COMMA)
              {
                if (depth == 0)
                  break;
                pos--;
              }
            else if (OPEN_PAREN(c))
              {
                if (depth == 0)
                  break;
                depth--;
                if (depth == 0)
                  {
                    pos--;
                    break;
                  }
              }
            else if (CLOSE_PAREN(c))
              {
                if (depth == 0)
                  break;
                depth++;
              }
            else if (STRING_DELIM(c))
              {
                if (depth == 0)
                  break;
                in_string = c;
              }

            pos--;
          }
        else
          {
            int i;

            i = --pos;

            while (i > 0 &&
                   buf[i-1] == UNICODE_BACKSLASH) /* find nb of backslashes */
              i--;

            if ((pos-i) & 1) /* character is escaped */
              pos = i;
            else if (buf[pos] == in_string) /* beginning of string reached */
              {
                if (depth == 0)
                  break;
                in_string = UNICODE_NUL;
              }
          }

      done_backward:;

      if (back == 2)
        {
          /* handle #(...), #f32(...), etc */

          if (pos < len)
            {
              c = buf[pos];
              if (OPEN_PAREN(c))
                {
                  if (pos >= 1 && buf[pos-1] == UNICODE_SHARP)
                    pos--;
                  else if (pos >= 3 && buf[pos-3] == UNICODE_SHARP)
                    {
                      c = buf[pos-2];
                      if (INT_VECTOR(c) && buf[pos-1] == UNICODE_8)
                        pos -= 3;
                    }
                  else if (pos >= 4 && buf[pos-4] == UNICODE_SHARP)
                    {
                      ___BOOL f;
                      c = buf[pos-3];
                      f = FLOAT_VECTOR(c);
                      if (f || INT_VECTOR(c))
                        {
                          c = buf[pos-2];
                          if ((!f &&
                               c == UNICODE_1 &&
                               buf[pos-1] == UNICODE_6) ||
                              (c == UNICODE_3 && buf[pos-1] == UNICODE_2) ||
                              (c == UNICODE_6 && buf[pos-1] == UNICODE_4))
                            pos -= 4;
                        }
                    }
                }
            }

          /* skip standard read-macros */

          skip_read_macros:

          while (pos > 0)
            {
              c = buf[pos-1];
              if (c == UNICODE_QUOTE || c == UNICODE_COMMA)
                pos--;
              else if (pos >= 2 &&
                       c == UNICODE_AT &&
                       buf[pos-2] == UNICODE_COMMA)
                pos -= 2;
              else
                break;
            }
        }
    }

  *final_pos = pos;

  return (depth == 0 && in_string == UNICODE_NUL);
}

___HIDDEN ___WORD lineedit_insert_chars
   ___P((lineedit_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
lineedit_state *state;
tty_char *buf;
int len;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  int edit_point = state->current.edit_point;
  int open_paren_point;
  tty_char c;

  if (len <= 0)
    return ___FIX(___NO_ERR);

  if ((e = extensible_string_insert (edited, edit_point, len, buf))
      != ___FIX(___NO_ERR))
    return e;

  if (state->current.mark_point > edit_point)
    state->current.mark_point += len;

  if ((e = lineedit_update_region (state, edit_point, edited->length, 1))
      != ___FIX(___NO_ERR))
    return e;

  edit_point += len;

  if (state->paren_balance_duration_nsecs > 0 &&
      state->ts->rlo >= state->ts->rhi &&
      (c = buf[len-1], CLOSE_PAREN(c)) &&
      lineedit_word_boundary (state, 1, edit_point, &open_paren_point) &&
      (c = edited->buffer[open_paren_point], OPEN_PAREN(c)))
    {
      if ((e = lineedit_move_edit_point (state, open_paren_point))
          != ___FIX(___NO_ERR))
        return e;

      if ((e = lineedit_output_drain (state)) != ___FIX(___NO_ERR))
        return e;

      state->current.edit_point = edit_point;
      state->current.paren_balance_trigger = 1;

      return ___FIX(___NO_ERR);
    }

  return lineedit_move_edit_point (state, edit_point);
}

___HIDDEN ___WORD lineedit_delete_chars
   ___P((lineedit_state *state,
         int pos,
         ___BOOL copy_to_clipboard),
        (state,
         pos,
         copy_to_clipboard)
lineedit_state *state;
int pos;
___BOOL copy_to_clipboard;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  int start;
  int end;
  int n;

  if (pos < 0 ||
      pos > edited->length)
    return ___FIX(___INVALID_OP_ERR);

  if (pos < state->current.edit_point)
    {
      start = pos;
      end = state->current.edit_point;
    }
  else
    {
      start = state->current.edit_point;
      end = pos;
    }

  n = end - start;

  if (n == 0)
    return ___FIX(___INVALID_OP_ERR);

  if (copy_to_clipboard &&
      ((e = lineedit_copy_to_clipboard (state, &edited->buffer[start], n))
       != ___FIX(___NO_ERR)))
    return e;

  if ((e = extensible_string_delete (edited, start, n)) != ___FIX(___NO_ERR))
    return e;

  if (state->current.mark_point >= start)
    {
      if (state->current.mark_point >= end)
        state->current.mark_point -= n;
      else
        state->current.mark_point = start;
    }

  if ((e = lineedit_update_region (state, start, edited->length + n, 1))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_move_edit_point (state, start);
}

___HIDDEN void lineedit_rotate_left_chars
   ___P((lineedit_state *state,
         int start,
         int end,
         int n),
        (state,
         start,
         end,
         n)
lineedit_state *state;
int start;
int end;
int n;)
{
  extensible_string *edited = &state->current.hist->edited;
  tty_char *p = &edited->buffer[start];
  tty_char temp;
  int len = end-start;
  int left = len;
  int i;
  int j;
  int k;

  i = 0;
  while (left > 0)
    {
      temp = p[i];
      j = i;
      k = (j+n) % len;
      left--;
      while (k != i)
        {
          p[j] = p[k];
          j = k;
          k = (j+n) % len;
          left--;
        }
      p[j] = temp;
      i++;
    }
}

___HIDDEN ___WORD lineedit_transpose_chars
   ___P((lineedit_state *state,
         int start1,
         int end1,
         int start2,
         int end2),
        (state,
         start1,
         end1,
         start2,
         end2)
lineedit_state *state;
int start1;
int end1;
int start2;
int end2;)
{
  ___WORD e;

  if (start1 > end1 ||
      end1 > start2 ||
      start2 > end2)
    return ___FIX(___INVALID_OP_ERR);

  lineedit_rotate_left_chars
    (state,
     start1,
     end2,
     end1-start1);

  lineedit_rotate_left_chars
    (state,
     start1,
     end2-(end1-start1),
     start2-end1);

  if ((e = lineedit_update_region (state, start1, end2, 1))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_move_edit_point (state, end2);
}

___HIDDEN ___WORD lineedit_move_history
   ___P((lineedit_state *state,
         lineedit_history *h),
        (state,
         h)
lineedit_state *state;
lineedit_history *h;)
{
  ___WORD e;
  int old_length;
  int new_length;

  if ((e = lineedit_history_begin_edit (state, h))
      != ___FIX(___NO_ERR))
    return e;

  if ((e = lineedit_move_edit_point (state, 0))
      != ___FIX(___NO_ERR))
    return e;

  old_length = state->current.hist->edited.length;
  new_length = h->edited.length;

  state->current.mark_point = 0;
  state->current.hist = h;

  if ((e = lineedit_update_region
             (state,
              0,
              (new_length > old_length) ? new_length : old_length,
              1))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_move_edit_point (state, new_length);
}

___HIDDEN ___WORD lineedit_move_history_relative
   ___P((lineedit_state *state,
         ___BOOL back),
        (state,
         back)
lineedit_state *state;
___BOOL back;)
{
  lineedit_history *h;
  lineedit_history *limit;

  if (back)
    {
      h = state->current.hist->prev;
      limit = state->hist_last;
    }
  else
    {
      h = state->current.hist->next;
      limit = state->hist_last->next;
    }

  if (h == limit)
    return ___FIX(___INVALID_OP_ERR);

  return lineedit_move_history (state, h);
}

___HIDDEN ___WORD lineedit_line_done
   ___P((lineedit_state *state,
         ___BOOL end_of_file),
        (state,
         end_of_file)
lineedit_state *state;
___BOOL end_of_file;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  extensible_string *input = &state->input;

  /*
   * Transfer the current edited line to the input.
   */

  if (input->buffer == NULL)
    {
      state->input_rlo = 0;

      if ((e = extensible_string_copy
                 (edited->buffer,
                  edited->length,
                  input,
                  1))
          != ___FIX(___NO_ERR))
        return e;

      if (!end_of_file)
        {
          int old_length = input->length;
          if ((e = extensible_string_set_length (input, old_length+1, 0))
              != ___FIX(___NO_ERR))
            {
              extensible_string_cleanup (input); /* ignore error */
              state->input.buffer = NULL;
              return e;
            }
          input->buffer[old_length] = UNICODE_LINEFEED;
        }
    }

  /*
   * Update history (empty lines are not added).
   */

  if (edited->length == 0 || end_of_file)
    {
      if ((e = lineedit_move_edit_point (state, edited->length))
          == ___FIX(___NO_ERR) &&
          (end_of_file ||
           (e = lineedit_newline (state))
           == ___FIX(___NO_ERR)))
        extensible_string_set_length
          (&state->hist_last->edited,
           0,
           1); /* ignore error */
    }
  else
    {
      /*
       * Make a copy of the current edited line for the history.
       */

      extensible_string actual;

      if ((e = extensible_string_copy
                 (edited->buffer,
                  edited->length,
                  &actual,
                  0))
          == ___FIX(___NO_ERR))
        {
          /*
           * Add the current edited line to the history.
           */

          lineedit_history *next_line;

          if ((e = lineedit_history_setup (state, &next_line))
              == ___FIX(___NO_ERR))
            {
              if ((e = lineedit_history_begin_edit (state, next_line))
                  == ___FIX(___NO_ERR) &&
                  (e = lineedit_move_edit_point (state, edited->length))
                  == ___FIX(___NO_ERR) &&
                  (e = lineedit_newline (state))
                  == ___FIX(___NO_ERR))
                {
                  extensible_string_cleanup
                    (&state->hist_last->actual); /* ignore error */
                  state->hist_last->actual = actual;
                  lineedit_history_add (state, next_line);
                  lineedit_history_trim (state);
                }
              else
                lineedit_history_cleanup (state, next_line); /* ignore error */
            }

          if (e != ___FIX(___NO_ERR))
            extensible_string_cleanup (&actual); /* ignore error */
        }
    }

  if (e == ___FIX(___NO_ERR))
    {
      /*
       * Revert history lines to their nonedited state.
       */

      lineedit_history *h = state->hist_last;

      while (1)
        {
          h = h->prev;
          if (h == state->hist_last)
            break;
          lineedit_history_end_edit (state, h);
        }

      /*
       * Prepare editing of next line.
       */

      state->current.edit_point = 0;
      state->current.mark_point = 0;
      state->current.cursor_to_point = -state->current.screen_cursor;
      state->current.hist = state->hist_last;
    }

  return e;
}

___HIDDEN ___WORD lineedit_end_paren_balance
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  if (state->current.paren_balance_in_progress)
    {
      state->current.paren_balance_in_progress = 0;
      return lineedit_move_edit_point
               (state,
                state->current.edit_point); /* ignore error */
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_process_single_event
   ___P((lineedit_state *state,
         lineedit_event *ev),
        (state,
         ev)
lineedit_state *state;
lineedit_event *ev;)
{
  extensible_string *edited = &state->current.hist->edited;

  switch (ev->event_kind)
    {
    case LINEEDIT_EVENT_NONE:
      return ___FIX(___NO_ERR);

    case LINEEDIT_EVENT_KEY:
      if (ev->event_char < UNICODE_SPACE || /* discard control characters */
          ev->event_char == UNICODE_RUBOUT)
        return ___FIX(___INVALID_OP_ERR);
      return lineedit_insert_chars (state, &ev->event_char, 1);

    case LINEEDIT_EVENT_RETURN:
      return lineedit_line_done (state, 0);

    case LINEEDIT_EVENT_BACK:
      return lineedit_delete_chars (state, state->current.edit_point-1, 0);

    case LINEEDIT_EVENT_BACK_WORD:
      {
        int i;
        if (!lineedit_word_boundary (state, 2, state->current.edit_point, &i))
          return ___FIX(___INVALID_OP_ERR);
        return lineedit_delete_chars (state, i, 0);
      }

    case LINEEDIT_EVENT_TAB:
      return lineedit_insert_chars
               (state,
                lem.tab_spaces,
                state->tab_stop-(state->current.edit_point%state->tab_stop));

    case LINEEDIT_EVENT_MARK:
      state->current.mark_point = state->current.edit_point;
      return ___FIX(___NO_ERR);

    case LINEEDIT_EVENT_PASTE:
      return lineedit_paste_from_clipboard (state);

    case LINEEDIT_EVENT_CUT:
      return lineedit_delete_chars (state, state->current.mark_point, 1);

    case LINEEDIT_EVENT_CUT_RIGHT:
      return lineedit_delete_chars (state, edited->length, 1);

    case LINEEDIT_EVENT_CUT_LEFT:
      return lineedit_delete_chars (state, 0, 1);

    case LINEEDIT_EVENT_REFRESH:
      return lineedit_refresh (state);

    case LINEEDIT_EVENT_TRANSPOSE:
      {
        int i = state->current.edit_point;
        if (i <= 0 ||
            edited->length < 2)
          return ___FIX(___INVALID_OP_ERR);
        if (i == edited->length)
          return lineedit_transpose_chars (state, i-2, i-1, i-1, i);
        else
          return lineedit_transpose_chars (state, i-1, i, i, i+1);
      }

    case LINEEDIT_EVENT_TRANSPOSE_WORD:
      {
        int start1;
        int end1;
        int start2;
        int end2;
        if (!lineedit_word_boundary
               (state,
                0,
                state->current.edit_point,
                &end2) ||
            !lineedit_word_boundary (state, 2, end2, &start2) ||
            !lineedit_word_boundary (state, 2, start2, &start1) ||
            !lineedit_word_boundary (state, 0, start1, &end1))
          return ___FIX(___INVALID_OP_ERR);
        return lineedit_transpose_chars (state, start1, end1, start2, end2);
      }

    case LINEEDIT_EVENT_UP:
      return lineedit_move_history_relative (state, 1);

    case LINEEDIT_EVENT_DOWN:
      return lineedit_move_history_relative (state, 0);

    case LINEEDIT_EVENT_RIGHT:
      return lineedit_move_edit_point (state, state->current.edit_point+1);

    case LINEEDIT_EVENT_RIGHT_WORD:
      {
        int i;
        lineedit_word_boundary (state, 0, state->current.edit_point, &i);
        return lineedit_move_edit_point (state, i);
      }

    case LINEEDIT_EVENT_LEFT:
      return lineedit_move_edit_point (state, state->current.edit_point-1);

    case LINEEDIT_EVENT_LEFT_WORD:
      {
        int i;
        lineedit_word_boundary (state, 2, state->current.edit_point, &i);
        return lineedit_move_edit_point (state, i);
      }

    case LINEEDIT_EVENT_HOME:
      return lineedit_move_edit_point (state, 0);

    case LINEEDIT_EVENT_HOME_DOC:
      return lineedit_move_history (state, state->hist_last->next);

    case LINEEDIT_EVENT_INSERT:
{ tty_char c = '*';/****************/
  return extensible_string_insert
             (edited,
              edited->length,
              1,
              &c);
}
      return ___FIX(___INVALID_OP_ERR);

    case LINEEDIT_EVENT_DELETE:
      if (edited->length == 0)
        return lineedit_line_done (state, 1);
      else
        return lineedit_delete_chars (state, state->current.edit_point+1, 0);

    case LINEEDIT_EVENT_DELETE_WORD:
      {
        int i;
        if (!lineedit_word_boundary (state, 0, state->current.edit_point, &i))
          return ___FIX(___INVALID_OP_ERR);
        return lineedit_delete_chars (state, i, 0);
      }

    case LINEEDIT_EVENT_END:
      return lineedit_move_edit_point (state, edited->length);

    case LINEEDIT_EVENT_END_DOC:
      return lineedit_move_history (state, state->hist_last);

    case LINEEDIT_EVENT_F1:
    case LINEEDIT_EVENT_META_F1:
    case LINEEDIT_EVENT_F2:
    case LINEEDIT_EVENT_META_F2:
    case LINEEDIT_EVENT_F3:
    case LINEEDIT_EVENT_META_F3:
    case LINEEDIT_EVENT_F4:
    case LINEEDIT_EVENT_META_F4:
#ifdef LINEEDIT_SUPPORT_F5_TO_F12
    case LINEEDIT_EVENT_F5:
    case LINEEDIT_EVENT_META_F5:
    case LINEEDIT_EVENT_F6:
    case LINEEDIT_EVENT_META_F6:
    case LINEEDIT_EVENT_F7:
    case LINEEDIT_EVENT_META_F7:
    case LINEEDIT_EVENT_F8:
    case LINEEDIT_EVENT_META_F8:
    case LINEEDIT_EVENT_F9:
    case LINEEDIT_EVENT_META_F9:
    case LINEEDIT_EVENT_F10:
    case LINEEDIT_EVENT_META_F10:
    case LINEEDIT_EVENT_F11:
    case LINEEDIT_EVENT_META_F11:
    case LINEEDIT_EVENT_F12:
    case LINEEDIT_EVENT_META_F12:
#endif
      break;
    }

  return ___FIX(___INVALID_OP_ERR);
}

___HIDDEN ___WORD lineedit_process_events
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
#ifdef LINEEDIT_WITH_NONBLOCKING_IO
#define AGGREGATION_BUFFER_SIZE (80*50)
#else
#define AGGREGATION_BUFFER_SIZE 1
#endif

  ___WORD e1;
  ___WORD e2;
  tty_char aggregation_buffer[AGGREGATION_BUFFER_SIZE];
  lineedit_event ev;
  int i;
  struct lineedit_state_undo_struct previous;

  /*
   * Process events only if no input line is currently available.
   */

  if (state->input.buffer != NULL)
    return ___FIX(___NO_ERR);

  /*
   * Process events only if no line editing output is pending.
   */

  if ((e1 = lineedit_output_drain (state)) != ___FIX(___NO_ERR))
    return e1;

  /*
   * Save state so that we can undo if there is an error.
   */

  previous = state->current;

  i = 0;

  ev.event_kind = LINEEDIT_EVENT_NONE;

  state->current.paren_balance_trigger = 0;

  while (1)
    {
      if ((e1 = lineedit_get_event (state, &ev)) != ___FIX(___NO_ERR) ||
          ev.event_kind == LINEEDIT_EVENT_NONE)
        break;

      if (ev.event_kind != LINEEDIT_EVENT_KEY ||
          ev.event_char < UNICODE_SPACE ||
          ev.event_char == UNICODE_RUBOUT)
        break;

      aggregation_buffer[i++] = ev.event_char;

      ev.event_kind = LINEEDIT_EVENT_NONE;

      if (i == AGGREGATION_BUFFER_SIZE)
        break;
    }

  e2 = lineedit_insert_chars (state, aggregation_buffer, i);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  e2 = lineedit_process_single_event (state, &ev);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  if (e1 == ___FIX(___INVALID_OP_ERR))
    {
      /*
       * Undo line editing operation.
       */

      tty_char c = UNICODE_BELL;

      extensible_string_set_length
        (&state->output,
         0,
         1); /* ignore error */

      state->output_wlo = state->output.length;

      state->current = previous;

      e1 = extensible_string_insert
             (&state->output,
              state->output.length,
              1,
              &c);
    }

  if (state->current.paren_balance_trigger)
    {
      /*
       * Parenthesis balancing started in this round of event
       * processing.
       */

      ___time duration;

      ___time_get_current_time (&state->current.paren_balance_end);
      ___time_from_nsecs (&duration, 0, state->paren_balance_duration_nsecs);
      ___time_add (&state->current.paren_balance_end, duration);

      state->current.paren_balance_in_progress = 1;

#ifndef LINEEDIT_WITH_NONBLOCKING_IO
      absolute_time_sleep (state->current.paren_balance_end);
      lineedit_end_paren_balance (state); /* ignore error */
#endif
    }
  else if (state->current.paren_balance_in_progress)
    {
      /*
       * Parenthesis balancing did not start in this round of event
       * processing and parenthesis balancing is in progress, check if
       * parenthesis balancing should end.
       */

      if (i > 0 || ev.event_kind != LINEEDIT_EVENT_NONE)
        lineedit_end_paren_balance (state); /* ignore error */
#ifdef LINEEDIT_WITH_NONBLOCKING_IO
      else
        {
          ___time now;
          ___time_get_current_time (&now);
          if (!___time_less (now, state->current.paren_balance_end))
            lineedit_end_paren_balance (state); /* ignore error */
        }
#endif
    }

  e2 = lineedit_output_drain (state);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  return e1;
}

___HIDDEN ___WORD lineedit_read
   ___P((lineedit_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
lineedit_state *state;
tty_char *buf;
int len;)
{
  ___WORD e;
  int rlo;
  int line_length;
  int avail;

  if ((e = lineedit_process_events (state)) != ___FIX(___NO_ERR))
    return e;

  if (state->input.buffer == NULL)
    return ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));

  rlo = state->input_rlo;
  line_length = state->input.length;
  avail = line_length - rlo;

  if (avail > 0)
    {
      tty_char *p = &state->input.buffer[rlo];
      tty_char *end;

      if (len > avail)
        len = avail;

      end = p + len;

      while (p < end)
        *buf++ = *p++;

      rlo += len;
    }
  else
    len = 0;

  state->input_rlo = rlo;

  if (rlo >= line_length)
    {
      extensible_string_cleanup (&state->input); /* ignore error */
      state->input.buffer = NULL;
    }

  return ___FIX(len);
}

___HIDDEN ___WORD lineedit_module_setup ___PVOID
{
  int i;

  for (i=0; i<LINEEDIT_MAX_TAB_STOP; i++)
    lem.tab_spaces[i] = UNICODE_SPACE;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_module_cleanup ___PVOID
{
  return ___FIX(___NO_ERR);
}

#endif

/*---------------------------------------------------------------------------*/

/* Tty management routines */

#define TTY_BIG_SCREEN_SIZE (80*50)

___HIDDEN void tty_broadcast_signal
   ___P((tty_signal s),
        (s)
tty_signal s;)
{
  tty_state *list = tm.list;

  if (list != NULL)
    {
      tty_state *state = list;
      do
        {
          state->received_signal[s] = 1;
          state = state->next;
        } while (state != list);
    }
}

___HIDDEN ___WORD tty_setup_common
   ___P((tty *t,
         tty_state *state),
        (t,
         state)
tty *t;
tty_state *state;)
{
  ___WORD e;
  tty_state *dest;

  state->refcount = 1;

  state->input_mode = LINEEDIT_INPUT_MODE;
  state->term_mode = INITIAL_TERM_MODE;

  state->rlo = 0;
  state->rhi = 0;
  state->wlo = 0;
  state->whi = 0;
  state->werr = ___FIX(___NO_ERR);

  state->received_signal[USER_INTERRUPT_SIGNAL] = 0;

#ifdef USE_LINEEDIT

  /*
   * Setup line editing state.
   */

  if ((e = lineedit_setup (&state->les, state)) != ___FIX(___NO_ERR))
    {
      ___free_mem (state);
      return e;
    }

  /*
   * Switch to CBREAK mode so that terminal driver does not perform
   * its own line editing.
   */

  if ((e = tty_set_term_mode (state, CBREAK_TERM_MODE))
      != ___FIX(___NO_ERR))
    return e;

#endif

  /* add tty_state to tail of circular list of tty_state's */

  dest = tm.list;

  if (dest == NULL)
    {
      state->prev = state;
      state->next = state;
      tm.list = state;
    }
  else
    {
      tty_state *before_dest = dest->prev;
      state->next = dest;
      state->prev = before_dest;
      dest->prev = state;
      before_dest->next = state;
    }

  t->state = state;

  return ___FIX(___NO_ERR);
}

#ifndef USE_POSIX
#ifndef USE_WIN32
#ifndef USE_MACOS

/**************************/

#endif
#endif
#endif

#ifdef USE_POSIX

___HIDDEN ___BOOL tty_equal
   ___P((int fd1,
         int fd2),
        (fd1,
         fd2)
int fd1;
int fd2;)
{
  struct stat fd1_stat;
  struct stat fd2_stat;

  if (fd1 == fd2)
    return 1;

  /* check if st_rdev fields are the same */

  if (fstat (fd1, &fd1_stat) == 0 &&
      fstat (fd2, &fd2_stat) == 0 &&
      fd1_stat.st_rdev == fd2_stat.st_rdev)
    return 1;

  /* check if the descriptors both refer to the controlling terminal */

  if (tcgetpgrp (fd1) >= 0 &&
      tcgetpgrp (fd2) >= 0)
    return 1;

  return 0;
}

___HIDDEN ___WORD tty_setup_os_specific
   ___P((tty *t,
         int fd),
        (t,
         fd)
tty *t;
int fd;)
{
  tty_state *state;

  t->state = NULL;

  /* check if the same terminal was previously opened */

  state = tm.list;

  if (state != NULL)
    {
      do
        {
          if (tty_equal (fd, state->fd))
            {
/******************              close (fd); /* ignore error */
              state->refcount++;
              t->state = state;
              return ___FIX(___NO_ERR);
            }
          state = state->next;
        } while (state != tm.list);
    }

  /* need to allocate a new tty_state */

  state = ___CAST(tty_state*,___alloc_mem (sizeof (tty_state)));

  if (state == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  state->received_signal[WINDOW_SIZE_SIGNAL] = 1; /* query window size */
  state->received_signal[CONTINUE_SIGNAL] = 0;

  state->fd = fd;

  if (tcgetattr (state->fd, &state->initial_termios) < 0)
    {
      ___WORD e = ___FIX(err_code_from_errno ());
      ___free_mem (state);
      return e;
    }

  state->encoding = LATIN1_ENCODING;

  return tty_setup_common (t, state);
}

___HIDDEN ___WORD tty_cleanup_os_specific
   ___P((tty_state *state),
        (state)
tty_state *state;)
{
  ___WORD e1 = ___FIX(___NO_ERR);

  if (close (state->fd) < 0)
    e1 = ___FIX(err_code_from_errno ());

  return e1;
}

___HIDDEN ___WORD tty_force_open
   ___P((tty *t),
        (t)
tty *t;)
{
  ___WORD e;
  int fd;
  char term_name[L_ctermid];
  char *term_type;
  tty_state *state = t->state;

  if (state != NULL)
    return ___FIX(___NO_ERR);

  /* tty refers to the console */

  ctermid (term_name); /* get controlling terminal's name */

#if 0
  strcpy (term_name, "/dev/pts/3");/***********************/
#endif

#ifdef LINEEDIT_WITH_NONBLOCKING_IO
  fd = open (term_name, O_RDWR|O_NONBLOCK);
#else
  fd = open (term_name, O_RDWR);
#endif

  if (fd < 0)
    return ___FIX(err_code_from_errno ());

  if ((e = tty_setup_os_specific (t, fd)) != ___FIX(___NO_ERR))
    close (fd); /* ignore error */

  term_type = getenv (CONSOLE_TERMINAL_TYPE_ENV_VAR);

  if ((e = tty_set_terminal_type (t, term_type, DEFAULT_EMACS_BINDINGS))
      != ___FIX(___NO_ERR) ||
      (e = tty_set_encoding (t, NATIVE_ENCODING))
      != ___FIX(___NO_ERR))
    tty_cleanup (t); /* ignore error */

  return e;
}

___HIDDEN ___WORD tty_set_term_mode
   ___P((tty_state *state,
         tty_term_mode term_mode),
        (state,
         term_mode)
tty_state *state;
tty_term_mode term_mode;)
{
  struct termios new_termios;

  new_termios = state->initial_termios;

  switch (term_mode)
    {
    case CBREAK_TERM_MODE:
      new_termios.c_lflag &= ~(ECHO | ICANON);
      new_termios.c_cc[VMIN] = 0;
      new_termios.c_cc[VTIME] = 0;
      break;

    case RAW_TERM_MODE:
      new_termios.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
      new_termios.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON | IXOFF);
      new_termios.c_cflag &= ~(CSIZE | PARENB);
      new_termios.c_cflag |= CS8;
      new_termios.c_oflag &= ~(OPOST);
      new_termios.c_cc[VMIN] = 0;
      new_termios.c_cc[VTIME] = 0;
      break;
    }

  if (tcsetattr (state->fd, TCSAFLUSH, &new_termios) < 0)
    return ___FIX(err_code_from_errno ());

  state->term_mode = term_mode;

#ifdef USE_LINEEDIT
  state->les->linefeed_moves_to_left_margin =
    ((new_termios.c_oflag & (OPOST | ONLCR)) == (OPOST | ONLCR));
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN void tty_signal_handler
   ___P((int sig),
        (sig)
int sig;)
{
  signal (sig, tty_signal_handler);

  switch (sig)
    {
    case SIGINT:
      tty_broadcast_signal (USER_INTERRUPT_SIGNAL);
      if (tm.user_interrupt_handler != NULL)
        tm.user_interrupt_handler ();
      break;

    case SIGWINCH:
      tty_broadcast_signal (WINDOW_SIZE_SIGNAL);
      {/**************/
        struct winsize size;
        if (ioctl (0, TIOCGWINSZ, &size) < 0)
          ;/************/
        printf ("%d rows, %d columns\n", size.ws_row, size.ws_col);
        fflush (stdout);
      }
      break;

    case SIGCONT:
      tty_broadcast_signal (CONTINUE_SIGNAL);
      tty_set_term_mode (tm.console.state, tm.console.state->term_mode);
      printf ("continued\n");
      fflush (stdout);
      break;
    }
}

___HIDDEN ___WORD tty_read_os_specific
   ___P((tty_state *state,
         ___U8 *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
___U8 *buf;
int len;)
{
  int n;

  if ((n = read (state->fd, buf, len)) < 0)
    return ___FIX(err_code_from_errno ());

  return ___FIX(n);
}

___HIDDEN ___WORD tty_write_os_specific
   ___P((tty_state *state,
         ___U8 *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
___U8 *buf;
int len;)
{
  int n;

#if 1
  n = write (state->fd, buf, len);
#else
  /*****************************/
  {
    static last_time = 0;
    int now;
    struct timeval tv;
    gettimeofday (&tv, NULL);
    now = tv.tv_usec / 33000; /* simulate 300 baud */
    if (now == last_time)
      n = 0;
    else
      {
        int x = len;
        if (x >= 1)
          x = 1;
        n = write (state->fd, buf, x);
        if (n == 1)
          {
#if 0
            printf ("%c", *buf); fflush(stdout);
#endif
          }
        last_time = now;
      }
  }
#endif

  if (n < 0)
    return ___FIX(err_code_from_errno ());

  return ___FIX(n);
}

___HIDDEN ___WORD tty_module_setup_os_specific ___PVOID
{
  signal (SIGINT, tty_signal_handler);
  signal (SIGWINCH, tty_signal_handler);
  signal (SIGCONT, tty_signal_handler);

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_module_cleanup_os_specific ___PVOID
{
  signal (SIGINT, SIG_DFL);
  signal (SIGWINCH, SIG_DFL);
  signal (SIGCONT, SIG_DFL);

  return ___FIX(___NO_ERR);
}

#endif

#ifdef USE_WIN32

___HIDDEN ___WORD tty_setup_os_specific
   ___P((tty *t,
         HANDLE input,
         HANDLE output),
        (t,
         input,
         output)
tty *t;
HANDLE input;
HANDLE output;)
{
  tty_state *state;

  t->state = NULL;

  /* check if the same terminal was previously opened */

  state = tm.list;

  if (state != NULL)
    {
      /* Windows only allows one console per process. */

      CloseHandle (input); /* ignore error */
      CloseHandle (output); /* ignore error */
      state->refcount++;
      t->state = state;
      return ___FIX(___NO_ERR);
    }

  /* need to allocate a new tty_state */

  state = ___CAST(tty_state*,___alloc_mem (sizeof (tty_state)));

  if (state == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  state->input = input;
  state->output = output;
  state->input_rlo = 0;
  state->input_rhi = 0;

  {
    DWORD n;
    if (WriteConsoleW (output, NULL, 0, &n, NULL))
      state->encoding = UCS2LE_ENCODING;
    else
      state->encoding = LATIN1_ENCODING;
  }

  return tty_setup_common (t, state);
}

___HIDDEN ___WORD tty_cleanup_os_specific
   ___P((tty_state *state),
        (state)
tty_state *state;)
{
  ___WORD e1 = ___FIX(___NO_ERR);
  ___WORD e2;

  if (!CloseHandle (state->input))
    e1 = ___FIX(err_code_from_GetLastError ());

  if (!CloseHandle (state->output))
    {
      e2 = ___FIX(err_code_from_GetLastError ());
      if (e1 == ___FIX(___NO_ERR))
        e1 = e2;
    }

  return e1;
}

___HIDDEN ___WORD tty_force_open
   ___P((tty *t),
        (t)
tty *t;)
{
  tty_state *state = t->state;

  if (state != NULL)
    return ___FIX(___NO_ERR);

  /* tty refers to the console */

  /********************/

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_set_term_mode
   ___P((tty_state *state,
         tty_term_mode term_mode),
        (state,
         term_mode)
tty_state *state;
tty_term_mode term_mode;)
{
  state->term_mode = term_mode;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_read_os_specific
   ___P((tty_state *state,
         ___U8 *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
___U8 *buf;
int len;)
{
  int i = 0;

  while (state->input_rlo >= state->input_rhi)
    {
      DWORD n;

#ifdef LINEEDIT_WITH_NONBLOCKING_IO

      if (!GetNumberOfConsoleInputEvents (state->input, &n))
        return ___FIX(err_code_from_GetLastError ());

      if (n <= 0)
        goto done;

      if (n > INPUT_RBUF_SIZE)
        n = INPUT_RBUF_SIZE;

#else

      n = INPUT_RBUF_SIZE;

#endif

      if (!((state->encoding == UCS2LE_ENCODING)
            ? ReadConsoleInputW
                (state->input,
                 state->input_rbuf,
                 n,
                 &n)
            : ReadConsoleInputA
                (state->input,
                 state->input_rbuf,
                 n,
                 &n)))
        return ___FIX(err_code_from_GetLastError ());

      state->input_rlo = 0;
      state->input_rhi = n;
    }

  while (state->input_rlo < state->input_rhi)
    {
      INPUT_RECORD *ir = &state->input_rbuf[state->input_rlo];

      switch (ir->EventType)
        {
        case KEY_EVENT:
          {
            DWORD ctrl_keys;
            ___BOOL ctrl;
            ___BOOL alt;
            ___BOOL shift;
            char *seq;
            ___UCS4 c;

            if (!(ir->Event.KeyEvent.bKeyDown &&
                  ir->Event.KeyEvent.wRepeatCount > 0))
              {
                state->input_rlo++;
                break;
              }

            ctrl_keys = ir->Event.KeyEvent.dwControlKeyState;
            ctrl = ctrl_keys & (LEFT_CTRL_PRESSED|RIGHT_CTRL_PRESSED);
            alt = ctrl_keys & (LEFT_ALT_PRESSED|RIGHT_ALT_PRESSED);
            shift = ctrl_keys & SHIFT_PRESSED;

            if (state->encoding == UCS2LE_ENCODING)
              c = ir->Event.KeyEvent.uChar.UnicodeChar;
            else
              c = ir->Event.KeyEvent.uChar.AsciiChar;

            seq = NULL;

            if (!ctrl && !alt && !shift && c == 0)
              switch (ir->Event.KeyEvent.wVirtualKeyCode)
                {
                case VK_BACK:   seq = "\010";     break;
                case VK_UP:     seq = "\033[A";   break;
                case VK_DOWN:   seq = "\033[B";   break;
                case VK_RIGHT:  seq = "\033[C";   break;
                case VK_LEFT:   seq = "\033[D";   break;
                case VK_HOME:   seq = "\033[H";   break;
                case VK_INSERT: seq = "\033[2~";  break;
                case VK_DELETE: seq = "\177";     break;
                case VK_END:    seq = "\033[F";   break;
                case VK_PRIOR:  seq = "\033[5~";  break;
                case VK_NEXT:   seq = "\033[6~";  break;
                case VK_F1:     seq = "\033OP";   break;
                case VK_F2:     seq = "\033OQ";   break;
                case VK_F3:     seq = "\033OR";   break;
                case VK_F4:     seq = "\033OS";   break;
#ifdef LINEEDIT_SUPPORT_F5_TO_F12
                case VK_F5:     seq = "\033[15~"; break;
                case VK_F6:     seq = "\033[17~"; break;
                case VK_F7:     seq = "\033[18~"; break;
                case VK_F8:     seq = "\033[19~"; break;
                case VK_F9:     seq = "\033[20~"; break;
                case VK_F10:    seq = "\033[21~"; break;
                case VK_F11:    seq = "\033[23~"; break;
                case VK_F12:    seq = "\033[24~"; break;
#endif
                }

            if (seq == NULL)
              {
                if (c == 0)
                  {
                    ir->Event.KeyEvent.wRepeatCount = 0;
                    break;
                  }

                if (ctrl && c == UNICODE_SPACE)
                  c = UNICODE_NUL;

                if (state->encoding == UCS2LE_ENCODING)
                  {
                    if (len < i + 2)
                      goto done;
                    buf[i++] = c & 0xff;
                    buf[i++] = c >> 8;
                  }
                else
                  {
                    if (len < i + 1)
                      goto done;
                    buf[i++] = c & 0xff;
                  }
              }
            else
              {
                while (*seq != '\0')
                  {
                    if (state->encoding == UCS2LE_ENCODING)
                      {
                        if (len < i + 2)
                          goto done;
                        buf[i++] = *seq++;
                        buf[i++] = 0;
                      }
                    else
                      {
                        if (len < i + 1)
                          goto done;
                        buf[i++] = *seq++;
                      }
                  }
              }

            ir->Event.KeyEvent.wRepeatCount--;

            break;
          }

        case MOUSE_EVENT:
          {
            ___WORD e;

            if (i > 0)
              goto done; /* in case paste returns an error */

            state->input_rlo++;

            if (ir->Event.MouseEvent.dwEventFlags == 0 &&
                (ir->Event.MouseEvent.dwButtonState ==
                 (FROM_LEFT_1ST_BUTTON_PRESSED|RIGHTMOST_BUTTON_PRESSED) ||
                 ir->Event.MouseEvent.dwButtonState ==
                 FROM_LEFT_2ND_BUTTON_PRESSED))
              if ((e = lineedit_paste_from_clipboard (state->les))
                  != ___FIX(___NO_ERR))
                return e;

            goto done; /* pasted text is available to be read */
          }

        default:
          state->input_rlo++;
          break;
        }
    }

  done:

  if (i == 0)
    return ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));

  return ___FIX(i);
}

___HIDDEN ___WORD tty_write_os_specific
   ___P((tty_state *state,
         ___U8 *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
___U8 *buf;
int len;)
{
  DWORD n;

  if (state->encoding == UCS2LE_ENCODING)
    {
      len = len >> 1;
      if (!WriteConsoleW (state->output, (wchar_t*)buf, len, &n, NULL))
        return ___FIX(err_code_from_GetLastError ());
      n = n << 1;
    }
  else
    {
      if (!WriteConsoleA (state->output, (char*)buf, len, &n, NULL))
        return ___FIX(err_code_from_GetLastError ());
    }

  return ___FIX(n);
}

___HIDDEN BOOL WINAPI tty_console_event_handler
   ___P((DWORD dwCtrlType),
        (dwCtrlType)
DWORD dwCtrlType;)
{
  tty_state *state = tm.console.state;

  switch (dwCtrlType)
    {
    case CTRL_C_EVENT:
      {
        if (state != NULL &&
            state->term_mode != CBREAK_TERM_MODE)
          return TRUE;

        tty_broadcast_signal (USER_INTERRUPT_SIGNAL);

        if (tm.user_interrupt_handler != NULL)
          tm.user_interrupt_handler ();

        return TRUE;
      }

    case CTRL_BREAK_EVENT:
      {
        if (state != NULL &&
            state->term_mode == RAW_TERM_MODE)
          return TRUE;

        break;
      }

    case CTRL_CLOSE_EVENT:
    case CTRL_LOGOFF_EVENT:
    case CTRL_SHUTDOWN_EVENT:
      break;
    }

  return FALSE; /* let system handle the console event */
}

___HIDDEN ___WORD tty_module_setup_os_specific ___PVOID
{
  if (!SetConsoleCtrlHandler (tty_console_event_handler, TRUE))
    return ___FIX(err_code_from_GetLastError ());

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_module_cleanup_os_specific ___PVOID
{
  SetConsoleCtrlHandler (tty_console_event_handler, FALSE); /* ignore error */
  return ___FIX(___NO_ERR);
}

#endif

___HIDDEN ___WORD tty_setup_console
   ___P((tty *t),
        (t)
tty *t;)
{
  t->state = NULL;
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_cleanup
   ___P((tty *t),
        (t)
tty *t;)
{
  ___WORD e1;
  ___WORD e2;
  tty_state *state = t->state;

  t->state = NULL; /* avoid dangling pointer */

  if (state == NULL || --state->refcount != 0)
    return ___FIX(___NO_ERR);

#ifdef USE_LINEEDIT
  e1 = lineedit_cleanup (state->les);
#else
  e1 = ___FIX(___NO_ERR);
#endif

  e2 = tty_set_term_mode (state, INITIAL_TERM_MODE);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  e2 = tty_cleanup_os_specific (state);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  /* remove tty_state from circular list of tty_state's */

  if (state->next == state) /* only tty_state in list? */
    tm.list = NULL;
  else
    {
      if (tm.list == state)
        tm.list = state->next;
      state->next->prev = state->prev;
      state->prev->next = state->next;
    }

  ___free_mem (state);

  return e1;
}

___HIDDEN ___WORD tty_set_terminal_type
   ___P((tty *t,
         char *terminal_type,
         ___BOOL emacs_bindings),
        (t,
         terminal_type,
         emacs_bindings)
tty *t;
char *terminal_type;
___BOOL emacs_bindings;)
{
  ___WORD e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT
  e = lineedit_set_terminal_type (state->les, terminal_type, emacs_bindings);
#endif

  return e;
}

___HIDDEN ___WORD tty_set_encoding
   ___P((tty *t,
         tty_encoding encoding),
        (t,
         encoding)
tty *t;
tty_encoding encoding;)
{
  ___WORD e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

  state->encoding = encoding;

  return e;
}

___HIDDEN ___WORD tty_set_input_mode
   ___P((tty *t,
         tty_input_mode input_mode),
        (t,
         input_mode)
tty *t;
tty_input_mode input_mode;)
{
  ___WORD e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

  if (input_mode & RAW_INPUT_MODE)
    e = tty_set_term_mode (state, RAW_TERM_MODE);
  else
    e = tty_set_term_mode (state, CBREAK_TERM_MODE);

  if (e == ___FIX(___NO_ERR))
    state->input_mode = input_mode;

  return e;
}

___HIDDEN ___WORD tty_set_paren_balance_duration
   ___P((tty *t,
         int nsecs),
        (t,
         nsecs)
tty *t;
int nsecs;)
{
  ___WORD e;

  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT
  e = lineedit_set_paren_balance_duration (state->les, nsecs);
#endif

  return e;
}

___HIDDEN ___WORD tty_set_max_history_length
   ___P((tty *t,
         int max_history_length),
        (t,
         max_history_length)
tty *t;
int max_history_length;)
{
  ___WORD e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT
  lineedit_set_max_history_length (state->les, max_history_length);
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_set_text_attrs
   ___P((tty *t,
         tty_text_attrs input_attrs,
         tty_text_attrs output_attrs),
        (t,
         input_attrs,
         output_attrs)
tty *t;
tty_text_attrs input_attrs;
tty_text_attrs output_attrs;)
{
  ___WORD e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT
  state->les->input_attrs = input_attrs;
  state->les->output_attrs = output_attrs;
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_read_from_rbuf
   ___P((tty_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
tty_char *buf;
int len;)
{
  int rlo = state->rlo;
  int rhi = state->rhi;
  ___U8 *rbuf = state->rbuf;
  int n = len;

  switch (state->encoding)
    {
    case ASCII_ENCODING:
      {
        ___U8 byte;

        while (rlo < rhi && n > 0)
          {
            byte = rbuf[rlo++];
            if (byte > 0x7f)
              {
                rbuf[--rlo] = byte & 0x7f;
                *buf++ = UNICODE_ESCAPE;
              }
            else
              *buf++ = byte;
            n--;
          }
        break;
      }

    default: /* in case state->encoding is set incorrectly */
    case LATIN1_ENCODING:
      while (rlo < rhi && n > 0)
        {
          *buf++ = rbuf[rlo++];
          n--;
        }
      break;

    case UTF8_ENCODING:
      {
        ___U8 byte;
        ___UCS4 c;

        while (rlo < rhi && n > 0)
          {
            int start = rlo; /* in case we need to backtrack */

            byte = rbuf[rlo++];

            if (byte <= 0x7f)
              c = byte; /* ASCII character */
            else if (byte <= 0xbf || byte > 0xfd)
              c = TTY_CHAR_MAX+1; /* illegal first byte of a UTF-8 */
            else
              {
                ___U8 next;
                ___UCS4 range;

                c = byte; /* upper bits are removed later */
                range = 1<<6;

                loop:
                if (rlo >= rhi)
                  {
                    /*
                     * Not enough bytes to form a UTF-8 encoded
                     * character, leave them there for later.
                     */
                    rlo = start;
                    break; /* stop short */
                  }
                next = rbuf[rlo++];
                if (next <= 0x7f || next > 0xbf)
                  {
                    /*
                     * The byte is not a legal "after-first" byte.
                     * Leave it there so that we will try to process it
                     * as a first byte later.
                     */
                    rlo--;
                    c = TTY_CHAR_MAX+1;
                  }
                else
                  {
                    c = (c << 6) + (next & 0x3f);
                    byte <<= 1;
                    range <<= 5;
                    if (byte & 0x40)
                      goto loop;
                    c &= range-1;
                  }
              }

            if (c > TTY_CHAR_MAX)
              {
#ifdef LINEEDIT_REPORT_CHAR_ENCODING_ERRORS
                if (n == len)
                  {
                    state->rlo = rlo; /* step over the faulty character */
                    return ___FIX(___CTOS_UTF8STRING_ERR);
                  }
                rlo = start; /* error will be reported later */
                break; /* stop short */
#else
                c = UNICODE_REPLACEMENT; /* Unicode replacement character */
#endif
              }

            *buf++ = c;
            n--;
          }
        break;
      }

    case UCS2LE_ENCODING:
      while (rlo+1 < rhi && n > 0)
        {
          ___UCS4 c = rbuf[rlo++];
          c += rbuf[rlo++] << 8;
          *buf++ = c;
          n--;
        }
      break;

    case UCS2BE_ENCODING:
      while (rlo+1 < rhi && n > 0)
        {
          ___UCS4 c = rbuf[rlo++];
          c = (c << 8) + rbuf[rlo++];
          *buf++ = c;
          n--;
        }
      break;

    case UCS4LE_ENCODING:
      while (rlo+3 < rhi && n > 0)
        {
          ___UCS4 c = rbuf[rlo++];
          c += rbuf[rlo++] << 8;
          c += rbuf[rlo++] << 16;
          c += rbuf[rlo++] << 24;
          *buf++ = c;
          n--;
        }
      break;

    case UCS4BE_ENCODING:
      while (rlo+3 < rhi && n > 0)
        {
          ___UCS4 c = rbuf[rlo++];
          c = (c << 8) + rbuf[rlo++];
          c = (c << 8) + rbuf[rlo++];
          c = (c << 8) + rbuf[rlo++];
          *buf++ = c;
          n--;
        }
      break;
    }

  state->rlo = rlo;

  return ___FIX(len-n);
}

___HIDDEN ___WORD tty_rbuf_fill
   ___P((tty_state *state),
        (state)
tty_state *state;)
{
  int rlo = state->rlo;
  int rhi = state->rhi;
  int end = TTY_RBUF_SIZE;
  ___U8 *rbuf = state->rbuf;
  ___WORD n;

  if (rlo != 0)
    {
      int i = 0;
      while (rlo < rhi)
        rbuf[i++] = rbuf[rlo++];
      rlo = 0;
      rhi = i;
      state->rlo = rlo;
      state->rhi = rhi;
    }

  if ((n = tty_read_os_specific (state, &rbuf[rhi], end-rhi)) < ___FIX(0))
    return n;

  state->rhi = rhi + ___INT(n);

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_read_noncanonical
   ___P((tty_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
tty_char *buf;
int len;)
{
  ___WORD e;
  ___WORD n;

#ifdef USE_LINEEDIT

  lineedit_state *les = state->les;
  tty_char *p = les->paste_text;

  /*
   * Try to get characters from pasted text first.
   */

  if (les->paste_cancel)
    {
      if (p != NULL)
        {
          ___free_mem (p);
          p = NULL;
          les->paste_text = NULL;
        }
      les->paste_cancel = 0;
    }

  if (p != NULL)
    {
      int start = les->paste_index;
      int end = start + len;
      int i = start;
      tty_char c;

      while (i < end && (c = p[i]) != UNICODE_NUL)
        {
          *buf++ = c;
          i++;
        }

      if (i < end)
        {
          /* reached null character at end of pasted text */

          ___free_mem (p);
          les->paste_text = NULL;
        }

      if (i > start)
        {
          les->paste_index = i;
          return ___FIX(i-start);
        }
    }

#endif

  /*
   * Get characters from rbuf.
   */

  if ((n = tty_read_from_rbuf (state, buf, len)) == ___FIX(0))
    {
      if ((e = tty_rbuf_fill (state)) != ___FIX(___NO_ERR))
        return e;
      n = tty_read_from_rbuf (state, buf, len);
    }

  return n;
}

___HIDDEN ___WORD tty_read
   ___P((tty *t,
         tty_char *buf,
         int len),
        (t,
         buf,
         len)
tty *t;
tty_char *buf;
int len;)
{
  ___WORD e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT

  if (state->input_mode & LINEEDIT_INPUT_MODE)
    return lineedit_read (state->les, buf, len);

#endif

  return tty_read_noncanonical (state, buf, len);
}

___HIDDEN ___WORD tty_write_to_wbuf
   ___P((tty_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
tty_char *buf;
int len;)
{
  int whi = state->whi;
  int end = TTY_WBUF_SIZE;
  ___U8 *wbuf = state->wbuf;
  int n = len;

  switch (state->encoding)
    {
    default: /* in case state->encoding is set incorrectly */
    case ASCII_ENCODING:
    case LATIN1_ENCODING:
      while (whi < end && n > 0)
        {
          ___UCS4 c = *buf++;
          if (c > 0xff)
            {
#ifdef LINEEDIT_REPORT_CHAR_ENCODING_ERRORS
              if (n == len)
                {
                  /* The caller must step over the faulty character. */
                  return ___FIX(___STOC_LATIN1STRING_ERR);
                }
              break; /* stop short */
#else
              c = UNICODE_BELL;
#endif
            }
          wbuf[whi++] = c;
          n--;
        }
      break;

    case UTF8_ENCODING:
      while (whi < end && n > 0)
        {
          ___UCS4 c = *buf++;
          ___U8 *p = &wbuf[whi];
          int bytes = 1;

          if (c <= 0x7f)
            *p = c; /* ASCII character */
          else
            {
              ___UCS4 x = c;
              if      (x <= 0x7ff)     bytes = 2;
              else if (x <= 0xffff)    bytes = 3;
              else if (x <= 0x1fffff)  bytes = 4;
              else if (x <= 0x3ffffff) bytes = 5;
              else                     bytes = 6;
              if (whi+bytes > end)
                break; /* stop short */
              p += bytes;
              switch (bytes)
                {
                case 6:  *--p = 0x80+(x&0x3f); x >>= 6;
                case 5:  *--p = 0x80+(x&0x3f); x >>= 6;
                case 4:  *--p = 0x80+(x&0x3f); x >>= 6;
                case 3:  *--p = 0x80+(x&0x3f); x >>= 6;
                default: *--p = 0x80+(x&0x3f); x >>= 6;
                }
              *--p = 0xff - (0xff>>bytes) + x;
            }
          whi += bytes;
          n--;
        }
      break;

    case UCS2LE_ENCODING:
      while (whi+1 < end && n > 0)
        {
          ___UCS4 c = *buf++;
          wbuf[whi++] = c & 0xff;
          wbuf[whi++] = (c >> 8) & 0xff;
          n--;
        }
      break;

    case UCS2BE_ENCODING:
      while (whi+1 < end && n > 0)
        {
          ___UCS4 c = *buf++;
          wbuf[whi++] = (c >> 8) & 0xff;
          wbuf[whi++] = c & 0xff;
          n--;
        }
      break;

    case UCS4LE_ENCODING:
      while (whi+3 < end && n > 0)
        {
          ___UCS4 c = *buf++;
          wbuf[whi++] = c & 0xff;
          wbuf[whi++] = (c >> 8) & 0xff;
          wbuf[whi++] = (c >> 16) & 0xff;
          wbuf[whi++] = (c >> 24) & 0xff;
          n--;
        }
      break;

    case UCS4BE_ENCODING:
      while (whi+3 < end && n > 0)
        {
          ___UCS4 c = *buf++;
          wbuf[whi++] = (c >> 24) & 0xff;
          wbuf[whi++] = (c >> 16) & 0xff;
          wbuf[whi++] = (c >> 8) & 0xff;
          wbuf[whi++] = c & 0xff;
          n--;
        }
      break;
    }

  state->whi = whi;

  return ___FIX(len-n);
}

___HIDDEN void tty_wbuf_drain
   ___P((tty_state *state),
        (state)
tty_state *state;)
{
  int wlo = state->wlo;
  int whi = state->whi;
  int end = TTY_WBUF_SIZE;
  ___U8 *wbuf = state->wbuf;
  ___WORD n;

  if (wlo < whi)
    {
      if ((n = tty_write_os_specific (state, &wbuf[wlo], whi-wlo)) < ___FIX(0))
        {
          state->werr = n;
          return;
        }

      wlo += ___INT(n);

      if (wlo >= whi || (wlo<<1) >= end)
        {
          /* shift buffer if first half is now empty */
          int i = 0;
          while (wlo < whi)
            wbuf[i++] = wbuf[wlo++];
          wlo = 0;
          state->whi = i;
        }

      state->wlo = wlo;
    }
}

___HIDDEN ___WORD tty_write
   ___P((tty *t,
         tty_char *buf,
         int len),
        (t,
         buf,
         len)
tty *t;
tty_char *buf;
int len;)
{
  ___WORD e;
  ___WORD n;
  tty_state *state;

#ifdef USE_LINEEDIT
  lineedit_state *les;
#endif

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT

  les = state->les;

  if ((e = lineedit_end_paren_balance (les)) != ___FIX(___NO_ERR))
    return e;

  if ((e = lineedit_output_set_text_attrs (les, les->output_attrs))
      != ___FIX(___NO_ERR))
    return e;

  if ((e = lineedit_output_drain (les)) != ___FIX(___NO_ERR))
    return e;

#endif

  if ((e = state->werr) != ___FIX(___NO_ERR))
    {
      state->werr = ___FIX(___NO_ERR);
      return e;
    }

  n = tty_write_to_wbuf (state, buf, len);

  tty_wbuf_drain (state); /* error code returned in state->werr */

#ifdef USE_LINEEDIT

  if (n >= ___FIX(0))
    {
      int pos = les->current.screen_cursor + les->current.cursor_to_point;
      lineedit_track_cursor_movement (les, buf, ___INT(n));
      les->current.cursor_to_point = pos - les->current.screen_cursor;
    }

#endif

  return n;
}

___HIDDEN ___WORD tty_module_setup
   ___P((void (*user_interrupt_handler) ___PVOID),
        (user_interrupt_handler)
void (*user_interrupt_handler) ___PVOID;)
{
  ___WORD e;

  tm.list = NULL;
  tm.user_interrupt_handler = user_interrupt_handler;

  if ((e = tty_module_setup_os_specific ()) != ___FIX(___NO_ERR))
    return e;

#ifdef USE_LINEEDIT
  if ((e = lineedit_module_setup ()) != ___FIX(___NO_ERR))
    {
      tty_module_cleanup_os_specific (); /* ignore error */
      return e;
    }
#endif

  return tty_setup_console (&tm.console);
}

___HIDDEN ___WORD tty_module_cleanup ___PVOID
{
  ___WORD e1;
  ___WORD e2;

  e1 = tty_cleanup (&tm.console);

#ifdef USE_LINEEDIT
  e2 = lineedit_module_cleanup ();
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;
#endif

  e2 = tty_module_cleanup_os_specific ();
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  return e1;
}

/*---------------------------------------------------------------------------*/

#if 0

void terminal_setup ___PVOID
{
  tty_module_setup (NULL); /* ignore error */
}

void terminal_cleanup ___PVOID
{
  tty_module_cleanup (); /* ignore error */
}

tty *terminal_open ___P((int fd),(fd)
int fd;)
{
  tty *t;
  ___WORD e;

 if (!isatty (fd))
   return NULL;

 t = ___CAST(tty*,___alloc_mem (sizeof (tty)));

  if (t != NULL)
    {
      int fd2 = dup (fd);

#if 0
      int fl;

      fl = fcntl (fd2, F_GETFL);
      fcntl (fd2, F_SETFL, fl | O_NONBLOCK); /* set nonblocking mode */
#endif

      if ((e = tty_setup_os_specific (t, fd2)) != ___FIX(___NO_ERR))
        {
          ___free_mem (t);
          t = NULL;
        }
      else
        {
          if ((e = tty_set_terminal_type
                     (t,
                      getenv (CONSOLE_TERMINAL_TYPE_ENV_VAR),
                      DEFAULT_EMACS_BINDINGS))
              != ___FIX(___NO_ERR))
            ;
          if ((e = tty_set_encoding
                     (t,
                      UTF8_ENCODING))
              != ___FIX(___NO_ERR))
            ;
        }
    }

  return t;
}

void terminal_close ___P((tty *t),(t)
tty *t;)
{
  if (t != NULL)
    tty_cleanup (t); /* ignore error */
}

___WORD ___io_set_terminal_type
   ___P((___STREAM *stream, char *term_type, ___BOOL emacs_bindings),
        (stream, term_type, emacs_bindings)
___STREAM *stream;
char *term_type;
___BOOL emacs_bindings;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___WORD e;
      if ((e = tty_set_terminal_type (stream->dummy1, term_type, emacs_bindings))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___WORD ___io_set_encoding
   ___P((___STREAM *stream, int encoding),
        (stream, encoding)
___STREAM *stream;
int encoding;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___WORD e;
      if ((e = tty_set_encoding (stream->dummy1, encoding))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___WORD ___io_set_text_attributes
   ___P((___STREAM *stream, int input, int output),
        (stream, input, output)
___STREAM *stream;
int input;
int output;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___WORD e;
      if ((e = tty_set_text_attrs (stream->dummy1, input, output))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___WORD ___io_set_max_history_length
   ___P((___STREAM *stream, int max_history_length),
        (stream, max_history_length)
___STREAM *stream;
int max_history_length;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___WORD e;
      if ((e = tty_set_max_history_length (stream->dummy1, max_history_length))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___WORD ___io_set_paren_balance_duration
   ___P((___STREAM *stream, int nsecs),
        (stream, nsecs)
___STREAM *stream;
int nsecs;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___WORD e;
      if (nsecs >= 0)
        if ((e = tty_set_paren_balance_duration (stream->dummy1, nsecs))
            != ___FIX(___NO_ERR))
          ;
    }

  return ___VOID;
}

___WORD ___io_set_input_mode
   ___P((___STREAM *stream, int input_mode),
        (stream, input_mode)
___STREAM *stream;
int input_mode;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___WORD e;
      if ((e = tty_set_input_mode (stream->dummy1, input_mode))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___WORD terminal_get ___P((tty *t),(t)
tty *t;)
{
  tty_char c;
  ___WORD n;

  do
    {
      n = tty_read (t, &c, 1);
    } while (n == ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)));

  if (n == ___FIX(0))
    return ___EOF;
  else if (n == ___FIX(1))
    return ___CHR(c);
  else
    return n;
}

___WORD terminal_put ___P((tty *t, ___UCS4 c),(t, c)
tty *t;
___UCS4 c;)
{
  tty_char ch = c;
  ___WORD n;

  do
    {
      n = tty_write (t, &ch, 1);
    } while (n == ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)) ||
             n == ___FIX(0));

  if (n == ___FIX(1))
    return ___FAL;

  return n;
}

#endif

/*================================================================*/

/* Tty stream device. */

typedef struct ___device_tty_struct
  {
    ___device_stream base;
    tty t;
  } ___device_tty;

typedef struct ___device_tty_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_tty_vtbl;

extern ___WORD ___device_tty_setup
   ___P((___device_tty **dev,
         ___device_group *dgroup,
         int fd,
         int direction),
        ());

___HIDDEN int ___device_tty_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___TTY_DEVICE_KIND;
}

___HIDDEN ___WORD ___device_tty_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___device_tty *d = (___device_tty*)self;
  int is_not_closed = 0;

  if (d->base.base.read_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_RD;

  if (d->base.base.write_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_WR;

  if (is_not_closed == 0)
    return ___FIX(___NO_ERR);

  if ((is_not_closed & ~direction) == 0)
    {
      /* Close tty when both sides are closed. */

      ___WORD e;

      d->base.base.read_stage = ___STAGE_CLOSED; /* avoid multiple closes */
      d->base.base.write_stage = ___STAGE_CLOSED;

      if ((e = tty_cleanup (&d->t)) != ___FIX(___NO_ERR))
        return e;
    }
  else if (direction & ___DIRECTION_RD)
    d->base.base.read_stage = ___STAGE_CLOSED;
  else if (direction & ___DIRECTION_WR)
    d->base.base.write_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tty_select_raw_virt
   ___P((___device_stream *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device_stream *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_tty *d = (___device_tty*)self;

  if ((for_writing ? d->base.base.write_stage : d->base.base.read_stage)
      != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (pass == ___SELECT_PASS_1)
    {
      if (d->t.state != NULL)
        {
          device_select_add_fd (state, d->t.state->fd, for_writing);

#ifdef USE_LINEEDIT

          if (d->t.state->les->current.paren_balance_in_progress)
            {
              if (___time_less (d->t.state->les->current.paren_balance_end,
                                state->timeout))
                state->timeout = d->t.state->les->current.paren_balance_end;
            }

#endif
        }
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  if (for_writing)
    {
      if (FD_ISSET(d->t.state->fd, &state->writefds))
        state->devs[i] = NULL;
    }
  else
    {
      if (FD_ISSET(d->t.state->fd, &state->readfds))
        state->devs[i] = NULL;
    }

#ifdef USE_LINEEDIT

  if (d->t.state->les->current.paren_balance_in_progress)
    {
      if (___time_equal (d->t.state->les->current.paren_balance_end,
                         state->timeout))
        state->devs[i] = NULL;
    }

#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tty_release_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  ___device_tty *d = (___device_tty*)self;
  return tty_cleanup (&d->t);
}

___HIDDEN ___WORD ___device_tty_flush_write_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  ___device_tty *d = (___device_tty*)self;

  if (d->base.base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  /*************************/
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tty_seek_raw_virt
   ___P((___device_stream *self,
         ___stream_index pos),
        (self,
         pos)
___device_stream *self;
___stream_index pos;)
{
  return ___FIX(___UNKNOWN_ERR);
}

___HIDDEN ___WORD ___device_tty_read_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_tty *d = (___device_tty*)self;
  ___WORD n;

  if (d->base.base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  n = tty_read (&d->t, (tty_char*)buf, len>>1);/***********add len_done here */

  if (n < ___FIX(0))
    return n;

  *len_done = 2*___INT(n);/********************/

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tty_write_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_tty *d = (___device_tty*)self;
  ___WORD n;

  if (d->base.base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  n = tty_write (&d->t, (tty_char*)buf, len>>1);/***********add len_done here */

  if (n < ___FIX(0))
    return n;

  *len_done = 2*___INT(n);/********************/

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tty_width_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(80);
}

___HIDDEN ___WORD ___device_tty_default_options_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
#define TTY_CHAR_ENCODING ___CHAR_ENCODING_LATIN1

  return ___STREAM_DEFAULT_OPTIONS(TTY_CHAR_ENCODING,
                                   ___EOL_ENCODING_LF,
                                   ___NO_BUFFERING,
                                   TTY_CHAR_ENCODING,
                                   ___EOL_ENCODING_LF,
                                   ___NO_BUFFERING);
}

___HIDDEN ___device_tty_vtbl ___device_tty_table =
{
  {
    {
      ___device_tty_kind,
      ___device_stream_select_virt,
      ___device_stream_release_virt,
      ___device_stream_flush_write_virt,
      ___device_tty_close_virt
    },
    ___device_tty_select_raw_virt,
    ___device_tty_release_raw_virt,
    ___device_tty_flush_write_raw_virt,
    ___device_tty_seek_raw_virt,
    ___device_tty_read_raw_virt,
    ___device_tty_write_raw_virt,
    ___device_tty_width_virt,
    ___device_tty_default_options_virt
  }
};

___WORD ___device_tty_setup_from_fd
   ___P((___device_tty **dev,
         ___device_group *dgroup,
         int fd,
         int direction),
        (dev,
         dgroup,
         fd,
         direction)
___device_tty **dev;
___device_group *dgroup;
int fd;
int direction;)
{
  ___device_tty *d;

  d = ___CAST(___device_tty*,___alloc_mem (sizeof (___device_tty)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  d->base.base.vtbl = &___device_tty_table;

  {/*****************/
    ___WORD e;
    int fd2 = dup (fd);

#if 0
    int fl;

    fl = fcntl (fd2, F_GETFL);
    fcntl (fd2, F_SETFL, fl | O_NONBLOCK); /* set nonblocking mode */
#endif

    if ((e = tty_setup_os_specific (&d->t, fd2)) != ___FIX(___NO_ERR))
      {
        ___free_mem (d);
        return e;
      }

    if ((e = tty_set_terminal_type
               (&d->t,
                getenv (CONSOLE_TERMINAL_TYPE_ENV_VAR),
                DEFAULT_EMACS_BINDINGS))
        != ___FIX(___NO_ERR))
      ;/*******************/
#if 0
    if ((e = tty_set_encoding
               (&d->t,
                UTF8_ENCODING))
        != ___FIX(___NO_ERR))
      ;/*******************/
    if ((e = tty_set_input_mode (&d->t, CBREAK_TERM_MODE))
        != ___FIX(___NO_ERR))
      ;/*******************/
#endif
  }

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            0);
}

#endif


#ifdef USE_POSIX

/* TCP client stream device */

typedef struct ___device_tcp_client_struct
  {
    ___device_stream base;
    int fd;
  } ___device_tcp_client;

typedef struct ___device_tcp_client_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_tcp_client_vtbl;

___HIDDEN int ___device_tcp_client_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___TCP_CLIENT_DEVICE_KIND;
}

___HIDDEN ___WORD ___device_tcp_client_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___device_tcp_client *d = (___device_tcp_client*)self;
  int is_not_closed = 0;

  if (d->base.base.read_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_RD;

  if (d->base.base.write_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_WR;

  if (is_not_closed == 0)
    return ___FIX(___NO_ERR);

  if ((is_not_closed & ~direction) == 0)
    {
      /* Close socket when both sides are closed. */

      d->base.base.read_stage = ___STAGE_CLOSED; /* avoid multiple closes */
      d->base.base.write_stage = ___STAGE_CLOSED;

      if (close (d->fd) < 0)
        return ___FIX(err_code_from_errno ());
    }
  else if (direction & ___DIRECTION_RD)
    {
      /* Shutdown receiving side. */

#ifndef SHUT_RD
#define SHUT_RD 0
#endif

      if (shutdown (d->fd, SHUT_RD) < 0)
        {
          ___WORD e = ___FIX(err_code_from_errno ());
          if (e != ___FIX(___ERR_CODE_FROM_ERRNO(ENOTCONN)))
            return e;
        }

      d->base.base.read_stage = ___STAGE_CLOSED;
    }
  else if (direction & ___DIRECTION_WR)
    {
      /* Shutdown sending side. */

#ifndef SHUT_WR
#define SHUT_WR 1
#endif

      if (shutdown (d->fd, SHUT_WR) < 0)
        {
          ___WORD e = ___FIX(err_code_from_errno ());
          if (e != ___FIX(___ERR_CODE_FROM_ERRNO(ENOTCONN)))
            return e;
        }

      d->base.base.write_stage = ___STAGE_CLOSED;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_select_raw_virt
   ___P((___device_stream *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device_stream *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_tcp_client *d = (___device_tcp_client*)self;

  if ((for_writing ? d->base.base.write_stage : d->base.base.read_stage)
      != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (pass == ___SELECT_PASS_1)
    {
      device_select_add_fd (state, d->fd, for_writing);
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  if (for_writing)
    {
      if (FD_ISSET(d->fd, &state->writefds))
        state->devs[i] = NULL;
    }
  else
    {
      if (FD_ISSET(d->fd, &state->readfds))
        state->devs[i] = NULL;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_release_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_flush_write_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_seek_raw_virt
   ___P((___device_stream *self,
         ___stream_index pos),
        (self,
         pos)
___device_stream *self;
___stream_index pos;)
{
  return ___FIX(___UNKNOWN_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_read_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_tcp_client *d = (___device_tcp_client*)self;
  int n;

  if (d->base.base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if ((n = recv (d->fd, buf, len, 0)) < 0)
    return ___FIX(err_code_from_errno ());

  *len_done = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_write_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_tcp_client *d = (___device_tcp_client*)self;
  int n;

  if (d->base.base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if ((n = send (d->fd, buf, len, 0)) < 0)
    return ___FIX(err_code_from_errno ());

  *len_done = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_client_width_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(80);
}

___HIDDEN ___WORD ___device_tcp_client_default_options_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___STREAM_DEFAULT_OPTIONS(___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING,
                                   ___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING);
}

___HIDDEN ___device_tcp_client_vtbl ___device_tcp_client_table =
{
  {
    {
      ___device_tcp_client_kind,
      ___device_stream_select_virt,
      ___device_stream_release_virt,
      ___device_stream_flush_write_virt,
      ___device_tcp_client_close_virt
    },
    ___device_tcp_client_select_raw_virt,
    ___device_tcp_client_release_raw_virt,
    ___device_tcp_client_flush_write_raw_virt,
    ___device_tcp_client_seek_raw_virt,
    ___device_tcp_client_read_raw_virt,
    ___device_tcp_client_write_raw_virt,
    ___device_tcp_client_width_virt,
    ___device_tcp_client_default_options_virt
  }
};

___WORD ___device_tcp_client_setup_from_fd
   ___P((___device_tcp_client **dev,
         ___device_group *dgroup,
         int fd,
         int direction),
        (dev,
         dgroup,
         fd,
         direction)
___device_tcp_client **dev;
___device_group *dgroup;
int fd;
int direction;)
{
  ___device_tcp_client *d;

  d = ___CAST(___device_tcp_client*,___alloc_mem (sizeof (___device_tcp_client)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  /*
   * Setup socket to perform nonblocking I/O.
   */

  if (fcntl (fd, F_SETFL, O_NONBLOCK) < 0) /* set nonblocking mode */
    {
      ___WORD e = ___FIX(err_code_from_errno ());
      ___free_mem (&d->base.base);
      return e;
    }

  d->base.base.vtbl = &___device_tcp_client_table;
  d->fd = fd;

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            0);
}

___WORD ___device_tcp_client_setup_from_sockaddr
   ___P((___device_tcp_client **dev,
         ___device_group *dgroup,
         struct sockaddr *server_addr,
         int server_addrlen,
         int direction),
        (dev,
         dgroup,
         server_addr,
         server_addrlen,
         direction)
___device_tcp_client **dev;
___device_group *dgroup;
struct sockaddr *server_addr;
int server_addrlen;
int direction;)
{
  ___WORD e;
  int fd;

  if ((fd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
    return ___FIX(err_code_from_errno ());

  if (connect (fd, server_addr, server_addrlen) < 0)
    {
      e = ___FIX(err_code_from_errno ());
      close (fd); /* ignore error */
      return e;
    }

  e = ___device_tcp_client_setup_from_fd
        (dev,
         dgroup,
         fd,
         direction);

  if (e != ___FIX(___NO_ERR))
    {
      close (fd); /* ignore error */
      return e;
    }

  return e;
}

#endif

#ifdef USE_POSIX

/* TCP server device. */

typedef struct ___device_tcp_server_struct
  {
    ___device base;
    int fd;
  } ___device_tcp_server;

typedef struct ___device_tcp_server_vtbl_struct
  {
    ___device_vtbl base;
  } ___device_tcp_server_vtbl;

___HIDDEN int ___device_tcp_server_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___TCP_SERVER_KIND;
}

___HIDDEN ___WORD ___device_tcp_server_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___device_tcp_server *d = (___device_tcp_server*)self;

  if (d->base.read_stage == ___STAGE_CLOSED)
    return ___FIX(___NO_ERR);

  if (direction & ___DIRECTION_RD)
    {
      d->base.read_stage = ___STAGE_CLOSED; /* avoid multiple closes */

      if (close (d->fd) < 0)
        return ___FIX(err_code_from_errno ());
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_server_select_virt
   ___P((___device *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_tcp_server *d = (___device_tcp_server*)self;

  if (pass == ___SELECT_PASS_1)
    {
      device_select_add_fd (state, d->fd, for_writing);
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  if (for_writing)
    {
      if (FD_ISSET(d->fd, &state->writefds))
        state->devs[i] = NULL;
    }
  else
    {
      if (FD_ISSET(d->fd, &state->readfds))
        state->devs[i] = NULL;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_server_release_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_tcp_server_flush_write_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___device_tcp_server_vtbl ___device_tcp_server_table =
{
  {
    ___device_tcp_server_kind,
    ___device_tcp_server_select_virt,
    ___device_tcp_server_release_virt,
    ___device_tcp_server_flush_write_virt,
    ___device_tcp_server_close_virt
  }
};

___WORD ___device_tcp_server_setup
   ___P((___device_tcp_server **dev,
         ___device_group *dgroup,
         int port_num,
         int backlog),
        (dev,
         dgroup,
         port_num,
         backlog)
___device_tcp_server **dev;
___device_group *dgroup;
int port_num;
int backlog)
{
  ___device_tcp_server *d;
  struct sockaddr_in addr;
  int fd;

  if ((fd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
    return ___FIX(err_code_from_errno ());

  addr.sin_family = AF_INET;
  addr.sin_port = htons (port_num);
  addr.sin_addr.s_addr = htonl (INADDR_ANY);

  if (fcntl (fd, F_SETFL, O_NONBLOCK) < 0 || /* set nonblocking mode */
      bind (fd, &addr, sizeof (addr)) < 0 ||
      listen (fd, backlog) < 0)
    {
      ___WORD e = ___FIX(err_code_from_errno ());
      close (fd); /* ignore error */
      return e;
    }

  d = ___CAST(___device_tcp_server*,___alloc_mem (sizeof (___device_tcp_server)));

  if (d == NULL)
    {
      close (fd); /* ignore error */
      return ___FIX(___HEAP_OVERFLOW_ERR);
    }

  d->base.vtbl = &___device_tcp_server_table;
  d->base.refcount = 1;
  d->base.direction = ___DIRECTION_RD;
  d->base.read_stage = ___STAGE_OPEN;
  d->base.write_stage = ___STAGE_CLOSED;

  d->fd = fd;

  *dev = d;

  device_add_to_group (dgroup, &d->base);

  return ___FIX(___NO_ERR);
}

___WORD ___device_tcp_server_read
   ___P((___device_tcp_server *dev,
         ___device_group *dgroup,
         ___device_tcp_client **client),
        (dev,
         dgroup,
         client)
___device_tcp_server *dev;
___device_group *dgroup;
___device_tcp_client **client;)
{
  ___WORD e;
  struct sockaddr_in addr;
  int addrlen;
  int sock_fd;

  if (dev->base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  addrlen = sizeof (addr);

  if ((sock_fd = accept (dev->fd, &addr, &addrlen)) < 0)
    return ___FIX(err_code_from_errno ());

  e = ___device_tcp_client_setup_from_fd
        (client,
         dgroup,
         sock_fd,
         ___DIRECTION_RD|___DIRECTION_WR);

  if (e != ___FIX(___NO_ERR))
    {
      close (sock_fd); /* ignore error */
      return e;
    }

  return ___FIX(___NO_ERR);
}

#endif

#ifdef USE_POSIX

/* Directory device. */

typedef struct ___device_directory_struct
  {
    ___device base;
    DIR *dir;
  } ___device_directory;

typedef struct ___device_directory_vtbl_struct
  {
    ___device_vtbl base;
  } ___device_directory_vtbl;

___HIDDEN int ___device_directory_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___DIRECTORY_KIND;
}

___HIDDEN ___WORD ___device_directory_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___device_directory *d = (___device_directory*)self;

  if (d->base.read_stage == ___STAGE_CLOSED)
    return ___FIX(___NO_ERR);

  if (direction & ___DIRECTION_RD)
    {
      d->base.read_stage = ___STAGE_CLOSED; /* avoid multiple closes */

      if (closedir (d->dir) < 0)
        return ___FIX(err_code_from_errno ());
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_directory_select_virt
   ___P((___device *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  if (pass == ___SELECT_PASS_1)
    {
      state->timeout = ___time_neg_infinity;
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  state->devs[i] = NULL;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_directory_release_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_directory_flush_write_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___device_directory_vtbl ___device_directory_table =
{
  {
    ___device_directory_kind,
    ___device_directory_select_virt,
    ___device_directory_release_virt,
    ___device_directory_flush_write_virt,
    ___device_directory_close_virt
  }
};

___WORD ___device_directory_setup
   ___P((___device_directory **dev,
         ___device_group *dgroup,
         char *path),
        (dev,
         dgroup,
         path)
___device_directory **dev;
___device_group *dgroup;
char *path;)
{
  ___device_directory *d;
  DIR *dir;

  dir = opendir (path);

  if (dir == NULL)
    return ___FIX(err_code_from_errno ());

  d = ___CAST(___device_directory*,___alloc_mem (sizeof (___device_directory)));

  if (d == NULL)
    {
      closedir (dir); /* ignore error */
      return ___FIX(___HEAP_OVERFLOW_ERR);
    }

  d->base.vtbl = &___device_directory_table;
  d->base.refcount = 1;
  d->base.direction = ___DIRECTION_RD;
  d->base.read_stage = ___STAGE_OPEN;
  d->base.write_stage = ___STAGE_CLOSED;

  d->dir = dir;

  *dev = d;

  device_add_to_group (dgroup, &d->base);

  return ___FIX(___NO_ERR);
}

___WORD ___device_directory_read
   ___P((___device_directory *dev,
         char **name),
        (dev,
         name)
___device_directory *dev;
char **name;)
{
  struct dirent *de;

  if (dev->base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  de = readdir (dev->dir);

  if (de == NULL)
    {
#if 0
      /* this seems to be broken, at least under Linux */
      if (errno != 0)
        return ___FIX(err_code_from_errno ());
#endif
      *name = NULL;
      return ___FIX(___NO_ERR);
    }

  *name = de->d_name;

  return ___FIX(___NO_ERR);
}

#endif

#ifdef USE_POSIX

/* File stream device */

typedef struct ___device_file_struct
  {
    ___device_stream base;
    int fd;
  } ___device_file;

typedef struct ___device_file_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_file_vtbl;

___HIDDEN int ___device_file_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FILE_DEVICE_KIND;
}

___HIDDEN ___WORD ___device_file_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___device_file *d = (___device_file*)self;
  int is_not_closed = 0;

  if (d->base.base.read_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_RD;

  if (d->base.base.write_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_WR;

  if (is_not_closed == 0)
    return ___FIX(___NO_ERR);

  if ((is_not_closed & ~direction) == 0)
    {
      /* Close file when both sides are closed. */

      d->base.base.read_stage = ___STAGE_CLOSED; /* avoid multiple closes */
      d->base.base.write_stage = ___STAGE_CLOSED;

      if (close (d->fd) < 0)
        return ___FIX(err_code_from_errno ());
    }
  else if (direction & ___DIRECTION_RD)
    d->base.base.read_stage = ___STAGE_CLOSED;
  else if (direction & ___DIRECTION_WR)
    d->base.base.write_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_select_raw_virt
   ___P((___device_stream *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device_stream *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_file *d = (___device_file*)self;

  if ((for_writing ? d->base.base.write_stage : d->base.base.read_stage)
      != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (pass == ___SELECT_PASS_1)
    {
      device_select_add_fd (state, d->fd, for_writing);
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  if (for_writing)
    {
      if (FD_ISSET(d->fd, &state->writefds))
        state->devs[i] = NULL;
    }
  else
    {
      if (FD_ISSET(d->fd, &state->readfds))
        state->devs[i] = NULL;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_release_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_flush_write_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  ___device_file *d = (___device_file*)self;

  if (d->base.base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

#if 0
  if (fsync (d->fd) < 0)/************only works on disk files!!!!!!!*/
    return ___FIX(err_code_from_errno ());
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_seek_raw_virt
   ___P((___device_stream *self,
         ___stream_index pos),
        (self,
         pos)
___device_stream *self;
___stream_index pos;)
{
  return ___FIX(___UNKNOWN_ERR);
}

___HIDDEN ___WORD ___device_file_read_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_file *d = (___device_file*)self;
  int n;

  if (d->base.base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if ((n = read (d->fd, buf, len)) < 0)
    return ___FIX(err_code_from_errno ());

  *len_done = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_write_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_file *d = (___device_file*)self;
  int n;

  if (d->base.base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  n = write (d->fd, buf, len);

  if (n < 0)
    return ___FIX(err_code_from_errno ());

  *len_done = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD ___device_file_width_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(80);
}

___HIDDEN ___WORD ___device_file_default_options_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___STREAM_DEFAULT_OPTIONS(___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING,
                                   ___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING);
}

___HIDDEN ___device_file_vtbl ___device_file_table =
{
  {
    {
      ___device_file_kind,
      ___device_stream_select_virt,
      ___device_stream_release_virt,
      ___device_stream_flush_write_virt,
      ___device_file_close_virt
    },
    ___device_file_select_raw_virt,
    ___device_file_release_raw_virt,
    ___device_file_flush_write_raw_virt,
    ___device_file_seek_raw_virt,
    ___device_file_read_raw_virt,
    ___device_file_write_raw_virt,
    ___device_file_width_virt,
    ___device_file_default_options_virt
  }
};

___HIDDEN ___WORD ___device_file_setup_from_fd
   ___P((___device_file **dev,
         ___device_group *dgroup,
         int fd,
         int direction),
        (dev,
         dgroup,
         fd,
         direction)
___device_file **dev;
___device_group *dgroup;
int fd;
int direction;)
{
  ___device_file *d;

  d = ___CAST(___device_file*,___alloc_mem (sizeof (___device_file)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  d->base.base.vtbl = &___device_file_table;
  d->fd = fd;

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            0);
}

#endif


#ifdef USE_POSIX

___HIDDEN int ___device_stream_kind_from_fd
   ___P((int fd),
        (fd)
int fd;)
{
  struct stat s;

  if (isatty (fd))
    return ___TTY_DEVICE_KIND;

  if (fstat (fd, &s) < 0)
    return ___NONE_KIND;

  if (S_ISSOCK(s.st_mode))
    return ___TCP_CLIENT_DEVICE_KIND;

  if (S_ISREG(s.st_mode))
    return ___FILE_DEVICE_KIND;

  return ___FILE_DEVICE_KIND;/****************/
}

___HIDDEN int ___device_stream_direction_from_fd
   ___P((int fd),
        (fd)
int fd;)
{
  int direction = 0;

  direction |= ___DIRECTION_RD;

  direction |= ___DIRECTION_WR;

  return direction;
}

___WORD ___device_stream_setup_from_fd
   ___P((___device_stream **dev,
         ___device_group *dgroup,
         int fd,
         int kind,
         int direction),
        (dev,
         dgroup,
         fd,
         kind,
         direction)
___device_stream **dev;
___device_group *dgroup;
int fd;
int kind;
int direction;)
{
  ___WORD e = ___FIX(___UNKNOWN_ERR);

  if (kind == ___NONE_KIND)
    kind = ___device_stream_kind_from_fd (fd);

  if (direction == 0)
    direction = ___device_stream_direction_from_fd (fd);

  switch (kind)
    {
    case ___TTY_DEVICE_KIND:
      {
        ___device_tty *d;
        e = ___device_tty_setup_from_fd (&d, dgroup, fd, direction);
        *dev = ___CAST(___device_stream*,d);
        break;
      }

    case ___TCP_CLIENT_DEVICE_KIND:
      {
        ___device_tcp_client *d;
        e = ___device_tcp_client_setup_from_fd (&d, dgroup, fd, direction);
        *dev = ___CAST(___device_stream*,d);
        break;
      }

    case ___FILE_DEVICE_KIND:
      {
        ___device_file *d;
        e = ___device_file_setup_from_fd (&d, dgroup, fd, direction);
        *dev = ___CAST(___device_stream*,d);
        break;
      }
    }

  return e;
}

___HIDDEN void ___device_translate_flags
   ___P((int flags,
         int *fl,
         int *direction),
        (flags,
         fl,
         direction)
int flags;
int *fl;
int *direction;)
{
  int f;

  switch ((flags >> 4) & 3)
    {
    default:
    case 1:
      f = O_RDONLY;
      *direction = ___DIRECTION_RD;
      break;
    case 2:
      f = O_WRONLY;
      *direction = ___DIRECTION_WR;
      break;
    case 3:
      f = O_RDWR;
      *direction = ___DIRECTION_RD|___DIRECTION_WR;
      break;
    }

  if (flags & 16)
    f |= O_APPEND;

  switch ((flags >> 1) & 3)
    {
    default:
    case 0: break;
    case 1: f |= O_CREAT; break;
    case 2: f |= O_CREAT|O_EXCL; break;
    }

  if (flags & 1)
    f |= O_TRUNC;

  f |= O_NONBLOCK;

  *fl = f;
}

___WORD ___device_stream_setup_from_path
   ___P((___device_stream **dev,
         ___device_group *dgroup,
         char *path,
         int flags,
         int mode),
        (dev,
         dgroup,
         path,
         flags,
         mode)
___device_stream **dev;
___device_group *dgroup;
char *path;
int flags;
int mode;)
{
  int fd;
  int fl;
  int direction;

  ___device_translate_flags (flags, &fl, &direction);

  fd = open (path, fl, mode);

  if (fd < 0)
    return ___FIX(err_code_from_errno ());

  return ___device_stream_setup_from_fd
           (dev,
            dgroup,
            fd,
            ___NONE_KIND,
            direction);
}

___LOCAL ___WORD ___dev_release
   ___P((void *ptr),
        (ptr)
void *ptr;)
{
  return ___device_cleanup (___CAST(___device*,ptr));
}

#define ___PORT_MUTEX                1
#define ___PORT_RKIND                2
#define ___PORT_WKIND                3
#define ___PORT_NAME                 4
#define ___PORT_READ_DATUM           5
#define ___PORT_WRITE_DATUM          6
#define ___PORT_NEWLINE              7
#define ___PORT_FORCE_OUTPUT         8
#define ___PORT_CLOSE                9
#define ___PORT_ROPTIONS             10
#define ___PORT_RTIMEOUT             11
#define ___PORT_RTIMEOUT_THUNK       12
#define ___PORT_SET_RTIMEOUT         13
#define ___PORT_WOPTIONS             14
#define ___PORT_WTIMEOUT             15
#define ___PORT_WTIMEOUT_THUNK       16
#define ___PORT_SET_WTIMEOUT         17

#define ___PORT_OBJECT_OTHER1        18
#define ___PORT_OBJECT_OTHER2        19
#define ___PORT_OBJECT_OTHER3        20

#define ___PORT_CHAR_RBUF            18
#define ___PORT_CHAR_RLO             19
#define ___PORT_CHAR_RHI             20
#define ___PORT_CHAR_RCHARS          21
#define ___PORT_CHAR_RLINES          22
#define ___PORT_CHAR_RCURLINE        23
#define ___PORT_CHAR_RBUF_FILL       24
#define ___PORT_CHAR_PEEK_EOFP       25

#define ___PORT_CHAR_WBUF            26
#define ___PORT_CHAR_WLO             27
#define ___PORT_CHAR_WHI             28
#define ___PORT_CHAR_WCHARS          29
#define ___PORT_CHAR_WLINES          30
#define ___PORT_CHAR_WCURLINE        31
#define ___PORT_CHAR_WBUF_DRAIN      32
#define ___PORT_OUTPUT_WIDTH         33

#define ___PORT_CHAR_OTHER1          34
#define ___PORT_CHAR_OTHER2          35
#define ___PORT_CHAR_OTHER3          36
#define ___PORT_CHAR_OTHER4          37
#define ___PORT_CHAR_OTHER5          38

#define ___PORT_BYTE_RBUF            34
#define ___PORT_BYTE_RLO             35
#define ___PORT_BYTE_RHI             36
#define ___PORT_BYTE_RBUF_FILL       37

#define ___PORT_BYTE_WBUF            38
#define ___PORT_BYTE_WLO             39
#define ___PORT_BYTE_WHI             40
#define ___PORT_BYTE_WBUF_DRAIN      41

#define ___PORT_BYTE_OTHER1          42
#define ___PORT_BYTE_OTHER2          43

#define ___PORT_RDEVICE_CONDVAR      42
#define ___PORT_WDEVICE_CONDVAR      43

#define ___PORT_DEVICE_OTHER1        44
#define ___PORT_DEVICE_OTHER2        45

#define ___C ___CS_SELECT(___U8,___U16,___U32)

static ___device_group *dgroup = NULL;/**************/
static void init_dgroup ()
{
  if (dgroup == NULL) /*************/
    ___device_group_setup (&dgroup);
}

___WORD ___dev_stream_open_predefined
   ___P((___WORD index,
         ___WORD flags),
        (index,
         flags)
___WORD index;
___WORD flags;)
{
  ___ERR_CODE err;
  ___WORD e;
  ___device_stream *dev;
  ___WORD result;
  int fl;
  int direction;

  init_dgroup ();/****************/

  ___device_translate_flags (___INT(flags), &fl, &direction);

  if (index == ___FIX(-2)) /* IDE device? */
    {
      return ___FAL;
    }
  else if (index == ___FIX(-1)) /* console device? */
    {
      return ___FAL;
    }
  else
    {
      int fd;

      switch (___INT(index))
        {
        default:
        case 0:
          fd = STDIN_FILENO;
          break;
        case 1:
          fd = STDOUT_FILENO;
          break;
        case 2:
          fd = STDERR_FILENO;
          break;
        }

      e = ___device_stream_setup_from_fd
            (&dev,
             dgroup,
             fd,
             ___NONE_KIND,
             direction);

      if (e != ___FIX(___NO_ERR))
        return e;
    }

  err = ___NONNULLPOINTER_to_SCMOBJ
          (dev,
           ___FAL,
           ___dev_release,
           &result,
           ___RETURN_POS);

  if (err != ___NO_ERR)
    {
      ___device_cleanup (___CAST(___device*,dev)); /* ignore error */
      return ___FIX(err);
    }

  ___release_scmobj (result);

  return result;
}

___WORD ___dev_stream_open_path
   ___P((___WORD path,
         ___WORD flags,
         ___WORD mode),
        (path,
         flags,
         mode)
___WORD path;
___WORD flags;
___WORD mode;)
{
  ___ERR_CODE err;
  ___WORD e;
  ___device_stream *dev;
  ___WORD result;
  char *p;

  init_dgroup ();

  /*
   * Convert the Scheme string to a C "char*" string.  If an invalid
   * character is seen then return ___FAL as if the call to "open"
   * could not find the file.
   */

  err = ___SCMOBJ_to_CHARSTRING (path, &p, 1);

  if (err == ___STOC_CHARSTRING_ERR+1 || p == 0)
    return ___FAL;

  if (err != ___NO_ERR)
    return ___FIX(err);

  e = ___device_stream_setup_from_path
        (&dev,
         dgroup,
         p,
         ___INT(flags),
         ___INT(mode));

  ___free_mem (p);

  if (e != ___FIX(___NO_ERR))
    {
      if (e == ___FIX(___ERR_CODE_FROM_ERRNO(ENOENT)))
        return ___FAL;
      return e;
    }

  err = ___NONNULLPOINTER_to_SCMOBJ
          (dev,
           ___FAL,
           ___dev_release,
           &result,
           ___RETURN_POS);

  if (err != ___NO_ERR)
    {
      ___device_cleanup (___CAST(___device*,dev)); /* ignore error */
      return ___FIX(err);
    }

  ___release_scmobj (result);

  return result;
}

#include <netdb.h>
/**************/

___WORD ___get_host_by_name
   ___P((___WORD host_name),
        (host_name)
___WORD host_name;)
{
  ___ERR_CODE err;
  struct hostent *he;
  char *hn;

  /*
   * Convert the Scheme string to a C "char*" string.  If an invalid
   * character is seen then return an error.
   */

  err = ___SCMOBJ_to_CHARSTRING (host_name, &hn, 1);

  if (err == ___STOC_CHARSTRING_ERR+1 || hn == 0)
    return ___FAL;

  if (err != ___NO_ERR)
    return ___FAL;

  he = gethostbyname (hn);

  ___free_mem (hn);

  if (he == 0)
    return ___FAL;

  switch (he->h_addrtype)
    {
    case AF_INET:
      {
        unsigned long host_addr;
        ___WORD result;
        host_addr = ntohl (___CAST(struct in_addr*,he->h_addr_list[0])->s_addr);
        result = ___alloc_scmobj (___sU8VECTOR, 4, ___STILL);
        if (result == ___FAL)
          return ___FAL;
        ___U8VECTORSET(result,___FIX(0),___FIX((host_addr>>24)&0xff))
        ___U8VECTORSET(result,___FIX(1),___FIX((host_addr>>16)&0xff))
        ___U8VECTORSET(result,___FIX(2),___FIX((host_addr>>8)&0xff))
        ___U8VECTORSET(result,___FIX(3),___FIX(host_addr&0xff))
        ___release_scmobj (result);
        return result;
      }
      break;
    default:
#ifdef AF_INET6
    case AF_INET6:
#endif
      return ___FAL;
    }
}

___WORD ___dev_tcp_client_open
   ___P((___WORD server_addr,
         ___WORD port_num),
        (server_addr,
         port_num)
___WORD server_addr;
___WORD port_num;)
{
  ___ERR_CODE err;
  ___WORD e;
  ___device_tcp_client *dev;
  ___WORD result;
  struct sockaddr *addr;
  int addrlen;
  struct sockaddr_in addr_in;
#ifdef AF_INET6
  struct sockaddr_in6 addr_in6;
#endif

  init_dgroup ();

  if (___U8VECTORLENGTH(server_addr) == ___FIX(4))
    {
      addr_in.sin_family = AF_INET;
      addr_in.sin_port = htons (___INT(port_num));
      addr_in.sin_addr.s_addr =
        htonl ((___INT(___U8VECTORREF(server_addr,___FIX(0)))<<24) +
               (___INT(___U8VECTORREF(server_addr,___FIX(1)))<<16) +
               (___INT(___U8VECTORREF(server_addr,___FIX(2)))<<8) +
               ___INT(___U8VECTORREF(server_addr,___FIX(3))));
      addr = ___CAST(struct sockaddr*,&addr_in);
      addrlen = sizeof (addr_in);
    }
  else
    {
#ifdef AF_INET6
      int i;
      addr_in6.sin6_family = AF_INET6;
      addr_in6.sin6_port = htons (___INT(port_num));
      for (i=0; i<16; i++)
        addr_in6.sin6_addr.s6_addr[i] =
          ___INT(___U8VECTORREF(server_addr,___FIX(i)));
      addr = ___CAST(struct sockaddr*,&addr_in6);
      addrlen = sizeof (addr_in6);
#else
      return ___FIX(___UNKNOWN_ERR);
#endif
    }

  e = ___device_tcp_client_setup_from_sockaddr
        (&dev,
         dgroup,
         addr,
         addrlen,
         ___DIRECTION_RD|___DIRECTION_WR);

  if (e != ___FIX(___NO_ERR))
    return e;

  err = ___NONNULLPOINTER_to_SCMOBJ
          (dev,
           ___FAL,
           ___dev_release,
           &result,
           ___RETURN_POS);

  if (err != ___NO_ERR)
    {
      ___device_cleanup (___CAST(___device*,dev)); /* ignore error */
      return ___FIX(err);
    }

  ___release_scmobj (result);

  return result;
}

___WORD ___dev_tcp_server_open
   ___P((___WORD port_num,
         ___WORD backlog),
        (port_num,
         backlog)
___WORD port_num;
___WORD backlog;)
{
  ___ERR_CODE err;
  ___WORD e;
  ___device_tcp_server *dev;
  ___WORD result;

  init_dgroup ();

  e = ___device_tcp_server_setup
        (&dev,
         dgroup,
         ___INT(port_num),
         ___INT(backlog));

  if (e != ___FIX(___NO_ERR))
    return e;

  err = ___NONNULLPOINTER_to_SCMOBJ
          (dev,
           ___FAL,
           ___dev_release,
           &result,
           ___RETURN_POS);

  if (err != ___NO_ERR)
    {
      ___device_cleanup (___CAST(___device*,dev)); /* ignore error */
      return ___FIX(err);
    }

  ___release_scmobj (result);

  return result;
}

___WORD ___dev_tcp_server_read
   ___P((___WORD dev),
        (dev)
___WORD dev;)
{
  ___device_tcp_server *d = ___CAST(___device_tcp_server*,___FIELD(dev,0));
  ___ERR_CODE err;
  ___WORD e;
  ___device_tcp_client *client;
  ___WORD result;

  if ((e = ___device_tcp_server_read (d, dgroup, &client))
      != ___FIX(___NO_ERR))
    return e;

  err = ___NONNULLPOINTER_to_SCMOBJ
          (client,
           ___FAL,
           ___dev_release,
           &result,
           ___RETURN_POS);

  if (err != ___NO_ERR)
    {
      ___device_cleanup (___CAST(___device*,d)); /* ignore error */
      return ___FIX(err);
    }

  ___release_scmobj (result);

  return result;
}

___WORD ___dev_directory_open_path
   ___P((___WORD path),
        (path)
___WORD path;)
{
  ___ERR_CODE err;
  ___WORD e;
  ___device_directory *dev;
  ___WORD result;
  char *p;

  init_dgroup ();

  /*
   * Convert the Scheme string to a C "char*" string.  If an invalid
   * character is seen then return ___FAL as if the call to "open"
   * could not find the file.
   */

  err = ___SCMOBJ_to_CHARSTRING (path, &p, 1);

  if (err == ___STOC_CHARSTRING_ERR+1 || p == 0)
    return ___FAL;

  if (err != ___NO_ERR)
    return ___FIX(err);

  e = ___device_directory_setup
        (&dev,
         dgroup,
         p);

  ___free_mem (p);

  if (e != ___FIX(___NO_ERR))
    {
      if (e == ___FIX(___ERR_CODE_FROM_ERRNO(ENOENT)))
        return ___FAL;
      return e;
    }

  err = ___NONNULLPOINTER_to_SCMOBJ
          (dev,
           ___FAL,
           ___dev_release,
           &result,
           ___RETURN_POS);

  if (err != ___NO_ERR)
    {
      ___device_cleanup (___CAST(___device*,dev)); /* ignore error */
      return ___FIX(err);
    }

  ___release_scmobj (result);

  return result;
}

___WORD ___dev_directory_read
   ___P((___WORD dev),
        (dev)
___WORD dev;)
{
  ___device_directory *d = ___CAST(___device_directory*,___FIELD(dev,0));
  ___WORD e;
  char *name;
  ___WORD result;

  if ((e = ___device_directory_read (d, &name)) != ___FIX(___NO_ERR))
    return e;

  if (name == NULL)
    return ___EOF;

  if ((e = ___CHARSTRING_to_SCMOBJ (name, &result, ___RETURN_POS))
      != ___FIX(___NO_ERR))
    return e;

  ___release_scmobj (result);

  return result;
}

___WORD ___dev_kind
   ___P((___WORD dev),
        (dev)
___WORD dev;)
{
  ___device *d = ___CAST(___device*,___FIELD(dev,0));

  return ___FIX(___device_kind (d));
}

___WORD ___dev_force_output
   ___P((___WORD dev),
        (dev)
___WORD dev;)
{
  ___device *d = ___CAST(___device*,___FIELD(dev,0));

  return ___device_flush_write (d);
}

___WORD ___dev_close
   ___P((___WORD dev,
         ___WORD direction),
        (dev,
         direction)
___WORD dev;
___WORD direction;)
{
  ___device *d = ___CAST(___device*,___FIELD(dev,0));

  return ___device_close (d, ___INT(direction));
}

___WORD ___dev_stream_read
   ___P((___WORD dev,
         ___WORD buffer,
         ___WORD lo,
         ___WORD hi),
        (dev,
         buffer,
         lo,
         hi)
___WORD dev;
___WORD buffer;
___WORD lo;
___WORD hi;)
{
  ___WORD e;
  ___stream_index len_done;

  if (___INT(hi) <= ___INT(lo))
    return ___FIX(0);

  e = ___device_stream_read
        (___CAST(___device_stream*,___FIELD(dev,0)),
         ___CAST(___U8*,___BODY_AS(buffer,___tSUBTYPED)) + ___INT(lo),
         ___INT(hi) - ___INT(lo),
         &len_done);

  if (e != ___FIX(___NO_ERR))
    return e;

  return ___FIX(len_done);
}

___WORD ___dev_stream_write
   ___P((___WORD dev,
         ___WORD buffer,
         ___WORD lo,
         ___WORD hi),
        (dev,
         buffer,
         lo,
         hi)
___WORD dev;
___WORD buffer;
___WORD lo;
___WORD hi;)
{
  ___WORD e;
  ___stream_index len_done;

  if (___INT(hi) <= ___INT(lo))
    return ___FIX(0);

  e = ___device_stream_write
        (___CAST(___device_stream*,___FIELD(dev,0)),
         ___CAST(___U8*,___BODY_AS(buffer,___tSUBTYPED)) + ___INT(lo),
         ___INT(hi) - ___INT(lo),
         &len_done);

  if (e != ___FIX(___NO_ERR))
    return e;

  return ___FIX(len_done);
}

___WORD ___dev_stream_width
   ___P((___WORD dev),
        (dev)
___WORD dev;)
{
  ___device_stream *d = ___CAST(___device_stream*,___FIELD(dev,0));

  return ___device_stream_width (d);
}

___WORD ___dev_stream_default_options
   ___P((___WORD dev),
        (dev)
___WORD dev;)
{
  ___device_stream *d = ___CAST(___device_stream*,___FIELD(dev,0));

  return ___device_stream_default_options (d);
}

___WORD ___dev_tty_type_set
   ___P((___WORD dev,
         ___WORD term_type,
         ___WORD emacs_bindings),
        (dev,
         term_type,
         emacs_bindings)
___WORD dev;
___WORD term_type;
___WORD emacs_bindings;)
{
  return ___FIX(___NO_ERR);
}

___WORD ___dev_tty_text_attributes_set
   ___P((___WORD dev,
         ___WORD input,
         ___WORD output),
        (dev,
         input,
         output)
___WORD dev;
___WORD input;
___WORD output;)
{
  return ___FIX(___NO_ERR);
}

___WORD ___dev_tty_history
   ___P((___WORD dev),
        (dev)
___WORD dev;)
{
  return ___FIX(___NO_ERR);
}

___WORD ___dev_tty_history_set
   ___P((___WORD dev,
         ___WORD history),
        (dev,
         history)
___WORD dev;
___WORD history;)
{
  return ___FIX(___NO_ERR);
}

___WORD ___dev_tty_max_history_length_set
   ___P((___WORD dev,
         ___WORD max_length),
        (dev,
         max_length)
___WORD dev;
___WORD max_length;)
{
  return ___FIX(___NO_ERR);
}

___WORD ___dev_tty_paren_balance_duration_set
   ___P((___WORD dev,
         ___WORD duration),
        (dev,
         duration)
___WORD dev;
___WORD duration;)
{
  return ___FIX(___NO_ERR);
}

___WORD ___dev_tty_mode_set
   ___P((___WORD dev,
         ___WORD mode),
        (dev,
         mode)
___WORD dev;
___WORD mode;)
{
  return ___FIX(___NO_ERR);
}



#define unicode_LF 10
#define unicode_CR 13
#define char_EOL unicode_LF
#define unicode_BOM 0xfeff

#define bytes_per_LATIN1 1
#define max_LATIN1       0xff

#define bytes_per_UTF8 1 /* optimization for single byte case */
#define max_UTF8       0x7f

#define bytes_per_UCS2 2
#define max_UCS2       0xffff

#define bytes_per_UCS4 4
#define max_UCS4       0x7fffffff

#define DECODE_EOL(loop_label) \
if (c != unicode_LF) \
  if (c != unicode_CR) \
    { \
      options += ___DECODE_STATE_NONE-___DECODE_STATE(options); \
      *chi++ = c; \
      if (chi < end) \
        goto loop_label; \
    } \
  else \
    { \
      int eol = ___EOL_ENCODING(options); \
      if (eol != ___EOL_ENCODING_LF) \
        { \
          if (eol != ___EOL_ENCODING_CR) \
            { \
              int rs = ___DECODE_STATE(options); \
              if (rs == ___DECODE_STATE_LF) \
                { \
                  options += \
                    ___DECODE_STATE_NONE-___DECODE_STATE_LF; \
                  goto loop_label; \
                } \
              options += ___DECODE_STATE_CR-rs; \
            } \
          c = char_EOL; \
        } \
      *chi++ = c; \
      if (chi < end) \
        goto loop_label; \
    } \
else \
  { \
    int eol = ___EOL_ENCODING(options); \
    if (eol != ___EOL_ENCODING_CR) \
      { \
        if (eol != ___EOL_ENCODING_LF) \
          { \
            int rs = ___DECODE_STATE(options); \
            if (rs == ___DECODE_STATE_CR) \
              { \
                options += \
                  ___DECODE_STATE_NONE-___DECODE_STATE_CR; \
                goto loop_label; \
              } \
            options += ___DECODE_STATE_LF-rs; \
          } \
        c = char_EOL; \
      } \
    *chi++ = c; \
    if (chi < end) \
      goto loop_label; \
  }

#define DECODE_CHARS_LOOP(loop_label,bytes_per_char,max_char,err_code,get_char) \
\
loop_label: \
blo += bytes_per_char; \
if (blo <= bhi) \
  { \
    c = get_char(-1); \
    if (max_char <= ___MAX_CHR || \
        c <= ___MAX_CHR) \
      { \
        DECODE_EOL(loop_label) \
      } \
    else \
      { \
        blo -= bytes_per_char; \
        err = err_code; \
      } \
  } \
else \
  { \
    blo -= bytes_per_char; \
    if (bytes_per_char > 1 && \
        blo < bhi &&   /* at least one byte left in buffer? */ \
        eof != ___FAL) /* end-of-file was reached? */ \
      err = err_code; /* trailing partial! */ \
  } \
break;

#define ENCODE_EOL(loop_label,bytes_per_char,put_char) \
switch (___EOL_ENCODING(options)) \
  { \
  case ___EOL_ENCODING_LFCR: \
    bhi += bytes_per_char; \
    if (bhi > end) \
      { \
        bhi -= 2*bytes_per_char; \
        clo--; \
        goto encode_chars_end; \
      } \
    put_char(-2,unicode_LF); \
  case ___EOL_ENCODING_CR: \
    put_char(-1,unicode_CR); \
    break; \
  case ___EOL_ENCODING_CRLF: \
    bhi += bytes_per_char; \
    if (bhi > end) \
      { \
        bhi -= 2*bytes_per_char; \
        clo--; \
        goto encode_chars_end; \
      } \
    put_char(-2,unicode_CR); \
  default: \
    put_char(-1,unicode_LF); \
    break; \
  } \
if (___LINE_BUFFERED(options)) \
  goto encode_chars_end; /* must empty byte buffer */ \
else if (clo < chi) \
  goto loop_label;

#define ENCODE_CHARS_LOOP(loop_label,bytes_per_char,max_char,err_code,put_char) \
\
loop_label: \
c = *clo++; \
if (___MAX_CHR <= max_char || \
    c <= max_char) \
  { \
    bhi += bytes_per_char; \
    if (bhi <= end) \
      if (c != char_EOL) \
        { \
          put_char(-1,c); \
          if (clo < chi) \
            goto loop_label; \
        } \
      else \
        { \
          ENCODE_EOL(loop_label,bytes_per_char,put_char); \
        } \
    else \
      { \
        bhi -= bytes_per_char; \
        clo--; \
        goto encode_chars_end; \
      } \
  } \
else \
  { \
    err = err_code; \
    goto encode_chars_end; \
  } \
break;

#define get_LATIN1(i) \
blo[(i)*bytes_per_LATIN1]

#define get_UTF8(i) \
blo[(i)*bytes_per_UTF8]

#define get_UCS2BE(i) \
((___UCS2)blo[(i)*bytes_per_UCS2+0] << 8) + \
(___UCS2)blo[(i)*bytes_per_UCS2+1]

#define get_UCS2LE(i) \
((___UCS2)blo[(i)*bytes_per_UCS2+1] << 8) + \
(___UCS2)blo[(i)*bytes_per_UCS2+0]

#define get_UCS4BE(i) \
((((((___UCS4)blo[(i)*bytes_per_UCS4+0] << 8) + \
    (___UCS4)blo[(i)*bytes_per_UCS4+1]) << 8) + \
  (___UCS4)blo[(i)*bytes_per_UCS4+2]) << 8) + \
(___UCS4)blo[(i)*bytes_per_UCS4+3]

#define get_UCS4LE(i) \
((((((___UCS4)blo[(i)*bytes_per_UCS4+3] << 8) + \
    (___UCS4)blo[(i)*bytes_per_UCS4+2]) << 8) + \
  (___UCS4)blo[(i)*bytes_per_UCS4+1]) << 8) + \
(___UCS4)blo[(i)*bytes_per_UCS4+0]

#define put_LATIN1(i,c) \
bhi[(i)*bytes_per_LATIN1] = (c);

#define put_UTF8(i,c) \
bhi[(i)*bytes_per_UTF8] = (c);

#define put_UCS2BE(i,c) \
bhi[(i)*bytes_per_UCS2+1] = (c) & 0xff; \
bhi[(i)*bytes_per_UCS2+0] = ((c)>>8) & 0xff;

#define put_UCS2LE(i,c) \
bhi[(i)*bytes_per_UCS2+0] = (c) & 0xff; \
bhi[(i)*bytes_per_UCS2+1] = ((c)>>8) & 0xff;

#define put_UCS4BE(i,c) \
bhi[(i)*bytes_per_UCS4+3] = (c) & 0xff; \
bhi[(i)*bytes_per_UCS4+2] = ((c)>>8) & 0xff; \
bhi[(i)*bytes_per_UCS4+1] = ((c)>>16) & 0xff; \
bhi[(i)*bytes_per_UCS4+0] = ((c)>>24) & 0xff;

#define put_UCS4LE(i,c) \
bhi[(i)*bytes_per_UCS4+0] = (c) & 0xff; \
bhi[(i)*bytes_per_UCS4+1] = ((c)>>8) & 0xff; \
bhi[(i)*bytes_per_UCS4+2] = ((c)>>16) & 0xff; \
bhi[(i)*bytes_per_UCS4+3] = ((c)>>24) & 0xff;

___WORD ___port_decode_chars
   ___P((___WORD port,
         ___WORD eof),
        (port,
         eof)
___WORD port;
___WORD eof;)
{
  ___ERR_CODE err;
  ___WORD result;
  ___C *char_rbuf =
     (___C*)___BODY_AS(___FIELD(port,___PORT_CHAR_RBUF),___tSUBTYPED);
  ___C *chi =
     char_rbuf + ___INT(___FIELD(port,___PORT_CHAR_RHI));
  ___C *end =
     char_rbuf + ___INT(___STRINGLENGTH(___FIELD(port,___PORT_CHAR_RBUF)));
  ___U8 *byte_rbuf =
     (___U8*)___BODY_AS(___FIELD(port,___PORT_BYTE_RBUF),___tSUBTYPED);
  ___U8 *blo =
     byte_rbuf + ___INT(___FIELD(port,___PORT_BYTE_RLO));
  ___U8 *bhi =
     byte_rbuf + ___INT(___FIELD(port,___PORT_BYTE_RHI));
  ___UCS4 c;
  int options = ___INT(___FIELD(port,___PORT_ROPTIONS));

  if ((___INT(___FIELD(port,___PORT_RKIND)) & ___TTY_DEVICE_KIND) ==
      ___TTY_DEVICE_KIND)
    {
      /*
       * Characters read from tty stream devices are represented as 8,
       * 16 or 32 bit unsigned integers of the machine's endianness.
       * It is the tty stream device driver that decodes the bytes
       * received from the device using the encoding specified in the
       * ___PORT_ROPTIONS field.
       */

      options = (options - ___CHAR_ENCODING(options)) +
                ___CS_SELECT(___CHAR_ENCODING_U8,
                             ___CHAR_ENCODING_U16,
                             ___CHAR_ENCODING_U32);
    }

  err = ___NO_ERR;

  ___FIELD(port,___PORT_CHAR_RCHARS) =
    ___FIELD(port,___PORT_CHAR_RCHARS) + ___FIX(chi-char_rbuf);

  chi = char_rbuf;

  /* fill character buffer as much as possible */

  if (chi < end)
    {
      switch (___CHAR_ENCODING(options))
        {
        default:
        case ___CHAR_ENCODING_ASCII:
        case ___CHAR_ENCODING_LATIN1:
          DECODE_CHARS_LOOP(decode_next_LATIN1,
                            bytes_per_LATIN1,
                            0xff,
                            ___CTOS_LATIN1STRING_ERR,
                            get_LATIN1);

        case ___CHAR_ENCODING_UTF8:
          {
            decode_next_UTF8:
            blo += bytes_per_UTF8;
            if (blo <= bhi)
              {
                c = get_UTF8(-1);
                if (c <= 0x7f)
                  {
                    DECODE_EOL(decode_next_UTF8);
                  }
                else if (c <= 0xbf || c > 0xfd)
                  {
                    blo -= bytes_per_UTF8;
                    err = ___CTOS_UTF8STRING_ERR;
                  }
                else
                  {
                    ___U8* orig_blo = blo;
                    ___U8 b0 = c;
                    int bits = 6;
                    while (b0 & 0x40)
                      {
                        ___U8 next = *blo++;
                        if (blo > bhi)
                          {
                            blo = orig_blo-bytes_per_UTF8;
                            if (eof != ___FAL) /* end-of-file reached? */
                              err = ___CTOS_UTF8STRING_ERR;
                            goto end_UTF8;
                          }
                        if (next <= 0x7f || next > 0xbf)
                          {
                            blo = orig_blo-bytes_per_UTF8;
                            err = ___CTOS_UTF8STRING_ERR;
                            goto end_UTF8;
                          }
                        c = (c << 6) + (next & 0x3f);
                        b0 <<= 1;
                        bits += 5;
                      }
                    c &= ((___UCS4)1<<bits)-1;
                    if (c >= 0x80 &&
                        c >= ((___UCS4)1<<(bits-5)) &&
                        c <= ___MAX_CHR)
                      {
                        options +=
                          ___DECODE_STATE_NONE-___DECODE_STATE(options);
                        *chi++ = c;
                        if (chi < end)
                          goto decode_next_UTF8;
                      }
                    else
                      {
                        blo = orig_blo-bytes_per_UTF8;
                        err = ___CTOS_UTF8STRING_ERR;
                      }
                    end_UTF8:;
                  }
              }
            else
              {
                blo -= bytes_per_UTF8;
                if (blo < bhi &&   /* at least one byte left in buffer? */
                    eof != ___FAL) /* end-of-file was reached? */
                  err = ___CTOS_UTF8STRING_ERR;
              }
          }

        case ___CHAR_ENCODING_UCS2:
          {
            blo += bytes_per_UCS2;
            if (blo <= bhi)
              {
                ___UCS4 cle;
                c = get_UCS2BE(-1);
                if (c == unicode_BOM)
                  {
                    options += ___CHAR_ENCODING_UCS2BE-___CHAR_ENCODING_UCS2;
                    goto decode_next_UCS2BE;
                  }
                cle = ((c&0xff) << 8) +
                      ((c>>8)&0xff);
                if (cle == unicode_BOM)
                  {
                    options += ___CHAR_ENCODING_UCS2LE-___CHAR_ENCODING_UCS2;
                    goto decode_next_UCS2LE;
                  }
                blo -= bytes_per_UCS2;
#ifdef DEFAULT_TO_BIGEND
                options += ___CHAR_ENCODING_UCS2BE-___CHAR_ENCODING_UCS2;
                goto decode_next_UCS2BE;
#else
                options += ___CHAR_ENCODING_UCS2LE-___CHAR_ENCODING_UCS2;
                goto decode_next_UCS2LE;
#endif
              }
            else
              {
                blo -= bytes_per_UCS2;
                if (bytes_per_UCS2 > 1 &&
                    blo < bhi &&   /* at least one byte left in buffer? */
                    eof != ___FAL) /* end-of-file was reached? */
                  err = ___CTOS_UCS2STRING_ERR; /* trailing partial! */
              }
            break;
          }

        case ___CHAR_ENCODING_UCS2BE:
          DECODE_CHARS_LOOP(decode_next_UCS2BE,
                            bytes_per_UCS2,
                            0xffff,
                            ___CTOS_UCS2STRING_ERR,
                            get_UCS2BE);

        case ___CHAR_ENCODING_UCS2LE:
          DECODE_CHARS_LOOP(decode_next_UCS2LE,
                            bytes_per_UCS2,
                            0xffff,
                            ___CTOS_UCS2STRING_ERR,
                            get_UCS2LE);

        case ___CHAR_ENCODING_UCS4:
          {
            blo += bytes_per_UCS4;
            if (blo <= bhi)
              {
                ___UCS4 cle;
                c = get_UCS4BE(-1);
                if (c == unicode_BOM)
                  {
                    options += ___CHAR_ENCODING_UCS4BE-___CHAR_ENCODING_UCS4;
                    goto decode_next_UCS4BE;
                  }
                cle = ((((((c&0xff) << 8) +
                          ((c>>8)&0xff)) << 8) +
                        ((c>>16)&0xff)) << 8) +
                      ((c>>24)&0xff);
                if (cle == unicode_BOM)
                  {
                    options += ___CHAR_ENCODING_UCS4LE-___CHAR_ENCODING_UCS4;
                    goto decode_next_UCS4LE;
                  }
                blo -= bytes_per_UCS4;
#ifdef DEFAULT_TO_BIGEND
                options += ___CHAR_ENCODING_UCS4BE-___CHAR_ENCODING_UCS4;
                goto decode_next_UCS4BE;
#else
                options += ___CHAR_ENCODING_UCS4LE-___CHAR_ENCODING_UCS4;
                goto decode_next_UCS4LE;
#endif
              }
            else
              {
                blo -= bytes_per_UCS4;
                if (bytes_per_UCS4 > 1 &&
                    blo < bhi &&   /* at least one byte left in buffer? */
                    eof != ___FAL) /* end-of-file was reached? */
                  err = ___CTOS_UCS4STRING_ERR; /* trailing partial! */
              }
            break;
          }

        case ___CHAR_ENCODING_UCS4BE:
          DECODE_CHARS_LOOP(decode_next_UCS4BE,
                            bytes_per_UCS4,
                            0xffffffff,
                            ___CTOS_UCS4STRING_ERR,
                            get_UCS4BE);

        case ___CHAR_ENCODING_UCS4LE:
          DECODE_CHARS_LOOP(decode_next_UCS4LE,
                            bytes_per_UCS4,
                            0xffffffff,
                            ___CTOS_UCS4STRING_ERR,
                            get_UCS4LE);
        }
    }

  /*
   * either the character buffer is full (chi == end) or no more
   * characters can be extracted from the byte buffer either because
   * the remaining bytes are not long enough to form a character (err
   * == ___NO_ERR) or don't form a valid character (err != ___NO_ERR).
   */

  if (chi == char_rbuf)
    {
      /* could not extract a single character from the byte buffer */

      if (err == ___NO_ERR)
        {
          /*
           * the remaining bytes in the buffer (possibly zero bytes)
           * are not long enough to form a character so make some room
           * at the end of the byte buffer to be filled with new bytes.
           */

          ___U8 *p = byte_rbuf;

          if (p != blo)
            {
              while (blo < bhi)
                *p++ = *blo++;
              blo = byte_rbuf;
              bhi = p;
            }
        }

      result = ___FAL; /* need to fill byte buffer */
    }
  else
    {
      err = ___NO_ERR; /* only report errors at head of character buffer */
      result = ___TRU; /* no need to fill byte buffer */
    }

  ___FIELD(port,___PORT_CHAR_RLO) = ___FIX(0);
  ___FIELD(port,___PORT_CHAR_RHI) = ___FIX(chi-char_rbuf);
  ___FIELD(port,___PORT_BYTE_RLO) = ___FIX(blo-byte_rbuf);
  ___FIELD(port,___PORT_BYTE_RHI) = ___FIX(bhi-byte_rbuf);
  ___FIELD(port,___PORT_ROPTIONS) = ___FIX(options);

  if (err != ___NO_ERR)
    result = ___FIX(err);

  return result;
}

___WORD ___port_encode_chars
   ___P((___WORD port),
        (port)
___WORD port;)
{
  ___ERR_CODE err;
  ___WORD result;
  ___C *char_wbuf =
     (___C*)___BODY_AS(___FIELD(port,___PORT_CHAR_WBUF),___tSUBTYPED);
  ___C *clo =
     char_wbuf + ___INT(___FIELD(port,___PORT_CHAR_WLO));
  ___C *chi =
     char_wbuf + ___INT(___FIELD(port,___PORT_CHAR_WHI));
  ___U8 *byte_wbuf =
     (___U8*)___BODY_AS(___FIELD(port,___PORT_BYTE_WBUF),___tSUBTYPED);
  ___U8 *bhi =
     byte_wbuf + ___INT(___FIELD(port,___PORT_BYTE_WHI));
  ___U8 *end =
     byte_wbuf + ___HD_BYTES(___HEADER(___FIELD(port,___PORT_BYTE_WBUF)));
  ___UCS4 c;
  int options = ___INT(___FIELD(port,___PORT_WOPTIONS));

  if ((___INT(___FIELD(port,___PORT_WKIND)) & ___TTY_DEVICE_KIND) ==
      ___TTY_DEVICE_KIND)
    {
      /*
       * Characters written to tty stream devices are represented as
       * 8, 16 or 32 bit unsigned integers of the machine's
       * endianness.  When the tty stream device driver transfers the
       * characters to the device, the characters will be encoded
       * using the encoding specified in the ___PORT_WOPTIONS field.
       */

      options = (options - ___CHAR_ENCODING(options)) +
                ___CS_SELECT(___CHAR_ENCODING_U8,
                             ___CHAR_ENCODING_U16,
                             ___CHAR_ENCODING_U32);
    }

  err = ___NO_ERR;

  result = ___TRU; /* default is to empty byte buffer */

{
#if 0
 ___C *p; printf ("<clo=%d,",clo-char_wbuf); for (p=clo; p<chi; p++) printf ("%c", *p); printf (">"); fflush(stdout);
#endif
}

  /* empty character buffer as much as possible */

  if (clo < chi)
    {
      switch (___CHAR_ENCODING(options))
        {
        default:
        case ___CHAR_ENCODING_ASCII:
        case ___CHAR_ENCODING_LATIN1:
          ENCODE_CHARS_LOOP(encode_next_LATIN1,
                            bytes_per_LATIN1,
                            max_LATIN1,
                            ___CTOS_LATIN1STRING_ERR,
                            put_LATIN1);

        case ___CHAR_ENCODING_UTF8:
          {
            encode_next_UTF8:
            c = *clo++;
            if (___MAX_CHR <= max_UTF8 ||
                c <= max_UTF8)
              {
                bhi += bytes_per_UTF8;
                if (bhi <= end)
                  if (c != char_EOL)
                    {
                      put_UTF8(-1,c);
                      if (clo < chi)
                        goto encode_next_UTF8;
                    }
                  else
                    {
                      ENCODE_EOL(encode_next_UTF8,bytes_per_UTF8,put_UTF8);
                    }
                else
                  {
                    bhi -= bytes_per_UTF8;
                    clo--;
                    goto encode_chars_end;
                  }
              }
            else
              {
                ___U8 *p;
                int bytes;
                if      (c <= 0x7ff)      bytes = 2;
                else if (c <= 0xffff)     bytes = 3;
                else if (c <= 0x1fffff)   bytes = 4;
                else if (c <= 0x3ffffff)  bytes = 5;
                else if (c <= 0x7fffffff) bytes = 6;
                else
                  {
                    err = ___STOC_UTF8STRING_ERR;
                    goto encode_chars_end;
                  }
                p = bhi + bytes;
                if (p <= end)
                  {
                    bhi = p;
                    switch (bytes)
                      {
                      case 6:  *--p = 0x80+(c&0x3f); c >>= 6;
                      case 5:  *--p = 0x80+(c&0x3f); c >>= 6;
                      case 4:  *--p = 0x80+(c&0x3f); c >>= 6;
                      case 3:  *--p = 0x80+(c&0x3f); c >>= 6;
                      default: *--p = 0x80+(c&0x3f); c >>= 6;
                      }
                    *--p = 0xff - (0xff>>bytes) + c;
                    if (clo < chi)
                      goto encode_next_UTF8;
                  }
                else
                  {
                    clo--;
                    goto encode_chars_end;
                  }
              }
            break;
          }

        case ___CHAR_ENCODING_UCS2:
          bhi += bytes_per_UCS2;
          if (bhi > end)
            {
              bhi -= bytes_per_UCS2;
              goto encode_chars_end;
            }
#ifdef DEFAULT_TO_BIGEND
          put_UCS2BE(-1,unicode_BOM);
          options += ___CHAR_ENCODING_UCS2BE-___CHAR_ENCODING_UCS2;
          ___FIELD(port,___PORT_WOPTIONS) = ___FIX(options);
          goto encode_next_UCS2BE;
#else
          put_UCS2LE(-1,unicode_BOM);
          options += ___CHAR_ENCODING_UCS2LE-___CHAR_ENCODING_UCS2;
          ___FIELD(port,___PORT_WOPTIONS) = ___FIX(options);
          goto encode_next_UCS2LE;
#endif

        case ___CHAR_ENCODING_UCS2BE:
          ENCODE_CHARS_LOOP(encode_next_UCS2BE,
                            bytes_per_UCS2,
                            max_UCS2,
                            ___CTOS_UCS2STRING_ERR,
                            put_UCS2BE);

        case ___CHAR_ENCODING_UCS2LE:
          ENCODE_CHARS_LOOP(encode_next_UCS2LE,
                            bytes_per_UCS2,
                            max_UCS2,
                            ___CTOS_UCS2STRING_ERR,
                            put_UCS2LE);

        case ___CHAR_ENCODING_UCS4:
          bhi += bytes_per_UCS4;
          if (bhi > end)
            {
              bhi -= bytes_per_UCS4;
              goto encode_chars_end;
            }
#ifdef DEFAULT_TO_BIGEND
          put_UCS4BE(-1,unicode_BOM);
          options += ___CHAR_ENCODING_UCS4BE-___CHAR_ENCODING_UCS4;
          ___FIELD(port,___PORT_WOPTIONS) = ___FIX(options);
          goto encode_next_UCS4BE;
#else
          put_UCS4LE(-1,unicode_BOM);
          options += ___CHAR_ENCODING_UCS4LE-___CHAR_ENCODING_UCS4;
          ___FIELD(port,___PORT_WOPTIONS) = ___FIX(options);
          goto encode_next_UCS4LE;
#endif

        case ___CHAR_ENCODING_UCS4BE:
          ENCODE_CHARS_LOOP(encode_next_UCS4BE,
                            bytes_per_UCS4,
                            max_UCS4,
                            ___CTOS_UCS4STRING_ERR,
                            put_UCS4BE);

        case ___CHAR_ENCODING_UCS4LE:
          ENCODE_CHARS_LOOP(encode_next_UCS4LE,
                            bytes_per_UCS4,
                            max_UCS4,
                            ___CTOS_UCS4STRING_ERR,
                            put_UCS4LE);
        }
    }

  /* character buffer has been fully emptied */

  ___FIELD(port,___PORT_CHAR_WCHARS) =
    ___FIELD(port,___PORT_CHAR_WCHARS) + ___FIX(chi-char_wbuf);

  clo = char_wbuf;
  chi = char_wbuf;

  result = ___FAL; /* do not empty byte buffer */

  encode_chars_end:

  ___FIELD(port,___PORT_CHAR_WLO) = ___FIX(clo-char_wbuf);
  ___FIELD(port,___PORT_CHAR_WHI) = ___FIX(chi-char_wbuf);
  ___FIELD(port,___PORT_BYTE_WHI) = ___FIX(bhi-byte_wbuf);

  if (err != ___NO_ERR)
    result = ___FIX(err);

  return result;
}

#endif

/*---------------------------------------------------------------------------*/

___device_stream *___device_stream_open
   ___P((char *name,
         int direction),
        (name,
         direction)
char *name;
int direction;)
{
  static ___BOOL initialized = 0;
  static ___device_group *dg;
  ___device_stream *dev;
  if (!initialized)
    {
      ___device_group_setup (&dg);
      initialized = 1;
    }
  ___device_stream_setup_from_path (&dev, dg, name, 0, 0);/**********/
  return dev;
}

___WORD ___condvar_select
   ___P((___WORD run_queue,
         ___WORD timeout),
        (run_queue,
         timeout)
___WORD run_queue;
___WORD timeout;)
{
#define MAX_CONDVARS SELECT_DEVS_MAX

#define ___BTQ_DEQ_NEXT 1
#define ___BTQ_DEQ_PREV 2
#define ___BTQ_COLOR    3
#define ___BTQ_PARENT   4
#define ___BTQ_LEFT     5
#define ___BTQ_RIGHT    6
#define ___BTQ_LEFTMOST 6
#define ___BTQ_OWNER    7
#define ___CONDVAR_NAME 8

  ___WORD e;
  ___time to;
  ___device *devs[MAX_CONDVARS];
  ___WORD condvars[MAX_CONDVARS];
  int read_pos;
  int write_pos;
  ___WORD condvar;
  int i;
  int j;

  if (timeout == ___FAL)
    to = ___time_neg_infinity;
  else if (timeout == ___TRU)
    to = ___time_pos_infinity;
  else
    ___time_from_seconds (&to, ___F64VECTORREF(timeout,___FIX(0)));

  read_pos = 0;
  write_pos = MAX_CONDVARS;
  condvar = ___FIELD(run_queue,___BTQ_DEQ_NEXT);

  while (condvar != run_queue)
    {
      ___WORD owner = ___FIELD(condvar,___BTQ_OWNER);
      if (read_pos < write_pos)
        {
          if (owner & ___FIX(2))
            condvars[--write_pos] = condvar;
          else
            condvars[read_pos++] = condvar;
          ___FIELD(condvar,___BTQ_OWNER) = owner & ~___FIX(1);
        }
      else
        {
          to = ___time_neg_infinity;
          ___FIELD(condvar,___BTQ_OWNER) = owner | ___FIX(1);
        }
      condvar = ___FIELD(condvar,___BTQ_DEQ_NEXT);
    }

  i = 0;

  while (i < read_pos)
    {
      devs[i] = ___CAST(___device*,
                        ___FIELD(___FIELD(condvars[i],___CONDVAR_NAME),0));
      i++;
    }

  j = MAX_CONDVARS;

  while (j > write_pos)
    {
      j--;
      devs[i] = ___CAST(___device*,
                        ___FIELD(___FIELD(condvars[j],___CONDVAR_NAME),0));
      i++;
    }

#if 0
  fprintf (stderr, "___device_select (devs, %d, %d, %20.16g)\n",
           read_pos, MAX_CONDVARS-write_pos, to);
  fflush (stderr);
#endif

  e = ___device_select (devs, read_pos, MAX_CONDVARS-write_pos, to);

  i = 0;

  while (i < read_pos)
    {
      if (devs[i] == NULL)
        {
          condvar = condvars[i];
          ___FIELD(condvar,___BTQ_OWNER) |= ___FIX(1);
        }
      i++;
    }

  j = MAX_CONDVARS;

  while (j > write_pos)
    {
      j--;
      if (devs[i] == NULL)
        {
          condvar = condvars[j];
          ___FIELD(condvar,___BTQ_OWNER) |= ___FIX(1);
        }
      i++;
    }

  return e;
}

#if 0

___WORD ___device_select ___P((___WORD timeout),(timeout)/*************************/
___WORD timeout;)
{
  fd_set readfds, writefds, exceptfds;
  struct timeval to;
  struct timeval *pto;
  struct timeval tv;

      FD_ZERO (&readfds);
      FD_ZERO (&writefds);
      FD_ZERO (&exceptfds);

      if (timeout == ___TRU)
        {
printf ("device_select (#t)\n");
          pto = NULL;
        }
      else
        {
          if (timeout == ___FAL ||
              gettimeofday (&tv, NULL) != 0)
            {
printf ("device_select (#f)\n");
              to.tv_sec  = 0;
              to.tv_usec = 0;
            }
          else
            {
              ___F64 x = ___F64VECTORREF(timeout,___FIX(0)) -
                         (tv.tv_sec * 1.0e6 + tv.tv_usec);
printf ("device_select (%g)  x=%g\n",___F64VECTORREF(timeout,___FIX(0)),x);
              if (x < 0.0)
                x = 0.0;
              {
                long secs = x / 1000000.0;
                long usecs = x - secs * 1000000.0;
                to.tv_sec  = secs;
                to.tv_usec = usecs;
              }
            }
          pto = &to;
        }

      select (0, &readfds, &writefds, &exceptfds, pto);

printf ("select done\n");

  return ___FAL;
}


#endif
/*****************************************************************************/

/*---------------------------------------------------------------------------*/

#if 0
/* Line-editing stuff.  ****TO BE CLEANED UP**** */

#define SUPPORT_F5_TO_F12zzzzzzz

#define LINEEDIT_WITH_NONBLOCKING_IO

#define USE_LINEEDIT

#define USE_UNIX
#define USE_WINDOWSzzzzzzzz
#define USE_TERMINFO
#define USE_TERMCAPzzzzzzzzzz
#define USE_LOCAL_CLIPBOARD

#define LINEEDIT_REPORT_CHAR_ENCODING_ERRORS

#define ___HEAP_OVERFLOW_ERR (___ERR_BASE+4)
#define ___INVALID_OP_ERR (___ERR_BASE+5)


#ifndef ___VERSION_MAJOR

/*
cl -Zi -MT console.c ws2_32.lib user32.lib
*/

#include <stdlib.h>

#define ___CTOS_UTF8STRING_ERR (-1000)
#define ___STOC_LATIN1STRING_ERR (-1001)

#define ___ERR_CODE_FACILITY_SYSTEM 448
#define ___ERR_CODE_FACILITY_ERRNO  384
#define ___ERR_CODE_FACILITY_MACOS  320

#define ___ERR_CODE_BUILD(facility,code) \
(((___ERR_CODE)-1<<29)+((___ERR_CODE)(facility)<<16)+(code))

#define ___ERR_CODE_FACILITY(err)(((err)>>16)&~(~(___ERR_CODE)0<<9))

#define ___ERR_CODE_FROM_HRESULT(hresult) \
((hresult)&~(~(___ERR_CODE)0<<25)|(((hresult)&~(___ERR_CODE)0<<27)>>2))

#define ___ERR_CODE_TO_HRESULT(err) \
((err)&~(~(___ERR_CODE)0<<25)|(((err)&~(___ERR_CODE)0<<25)<<2))

#define ___ERR_CODE_FROM_ERRNO(errno) \
((errno)==0?___NO_ERR:___ERR_CODE_BUILD(___ERR_CODE_FACILITY_ERRNO,errno))

#define ___ERR_CODE_TO_ERRNO(err) \
((err)==___NO_ERR?0:((err)-___ERR_CODE_BUILD(___ERR_CODE_FACILITY_ERRNO,0)))

#define ___ERR_CODE_FROM_MACOS(macos) \
((macos)==0?___NO_ERR:___ERR_CODE_BUILD(___ERR_CODE_FACILITY_MACOS,macos))

#define ___ERR_CODE_TO_MACOS(err) \
((err)==___NO_ERR?0:((err)-___ERR_CODE_BUILD(___ERR_CODE_FACILITY_MACOS,0)))

#define ___ERR_BASE ___ERR_CODE_BUILD(___ERR_CODE_FACILITY_SYSTEM,0)

#define ___NO_ERR                  0
#define ___UNWIND_C_STACK          (___ERR_BASE+0)
#define ___CDEF_HEAP_OVERFLOW_ERR  (___ERR_BASE+1)
#define ___IMPL_LIMIT_ERR          (___ERR_BASE+2)
#define ___UNKNOWN_ERR             (___ERR_BASE+3)


#define ___HIDDEN
#define __STDC__ 1
#undef __STDC__zzz

#ifdef __cplusplus
#define ___P(ansi,kr)ansi
#define ___PVOID ()
#define ___BOOL int
#define ___BEGIN_C_LINKAGE extern "C" {
#define ___END_C_LINKAGE }
#else
#ifdef __STDC__
#define ___P(ansi,kr)ansi
#define ___PVOID (void)
#else
#define ___P(ansi,kr)kr
#define ___PVOID ()
#endif
#define ___BOOL int
#define ___BEGIN_C_LINKAGE
#define ___END_C_LINKAGE
#define ___USE_SETJMP
#endif

#define ___WORD long
#define ___FIX(n)((n)<<2)
#define ___INT(n)((n)>>2)

#ifndef ___CS
#define ___CS 2
#endif

#if ___CS == 1
#define ___CS_SELECT(cs1,cs2,cs4)cs1
#else
#if ___CS == 2
#define ___CS_SELECT(cs1,cs2,cs4)cs2
#else
#define ___CS_SELECT(cs1,cs2,cs4)cs4
#endif
#endif

#define ___LCS ___CS_SELECT(0,1,2)

#define ___TB 2
#define ___MAX_FIX32 (((___WORD)1<<(32-___TB-1))-1)

#if ___CS == 4
#define ___MAX_CHR ___MAX_FIX32
#else
#define ___MAX_CHR (((___WORD)1<<(8*___CS))-1)
#endif

#define ___ERR_CODE int
#define ___ERR_CODE_FROM_WSAERROR(err) err
#define ___ERR_CODE_FROM_HRESULT(err) err
#define ___ERR_CODE_FROM_ERRNO(err) (-100000-(err))
#define ___NO_ERR 0
#define ___ERR_KILL_PUMP ___ERR_CODE_FROM_HRESULT((1<<29))

#define ___U8 unsigned char
#define ___U16 unsigned short
#define ___U32 unsigned int
#define ___LATIN1 unsigned char
#define ___UCS2 unsigned short
#define ___UCS4 unsigned long

/*---------------------------------------------------------------------------*/

/* Low-level memory allocation. */

long allocs = 0;
long frees = 0;

void *___alloc_mem ___P((unsigned long bytes),(bytes)
unsigned long bytes;)
{
#ifdef USE_WINDOWS
  InterlockedIncrement (&allocs);
#else
  allocs++;
#endif
#ifdef USE_MACOS
  if (has_OSDispatch)
    {
      OSErr err;
      Ptr p;
      Handle h = TempNewHandle (sizeof (Handle) + bytes, &err);
      if (err != noErr || h == 0)
        return 0;
      HLock (h);
      p = *h;
      *(Handle*)p = h;
      return p + sizeof (Handle);
    }
  else
    return malloc (bytes);
#else
  return malloc (bytes);
#endif
}

void ___free_mem ___P((void *ptr),(ptr)
void *ptr;)
{
#ifdef USE_WINDOWS
  InterlockedIncrement (&frees);
#else
  frees++;
#endif
#ifdef USE_MACOS
  if (has_OSDispatch)
    {
      OSErr err;
      Handle h = *(Handle*)((Ptr)ptr - sizeof (Handle));
      HUnlock (h);
      TempDisposeHandle (h, &err);
    }
  else
    free (ptr);
#else
  free (ptr);
#endif
}

#endif






#define UNICODE_NUL         0
#define UNICODE_BELL        7
#define UNICODE_BACKSPACE   8
#define UNICODE_TAB         9
#define UNICODE_LINEFEED    10
#define UNICODE_RETURN      13
#define UNICODE_ESCAPE      27
#define UNICODE_SPACE       32
#define UNICODE_DOUBLEQUOTE 34
#define UNICODE_SHARP       35
#define UNICODE_QUOTE       39
#define UNICODE_COMMA       44
#define UNICODE_1           49
#define UNICODE_2           50
#define UNICODE_3           51
#define UNICODE_4           52
#define UNICODE_6           54
#define UNICODE_8           56
#define UNICODE_SEMICOLON   59
#define UNICODE_AT          64
#define UNICODE_UPPER_F     70
#define UNICODE_LOWER_F     102
#define UNICODE_UPPER_S     83
#define UNICODE_LOWER_S     115
#define UNICODE_UPPER_U     85
#define UNICODE_LOWER_U     117
#define UNICODE_LPAREN      40
#define UNICODE_RPAREN      41
#define UNICODE_LBRACKET    91
#define UNICODE_BACKSLASH   92
#define UNICODE_RBRACKET    93
#define UNICODE_LBRACE      123
#define UNICODE_VBAR        124
#define UNICODE_RBRACE      125
#define UNICODE_RUBOUT      127
#define UNICODE_REPLACEMENT 0xfffd

#ifdef USE_TERMCAP
#define USE_TERMCAP_OR_TERMINFO
#endif

#ifdef USE_TERMINFO
#define USE_TERMCAP_OR_TERMINFO
#endif



#ifdef USE_WINDOWS

#include <windows.h>
#include <tchar.h>
#include <sys/types.h>
#include <sys/timeb.h>

#undef COLOR_CYAN
#undef COLOR_BLUE
#define COLOR_CYAN 6
#define COLOR_BLUE 4

#endif

#ifdef USE_UNIX

#include <curses.h>
#include <term.h>

#include <signal.h>
#include <sys/ioctl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/time.h>

#include <termios.h>
#include <unistd.h>

#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>


#ifdef USE_UNIX

___HIDDEN ___ERR_CODE err_code_from_errno ___PVOID
{
  int e = errno;

  if (e == 0)
    return ___UNKNOWN_ERR;

  return ___ERR_CODE_FROM_ERRNO(e);
}

#endif

#ifdef USE_WINDOWS

___HIDDEN ___ERR_CODE err_code_from_GetLastError ___PVOID
{
  DWORD e = GetLastError ();

  if (e == ERROR_SUCCESS)
    return ___UNKNOWN_ERR;

  return ___ERR_CODE_FROM_HRESULT(e);
}

#endif

/*****************************************************************************/
#ifdef USE_UNIXzzzzzzzzzz

extern char *ptsname (int);

int ptym_open (char *pts_name)
{
  char *ptr;
  int fdm;

  strcpy (pts_name, "/dev/ptmx");

  if ((fdm = open (pts_name, O_RDWR)) < 0)
    return -1;

  if (grantpt (fdm) < 0)
    {
      close (fdm);
      return -2;
    }

  if (unlockpt (fdm) < 0)
    {
      close (fdm);
      return -3;
    }

  if ((ptr = ptsname (fdm)) == NULL)
    {
      close (fdm);
      return -4;
    }

  strcpy (pts_name, ptr);

  return fdm;
}

int ptys_open (int fdm, char *pts_name)
{
  int fds;

  if ((fds = open (pts_name, O_RDWR)) < 0)
    {
      close (fdm);
      return -5;
    }

#if 0
  if (ioctl (fds, I_PUSH, "ptem") < 0)
    {
      close (fdm);
      close (fds);
      return -6;
    }

  if (ioctl (fds, I_PUSH, "ldterm") < 0)
    {
      close (fdm);
      close (fds);
      return -7;
    }

  if (ioctl (fds, I_PUSH, "ttcompat") < 0)
    {
      close (fdm);
      close (fds);
      return -8;
    }
#endif

  return fds;
}

pid_t pty_fork (int *ptrfdm)
{
  int fdm;
  int fds;
  pid_t pid;
  char pts_name[100];

  if ((fdm = ptym_open (pts_name)) < 0)
    exit (999);

  *ptrfdm = fdm;

  if ((pid = fork ()) < 0)
    return -1;

  if (pid == 0)
    {
      /* child */

      if (setsid () < 0)
        exit (999);

      if ((fds = ptys_open (fdm, pts_name)) < 0)
        exit (999);

      close (fdm);

      if (dup2 (fds, 0) != 0)
        exit (999);
      if (dup2 (fds, 1) != 1)
        exit (999);
      if (dup2 (fds, 2) != 2)
        exit (999);

      if (fds > 2)
        close (fds);
    }

  return pid;
}

int open_xterm (void)
{
  int fd;
  pid_t pid;
  char buf[101];
  int n;
  struct termios tios;

  pid = pty_fork (&fd);

  if (pid < 0)
    exit (999);

  if (pid == 0)
    {
      /* child */

      if (tcgetattr (0, &tios) < 0)
        exit (999);
      tios.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL | ICANON);
      tios.c_iflag &= ~(INLCR | ICRNL);
      tios.c_oflag &= ~(OPOST | ONLCR | OCRNL);
      if (tcsetattr (0, TCSANOW, &tios) < 0)
        exit (999);

      execlp ("/usr/local/bin/xterm",
              "/usr/local/bin/xterm",
              "-S001",
              "-u8",
              "-fn",
              "-misc-fixed-medium-r-normal--20-200-75-75-c-100-iso10646-1",
              "-ls",
              "-sl",
              "1000",
              "-sb",
              0);

      exit (999);
    }

  /* parent */

  if (tcgetattr (fd, &tios) < 0)
    exit (999);
  tios.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL | ICANON);
  tios.c_iflag &= ~(INLCR | ICRNL);
  tios.c_oflag &= ~(OPOST | ONLCR | OCRNL);
  if (tcsetattr (fd, TCSANOW, &tios) < 0)
    exit (999);

  n = read (fd, buf, 100); /* wait until xterm starts */

  fcntl (fd, F_SETFL, O_NONBLOCK); /* set nonblocking mode */

  return fd;
}
#endif
/*****************************************************************************/


/*---------------------------------------------------------------------------*/

/* Data types.  */

/* Type of characters that can be read from and written to a tty */

typedef ___CS_SELECT(___LATIN1,___UCS2,___UCS4) tty_char;

#define TTY_CHAR_MAX ___MAX_CHR

/* Tty management data structures */

#define TTY_TAB_STOP 8

#define TTY_RBUF_SIZE 100
#define TTY_WBUF_SIZE 100

#define DEFAULT_EMACS_BINDINGS 1
#define DEFAULT_PAREN_BALANCE_DURATION_MSECS 500 /* 1/2 second */

typedef int tty_input_mode;

#define CBREAK_LINEEDIT_INPUT_MODE 0
#define CBREAK_INPUT_MODE          1
#define RAW_LINEEDIT_INPUT_MODE    2
#define RAW_INPUT_MODE             3

typedef int tty_term_mode;

#define INITIAL_TERM_MODE 0
#define CBREAK_TERM_MODE  1
#define RAW_TERM_MODE     2

typedef int tty_encoding;

#define ASCII_ENCODING  0
#define LATIN1_ENCODING 1
#define UTF8_ENCODING   2
#define UCS2LE_ENCODING 3
#define UCS2BE_ENCODING 4
#define UCS4LE_ENCODING 5
#define UCS4BE_ENCODING 6
#define NATIVE_ENCODING 7

typedef int tty_text_attrs;

#define NORMAL_STYLE  0
#define BOLD_STYLE    1
#define REVERSE_STYLE 2

#define DEFAULT_COLOR 8

#define GET_STYLE(x)(((x)>>8)&3)
#define GET_FOREGROUND_COLOR(x)((x)&15)
#define GET_BACKGROUND_COLOR(x)(((x)>>4)&15)

#define MAKE_TEXT_ATTRS(s,f,b)(((s)<<8)+(f)+((b)<<4))

#define DEFAULT_TEXT_ATTRS \
MAKE_TEXT_ATTRS(NORMAL_STYLE,DEFAULT_COLOR,DEFAULT_COLOR)

typedef int tty_signal;

#define USER_INTERRUPT_SIGNAL 0

#ifndef USE_UNIX
#ifndef USE_WINDOWS
#ifndef USE_MACOS

#define NB_TTY_SIGNALS 1

#endif
#endif
#endif

#ifdef USE_UNIX

#define WINDOW_SIZE_SIGNAL 1
#define CONTINUE_SIGNAL    2
#define NB_TTY_SIGNALS     3

#define CONSOLE_TERMINAL_TYPE_ENV_VAR "TERM"

#endif

#ifdef USE_WINDOWS

#define NB_TTY_SIGNALS 1

#endif

#ifdef USE_MACOS

/********************/

#endif

typedef struct tty_state_struct
  {
    struct tty_state_struct *prev; /* circular list links */
    struct tty_state_struct *next;

    int refcount; /* nb of "tty" structures that point to this "tty_state" */

    tty_input_mode input_mode; /* input mode (CANONICAL, CBREAK, RAW) */
    tty_term_mode term_mode;   /* terminal mode (INITIAL, CBREAK, RAW) */
    tty_encoding encoding;     /* external char encoding (ASCII, UTF8, ...) */

    /* read buffer, needed for decoding characters */
    int rlo;
    int rhi;
    ___U8 rbuf[TTY_RBUF_SIZE];

    /* write buffer, needed for encoding characters */
    int wlo;
    int whi;
    ___U8 wbuf[TTY_WBUF_SIZE];
    ___WORD werr; /* write error flag for latest write operation */

    ___BOOL received_signal[NB_TTY_SIGNALS]; /* signals that were received */

#ifdef USE_UNIX
    int fd;                         /* file descriptor of the terminal */
    struct termios initial_termios; /* initial terminal driver settings */
#endif

#ifdef USE_WINDOWS
    HANDLE input;
    HANDLE output;
#define INPUT_RBUF_SIZE 10
    INPUT_RECORD input_rbuf[INPUT_RBUF_SIZE];
    int input_rlo;
    int input_rhi;
#endif

#ifdef USE_MACOS
    /*****************/
#endif

#ifdef USE_LINEEDIT
    struct lineedit_state_struct *les; /* line edit state */
#endif
  } tty_state;

typedef struct tty_struct
  {
    tty_state *state; /* NULL means not yet open console */
  } tty;

typedef struct tty_module_struct
  {
    tty_state *list; /* circular list of tty_state's */
    tty console;     /* the console is a special tty */
    void (*user_interrupt_handler) ___PVOID; /* called on a user interrupt */
  } tty_module;

___HIDDEN tty_module tm; /* tty module's global state */

/*---------------------------------------------------------------------------*/

/*********************/

___HIDDEN ___WORD tty_set_term_mode
   ___P((tty_state *state,
         tty_term_mode mode),
        ());

___HIDDEN ___WORD tty_set_terminal_type
   ___P((tty *t,
         char *terminal_type,
         ___BOOL emacs_bindings),
        ());

___HIDDEN ___WORD tty_set_encoding
   ___P((tty *t,
         tty_encoding encoding),
        ());

___HIDDEN ___WORD tty_cleanup
   ___P((tty *t),
        ());

/*---------------------------------------------------------------------------*/

#ifdef USE_LINEEDIT

/* Extensible string data structure */

typedef tty_char extensible_string_char;

typedef struct extensible_string_struct
  {
    extensible_string_char *buffer;
    int length;
    int max_length;
  } extensible_string;

/* Tty line editing data structures */

typedef struct lineedit_time_struct
  {
    int secs;
    int msecs;
  } lineedit_time;

#define LINEEDIT_EVENT_NONE           0
#define LINEEDIT_EVENT_KEY            1
#define LINEEDIT_EVENT_RETURN         2
#define LINEEDIT_EVENT_BACK           3
#define LINEEDIT_EVENT_BACK_WORD      4
#define LINEEDIT_EVENT_TAB            5
#define LINEEDIT_EVENT_MARK           6
#define LINEEDIT_EVENT_PASTE          7
#define LINEEDIT_EVENT_CUT            8
#define LINEEDIT_EVENT_CUT_RIGHT      9
#define LINEEDIT_EVENT_CUT_LEFT       10
#define LINEEDIT_EVENT_REFRESH        11
#define LINEEDIT_EVENT_TRANSPOSE      12
#define LINEEDIT_EVENT_TRANSPOSE_WORD 13
#define LINEEDIT_EVENT_UP             14
#define LINEEDIT_EVENT_DOWN           15
#define LINEEDIT_EVENT_RIGHT          16
#define LINEEDIT_EVENT_RIGHT_WORD     17
#define LINEEDIT_EVENT_LEFT           18
#define LINEEDIT_EVENT_LEFT_WORD      19
#define LINEEDIT_EVENT_HOME           20
#define LINEEDIT_EVENT_HOME_DOC       21
#define LINEEDIT_EVENT_INSERT         22
#define LINEEDIT_EVENT_DELETE         23
#define LINEEDIT_EVENT_DELETE_WORD    24
#define LINEEDIT_EVENT_END            25
#define LINEEDIT_EVENT_END_DOC        26
#define LINEEDIT_EVENT_F1             27
#define LINEEDIT_EVENT_META_F1        28
#define LINEEDIT_EVENT_F2             29
#define LINEEDIT_EVENT_META_F2        30
#define LINEEDIT_EVENT_F3             31
#define LINEEDIT_EVENT_META_F3        32
#define LINEEDIT_EVENT_F4             33
#define LINEEDIT_EVENT_META_F4        34

#ifdef SUPPORT_F5_TO_F12
#define LINEEDIT_EVENT_F5             35
#define LINEEDIT_EVENT_META_F5        36
#define LINEEDIT_EVENT_F6             37
#define LINEEDIT_EVENT_META_F6        38
#define LINEEDIT_EVENT_F7             39
#define LINEEDIT_EVENT_META_F7        40
#define LINEEDIT_EVENT_F8             41
#define LINEEDIT_EVENT_META_F8        42
#define LINEEDIT_EVENT_F9             43
#define LINEEDIT_EVENT_META_F9        44
#define LINEEDIT_EVENT_F10            45
#define LINEEDIT_EVENT_META_F10       46
#define LINEEDIT_EVENT_F11            47
#define LINEEDIT_EVENT_META_F11       48
#define LINEEDIT_EVENT_F12            49
#define LINEEDIT_EVENT_META_F12       50
#define LINEEDIT_EVENT_LAST           LINEEDIT_EVENT_META_F12
#else
#define LINEEDIT_EVENT_LAST           LINEEDIT_EVENT_META_F4
#endif

#define LINEEDIT_CAP_CLEAR 0
#define LINEEDIT_CAP_ED    1
#define LINEEDIT_CAP_CUU1  2
#define LINEEDIT_CAP_CUD1  3
#define LINEEDIT_CAP_CUU   4
#define LINEEDIT_CAP_CUD   5
#define LINEEDIT_CAP_CUF   6
#define LINEEDIT_CAP_CUB   7
#define LINEEDIT_CAP_SGR0  8
#define LINEEDIT_CAP_BOLD  9
#define LINEEDIT_CAP_REV   10
#define LINEEDIT_CAP_SETAF 11
#define LINEEDIT_CAP_SETAB 12
#define LINEEDIT_CAP_LAST  12

typedef struct lineedit_event_struct
  {
    int event_kind;
    tty_char event_char;
  } lineedit_event;

typedef struct lineedit_history_struct
  {
    struct lineedit_history_struct *prev; /* circular list links */
    struct lineedit_history_struct *next;
    extensible_string actual;
    extensible_string edited; /* buffer != NULL when being edited */
  } lineedit_history;

#define WITH_ESC_PREFIX ((___U8)(1<<7))

/*
 * LARGEST_DECODER is the size of the largest input decoder over all
 * terminals defined in the termcap database.  The largest input
 * decoder is for the "ti926-8" terminal.  Note that this value must
 * be updated when the keyboard bindings are changed.
 */

#ifdef SUPPORT_F5_TO_F12
#define LARGEST_DECODER 177
#else
#define LARGEST_DECODER 128
#endif

/*
 * The type lineedit_input_decoder_state must at least contain the
 * range 0 .. LINEEDIT_EVENT_LAST + LARGEST_DECODER + 1.
 */

#if LINEEDIT_EVENT_LAST + LARGEST_DECODER + 1 <= 255

typedef ___U8 lineedit_input_decoder_state;
#define LINEEDIT_INPUT_DECODER_STATE_MAX 255

#else

typedef ___U16 lineedit_input_decoder_state;
#define LINEEDIT_INPUT_DECODER_STATE_MAX 65535

#endif

#define LINEEDIT_INPUT_DECODER_MAX_LENGTH \
(LINEEDIT_INPUT_DECODER_STATE_MAX-LINEEDIT_EVENT_LAST-1)

typedef struct lineedit_input_test_struct
  {
    ___U8 trigger; /* when this byte is read, action is performed */
    lineedit_input_decoder_state action; /* event type or new state */
    lineedit_input_decoder_state next;   /* next test when byte != trigger */
  } lineedit_input_test;

typedef struct lineedit_input_decoder_struct
  {
    lineedit_input_test *buffer;
    int length;
    int max_length;
  } lineedit_input_decoder;

#define LINEEDIT_MODE_SCHEME 1

typedef struct lineedit_state_struct
  {
    struct tty_state_struct *ts;

    int edit_mode;

    ___BOOL paste_cancel;
    int paste_index;
    tty_char *paste_text;

    int input_rlo;           /* index of next character to read */
    extensible_string input; /* last input line that was entered */

    int output_wlo;
    extensible_string output;

    int max_history_length;
    int history_length; /* number of history lines, not counting last */
    lineedit_history *hist_last;

    struct lineedit_state_undo_struct
    {
      lineedit_history *hist;
      int edit_point;     /* position in buffer where edit ops take place */
      int mark_point;     /* position in buffer of mark */
      int screen_cursor;  /* position in screen of cursor */
      int cursor_to_point;/* offset to add to cursor to get buffer position */
      ___BOOL delayed_wrap;
      ___BOOL paren_balance_trigger;
      ___BOOL paren_balance_in_progress;
      lineedit_time paren_balance_end;
      tty_text_attrs attrs;
    } current;

    int paren_balance_duration_msecs;

    tty_text_attrs input_attrs;
    tty_text_attrs output_attrs;

    int screen_rows;
    int screen_cols;
    int screen_size;
    ___BOOL has_auto_left_margin;
    ___BOOL has_auto_right_margin;
    ___BOOL has_eat_newline_glitch;
    ___BOOL linefeed_moves_to_left_margin;
    int tab_stop;

    lineedit_input_decoder input_decoder;
    lineedit_input_decoder_state input_decoder_state;

#ifdef USE_LOCAL_CLIPBOARD
    extensible_string clipboard;
#endif

#ifdef USE_TERMCAP_OR_TERMINFO
    char *capability[LINEEDIT_CAP_LAST+1];
#endif

#ifdef USE_WINDOWS
    /***********************/
#endif

#ifdef USE_MACOS
    /***********************/
#endif
  } lineedit_state;

#define LINEEDIT_MAX_TAB_STOP 8

typedef struct lineedit_module_struct
  {
    tty_char tab_spaces[LINEEDIT_MAX_TAB_STOP];
#ifdef USE_TERMCAP_OR_TERMINFO
    lineedit_state *current_state; /* parameter of lineedit_output_putc */
    ___WORD last_err;              /* parameter of lineedit_output_putc */
#endif
  } lineedit_module;

___HIDDEN lineedit_module lem;

/*************************/

/* Forward declarations */

___HIDDEN ___WORD tty_read_noncanonical
   ___P((tty_state *state,
         tty_char *buf,
         int len),
        ());

___HIDDEN void lineedit_track_cursor_movement
   ___P((lineedit_state *state,
         tty_char *buf,
         int len),
        ());

___HIDDEN ___WORD tty_write_to_wbuf
   ___P((tty_state *state,
         tty_char *buf,
         int len),
        ());

___HIDDEN void tty_wbuf_drain
   ___P((tty_state *state),
        ());

/*---------------------------------------------------------------------------*/

/* Extensible string routines */

___HIDDEN ___WORD extensible_string_setup
   ___P((extensible_string *str,
         int n),
        (str,
         n)
extensible_string *str;
int n;)
{
#define EXTENSIBLE_STRING_INITIAL_BUFFER_SIZE 1

  if (n < EXTENSIBLE_STRING_INITIAL_BUFFER_SIZE)
    n = EXTENSIBLE_STRING_INITIAL_BUFFER_SIZE;

  str->buffer = ___alloc_mem (n * sizeof (extensible_string_char));

  if (str->buffer == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  str->length = 0;
  str->max_length = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD extensible_string_cleanup
   ___P((extensible_string *str),
        (str)
extensible_string *str;)
{
  ___free_mem (str->buffer);
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD extensible_string_copy
   ___P((extensible_string_char *src,
         int len,
         extensible_string *dst,
         int fudge),
        (src,
         len,
         dst,
         fudge)
extensible_string_char *src;
int len;
extensible_string *dst;
int fudge;)
{
  extensible_string_char *buf;

  buf = ___alloc_mem ((len+fudge) * sizeof (extensible_string_char));

  if (buf == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  dst->buffer = buf;
  dst->length = len;
  dst->max_length = len+fudge;

  while (len-- > 0)
    buf[len] = src[len];

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD extensible_string_set_length
   ___P((extensible_string *str,
         int len,
         int fudge),
        (str,
         len,
         fudge)
extensible_string *str;
int len;
int fudge;)
{
  if (len > str->max_length || 2*len+1 < str->max_length)
    {
      int i;
      int new_max_length = (fudge<0) ? 3*len/2+1 : len+fudge;
      extensible_string_char *old_buffer = str->buffer;
      extensible_string_char *new_buffer = (new_max_length>10000000)?NULL: ___alloc_mem (new_max_length * sizeof (extensible_string_char));/*********************/

      if (new_buffer == NULL)
        return ___FIX(___HEAP_OVERFLOW_ERR);

      i = str->length;
      if (i > len)
        i = len;

      while (i-- > 0)
        new_buffer[i] = old_buffer[i];

      ___free_mem (old_buffer);
      str->buffer = new_buffer;
      str->max_length = new_max_length;
    }

  str->length = len;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD extensible_string_delete
   ___P((extensible_string *str,
         int pos,
         int len),
        (str,
         pos,
         len)
extensible_string *str;
int pos;
int len;)
{
  int i;

  if (pos < 0)
    pos = 0;
  else if (pos > str->length)
    pos = str->length;

  i = str->length - pos;
  if (len < 0)
    len = 0;
  else if (len > i)
    len = i;

  for (i=pos; i<str->length-len; i++)
    str->buffer[i] = str->buffer[i+len];

  str->length -= len;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD extensible_string_insert
   ___P((extensible_string *str,
         int pos,
         int len,
         extensible_string_char *chars),
        (str,
         pos,
         len,
         chars)
extensible_string *str;
int pos;
int len;
extensible_string_char *chars;)
{
  ___WORD e;
  int i;

  if (pos < 0)
    pos = 0;
  else if (pos > str->length)
    pos = str->length;

  if (len < 0)
    len = 0;

  if ((e = extensible_string_set_length (str, str->length+len, -1))
      == ___FIX(___NO_ERR))
    {
      for (i=str->length-len-1; i>=pos; i--)
        str->buffer[i+len] = str->buffer[i];

      for (i=len-1; i>=0; i--)
        str->buffer[i+pos] = chars[i];
    }

  return e;
}

/*---------------------------------------------------------------------------*/

/* Tty line editing time routines */

___HIDDEN void lineedit_init_time
   ___P((lineedit_time *tim,
         int secs,
         int msecs),
        (tim,
         secs,
         msecs)
lineedit_time *tim;
int secs;
int msecs;)
{
  tim->secs = secs + msecs / 1000;
  tim->msecs = msecs % 1000;
}

___HIDDEN void lineedit_add_time
   ___P((lineedit_time *tim1,
         lineedit_time *tim2),
        (tim1,
         tim2)
lineedit_time *tim1;
lineedit_time *tim2;)
{
  int total_msecs = tim1->msecs + tim2->msecs;
  tim1->secs += tim2->secs + total_msecs / 1000;
  tim1->msecs = total_msecs % 1000;
}

___HIDDEN void lineedit_sub_time
   ___P((lineedit_time *tim1,
         lineedit_time *tim2),
        (tim1,
         tim2)
lineedit_time *tim1;
lineedit_time *tim2;)
{
  int total_msecs;

  if (tim2->msecs > tim1->msecs)
    {
      tim1->secs -= tim2->secs + 1;
      tim1->msecs += 1000 - tim2->msecs;
    }
  else
    {
      tim1->secs -= tim2->secs;
      tim1->msecs -= tim2->msecs;
    }
}

___HIDDEN ___BOOL lineedit_negative_time
   ___P((lineedit_time *tim),
        (tim)
lineedit_time *tim;)
{
  return (tim->secs < 0);
}

___HIDDEN void lineedit_get_time
   ___P((lineedit_time *tim),
        (tim)
lineedit_time *tim;)
{
#ifndef USE_UNIX
#ifndef USE_WINDOWS
#ifndef USE_MACOS
  lineedit_init_time (tim, 0, 0);
#endif
#endif
#endif

#ifdef USE_UNIX
  struct timeval tv;

  if (gettimeofday (&tv, NULL) == 0)
    lineedit_init_time (tim, tv.tv_sec, tv.tv_usec / 1000);
  else
    lineedit_init_time (tim, 0, 0);
#endif

#ifdef USE_WINDOWS
  struct _timeb tb;

  _ftime (&tb);

  lineedit_init_time (tim, tb.time, tb.millitm);
#endif

#ifdef USE_MACOS
  /****************/
  lineedit_init_time (tim, 0, 0);
#endif
}

#ifndef LINEEDIT_WITH_NONBLOCKING_IO

___HIDDEN void lineedit_sleep_until
   ___P((lineedit_time *tim),
        (tim)
lineedit_time *tim;)
{
  lineedit_time now;
  lineedit_time delta = *tim;

  lineedit_get_time (&now);
  lineedit_sub_time (&delta, &now);

  if (!lineedit_negative_time (&delta))
    {
      int msecs;

      if (delta.secs > 5)
        msecs = 5000;
      else
        msecs = 1000 * delta.secs + delta.msecs;

#ifdef USE_nanosleep
      {
        struct timespec ts;
        ts.tv_sec = msecs / 1000;
        ts.tv_nsec = msecs * 1000000;
        nanosleep (&ts, NULL);
      }
#endif

#ifdef USE_WINDOWS
      Sleep (msecs);
#endif

#ifdef USE_MACOS
      /***************/
#endif
    }
}

#endif

/*---------------------------------------------------------------------------*/

/* Tty line editing input decoder routines */

___HIDDEN ___WORD lineedit_input_decoder_setup
   ___P((lineedit_input_decoder *decoder),
        (decoder)
lineedit_input_decoder *decoder;)
{
#define LINEEDIT_INPUT_DECODER_INITIAL_BUFFER_SIZE 1 /******************/

  int n = LINEEDIT_INPUT_DECODER_INITIAL_BUFFER_SIZE;

  decoder->buffer = ___alloc_mem (n * sizeof (lineedit_input_test));

  if (decoder->buffer == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  decoder->length = 0;
  decoder->max_length = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_input_decoder_cleanup
   ___P((lineedit_input_decoder *decoder),
        (decoder)
lineedit_input_decoder *decoder;)
{
  ___free_mem (decoder->buffer);
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_input_decoder_set_length
   ___P((lineedit_input_decoder *decoder,
         int len,
         int fudge),
        (decoder,
         len,
         fudge)
lineedit_input_decoder *decoder;
int len;
int fudge;)
{
  if (len > LINEEDIT_INPUT_DECODER_MAX_LENGTH)
    return ___FIX(___UNKNOWN_ERR);

  if (len > decoder->max_length || 2*len+1 < decoder->max_length)
    {
      int i;
      int new_max_length = (fudge<0) ? 3*len/2+1 : len+fudge;
      lineedit_input_test *old_buffer = decoder->buffer;
      lineedit_input_test *new_buffer =
        ___alloc_mem (new_max_length * sizeof (lineedit_input_test));

      if (new_buffer == NULL)
        return ___FIX(___HEAP_OVERFLOW_ERR);

      i = decoder->length;
      if (i > len)
        i = len;

      while (i-- > 0)
        new_buffer[i] = old_buffer[i];

      ___free_mem (old_buffer);
      decoder->buffer = new_buffer;
      decoder->max_length = new_max_length;
    }

  decoder->length = len;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_input_decoder_add
   ___P((lineedit_input_decoder *decoder,
         char *seq,
         ___U8 event),
        (decoder,
         seq,
         event)
lineedit_input_decoder *decoder;
char *seq;
___U8 event;)
{
  ___WORD e;
  ___U8 b;
  int i;
  char *p = seq;

  if (event & WITH_ESC_PREFIX)
    b = UNICODE_ESCAPE;
  else
    b = (___U8)*p++;

if (0)
  {/****************************/
    char *p = seq;
    if (b < 32)
      printf ("(%d)", b);
    else
      printf ("%c", b);
    while (*p != 0)
      if (*p < 32)
        printf ("(%d)", *p++);
      else
        printf ("%c", *p++);
    printf ("\n");
    fflush (stdout);
  }

  if (decoder->length > 0)
    {
      i = 0;
      while (b != (___U8)'\0' || p == seq+1)
        {
          if (decoder->buffer[i].trigger == b)
            {
              int a = decoder->buffer[i].action;
              if (a < decoder->length)
                {
                  i = a;
                  b = (___U8)*p++;
                }
              else
                return ___FIX(___NO_ERR); /* prefix of sequence exists */
            }
          else
            {
              int n = decoder->buffer[i].next;
              if (n < decoder->length)
                i = n;
              else
                {
                  decoder->buffer[i].next = decoder->length;
                  break;
                }
            }
        }
    }

  if (b != (___U8)'\0' || p == seq+1)
    {
      /* sequence is not a prefix of a preexisting sequence */

      while (b != (___U8)'\0' || p == seq+1)
        {
          i = decoder->length;

          if ((e = lineedit_input_decoder_set_length (decoder, i+1, -1))
              != ___FIX(___NO_ERR))
            return e;

          decoder->buffer[i].trigger = b;
          decoder->buffer[i].action = i+1;
          decoder->buffer[i].next = LINEEDIT_INPUT_DECODER_STATE_MAX;

          b = (___U8)*p++;
        }
      decoder->buffer[i].action =
        LINEEDIT_INPUT_DECODER_STATE_MAX-(event&~WITH_ESC_PREFIX);
    }

  return ___FIX(___NO_ERR);
}

typedef struct lineedit_defseq_struct
  {
    char *seq;
    ___U8 event;
  } lineedit_defseq;

___HIDDEN lineedit_defseq lineedit_defseq_common[] =
{
  /* sequences that all terminals support */

   { "\012",     LINEEDIT_EVENT_RETURN                           }
  ,{ "\015",     LINEEDIT_EVENT_RETURN                           }
  ,{ "\010",     LINEEDIT_EVENT_BACK                             }
  ,{ "\010",     LINEEDIT_EVENT_BACK_WORD      | WITH_ESC_PREFIX }
  ,{ "\011",     LINEEDIT_EVENT_TAB                              }
};

___HIDDEN lineedit_defseq lineedit_defseq_map_rubout_to_backspace[] =
{
  /* sequences that map the rubout key to backspace */

   { "\177",     LINEEDIT_EVENT_BACK                             }
  ,{ "\177",     LINEEDIT_EVENT_BACK_WORD      | WITH_ESC_PREFIX }
};

___HIDDEN lineedit_defseq lineedit_defseq_emacs[] =
{
  /* sequences specific to emacs mode */

   { "\0",       LINEEDIT_EVENT_MARK                             }
  ,{ "\031",     LINEEDIT_EVENT_PASTE                            }
  ,{ "\027",     LINEEDIT_EVENT_CUT                              }
  ,{ "\013",     LINEEDIT_EVENT_CUT_RIGHT                        }
  ,{ "\025",     LINEEDIT_EVENT_CUT_LEFT                         }
  ,{ "\014",     LINEEDIT_EVENT_REFRESH                          }
  ,{ "\024",     LINEEDIT_EVENT_TRANSPOSE                        }
  ,{ "\024",     LINEEDIT_EVENT_TRANSPOSE_WORD | WITH_ESC_PREFIX }
  ,{ "t",        LINEEDIT_EVENT_TRANSPOSE_WORD | WITH_ESC_PREFIX }
  ,{ "T",        LINEEDIT_EVENT_TRANSPOSE_WORD | WITH_ESC_PREFIX }

  ,{ "\020",     LINEEDIT_EVENT_UP                               }
  ,{ "\016",     LINEEDIT_EVENT_DOWN                             }
  ,{ "\006",     LINEEDIT_EVENT_RIGHT                            }
  ,{ "\002",     LINEEDIT_EVENT_LEFT                             }
  ,{ "\006",     LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "f",        LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "F",        LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "\002",     LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }
  ,{ "b",        LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }
  ,{ "B",        LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }

  ,{ "\001",     LINEEDIT_EVENT_HOME                             }
  ,{ "\004",     LINEEDIT_EVENT_DELETE                           }
  ,{ "\005",     LINEEDIT_EVENT_END                              }
  ,{ "\004",     LINEEDIT_EVENT_DELETE_WORD    | WITH_ESC_PREFIX }
};

___HIDDEN lineedit_defseq lineedit_defseq_widespread[] =
{
  /* sequences that many types of terminals support */

   { "[A",       LINEEDIT_EVENT_UP             | WITH_ESC_PREFIX }
  ,{ "[B",       LINEEDIT_EVENT_DOWN           | WITH_ESC_PREFIX }
  ,{ "[C",       LINEEDIT_EVENT_RIGHT          | WITH_ESC_PREFIX }
  ,{ "[D",       LINEEDIT_EVENT_LEFT           | WITH_ESC_PREFIX }

  ,{ "\033[A",   LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033[B",   LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033[C",   LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "\033[D",   LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }

  ,{ "OA",       LINEEDIT_EVENT_UP             | WITH_ESC_PREFIX }
  ,{ "OB",       LINEEDIT_EVENT_DOWN           | WITH_ESC_PREFIX }
  ,{ "OC",       LINEEDIT_EVENT_RIGHT          | WITH_ESC_PREFIX }
  ,{ "OD",       LINEEDIT_EVENT_LEFT           | WITH_ESC_PREFIX }

  ,{ "\033OA",   LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033OB",   LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033OC",   LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "\033OD",   LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }

  ,{ "A",        LINEEDIT_EVENT_UP             | WITH_ESC_PREFIX }
  ,{ "B",        LINEEDIT_EVENT_DOWN           | WITH_ESC_PREFIX }
  ,{ "C",        LINEEDIT_EVENT_RIGHT          | WITH_ESC_PREFIX }
  ,{ "D",        LINEEDIT_EVENT_LEFT           | WITH_ESC_PREFIX }

  ,{ "\033A",    LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033B",    LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033C",    LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "\033D",    LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "[5A",      LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "[5B",      LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "[5C",      LINEEDIT_EVENT_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "[5D",      LINEEDIT_EVENT_LEFT_WORD      | WITH_ESC_PREFIX }
#endif

  ,{ "[1~",      LINEEDIT_EVENT_HOME           | WITH_ESC_PREFIX }
  ,{ "[2~",      LINEEDIT_EVENT_INSERT         | WITH_ESC_PREFIX }
  ,{ "[3~",      LINEEDIT_EVENT_DELETE         | WITH_ESC_PREFIX }
  ,{ "[4~",      LINEEDIT_EVENT_END            | WITH_ESC_PREFIX }
  ,{ "[5~",      LINEEDIT_EVENT_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "[6~",      LINEEDIT_EVENT_END_DOC        | WITH_ESC_PREFIX }

  ,{ "\033[1~",  LINEEDIT_EVENT_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "\033[2~",  LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033[3~",  LINEEDIT_EVENT_DELETE_WORD    | WITH_ESC_PREFIX }
  ,{ "\033[4~",  LINEEDIT_EVENT_END_DOC        | WITH_ESC_PREFIX }
  ,{ "\033[5~",  LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "\033[6~",  LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }

  ,{ "[H",       LINEEDIT_EVENT_HOME           | WITH_ESC_PREFIX }
  ,{ "[L",       LINEEDIT_EVENT_INSERT         | WITH_ESC_PREFIX }
  ,{ "[F",       LINEEDIT_EVENT_END            | WITH_ESC_PREFIX }
  ,{ "\177",     LINEEDIT_EVENT_DELETE                           }
  ,{ "\033[H",   LINEEDIT_EVENT_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "\177",     LINEEDIT_EVENT_DELETE_WORD    | WITH_ESC_PREFIX }
  ,{ "\033[F",   LINEEDIT_EVENT_END_DOC        | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "[1;5~",    LINEEDIT_EVENT_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "[2;5~",    LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "[3;5~",    LINEEDIT_EVENT_DELETE_WORD    | WITH_ESC_PREFIX }
  ,{ "[4;5~",    LINEEDIT_EVENT_END_DOC        | WITH_ESC_PREFIX }
  ,{ "[5;5~",    LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "[6;5~",    LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }

  ,{ "[5H",      LINEEDIT_EVENT_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "[5L",      LINEEDIT_EVENT_NONE           | WITH_ESC_PREFIX }
  ,{ "[5F",      LINEEDIT_EVENT_END_DOC        | WITH_ESC_PREFIX }
#endif

  ,{ "OP",       LINEEDIT_EVENT_F1             | WITH_ESC_PREFIX }
  ,{ "OQ",       LINEEDIT_EVENT_F2             | WITH_ESC_PREFIX }
  ,{ "OR",       LINEEDIT_EVENT_F3             | WITH_ESC_PREFIX }
  ,{ "OS",       LINEEDIT_EVENT_F4             | WITH_ESC_PREFIX }

  ,{ "\033OP",   LINEEDIT_EVENT_META_F1        | WITH_ESC_PREFIX }
  ,{ "\033OQ",   LINEEDIT_EVENT_META_F2        | WITH_ESC_PREFIX }
  ,{ "\033OR",   LINEEDIT_EVENT_META_F3        | WITH_ESC_PREFIX }
  ,{ "\033OS",   LINEEDIT_EVENT_META_F4        | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "O5P",      LINEEDIT_EVENT_META_F1        | WITH_ESC_PREFIX }
  ,{ "O5Q",      LINEEDIT_EVENT_META_F2        | WITH_ESC_PREFIX }
  ,{ "O5R",      LINEEDIT_EVENT_META_F3        | WITH_ESC_PREFIX }
  ,{ "O5S",      LINEEDIT_EVENT_META_F4        | WITH_ESC_PREFIX }
#endif

  ,{ "[11~",     LINEEDIT_EVENT_F1             | WITH_ESC_PREFIX }
  ,{ "[12~",     LINEEDIT_EVENT_F2             | WITH_ESC_PREFIX }
  ,{ "[13~",     LINEEDIT_EVENT_F3             | WITH_ESC_PREFIX }
  ,{ "[14~",     LINEEDIT_EVENT_F4             | WITH_ESC_PREFIX }

  ,{ "\033[11~", LINEEDIT_EVENT_META_F1        | WITH_ESC_PREFIX }
  ,{ "\033[12~", LINEEDIT_EVENT_META_F2        | WITH_ESC_PREFIX }
  ,{ "\033[13~", LINEEDIT_EVENT_META_F3        | WITH_ESC_PREFIX }
  ,{ "\033[14~", LINEEDIT_EVENT_META_F4        | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "[11;5~",   LINEEDIT_EVENT_META_F1        | WITH_ESC_PREFIX }
  ,{ "[12;5~",   LINEEDIT_EVENT_META_F2        | WITH_ESC_PREFIX }
  ,{ "[13;5~",   LINEEDIT_EVENT_META_F3        | WITH_ESC_PREFIX }
  ,{ "[14;5~",   LINEEDIT_EVENT_META_F4        | WITH_ESC_PREFIX }
#endif

#ifdef SUPPORT_F5_TO_F12

  ,{ "[15~",     LINEEDIT_EVENT_F5             | WITH_ESC_PREFIX }
  ,{ "[17~",     LINEEDIT_EVENT_F6             | WITH_ESC_PREFIX }
  ,{ "[18~",     LINEEDIT_EVENT_F7             | WITH_ESC_PREFIX }
  ,{ "[19~",     LINEEDIT_EVENT_F8             | WITH_ESC_PREFIX }
  ,{ "[20~",     LINEEDIT_EVENT_F9             | WITH_ESC_PREFIX }
  ,{ "[21~",     LINEEDIT_EVENT_F10            | WITH_ESC_PREFIX }
  ,{ "[23~",     LINEEDIT_EVENT_F11            | WITH_ESC_PREFIX }
  ,{ "[24~",     LINEEDIT_EVENT_F12            | WITH_ESC_PREFIX }

  ,{ "\033[15~", LINEEDIT_EVENT_META_F5        | WITH_ESC_PREFIX }
  ,{ "\033[17~", LINEEDIT_EVENT_META_F6        | WITH_ESC_PREFIX }
  ,{ "\033[18~", LINEEDIT_EVENT_META_F7        | WITH_ESC_PREFIX }
  ,{ "\033[19~", LINEEDIT_EVENT_META_F8        | WITH_ESC_PREFIX }
  ,{ "\033[20~", LINEEDIT_EVENT_META_F9        | WITH_ESC_PREFIX }
  ,{ "\033[21~", LINEEDIT_EVENT_META_F10       | WITH_ESC_PREFIX }
  ,{ "\033[23~", LINEEDIT_EVENT_META_F11       | WITH_ESC_PREFIX }
  ,{ "\033[24~", LINEEDIT_EVENT_META_F12       | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "[15;5~",   LINEEDIT_EVENT_META_F5        | WITH_ESC_PREFIX }
  ,{ "[17;5~",   LINEEDIT_EVENT_META_F6        | WITH_ESC_PREFIX }
  ,{ "[18;5~",   LINEEDIT_EVENT_META_F7        | WITH_ESC_PREFIX }
  ,{ "[19;5~",   LINEEDIT_EVENT_META_F8        | WITH_ESC_PREFIX }
  ,{ "[20;5~",   LINEEDIT_EVENT_META_F9        | WITH_ESC_PREFIX }
  ,{ "[21;5~",   LINEEDIT_EVENT_META_F10       | WITH_ESC_PREFIX }
  ,{ "[23;5~",   LINEEDIT_EVENT_META_F11       | WITH_ESC_PREFIX }
  ,{ "[24;5~",   LINEEDIT_EVENT_META_F12       | WITH_ESC_PREFIX }
#endif

#endif
};

___HIDDEN ___WORD lineedit_defseq_install_table
   ___P((lineedit_input_decoder *decoder,
         lineedit_defseq *table,
         int len),
        (decoder,
         table,
         len)
lineedit_input_decoder *decoder;
lineedit_defseq *table;
int len;)
{
  ___WORD e;
  int i;

  for (i=0; i<len; i++)
    if ((e = lineedit_input_decoder_add
               (decoder,
                table[i].seq,
                table[i].event))
        != ___FIX(___NO_ERR))
      return e;

  return ___FIX(___NO_ERR);
}

#ifdef USE_TERMCAP_OR_TERMINFO

typedef struct lineedit_defkey_struct
  {
    char *cap;
    ___U8 event_no_escape;
    ___U8 event_with_escape;
  } lineedit_defkey;

#ifdef USE_TERMCAP
#define DEFKEY(termcap_name,terminfo_name,event_no_escape,event_with_escape) \
{ termcap_name, event_no_escape, event_with_escape | WITH_ESC_PREFIX }
#else
#define DEFKEY(termcap_name,terminfo_name,event_no_escape,event_with_escape) \
{ terminfo_name, event_no_escape, event_with_escape | WITH_ESC_PREFIX }
#endif

typedef struct lineedit_defcap_struct
  {
    char *cap;
  } lineedit_defcap;

#ifdef USE_TERMCAP
#define DEFCAP(termcap_name,terminfo_name) { termcap_name }
#else
#define DEFCAP(termcap_name,terminfo_name) { terminfo_name }
#endif

___HIDDEN lineedit_defcap lineedit_defcap_table[] =
{
  /* order must match LINEEDIT_CAP_XXX */

   DEFCAP( "cl", "clear" ) /* erase screen and home cursor */
  ,DEFCAP( "cd", "ed"    ) /* erase to end of screen       */
  ,DEFCAP( "up", "cuu1"  ) /* move up one line             */
  ,DEFCAP( "do", "cud1"  ) /* move down one line           */
  ,DEFCAP( "UP", "cuu"   ) /* move up N lines              */
  ,DEFCAP( "DO", "cud"   ) /* move down N lines            */
  ,DEFCAP( "RI", "cuf"   ) /* move right N columns         */
  ,DEFCAP( "LE", "cub"   ) /* move left N columns          */
  ,DEFCAP( "me", "sgr0"  ) /* turn off all attributes      */
  ,DEFCAP( "md", "bold"  ) /* turn on bold attribute       */
  ,DEFCAP( "mr", "rev"   ) /* turn on reverse attribute    */
  ,DEFCAP( "AF", "setaf" ) /* ANSI set foreground color    */
  ,DEFCAP( "AB", "setab" ) /* ANSI set background color    */
};

___HIDDEN lineedit_defkey lineedit_defkey_table[] =
{
   DEFKEY( "ku", "kcuu1", LINEEDIT_EVENT_UP,       LINEEDIT_EVENT_NONE        )
  ,DEFKEY( "kd", "kcud1", LINEEDIT_EVENT_DOWN,     LINEEDIT_EVENT_NONE        )
  ,DEFKEY( "kr", "kcuf1", LINEEDIT_EVENT_RIGHT,    LINEEDIT_EVENT_RIGHT_WORD  )
  ,DEFKEY( "kl", "kcub1", LINEEDIT_EVENT_LEFT,     LINEEDIT_EVENT_LEFT_WORD   )
  ,DEFKEY( "kh", "khome", LINEEDIT_EVENT_HOME,     LINEEDIT_EVENT_HOME_DOC    )
  ,DEFKEY( "kI", "kich1", LINEEDIT_EVENT_INSERT,   LINEEDIT_EVENT_NONE        )
  ,DEFKEY( "kD", "kdch1", LINEEDIT_EVENT_DELETE,   LINEEDIT_EVENT_DELETE_WORD )
  ,DEFKEY( "@7", "kend",  LINEEDIT_EVENT_END,      LINEEDIT_EVENT_END_DOC     )
  ,DEFKEY( "kP", "kpp",   LINEEDIT_EVENT_HOME_DOC, LINEEDIT_EVENT_NONE        )
  ,DEFKEY( "kN", "knp",   LINEEDIT_EVENT_END_DOC,  LINEEDIT_EVENT_NONE        )
  ,DEFKEY( "k1", "kf1",   LINEEDIT_EVENT_F1,       LINEEDIT_EVENT_META_F1     )
  ,DEFKEY( "k2", "kf2",   LINEEDIT_EVENT_F2,       LINEEDIT_EVENT_META_F2     )
  ,DEFKEY( "k3", "kf3",   LINEEDIT_EVENT_F3,       LINEEDIT_EVENT_META_F3     )
  ,DEFKEY( "k4", "kf4",   LINEEDIT_EVENT_F4,       LINEEDIT_EVENT_META_F4     )
#ifdef SUPPORT_F5_TO_F12
  ,DEFKEY( "k5", "kf5",   LINEEDIT_EVENT_F5,       LINEEDIT_EVENT_META_F5     )
  ,DEFKEY( "k6", "kf6",   LINEEDIT_EVENT_F6,       LINEEDIT_EVENT_META_F6     )
  ,DEFKEY( "k7", "kf7",   LINEEDIT_EVENT_F7,       LINEEDIT_EVENT_META_F7     )
  ,DEFKEY( "k8", "kf8",   LINEEDIT_EVENT_F8,       LINEEDIT_EVENT_META_F8     )
  ,DEFKEY( "k9", "kf9",   LINEEDIT_EVENT_F9,       LINEEDIT_EVENT_META_F9     )
  ,DEFKEY( "k;", "kf10",  LINEEDIT_EVENT_F10,      LINEEDIT_EVENT_META_F10    )
  ,DEFKEY( "F1", "kf11",  LINEEDIT_EVENT_F11,      LINEEDIT_EVENT_META_F11    )
  ,DEFKEY( "F2", "kf12",  LINEEDIT_EVENT_F12,      LINEEDIT_EVENT_META_F12    )
#endif
};

___HIDDEN ___WORD lineedit_defkey_install
   ___P((lineedit_input_decoder *decoder,
         char *cap,
         ___U8 event_no_escape,
         ___U8 event_with_escape),
        (decoder,
         cap,
         event_no_escape,
         event_with_escape)
lineedit_input_decoder *decoder;
char *cap;
___U8 event_no_escape;
___U8 event_with_escape;)
{
  ___WORD e = ___FIX(___NO_ERR);
  char *seq;

#ifdef USE_TERMCAP
  seq = tgetstr (cap, NULL);
#else
  seq = tigetstr (cap);
#endif

  if (seq != (char*)-1 && seq != NULL)
    if ((e = lineedit_input_decoder_add (decoder, seq, event_no_escape))
        != ___FIX(___NO_ERR) &&
        event_with_escape != LINEEDIT_EVENT_NONE|WITH_ESC_PREFIX)
      e = lineedit_input_decoder_add (decoder, seq, event_with_escape);

  return e;
}

___HIDDEN ___WORD lineedit_defkey_install_table
   ___P((lineedit_input_decoder *decoder,
         lineedit_defkey *table,
         int len),
        (decoder,
         table,
         len)
lineedit_input_decoder *decoder;
lineedit_defkey *table;
int len;)
{
  ___WORD e;
  int i;

  for (i=0; i<len; i++)
    if ((e = lineedit_defkey_install
               (decoder,
                table[i].cap,
                table[i].event_no_escape,
                table[i].event_with_escape))
        != ___FIX(___NO_ERR))
      return e;

  return ___FIX(___NO_ERR);
}

#endif

#define NBELEMS(table)(sizeof (table) / sizeof (table[0]))

___HIDDEN ___WORD lineedit_input_decoder_init
   ___P((lineedit_input_decoder *decoder,
         ___BOOL map_rubout_to_backspace,
         ___BOOL emacs_bindings,
         ___BOOL termcap_or_terminfo_keys),
        (decoder,
         map_rubout_to_backspace,
         emacs_bindings,
         termcap_or_terminfo_keys)
lineedit_input_decoder *decoder;
___BOOL map_rubout_to_backspace;
___BOOL emacs_bindings;
___BOOL termcap_or_terminfo_keys;)
{
  ___WORD e;

  decoder->length = 0;

  if ((e = lineedit_defseq_install_table
             (decoder,
              lineedit_defseq_common,
              NBELEMS(lineedit_defseq_common)))
      != ___FIX(___NO_ERR))
    return e;

  if (map_rubout_to_backspace)
    if ((e = lineedit_defseq_install_table
               (decoder,
                lineedit_defseq_map_rubout_to_backspace,
                NBELEMS(lineedit_defseq_map_rubout_to_backspace)))
        != ___FIX(___NO_ERR))
      return e;

  if (emacs_bindings)
    if ((e = lineedit_defseq_install_table
               (decoder,
                lineedit_defseq_emacs,
                NBELEMS(lineedit_defseq_emacs)))
        != ___FIX(___NO_ERR))
      return e;

#ifdef USE_TERMCAP_OR_TERMINFO
  if (termcap_or_terminfo_keys)
    if ((e = lineedit_defkey_install_table
               (decoder,
                lineedit_defkey_table,
                NBELEMS(lineedit_defkey_table)))
        != ___FIX(___NO_ERR))
      return e;
#endif

  if ((e = lineedit_defseq_install_table
             (decoder,
              lineedit_defseq_widespread,
              NBELEMS(lineedit_defseq_widespread)))
      != ___FIX(___NO_ERR))
    return e;

#if 0
#if 0
  {int i;
  /***************************/
  i = 0;
  while (i<decoder->length)
    {
      if (decoder->buffer[i].trigger < UNICODE_SPACE)
      printf ("%d trigger=%d action=%d next=%d\n",
              i,
              decoder->buffer[i].trigger,
              decoder->buffer[i].action,
              decoder->buffer[i].next);
      else
      printf ("%d trigger='%c' action=%d next=%d\n",
              i,
              decoder->buffer[i].trigger,
              decoder->buffer[i].action,
              decoder->buffer[i].next);
      i++;
    }
  }
#else
  printf ("%d\n",decoder->length);
#endif
  fflush (stdout);
#endif

  return ___FIX(___NO_ERR);
}

/*---------------------------------------------------------------------------*/

/* Tty line editing routines */

___HIDDEN ___WORD lineedit_history_begin_edit
   ___P((lineedit_state *state,
         lineedit_history *h),
        (state,
         h)
lineedit_state *state;
lineedit_history *h;)
{
#define LINEEDIT_FUDGE 80

  ___WORD e;

  if (h->edited.buffer == NULL)
    {
      if ((e = extensible_string_copy
                 (h->actual.buffer,
                  h->actual.length,
                  &h->edited,
                  LINEEDIT_FUDGE))
          != ___FIX(___NO_ERR))
        return e;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN void lineedit_history_end_edit
   ___P((lineedit_state *state,
         lineedit_history *h),
        (state,
         h)
lineedit_state *state;
lineedit_history *h;)
{
  if (h->edited.buffer != NULL)
    {
      extensible_string_cleanup (&h->edited); /* ignore error */
      h->edited.buffer = NULL;
    }
}

___HIDDEN ___WORD lineedit_history_setup
   ___P((lineedit_state *state,
         lineedit_history **hist),
        (state,
         hist)
lineedit_state *state;
lineedit_history **hist;)
{
  ___WORD e;
  lineedit_history *h;

  h = ___alloc_mem (sizeof (lineedit_history));

  if (h == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  if ((e = extensible_string_setup (&h->actual, 0)) != ___FIX(___NO_ERR))
    {
      ___free_mem (h);
      return e;
    }

  h->edited.buffer = NULL;

  h->prev = h; /* create a circular list */
  h->next = h;

  *hist = h;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_history_cleanup
   ___P((lineedit_state *state,
         lineedit_history *h),
        (state,
         h)
lineedit_state *state;
lineedit_history *h;)
{
  lineedit_history_end_edit (state, h);

  extensible_string_cleanup (&h->actual); /* ignore error */

  ___free_mem (h);

  return ___FIX(___NO_ERR);
}

___HIDDEN void lineedit_history_remove
   ___P((lineedit_state *state,
         lineedit_history *item),
        (state,
         item)
lineedit_state *state;
lineedit_history *item;)
{
  lineedit_history *prev = item->prev;
  lineedit_history *next = item->next;

  if (prev == item)
    state->hist_last = NULL;
  else
    {
      next->prev = prev;
      prev->next = next;
      item->prev = item;
      item->next = item;
      if (state->hist_last == item)
        state->hist_last = prev;
    }
}

___HIDDEN void lineedit_history_trim
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  while (state->history_length > state->max_history_length)
    {
      lineedit_history *first = state->hist_last->next;
      lineedit_history_remove (state, first);
      lineedit_history_cleanup (state, first); /* ignore error */
      state->history_length--;
    }
}

___HIDDEN void lineedit_set_max_history_length
   ___P((lineedit_state *state,
         int max_history_length),
        (state,
         max_history_length)
lineedit_state *state;
int max_history_length;)
{
  if (max_history_length >= 0)
    {
      state->max_history_length = max_history_length;
      lineedit_history_trim (state);
    }
}

___HIDDEN void lineedit_history_add
   ___P((lineedit_state *state,
         lineedit_history *item),
        (state,
         item)
lineedit_state *state;
lineedit_history *item;)
{
  lineedit_history *dest = state->hist_last;

  if (dest == NULL)
    {
      item->prev = item;
      item->next = item;
    }
  else
    {
      lineedit_history *after_dest = dest->next;
      item->next = after_dest;
      item->prev = dest;
      dest->next = item;
      after_dest->prev = item;
    }

  state->hist_last = item;
  state->history_length++;
}

___HIDDEN ___WORD lineedit_output_drain
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  /*
   * Drain the output generated by the previous line editing
   * operation.
   */

  ___WORD e;
  ___WORD n;
  int wlo;
  int length;
  tty_state *ts = state->ts;

  if ((e = ts->werr) != ___FIX(___NO_ERR))
    {
      ts->werr = ___FIX(___NO_ERR);
      return e;
    }

  tty_wbuf_drain (ts); /* error code returned in ts->werr */

  wlo = state->output_wlo;
  length = state->output.length;

  while (wlo < length)
    {
      if ((e = ts->werr) != ___FIX(___NO_ERR))
        {
          ts->werr = ___FIX(___NO_ERR);
          state->output_wlo = wlo;
          return e;
        }

      n = tty_write_to_wbuf (ts, &state->output.buffer[wlo], length - wlo);

      if (n <= ___FIX(0))
        {
          if (n == ___FIX(0))
            n = ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));
#ifdef LINEEDIT_REPORT_CHAR_ENCODING_ERRORS
          else if (n == ___FIX(___STOC_LATIN1STRING_ERR))
            wlo++;
#endif
          state->output_wlo = wlo;
          return n;
        }

      wlo += ___INT(n);

      tty_wbuf_drain (ts); /* error code returned in ts->werr */
    }

  if (extensible_string_set_length (&state->output, 0, 1) == ___FIX(___NO_ERR))
    wlo = 0;

  state->output_wlo = wlo;

  return ___FIX(___NO_ERR);
}

___HIDDEN void lineedit_track_cursor_movement
   ___P((lineedit_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
lineedit_state *state;
tty_char *buf;
int len;)
{
  /*
   * Even with the termcap/terminfo database it is impossible to tell
   * exactly how the cursor will move when an arbitrary sequence of
   * characters is sent to the terminal.  So we make general
   * assumptions about the terminal behavior with a bias towards
   * "xterm", which is probably the terminal that will most often be
   * used under UNIX.
   */

  /*
   * The fields state->current.screen_cursor,
   * state->current.cursor_to_point, and state->current.delayed_wrap
   * are updated accordingly.
   */

  int screen_cols = state->screen_cols;
  int screen_cursor = state->current.screen_cursor;
  int cursor_to_point = state->current.cursor_to_point;
  ___BOOL delayed_wrap = state->current.delayed_wrap;
  tty_char c;
  int cursor;
  int col;

  while (len-- > 0)
    {
      c = *buf++;

      switch (c)
        {
        case UNICODE_BELL:
          break;

        case UNICODE_BACKSPACE:
          if (screen_cursor % screen_cols > 0 ||
              (screen_cursor > 0 && state->has_auto_left_margin))
            screen_cursor--;
          delayed_wrap = 0;
          break;

        case UNICODE_TAB:
          col = ((screen_cursor % screen_cols) / state->tab_stop + 1)
                * state->tab_stop;
          if (col >= screen_cols)
            {
              if (state->has_auto_right_margin &&
                  !state->has_eat_newline_glitch)
                {
                  screen_cursor = (screen_cursor / screen_cols + 1)
                                  * screen_cols;
                  if (screen_cursor >= state->screen_size)
                    {
                      screen_cursor -= screen_cols;
                      cursor_to_point += screen_cols;
                    }
                }
              else
                {
                  col = screen_cols-1; /* don't wrap */
                  screen_cursor = (screen_cursor / screen_cols) * screen_cols
                                  + col;
                }
            }
          else
            screen_cursor = (screen_cursor / screen_cols) * screen_cols + col;
          break;

        case UNICODE_LINEFEED:
          if (state->linefeed_moves_to_left_margin)
            screen_cursor = (screen_cursor / screen_cols + 1) * screen_cols;
          else
            screen_cursor = screen_cursor + screen_cols;
          if (screen_cursor >= state->screen_size)
            {
              screen_cursor -= screen_cols;
              cursor_to_point += screen_cols;
            }
          delayed_wrap = 0;
          break;

        case UNICODE_RETURN:
          screen_cursor = (screen_cursor / screen_cols) * screen_cols;
          delayed_wrap = 0;
          break;

        default:
          if (delayed_wrap)
            {
              screen_cursor++;
              if (screen_cursor >= state->screen_size)
                {
                  screen_cursor -= screen_cols;
                  cursor_to_point += screen_cols;
                }
            }

          screen_cursor++;

          if (screen_cursor % screen_cols != 0)
            delayed_wrap = 0;
          else if (state->has_auto_right_margin)
            {
              if (state->has_eat_newline_glitch)
                {
                  screen_cursor--;
                  delayed_wrap = 1; /* wrap later */
                }
              else
                {
                  if (screen_cursor >= state->screen_size)
                    {
                      screen_cursor -= screen_cols;
                      cursor_to_point += screen_cols;
                    }
                  delayed_wrap = 0;
                }
            }
          else
            {
              screen_cursor--;
              delayed_wrap = 0;
            }
        }
    }

  state->current.screen_cursor = screen_cursor;
  state->current.cursor_to_point = cursor_to_point;
  state->current.delayed_wrap = delayed_wrap;
}

___HIDDEN ___WORD lineedit_output_string
   ___P((lineedit_state *state,
         char *str,
         int rep),
        (state,
         str,
         rep)
lineedit_state *state;
char *str;
int rep;)
{
  /*
   * Output the C string "str" multiple times ("rep" is the repeat
   * factor).
   */

  while (rep-- > 0)
    {
      char *p = str;
      while (*p != '\0')
        {
          ___WORD e;
          tty_char c = *p++;
          if ((e = extensible_string_insert
                     (&state->output,
                      state->output.length,
                      1,
                      &c))
              != ___FIX(___NO_ERR))
            return e;
#if 0
printf ("%c", c); fflush(stdout);/******************/
#endif
        }
    }

  return ___FIX(___NO_ERR);
}

#ifdef USE_TERMCAP_OR_TERMINFO

___HIDDEN int lineedit_output_putc
   ___P((int c),
        (c)
int c;)
{
  /*
   * This routine outputs a C character.  It is needed for the call to
   * the termcap/terminfo "tputs" routine.  The parameters it uses are
   * stored in the lineedit module structure "lem".
   */

  if (lem.last_err == ___FIX(___NO_ERR))
    {
      tty_char ch = c;
      lem.last_err = extensible_string_insert
                       (&lem.current_state->output,
                        lem.current_state->output.length,
                        1,
                        &ch);
    }

  return c;
}

___HIDDEN ___WORD lineedit_output_cap
   ___P((lineedit_state *state,
         int cap,
         int arg,
         int rep),
        (state,
         cap,
         arg,
         rep)
lineedit_state *state;
int cap;
int arg;
int rep;)
{
  /*
   * Output the sequence for the terminal capability "cap" with
   * integer parameter "arg" (-1 means no parameter).  "rep" is the
   * repeat factor.
   */

  char *str = state->capability[cap];

  if (arg >= 0)
    {
#ifdef USE_TERMCAP
      str = tgoto (str, arg, arg);
#else
      str = tparm (str, arg);
#endif
    }

  if (cap < LINEEDIT_CAP_SGR0)
    state->current.delayed_wrap = 0; /* cursor commands cancel delayed wrap */

  lem.current_state = state;
  lem.last_err = ___FIX(___NO_ERR);

  if (tputs (str, 1, lineedit_output_putc) == ERR &&
      lem.last_err == ___FIX(___NO_ERR))
    lem.last_err = ___FIX(___UNKNOWN_ERR);

  return lem.last_err;
}

#endif

___HIDDEN ___WORD lineedit_output_set_text_attrs
   ___P((lineedit_state *state,
         tty_text_attrs attrs),
        (state,
         attrs)
lineedit_state *state;
tty_text_attrs attrs;)
{
  /*
   * Output the control sequence to set the text attributes of the
   * next characters sent to the terminal.
   */

#ifdef USE_TERMCAP_OR_TERMINFO

  ___WORD e;
  int turn_on;
  tty_text_attrs current_attrs = state->current.attrs;

  if (current_attrs == attrs)
    return ___FIX(___NO_ERR);

  turn_on = GET_STYLE(attrs);

  if ((GET_STYLE(current_attrs) & ~turn_on) != 0 ||
      (GET_FOREGROUND_COLOR(attrs) >= DEFAULT_COLOR &&
       GET_FOREGROUND_COLOR(current_attrs) < DEFAULT_COLOR) ||
      (GET_BACKGROUND_COLOR(attrs) >= DEFAULT_COLOR &&
       GET_BACKGROUND_COLOR(current_attrs) < DEFAULT_COLOR))
    {
      if (state->capability[LINEEDIT_CAP_SGR0] != NULL)
        if ((e = lineedit_output_cap (state, LINEEDIT_CAP_SGR0, -1, 1))
            != ___FIX(___NO_ERR))
          return e;
      current_attrs = MAKE_TEXT_ATTRS(0,DEFAULT_COLOR,DEFAULT_COLOR);
    }
  else
    turn_on = (~GET_STYLE(current_attrs) & turn_on);

  if ((turn_on & BOLD_STYLE) &&
      state->capability[LINEEDIT_CAP_BOLD] != NULL)
    if ((e = lineedit_output_cap (state, LINEEDIT_CAP_BOLD, -1, 1))
        != ___FIX(___NO_ERR))
      return e;

  if ((turn_on & REVERSE_STYLE) &&
      state->capability[LINEEDIT_CAP_REV] != NULL)
    if ((e = lineedit_output_cap (state, LINEEDIT_CAP_REV, -1, 1))
        != ___FIX(___NO_ERR))
      return e;

  if (GET_FOREGROUND_COLOR(attrs) < DEFAULT_COLOR &&
      GET_FOREGROUND_COLOR(attrs) != GET_FOREGROUND_COLOR(current_attrs) &&
      state->capability[LINEEDIT_CAP_SETAF] != NULL)
    if ((e = lineedit_output_cap
               (state,
                LINEEDIT_CAP_SETAF,
                GET_FOREGROUND_COLOR(attrs),
                1))
        != ___FIX(___NO_ERR))
      return e;

  if (GET_BACKGROUND_COLOR(attrs) < DEFAULT_COLOR &&
      GET_BACKGROUND_COLOR(attrs) != GET_BACKGROUND_COLOR(current_attrs) &&
      state->capability[LINEEDIT_CAP_SETAB] != NULL)
    if ((e = lineedit_output_cap
               (state,
                LINEEDIT_CAP_SETAB,
                GET_BACKGROUND_COLOR(attrs),
                1))
        != ___FIX(___NO_ERR))
      return e;

  state->current.attrs = attrs;

#endif

#ifdef USE_WINDOWS

  {
    /*********************/

    WORD attributes = 0;

    if (GET_STYLE(attrs) & BOLD_STYLE)
      attributes |= FOREGROUND_INTENSITY;

    if (GET_FOREGROUND_COLOR(attrs) == DEFAULT_COLOR)
      attributes |= (FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE);
    else
      {
        if (GET_FOREGROUND_COLOR(attrs) & 1)
          attributes |= FOREGROUND_RED;
        if (GET_FOREGROUND_COLOR(attrs) & 2)
          attributes |= FOREGROUND_GREEN;
        if (GET_FOREGROUND_COLOR(attrs) & 4)
          attributes |= FOREGROUND_BLUE;
      }

    if (GET_BACKGROUND_COLOR(attrs) == DEFAULT_COLOR)
      attributes |= 0;
    else
      {
        if (GET_BACKGROUND_COLOR(attrs) & 1)
          attributes |= BACKGROUND_RED;
        if (GET_BACKGROUND_COLOR(attrs) & 2)
          attributes |= BACKGROUND_GREEN;
        if (GET_BACKGROUND_COLOR(attrs) & 4)
          attributes |= BACKGROUND_BLUE;
      }

    lineedit_output_drain (state); /* ignore error */

    if (!SetConsoleTextAttribute
           (state->ts->output,
            attributes))
      return ___FIX(err_code_from_GetLastError ());
  }
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_output_chars
   ___P((lineedit_state *state,
         tty_char *buf,
         int len,
         tty_text_attrs attrs),
        (state,
         buf,
         len,
         attrs)
lineedit_state *state;
tty_char *buf;
int len;
tty_text_attrs attrs;)
{
  /*
   * Output "len" characters from the buffer "buf" and keep track of
   * cursor motion.
   */

  ___WORD e;

  if ((e = lineedit_output_set_text_attrs (state, attrs))
      != ___FIX(___NO_ERR))
    return e;

  e = extensible_string_insert
        (&state->output,
         state->output.length,
         len,
         buf);
  if (0) {int i; for (i=0;i<len;i++) printf ("%c", buf[i]); fflush(stdout);/******************/}

  lineedit_track_cursor_movement (state, buf, len);

  return e;
}

___HIDDEN ___WORD lineedit_output_char_repetition
   ___P((lineedit_state *state,
         tty_char c,
         int rep,
         tty_text_attrs attrs),
        (state,
         c,
         rep,
         attrs)
lineedit_state *state;
tty_char c;
int rep;
tty_text_attrs attrs;)
{
  /*
   * Output a character "rep" times and keep track of cursor motion.
   */

#define CHAR_BUFFER_SIZE (80*50)

  ___WORD e;
  tty_char char_buffer[CHAR_BUFFER_SIZE];
  int n;

  n = rep;
  if (n > CHAR_BUFFER_SIZE)
    n = CHAR_BUFFER_SIZE;

  while (n > 0)
    char_buffer[--n] = c;

  while (rep > 0)
    {
      n = rep;
      if (n > CHAR_BUFFER_SIZE)
        n = CHAR_BUFFER_SIZE;
      if ((e = lineedit_output_chars (state, char_buffer, n, attrs))
          != ___FIX(___NO_ERR))
        return e;
      rep -= n;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN tty_text_attrs lineedit_erase_attrs
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  tty_text_attrs output_attrs = state->output_attrs;
  int output_style = GET_STYLE(output_attrs);
  int current_style = GET_STYLE(state->current.attrs);

  return MAKE_TEXT_ATTRS((current_style & BOLD_STYLE) |
                         (output_style & REVERSE_STYLE),
                         GET_FOREGROUND_COLOR(output_attrs),
                         GET_BACKGROUND_COLOR(output_attrs));
}

___HIDDEN ___WORD lineedit_output_spaces
   ___P((lineedit_state *state,
         int rep),
        (state,
         rep)
lineedit_state *state;
int rep;)
{
  return lineedit_output_char_repetition
           (state,
            UNICODE_SPACE,
            rep,
            lineedit_erase_attrs (state));
}

___HIDDEN ___WORD lineedit_erase_to_end_of_screen
   ___P((lineedit_state *state,
         int pos,
         int chars_to_erase),
        (state,
         pos,
         chars_to_erase)
lineedit_state *state;
int pos;
int chars_to_erase;)
{
  /*
   * Erases from the current position up to the end of the screen.
   * "pos" is the start position and "chars_to_erase" is the number of
   * characters that are erased.
   */

/***************************strange interface**/

  ___WORD e;

#ifdef USE_TERMCAP_OR_TERMINFO
  if (chars_to_erase > 3 &&
      state->capability[LINEEDIT_CAP_ED] != NULL)
    {
      if ((e = lineedit_output_set_text_attrs
                 (state,
                  lineedit_erase_attrs (state)))
          == ___FIX(___NO_ERR))
        e = lineedit_output_cap (state, LINEEDIT_CAP_ED, -1, 1);
    }
  else
#endif
    e = lineedit_output_spaces (state, chars_to_erase);

  return e;
}

___HIDDEN ___WORD lineedit_output_chars_at
   ___P((lineedit_state *state,
         tty_char *str,
         int chars,
         int chars_to_erase,
         int pos),
        (state,
         str,
         chars,
         chars_to_erase,
         pos)
lineedit_state *state;
tty_char *str;
int chars;
int chars_to_erase;
int pos;)
{
  ___WORD e;

  /******************not used!

  if ((e = lineedit_prepare_to_write_at (state, pos))
      != ___FIX(___NO_ERR))
    return e;

  if ((e = lineedit_output_chars (state, str, chars, state->input_attrs))
       != ___FIX(___NO_ERR))
    return e;
  */

  return lineedit_erase_to_end_of_screen (state, pos+chars, chars_to_erase);
}

___HIDDEN ___WORD lineedit_output_current_hist
   ___P((lineedit_state *state,
         int start,
         int len),
        (state,
         start,
         len)
lineedit_state *state;
int start;
int len;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  int to_end;
  int n;

  if (start < 0)
    {
      n = -start;
      if (n > len)
        n = len;
      if ((e = lineedit_output_spaces (state, n)) != ___FIX(___NO_ERR))
        return e;
      start += n;
      len -= n;
    }

  to_end = edited->length - start;

  if (to_end > 0)
    {
      if (to_end > len)
        n = len;
      else
        n = to_end;

      if ((e = lineedit_output_chars
                 (state,
                  &edited->buffer[start],
                  n,
                  state->input_attrs))
          != ___FIX(___NO_ERR))
        return e;

      len -= n;

      if (n == to_end &&
          state->current.delayed_wrap &&
          state->current.screen_cursor + 1 < state->screen_size &&
          len < 1)
        len = 1;
    }

  return lineedit_erase_to_end_of_screen (state, 999999, len);/*******/
}

___HIDDEN ___WORD lineedit_output_force_delayed_wrap
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  /*
   * Forces the cursor to wrap to the next line if the wrap was
   * delayed.
   */

  if (state->current.delayed_wrap)
    return lineedit_output_current_hist
             (state,
              state->current.screen_cursor+1+state->current.cursor_to_point,
              1);

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_move_screen_cursor
   ___P((lineedit_state *state,
         int pos),
        (state,
         pos)
lineedit_state *state;
int pos;)
{
  /*
   * This routine sends the appropriate commands to the terminal to
   * move the cursor to position "pos".
   */

  /*
   * The fields state->current.screen_cursor and
   * state->current.delayed_wrap are updated accordingly.
   */

  ___WORD e;
  int screen_cursor = state->current.screen_cursor;
  int screen_cols;
  int pos_col;
  int col_dist;
  int row_dist;
  int dist;

  if (pos == screen_cursor)
    return ___FIX(___NO_ERR);

  screen_cols = state->screen_cols;
  pos_col = pos % screen_cols;
  col_dist = pos_col - screen_cursor % screen_cols;
  row_dist = pos / screen_cols - screen_cursor / screen_cols;

#ifdef USE_TERMCAP_OR_TERMINFO

  while (1)
    {
      /*
       * Check if the cursor can be moved backward by writing no more
       * than 4 backspace characters (if on the same row or column) or
       * 8 backspace characters (otherwise).  This is probably no more
       * than is required by cursor movement commands.
       */

      if (pos < screen_cursor &&
          (row_dist == 0 || state->has_auto_left_margin) &&
          screen_cursor - pos <= ((row_dist == 0 || col_dist == 0) ? 4 : 8))
        break;

      /*
       * Check if the cursor can be moved forward by writing no more
       * than 4 plain characters (if on the same row or column) or 8
       * plain characters (otherwise).  This is probably no more than
       * is required by cursor movement commands.
       */

      if (pos > screen_cursor &&
          screen_cursor + state->current.cursor_to_point >= 0 &&
          (row_dist == 0 || state->has_auto_right_margin) &&
          pos - (screen_cursor + state->current.delayed_wrap)
              + ((pos_col == 0 && state->has_eat_newline_glitch) ? 2 : 0) <=
              ((row_dist == 0 || col_dist == 0) ? 4 : 8))
        break;

      /*
       * If the cursor is on the target row, move the cursor to the
       * target column using cursor movement commands, if the terminal
       * supports this.
       */

      if (row_dist == 0)
        {
          if (col_dist == 0 ||
              ((col_dist > 0)
               ? (state->capability[LINEEDIT_CAP_CUF] == NULL)
               : (state->capability[LINEEDIT_CAP_CUB] == NULL)))
            break;

          if ((e = lineedit_output_cap
                     (state,
                      (col_dist > 0) ? LINEEDIT_CAP_CUF : LINEEDIT_CAP_CUB,
                      (col_dist > 0) ? col_dist : -col_dist,
                      1))
              != ___FIX(___NO_ERR))
            return e;

          screen_cursor += col_dist;
          state->current.screen_cursor = screen_cursor;
          break;
        }

      /*
       * Move the cursor to the target row using cursor movement
       * commands, if the terminal supports this.
       */

      if ((row_dist > 0)
          ? (state->capability[LINEEDIT_CAP_CUD] == NULL &&
             state->capability[LINEEDIT_CAP_CUD1] == NULL)
          : (state->capability[LINEEDIT_CAP_CUU] == NULL &&
             state->capability[LINEEDIT_CAP_CUU1] == NULL))
        break;

      if ((row_dist > 0)
          ? (state->capability[LINEEDIT_CAP_CUD] == NULL ||
             state->capability[LINEEDIT_CAP_CUD1] != NULL && row_dist == 1)
          : (state->capability[LINEEDIT_CAP_CUU] == NULL ||
             state->capability[LINEEDIT_CAP_CUU1] != NULL && row_dist == -1))
        e = lineedit_output_cap
              (state,
               (row_dist > 0) ? LINEEDIT_CAP_CUD1 : LINEEDIT_CAP_CUU1,
               -1,
               (row_dist > 0) ? row_dist : -row_dist);
      else
        e = lineedit_output_cap
              (state,
               (row_dist > 0) ? LINEEDIT_CAP_CUD : LINEEDIT_CAP_CUU,
               (row_dist > 0) ? row_dist : -row_dist,
               1);

      if (e != ___FIX(___NO_ERR))
        return e;

      /* Cursor commands to change row were successful. */

      screen_cursor += row_dist * screen_cols;
      state->current.screen_cursor = screen_cursor;
      row_dist = 0;
    }

#endif

  /*
   * If the cursor is still not at the target position, move the
   * cursor by writing plain characters or backspace characters.  We
   * assume that the terminal has auto left/right margins if the
   * cursor needs to change row.
   */

  dist = pos - screen_cursor;

  if (dist != 0)
    {
      if (dist < 0)
        {
          if ((e = lineedit_output_char_repetition
                     (state,
                      UNICODE_BACKSPACE,
                      -dist,
                      state->current.attrs))
              != ___FIX(___NO_ERR))
            return e;
        }
      else
        {
          int start = screen_cursor + state->current.cursor_to_point +
                      state->current.delayed_wrap;
          int len = dist - state->current.delayed_wrap;
          ___BOOL need_backspace =
            (pos_col == 0 && state->has_eat_newline_glitch);

          if ((e = lineedit_output_current_hist
                     (state,
                      start,
                      len+need_backspace))
              != ___FIX(___NO_ERR))
            return e;

          if (need_backspace)
            {
              if ((e = lineedit_output_char_repetition
                         (state,
                          UNICODE_BACKSPACE,
                          1,
                          state->current.attrs))
                  != ___FIX(___NO_ERR))
                return e;
            }
        }

      screen_cursor += dist;
      state->current.screen_cursor = screen_cursor;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_left_margin_of_next_row
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  ___WORD e;

  if (!state->linefeed_moves_to_left_margin)
    if ((e = lineedit_output_char_repetition
               (state,
                UNICODE_RETURN,
                1,
                state->output_attrs))
        != ___FIX(___NO_ERR))
      return e;

  return lineedit_output_char_repetition
           (state,
            UNICODE_LINEFEED,
            1,
            state->output_attrs);
}

___HIDDEN ___WORD lineedit_prepare_to_write_at
   ___P((lineedit_state *state,
         int pos),
        (state,
         pos)
lineedit_state *state;
int pos;)
{
  /*
   * This routine sends commands to the terminal such that the next
   * character sent to the terminal will show up at position "pos".
   * The position can be equal to state->screen_size, in which case
   * either the screen is scrolled right away or the next character
   * sent to the terminal will cause the screen to scroll.
   */

  ___WORD e;

  if (state->current.screen_cursor + state->current.delayed_wrap == pos)
    return ___FIX(___NO_ERR);

  if (pos >= state->screen_size)
    {
      if ((e = lineedit_move_screen_cursor (state, state->screen_size-1))
          != ___FIX(___NO_ERR))
        return e;
    }
  else
    {
      if ((e = lineedit_move_screen_cursor (state, pos))
          != ___FIX(___NO_ERR))
        return e;

      if (!state->current.delayed_wrap)
        return ___FIX(___NO_ERR);

      /*
       * Output a backspace and a character.  Note that this assumes
       * the screen is at least 2 columns wide.
       */

      if ((e = lineedit_output_char_repetition
                 (state,
                  UNICODE_BACKSPACE,
                  1,
                  state->current.attrs))
          != ___FIX(___NO_ERR))
        return e;
    }

  return lineedit_output_current_hist
           (state,
            state->current.screen_cursor + state->current.cursor_to_point,
            1);
}

___HIDDEN ___WORD lineedit_newline
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  int pos = edited->length - state->current.cursor_to_point;

  if (pos < 0)
    pos = 0;
  else if (pos >= state->screen_size)
    pos = state->screen_size-1;

  if ((e = lineedit_prepare_to_write_at (state, pos))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_left_margin_of_next_row (state);
}

___HIDDEN ___WORD lineedit_erase_screen
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  ___WORD e;

#ifdef USE_TERMCAP_OR_TERMINFO
  if (state->capability[LINEEDIT_CAP_CLEAR] != NULL)
    {
      if ((e = lineedit_output_set_text_attrs
                 (state,
                  lineedit_erase_attrs (state)))
          == ___FIX(___NO_ERR))
        e = lineedit_output_cap (state, LINEEDIT_CAP_CLEAR, -1, 1);
    }
  else
#endif
    e = lineedit_newline (state);

  if (e == ___FIX(___NO_ERR))
    {
      state->current.screen_cursor = 0; /****************/
    }

  return e;
}

___HIDDEN ___WORD lineedit_copy_to_clipboard
   ___P((lineedit_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
lineedit_state *state;
tty_char *buf;
int len;)
{
  /*
   * Copies "len" characters from the buffer "buf" to the clipboard.
   */

#ifdef USE_LOCAL_CLIPBOARD

  ___WORD e;
  extensible_string str;

  if ((e = extensible_string_copy (buf, len, &str, 0)) != ___FIX(___NO_ERR))
    return e;

  extensible_string_cleanup (&state->clipboard); /* ignore error */

  state->clipboard = str;

#else

  /***********************/

#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_paste_from_clipboard
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  /*
   * Paste the content of the clipboard to the input stream (the next
   * characters read from the terminal will in fact come from the
   * pasted text).
   */

#ifdef USE_LOCAL_CLIPBOARD

  int len = state->clipboard.length;
  tty_char *str;

  str = ___alloc_mem ((len+1) * sizeof (tty_char));

  if (str == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  str[len] = 0;
  while (len-- > 0)
    str[len] = state->clipboard.buffer[len];

  if (state->paste_text != NULL)
    ___free_mem (state->paste_text);

  state->paste_index = 0;
  state->paste_text = str;

#else

  /*************************/

#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_get_event
   ___P((lineedit_state *state,
         lineedit_event *ev),
        (state,
         ev)
lineedit_state *state;
lineedit_event *ev;)
{
  /*
   * This routine obtains the next line editing event from the
   * terminal.  An event can be a key press or some editing operation
   * (like "delete next word").
   */

  lineedit_input_decoder_state s = state->input_decoder_state;
  tty_char c;
  ___BOOL first_char = (s == 0);
  ___WORD n;

  next_char:

  n = tty_read_noncanonical (state->ts, &c, 1);

  if (n == ___FIX(1))
    {
      while (s < state->input_decoder.length)
        {
          if (state->input_decoder.buffer[s].trigger == c)
            {
              int a = state->input_decoder.buffer[s].action;
              if (a < LINEEDIT_INPUT_DECODER_MAX_LENGTH)
                {
                  s = a;
                  first_char = 0;
                  goto next_char;
                }
              state->input_decoder_state = 0;
              ev->event_kind = LINEEDIT_INPUT_DECODER_STATE_MAX-a;
              return ___FIX(___NO_ERR);
            }
          else
            s = state->input_decoder.buffer[s].next;
        }
      if (first_char)
        {
          ev->event_kind = LINEEDIT_EVENT_KEY;
          ev->event_char = c;
          return ___FIX(___NO_ERR);
        }
      s = 0; /* ignore the sequence including last character read */
    }
  else if (n != ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)))
    return n;

  state->input_decoder_state = s;
  ev->event_kind = LINEEDIT_EVENT_NONE;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_set_terminal_type
   ___P((lineedit_state *state,
         char *terminal_type,
         ___BOOL emacs_bindings),
        (state,
         terminal_type,
         emacs_bindings)
lineedit_state *state;
char *terminal_type;
___BOOL emacs_bindings;)
{
  ___WORD e;
  int screen_rows = 24;
  int screen_cols = 80;
  ___BOOL has_auto_left_margin = 1;
  ___BOOL has_auto_right_margin = 1;
  ___BOOL has_eat_newline_glitch = 0;

#ifdef USE_TERMCAP_OR_TERMINFO

  int i;

  if (terminal_type != NULL)
    {
#ifdef USE_TERMCAP

      char termcap_buffer[4080]; /* the value 4080 is used by GNU's readline */

      if (tgetent (termcap_buffer, terminal_type) != 1)
        return ___FIX(___UNKNOWN_ERR);

      screen_rows = tgetnum ("li");
      screen_cols = tgetnum ("co");
      has_auto_left_margin = (tgetflag ("bw") != 0);
      has_auto_right_margin = (tgetflag ("am") != 0);
      has_eat_newline_glitch = (tgetflag ("xn") != 0);

#else

      int errret;

      if (setupterm (terminal_type, 1, &errret) == ERR)
        return ___FIX(___UNKNOWN_ERR);

      screen_rows = tigetnum ("lines");
      screen_cols = tigetnum ("cols");
      has_auto_left_margin = (tigetflag ("bw") != 0);
      has_auto_right_margin = (tigetflag ("am") != 0);
      has_eat_newline_glitch = (tigetflag ("xenl") != 0);

#endif
    }

  /*
   * Setup the capability table.
   */

  for (i=0; i<LINEEDIT_CAP_LAST+1; i++)
    {
      char *seq = state->capability[i];

      if (seq != NULL)
        {
          ___free_mem (seq);
          state->capability[i] = NULL;
        }
    }

  if (terminal_type != NULL)
    for (i=0; i<LINEEDIT_CAP_LAST+1; i++)
      {
        char *seq;

#ifdef USE_TERMCAP
        seq = tgetstr (lineedit_defcap_table[i].cap, NULL);
#else
        seq = tigetstr (lineedit_defcap_table[i].cap);
#endif

        if (seq != (char*)-1 && seq != NULL)
          {
            char *p;

            /*
             * Reject any sequence that includes a linefeed if the
             * terminal driver automatically adds a carriage return.
             */

            if (state->linefeed_moves_to_left_margin)
              {
                p = seq;
                while (*p != '\0')
                  if (*p != UNICODE_LINEFEED)
                    p++;
                  else
                    break;
                if (*p != '\0')
                  seq = NULL;
              }

            /*
             * Keep a copy of the sequence.
             */

            if (seq != NULL)
              {
                int len = 0;
                while (seq[len] != '\0')
                  len++;
                p = ___alloc_mem (len);
                if (p != NULL)
                  {
                    p[len] = '\0';
                    while (len-- > 0)
                      p[len] = seq[len];
                  }
                seq = p;
              }
          }
        else
          seq = NULL;

        state->capability[i] = seq;
      }

#endif

  /*
   * Initialize input decoder so that it recognizes special keys.  Map
   * rubout key to backspace if terminal driver is setup to erase when
   * a rubout is typed.
   */

  state->input_decoder_state = 0;

  if ((e = lineedit_input_decoder_init
             (&state->input_decoder,
#ifdef USE_UNIX
              (state->ts->initial_termios.c_cc[VERASE] == UNICODE_RUBOUT),
#else
              0,
#endif
              emacs_bindings,
              terminal_type != NULL))
      != ___FIX(___NO_ERR))
    return e;

  if (screen_rows <= 0)
    screen_rows = 24;
  if (screen_cols <= 0)
    screen_cols = 80;

  state->screen_rows = screen_rows;
  state->screen_cols = screen_cols;
  state->screen_size = screen_rows * screen_cols;
  state->has_auto_left_margin = has_auto_left_margin;
  state->has_auto_right_margin = has_auto_right_margin;
  state->has_eat_newline_glitch = has_eat_newline_glitch;

#if 0
  printf ("%s screen_rows=%d screen_cols=%d alm=%d arm=%d eng=%d\n", terminal_type,screen_rows, screen_cols,
          has_auto_left_margin,
          has_auto_right_margin,
          has_eat_newline_glitch);fflush (stdout);
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_set_paren_balance_duration
   ___P((lineedit_state *state,
         int msecs),
        (state,
         msecs)
lineedit_state *state;
int msecs;)
{
  state->paren_balance_duration_msecs = msecs;
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_setup
   ___P((lineedit_state **les,
         tty_state *ts),
        (les,
         ts)
lineedit_state **les;
tty_state *ts;)
{
  ___WORD e;
  lineedit_history *h;
  lineedit_state *state;

  *les = NULL;

  /* allocate a new lineedit_state */

  state = ___alloc_mem (sizeof (lineedit_state));

  if (state == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  *les = state;

  state->ts = ts;

  state->edit_mode = LINEEDIT_MODE_SCHEME;

  state->paste_cancel = 0;
  state->paste_index = 0;
  state->paste_text = NULL;

  state->input_rlo = 0;
  state->input.buffer = NULL;

  state->output_wlo = 0;

  state->max_history_length = 1000000; /* set a very high limit */
  state->history_length = -1;
  state->hist_last = NULL;

  state->current.edit_point = 0;
  state->current.mark_point = 0;
  state->current.screen_cursor = 0;
  state->current.cursor_to_point = 0;
  state->current.delayed_wrap = 0;
  state->current.paren_balance_in_progress = 0;
  state->current.attrs = DEFAULT_TEXT_ATTRS;

  state->paren_balance_duration_msecs = DEFAULT_PAREN_BALANCE_DURATION_MSECS;

  state->input_attrs = DEFAULT_TEXT_ATTRS;
  state->output_attrs = DEFAULT_TEXT_ATTRS;

  state->screen_rows = 24;
  state->screen_cols = 80;
  state->screen_size = 24*80;

  state->has_auto_left_margin = 1;
  state->has_auto_right_margin = 1;
  state->has_eat_newline_glitch = 0;
  state->linefeed_moves_to_left_margin = 1;

  state->tab_stop = TTY_TAB_STOP;

  if ((e = extensible_string_setup (&state->output, 0)) == ___FIX(___NO_ERR))
    {
      if ((e = lineedit_history_setup (state, &h)) == ___FIX(___NO_ERR))
        {
          lineedit_history_add (state, h);
          state->current.hist = h;
          if ((e = lineedit_history_begin_edit (state, h))
              == ___FIX(___NO_ERR))
            {
              if ((e = lineedit_input_decoder_setup (&state->input_decoder))
                  == ___FIX(___NO_ERR))
                {
                  if ((e = lineedit_set_terminal_type
                             (state,
                              NULL,
                              DEFAULT_EMACS_BINDINGS))
                      == ___FIX(___NO_ERR))
                    {
#ifdef USE_LOCAL_CLIPBOARD
                      if ((e = extensible_string_setup (&state->clipboard, 0))
                          == ___FIX(___NO_ERR))
                        {
#endif
#ifdef USE_TERMCAP_OR_TERMINFO
                          int i;
                          for (i=0; i<LINEEDIT_CAP_LAST+1; i++)
                            state->capability[i] = NULL;
#endif
#ifdef USE_WINDOWS
                          /***********************/
#endif
#ifdef USE_MACOS
                          /***********************/
#endif
#ifdef USE_LOCAL_CLIPBOARD
                          if (e != ___FIX(___NO_ERR))
                            extensible_string_cleanup
                              (&state->clipboard); /* ignore error */
                        }
#endif
                    }
                  if (e != ___FIX(___NO_ERR))
                    lineedit_input_decoder_cleanup
                      (&state->input_decoder); /* ignore error */
                }
            }
          if (e != ___FIX(___NO_ERR))
            lineedit_history_cleanup (state, h); /* ignore error */
        }
      if (e != ___FIX(___NO_ERR))
        extensible_string_cleanup (&state->output); /* ignore error */
    }

  if (e != ___FIX(___NO_ERR))
    ___free_mem (state);

  return e;
}

___HIDDEN ___WORD lineedit_cleanup
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  lineedit_output_set_text_attrs
    (state,
     DEFAULT_TEXT_ATTRS); /* ignore error */

  lineedit_output_drain (state); /* ignore error */

  if (state->paste_text != NULL)
    ___free_mem (state->paste_text); /* discard paste text */

  if (state->input.buffer != NULL)
    extensible_string_cleanup (&state->input); /* ignore error */

  extensible_string_cleanup (&state->output); /* ignore error */

  state->max_history_length = -1;
  lineedit_history_trim (state);

#ifdef USE_LOCAL_CLIPBOARD
  extensible_string_cleanup (&state->clipboard); /* ignore error */
#endif

#ifdef USE_TERMCAP_OR_TERMINFO
  {
    int i;

    lineedit_input_decoder_cleanup (&state->input_decoder); /* ignore error */

    for (i=0; i<LINEEDIT_CAP_LAST+1; i++)
      {
        char *seq = state->capability[i];
        if (seq != NULL)
          ___free_mem (seq);
      }
  }
#endif

#ifdef USE_WINDOWS
  /*********************/
#endif

#ifdef USE_MACOS
  /*********************/
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_move_edit_point
   ___P((lineedit_state *state,
         int pos),
        (state,
         pos)
lineedit_state *state;
int pos;)
{
  /*
   * Moves the edit point of the current edited line.  The screen may
   * get scrolled in order to keep the edit point visible.
   */

  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  int screen_cursor;
  int screen_cols;
  int screen_rows;
  int screen_size;
  int cursor_to_point;
  int cursor_row;
  int first_row;
  int last_row;
  int nb_rows;
  int all_fits;
  int n;
  int start;

  if (pos < 0 ||
      pos > edited->length)
    return ___FIX(___INVALID_OP_ERR);

  screen_cols = state->screen_cols;
  screen_rows = state->screen_rows;
  screen_size = state->screen_size;
  cursor_to_point = state->current.cursor_to_point;
  screen_cursor = pos - cursor_to_point; /* new cursor position in screen */

  /*
   * Check if the text from the cursor to the position just after end
   * of the edited line can all fit on the screen (while maintaining
   * column alignment).
   */

  cursor_row = screen_cursor;
  if (cursor_row < 0)
    cursor_row -= screen_cols - 1;
  cursor_row /= screen_cols;

  first_row = -cursor_to_point;
  if (first_row < 0)
    first_row -= screen_cols - 1;
  first_row /= screen_cols;

  last_row = edited->length - cursor_to_point;
  if (last_row < 0)
    last_row -= screen_cols - 1;
  last_row /= screen_cols;

  all_fits = (last_row - cursor_row < screen_rows);

  /*
   * Determine how many rows to scroll and in which direction (nb_rows
   * is positive when scrolling forward and negative when scrolling
   * backward).
   */

  if (all_fits)
    {
      if (cursor_row >= 0 &&
          last_row < screen_rows)
        /* Don't scroll if text from cursor to end is visible right now. */
        nb_rows = 0;
      else
        /* Scroll so that last row of edited line is at bottom of screen. */
        nb_rows = last_row - (screen_rows - 1);
    }
  else
    {
      if (cursor_row >= 0 &&
          cursor_row < screen_rows)
        /* Don't scroll if cursor is visible right now. */
        nb_rows = 0;
      else
        {
          /*
           * Scroll cursor row to bottom of screen unless this makes
           * the beginning of the edited line visible, in which case
           * align the beginning of the edited line to the top of the
           * screen.
           */

          if (cursor_row < first_row + (screen_rows - 1))
            nb_rows = first_row;
          else
            nb_rows = cursor_row - (screen_rows - 1);
        }
    }

  /*
   * Perform scrolling of screen.
   */

  if (nb_rows != 0)
    {
      screen_cursor -= screen_cols * nb_rows;

      if (nb_rows < 0)
        {
          /* scroll screen backward by whole rows */

          if (pos < screen_cursor)
            screen_cursor = pos; /* align start of buffer with screen home */

          start = pos - screen_cursor;

          n = screen_size; /* write to the whole screen */

          if ((e = lineedit_prepare_to_write_at (state, 0))
              != ___FIX(___NO_ERR))
            return e;
        }
      else
        {
          /* scroll screen forward by whole rows */

          start = screen_size + cursor_to_point;

          n = screen_cols * nb_rows; /* write characters that scroll in */

          if ((e = lineedit_prepare_to_write_at (state, screen_size))
              != ___FIX(___NO_ERR))
            return e;
        }

      if (!state->has_eat_newline_glitch)/*****************/
        n--; /* to avoid wraparound at end of screen */

      if (n > edited->length - start)
        n = edited->length - start;

      if ((e = lineedit_output_current_hist (state, start, n))
          != ___FIX(___NO_ERR))
        return e;

      if (all_fits)
        if ((e = lineedit_output_force_delayed_wrap (state))
            != ___FIX(___NO_ERR))
          return e;
    }

  if ((e = lineedit_move_screen_cursor (state, screen_cursor))
      != ___FIX(___NO_ERR))
    return e;

  state->current.edit_point = pos;
  state->current.cursor_to_point = pos - screen_cursor;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_update_region
   ___P((lineedit_state *state,
         int start,
         int end,
         ___BOOL allow_scrolling),
        (state,
         start,
         end,
         allow_scrolling)
lineedit_state *state;
int start;
int end;
___BOOL allow_scrolling;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  int screen_size = state->screen_size;
  int cursor_to_point = state->current.cursor_to_point;
  int pos = start - cursor_to_point;

  if (!state->has_eat_newline_glitch)/*****************/
    screen_size--; /* to avoid wraparound at end of screen */

  if (pos >= screen_size ||
      end - cursor_to_point <= 0)
    return ___FIX(___NO_ERR);

  if (pos < 0)
    pos = 0;

  start = pos + cursor_to_point;

  if ((e = lineedit_prepare_to_write_at (state, pos))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_output_current_hist (state, start, end - start);
}

___HIDDEN ___WORD lineedit_refresh
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;

  if ((e = lineedit_erase_screen (state))
      != ___FIX(___NO_ERR))
    return e;

  state->current.cursor_to_point = -state->current.screen_cursor;

  if ((e = lineedit_update_region (state, 0, edited->length, 1))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_move_edit_point (state, state->current.edit_point);
}

___HIDDEN ___BOOL lineedit_word_boundary
   ___P((lineedit_state *state,
         int back,
         int pos,
         int *final_pos),
        (state,
         back,
         pos,
         final_pos)
lineedit_state *state;
int back;
int pos;
int *final_pos;)
{
  extensible_string *edited = &state->current.hist->edited;
  tty_char *buf = edited->buffer;
  int len = edited->length;
  int depth = 0;
  tty_char in_string = UNICODE_NUL; /* assume not inside a string */
  tty_char c;

#define OPEN_PAREN(x) \
((x) == UNICODE_LPAREN || (x) == UNICODE_LBRACKET || (x) == UNICODE_LBRACE)

#define CLOSE_PAREN(x) \
((x) == UNICODE_RPAREN || (x) == UNICODE_RBRACKET || (x) == UNICODE_RBRACE)

#define STRING_DELIM(x) \
((x) == UNICODE_DOUBLEQUOTE || (x) == UNICODE_VBAR)

#define INT_VECTOR(x) \
((x) == UNICODE_LOWER_S || (x) == UNICODE_UPPER_S || \
 (x) == UNICODE_LOWER_U || (x) == UNICODE_UPPER_U)

#define FLOAT_VECTOR(x) \
((x) == UNICODE_LOWER_F || (x) == UNICODE_UPPER_F)

  if (back == 0)
    {
      /* skip whitespace */

      while (pos < len)
        {
          c = buf[pos];
          if (c <= UNICODE_SPACE)
            pos++;
          else
            break;
        }

      /* skip standard read-macros */

      while (pos < len)
        {
          c = buf[pos];
          if (c == UNICODE_QUOTE)
            pos++;
          else if (c == UNICODE_COMMA)
            {
              pos++;
              if (pos < len && buf[pos] == UNICODE_AT)
                pos++;
            }
          else
            break;
        }

      if (pos == len)
        {
          depth = -1; /* force error */
          goto done_forward;
        }

      /* handle head specially */

      c = buf[pos++];

      if (pos < len && c == UNICODE_SHARP)
        {
          /* handle #\xxx, #(...), #f32(...), etc */

          c = buf[pos++];
          if (c == UNICODE_BACKSLASH)
            {
              if (pos < len)
                pos++;
            }
          else if (OPEN_PAREN(c))
            depth++;
          else
            {
              ___BOOL f = FLOAT_VECTOR(c);
              if (pos < len && (f || INT_VECTOR(c)))
                {
                  c = buf[pos++];
                  if ((!f && pos < len && c == UNICODE_8) ||
                      (pos+1 < len &&
                       ((!f && c == UNICODE_1 && buf[pos] == UNICODE_6) ||
                        (c == UNICODE_3 && buf[pos] == UNICODE_2) ||
                        (c == UNICODE_6 && buf[pos] == UNICODE_4)) &&
                       (pos++, 1)))
                    {
                      c = buf[pos++];
                      if (OPEN_PAREN(c))
                        depth++;
                      else
                        pos--;
                    }
                }
            }
        }
      else if (OPEN_PAREN(c))
        depth++;
      else if (CLOSE_PAREN(c))
        {
          depth = -1; /* force error */
          goto done_forward;
        }
      else if (STRING_DELIM(c))
        in_string = c;

      while (pos < len)
        if (in_string == UNICODE_NUL)
          {
            c = buf[pos];

            if (c == UNICODE_SHARP)
              {
                if (pos+2 < len && buf[pos+1] == UNICODE_BACKSLASH)
                  pos += 2;
              }
            else if (c <= UNICODE_SPACE ||
                     c == UNICODE_SEMICOLON ||
                     c == UNICODE_QUOTE ||
                     c == UNICODE_COMMA)
              {
                if (depth == 0)
                  break;
              }
            else if (OPEN_PAREN(c))
              {
                if (depth == 0)
                  break;
                depth++;
              }
            else if (CLOSE_PAREN(c))
              {
                if (depth == 0)
                  break;
                depth--;
                if (depth == 0)
                  {
                    pos++;
                    break;
                  }
              }
            else if (STRING_DELIM(c))
              {
                if (depth == 0)
                  break;
                in_string = c;
              }

            pos++;
          }
        else
          {
            c = buf[pos++];

            if (c == UNICODE_BACKSLASH) /* character is escaped */
              {
                if (pos < len)
                  pos++;
              }
            else if (c == in_string) /* end of string reached */
              {
                if (depth == 0)
                  break;
                in_string = UNICODE_NUL;
              }
          }

      done_forward:;
    }
  else
    {
      /* skip whitespace */

      while (pos > 0)
        {
          c = buf[pos-1];
          if (c <= UNICODE_SPACE)
            pos--;
          else
            break;
        }

      if (pos == 0)
        {
          depth = -1; /* force error */
          goto done_backward;
        }

      /* handle tail specially */

      c = buf[--pos];

      if (pos >= 2 &&
          buf[pos-2] == UNICODE_SHARP &&
          buf[pos-1] == UNICODE_BACKSLASH)
        pos -= 2;
      else if (OPEN_PAREN(c))
        {
          depth = -1; /* force error */
          goto done_backward;
        }
      else if (CLOSE_PAREN(c))
        depth++;
      else if (STRING_DELIM(c))
        in_string = c;
      else if (c == UNICODE_QUOTE ||
               c == UNICODE_COMMA ||
               (pos >= 1 && c == UNICODE_AT && buf[pos-1] == UNICODE_COMMA))
        {
          pos++;
          goto skip_read_macros;
        }

      while (pos > 0)
        if (in_string == UNICODE_NUL)
          {
            c = buf[pos-1];

            if (pos >= 3 &&
                buf[pos-3] == UNICODE_SHARP &&
                buf[pos-2] == UNICODE_BACKSLASH)
              pos -= 2;
            else if (c <= UNICODE_SPACE ||
                     c == UNICODE_SEMICOLON ||
                     c == UNICODE_QUOTE ||
                     c == UNICODE_COMMA)
              {
                if (depth == 0)
                  break;
              }
            else if (pos >= 2 &&
                     c == UNICODE_AT &&
                     buf[pos-2] == UNICODE_COMMA)
              {
                if (depth == 0)
                  break;
                pos--;
              }
            else if (OPEN_PAREN(c))
              {
                if (depth == 0)
                  break;
                depth--;
                if (depth == 0)
                  {
                    pos--;
                    break;
                  }
              }
            else if (CLOSE_PAREN(c))
              {
                if (depth == 0)
                  break;
                depth++;
              }
            else if (STRING_DELIM(c))
              {
                if (depth == 0)
                  break;
                in_string = c;
              }

            pos--;
          }
        else
          {
            int i;

            i = --pos;

            while (i > 0 &&
                   buf[i-1] == UNICODE_BACKSLASH) /* find nb of backslashes */
              i--;

            if ((pos-i) & 1) /* character is escaped */
              pos = i;
            else if (buf[pos] == in_string) /* beginning of string reached */
              {
                if (depth == 0)
                  break;
                in_string = UNICODE_NUL;
              }
          }

      done_backward:;

      if (back == 2)
        {
          /* handle #(...), #f32(...), etc */

          if (pos < len)
            {
              c = buf[pos];
              if (OPEN_PAREN(c))
                {
                  if (pos >= 1 && buf[pos-1] == UNICODE_SHARP)
                    pos--;
                  else if (pos >= 3 && buf[pos-3] == UNICODE_SHARP)
                    {
                      c = buf[pos-2];
                      if (INT_VECTOR(c) && buf[pos-1] == UNICODE_8)
                        pos -= 3;
                    }
                  else if (pos >= 4 && buf[pos-4] == UNICODE_SHARP)
                    {
                      ___BOOL f;
                      c = buf[pos-3];
                      f = FLOAT_VECTOR(c);
                      if (f || INT_VECTOR(c))
                        {
                          c = buf[pos-2];
                          if ((!f &&
                               c == UNICODE_1 &&
                               buf[pos-1] == UNICODE_6) ||
                              (c == UNICODE_3 && buf[pos-1] == UNICODE_2) ||
                              (c == UNICODE_6 && buf[pos-1] == UNICODE_4))
                            pos -= 4;
                        }
                    }
                }
            }

          /* skip standard read-macros */

          skip_read_macros:

          while (pos > 0)
            {
              c = buf[pos-1];
              if (c == UNICODE_QUOTE || c == UNICODE_COMMA)
                pos--;
              else if (pos >= 2 &&
                       c == UNICODE_AT &&
                       buf[pos-2] == UNICODE_COMMA)
                pos -= 2;
              else
                break;
            }
        }
    }

  *final_pos = pos;

  return (depth == 0 && in_string == UNICODE_NUL);
}

___HIDDEN ___WORD lineedit_insert_chars
   ___P((lineedit_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
lineedit_state *state;
tty_char *buf;
int len;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  int edit_point = state->current.edit_point;
  int open_paren_point;
  tty_char c;

  if (len <= 0)
    return ___FIX(___NO_ERR);

  if ((e = extensible_string_insert (edited, edit_point, len, buf))
      != ___FIX(___NO_ERR))
    return e;

  if (state->current.mark_point > edit_point)
    state->current.mark_point += len;

  if ((e = lineedit_update_region (state, edit_point, edited->length, 1))
      != ___FIX(___NO_ERR))
    return e;

  edit_point += len;

  if (state->paren_balance_duration_msecs > 0 &&
      state->ts->rlo >= state->ts->rhi &&
      (c = buf[len-1], CLOSE_PAREN(c)) &&
      lineedit_word_boundary (state, 1, edit_point, &open_paren_point) &&
      (c = edited->buffer[open_paren_point], OPEN_PAREN(c)))
    {
      if ((e = lineedit_move_edit_point (state, open_paren_point))
          != ___FIX(___NO_ERR))
        return e;

      if ((e = lineedit_output_drain (state)) != ___FIX(___NO_ERR))
        return e;

      state->current.edit_point = edit_point;
      state->current.paren_balance_trigger = 1;

      return ___FIX(___NO_ERR);
    }

  return lineedit_move_edit_point (state, edit_point);
}

___HIDDEN ___WORD lineedit_delete_chars
   ___P((lineedit_state *state,
         int pos,
         ___BOOL copy_to_clipboard),
        (state,
         pos,
         copy_to_clipboard)
lineedit_state *state;
int pos;
___BOOL copy_to_clipboard;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  int start;
  int end;
  int n;

  if (pos < 0 ||
      pos > edited->length)
    return ___FIX(___INVALID_OP_ERR);

  if (pos < state->current.edit_point)
    {
      start = pos;
      end = state->current.edit_point;
    }
  else
    {
      start = state->current.edit_point;
      end = pos;
    }

  n = end - start;

  if (n == 0)
    return ___FIX(___INVALID_OP_ERR);

  if (copy_to_clipboard &&
      ((e = lineedit_copy_to_clipboard (state, &edited->buffer[start], n))
       != ___FIX(___NO_ERR)))
    return e;

  if ((e = extensible_string_delete (edited, start, n)) != ___FIX(___NO_ERR))
    return e;

  if (state->current.mark_point >= start)
    {
      if (state->current.mark_point >= end)
        state->current.mark_point -= n;
      else
        state->current.mark_point = start;
    }

  if ((e = lineedit_update_region (state, start, edited->length + n, 1))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_move_edit_point (state, start);
}

___HIDDEN void lineedit_rotate_left_chars
   ___P((lineedit_state *state,
         int start,
         int end,
         int n),
        (state,
         start,
         end,
         n)
lineedit_state *state;
int start;
int end;
int n;)
{
  extensible_string *edited = &state->current.hist->edited;
  tty_char *p = &edited->buffer[start];
  tty_char temp;
  int len = end-start;
  int left = len;
  int i;
  int j;
  int k;

  i = 0;
  while (left > 0)
    {
      temp = p[i];
      j = i;
      k = (j+n) % len;
      left--;
      while (k != i)
        {
          p[j] = p[k];
          j = k;
          k = (j+n) % len;
          left--;
        }
      p[j] = temp;
      i++;
    }
}

___HIDDEN ___WORD lineedit_transpose_chars
   ___P((lineedit_state *state,
         int start1,
         int end1,
         int start2,
         int end2),
        (state,
         start1,
         end1,
         start2,
         end2)
lineedit_state *state;
int start1;
int end1;
int start2;
int end2;)
{
  ___WORD e;

  if (start1 > end1 ||
      end1 > start2 ||
      start2 > end2)
    return ___FIX(___INVALID_OP_ERR);

  lineedit_rotate_left_chars
    (state,
     start1,
     end2,
     end1-start1);

  lineedit_rotate_left_chars
    (state,
     start1,
     end2-(end1-start1),
     start2-end1);

  if ((e = lineedit_update_region (state, start1, end2, 1))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_move_edit_point (state, end2);
}

___HIDDEN ___WORD lineedit_move_history
   ___P((lineedit_state *state,
         lineedit_history *h),
        (state,
         h)
lineedit_state *state;
lineedit_history *h;)
{
  ___WORD e;
  int old_length;
  int new_length;

  if ((e = lineedit_history_begin_edit (state, h))
      != ___FIX(___NO_ERR))
    return e;

  if ((e = lineedit_move_edit_point (state, 0))
      != ___FIX(___NO_ERR))
    return e;

  old_length = state->current.hist->edited.length;
  new_length = h->edited.length;

  state->current.mark_point = 0;
  state->current.hist = h;

  if ((e = lineedit_update_region
             (state,
              0,
              (new_length > old_length) ? new_length : old_length,
              1))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_move_edit_point (state, new_length);
}

___HIDDEN ___WORD lineedit_move_history_relative
   ___P((lineedit_state *state,
         ___BOOL back),
        (state,
         back)
lineedit_state *state;
___BOOL back;)
{
  lineedit_history *h;
  lineedit_history *limit;

  if (back)
    {
      h = state->current.hist->prev;
      limit = state->hist_last;
    }
  else
    {
      h = state->current.hist->next;
      limit = state->hist_last->next;
    }

  if (h == limit)
    return ___FIX(___INVALID_OP_ERR);

  return lineedit_move_history (state, h);
}

___HIDDEN ___WORD lineedit_line_done
   ___P((lineedit_state *state,
         ___BOOL end_of_file),
        (state,
         end_of_file)
lineedit_state *state;
___BOOL end_of_file;)
{
  ___WORD e;
  extensible_string *edited = &state->current.hist->edited;
  extensible_string *input = &state->input;

  /*
   * Transfer the current edited line to the input.
   */

  if (input->buffer == NULL)
    {
      state->input_rlo = 0;

      if ((e = extensible_string_copy
                 (edited->buffer,
                  edited->length,
                  input,
                  1))
          != ___FIX(___NO_ERR))
        return e;

      if (!end_of_file)
        {
          int old_length = input->length;
          if ((e = extensible_string_set_length (input, old_length+1, 0))
              != ___FIX(___NO_ERR))
            {
              extensible_string_cleanup (input); /* ignore error */
              state->input.buffer = NULL;
              return e;
            }
          input->buffer[old_length] = UNICODE_LINEFEED;
        }
    }

  /*
   * Update history (empty lines are not added).
   */

  if (edited->length == 0 || end_of_file)
    {
      if ((e = lineedit_move_edit_point (state, edited->length))
          == ___FIX(___NO_ERR) &&
          (end_of_file ||
           (e = lineedit_newline (state))
           == ___FIX(___NO_ERR)))
        extensible_string_set_length
          (&state->hist_last->edited,
           0,
           1); /* ignore error */
    }
  else
    {
      /*
       * Make a copy of the current edited line for the history.
       */

      extensible_string actual;

      if ((e = extensible_string_copy
                 (edited->buffer,
                  edited->length,
                  &actual,
                  0))
          == ___FIX(___NO_ERR))
        {
          /*
           * Add the current edited line to the history.
           */

          lineedit_history *next_line;

          if ((e = lineedit_history_setup (state, &next_line))
              == ___FIX(___NO_ERR))
            {
              if ((e = lineedit_history_begin_edit (state, next_line))
                  == ___FIX(___NO_ERR) &&
                  (e = lineedit_move_edit_point (state, edited->length))
                  == ___FIX(___NO_ERR) &&
                  (e = lineedit_newline (state))
                  == ___FIX(___NO_ERR))
                {
                  extensible_string_cleanup
                    (&state->hist_last->actual); /* ignore error */
                  state->hist_last->actual = actual;
                  lineedit_history_add (state, next_line);
                  lineedit_history_trim (state);
                }
              else
                lineedit_history_cleanup (state, next_line); /* ignore error */
            }

          if (e != ___FIX(___NO_ERR))
            extensible_string_cleanup (&actual); /* ignore error */
        }
    }

  if (e == ___FIX(___NO_ERR))
    {
      /*
       * Revert history lines to their nonedited state.
       */

      lineedit_history *h = state->hist_last;

      while (1)
        {
          h = h->prev;
          if (h == state->hist_last)
            break;
          lineedit_history_end_edit (state, h);
        }

      /*
       * Prepare editing of next line.
       */

      state->current.edit_point = 0;
      state->current.mark_point = 0;
      state->current.cursor_to_point = -state->current.screen_cursor;
      state->current.hist = state->hist_last;
    }

  return e;
}

___HIDDEN ___WORD lineedit_end_paren_balance
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  if (state->current.paren_balance_in_progress)
    {
      state->current.paren_balance_in_progress = 0;
      return lineedit_move_edit_point
               (state,
                state->current.edit_point); /* ignore error */
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_process_single_event
   ___P((lineedit_state *state,
         lineedit_event *ev),
        (state,
         ev)
lineedit_state *state;
lineedit_event *ev;)
{
  extensible_string *edited = &state->current.hist->edited;

  switch (ev->event_kind)
    {
    case LINEEDIT_EVENT_NONE:
      return ___FIX(___NO_ERR);

    case LINEEDIT_EVENT_KEY:
      if (ev->event_char < UNICODE_SPACE || /* discard control characters */
          ev->event_char == UNICODE_RUBOUT)
        return ___FIX(___INVALID_OP_ERR);
      return lineedit_insert_chars (state, &ev->event_char, 1);

    case LINEEDIT_EVENT_RETURN:
      return lineedit_line_done (state, 0);

    case LINEEDIT_EVENT_BACK:
      return lineedit_delete_chars (state, state->current.edit_point-1, 0);

    case LINEEDIT_EVENT_BACK_WORD:
      {
        int i;
        if (!lineedit_word_boundary (state, 2, state->current.edit_point, &i))
          return ___FIX(___INVALID_OP_ERR);
        return lineedit_delete_chars (state, i, 0);
      }

    case LINEEDIT_EVENT_TAB:
      return lineedit_insert_chars
               (state,
                lem.tab_spaces,
                state->tab_stop-(state->current.edit_point%state->tab_stop));

    case LINEEDIT_EVENT_MARK:
      state->current.mark_point = state->current.edit_point;
      return ___FIX(___NO_ERR);

    case LINEEDIT_EVENT_PASTE:
      return lineedit_paste_from_clipboard (state);

    case LINEEDIT_EVENT_CUT:
      return lineedit_delete_chars (state, state->current.mark_point, 1);

    case LINEEDIT_EVENT_CUT_RIGHT:
      return lineedit_delete_chars (state, edited->length, 1);

    case LINEEDIT_EVENT_CUT_LEFT:
      return lineedit_delete_chars (state, 0, 1);

    case LINEEDIT_EVENT_REFRESH:
      return lineedit_refresh (state);

    case LINEEDIT_EVENT_TRANSPOSE:
      {
        int i = state->current.edit_point;
        if (i <= 0 ||
            edited->length < 2)
          return ___FIX(___INVALID_OP_ERR);
        if (i == edited->length)
          return lineedit_transpose_chars (state, i-2, i-1, i-1, i);
        else
          return lineedit_transpose_chars (state, i-1, i, i, i+1);
      }

    case LINEEDIT_EVENT_TRANSPOSE_WORD:
      {
        int start1;
        int end1;
        int start2;
        int end2;
        if (!lineedit_word_boundary
               (state,
                0,
                state->current.edit_point,
                &end2) ||
            !lineedit_word_boundary (state, 2, end2, &start2) ||
            !lineedit_word_boundary (state, 2, start2, &start1) ||
            !lineedit_word_boundary (state, 0, start1, &end1))
          return ___FIX(___INVALID_OP_ERR);
        return lineedit_transpose_chars (state, start1, end1, start2, end2);
      }

    case LINEEDIT_EVENT_UP:
      return lineedit_move_history_relative (state, 1);

    case LINEEDIT_EVENT_DOWN:
      return lineedit_move_history_relative (state, 0);

    case LINEEDIT_EVENT_RIGHT:
      return lineedit_move_edit_point (state, state->current.edit_point+1);

    case LINEEDIT_EVENT_RIGHT_WORD:
      {
        int i;
        lineedit_word_boundary (state, 0, state->current.edit_point, &i);
        return lineedit_move_edit_point (state, i);
      }

    case LINEEDIT_EVENT_LEFT:
      return lineedit_move_edit_point (state, state->current.edit_point-1);

    case LINEEDIT_EVENT_LEFT_WORD:
      {
        int i;
        lineedit_word_boundary (state, 2, state->current.edit_point, &i);
        return lineedit_move_edit_point (state, i);
      }

    case LINEEDIT_EVENT_HOME:
      return lineedit_move_edit_point (state, 0);

    case LINEEDIT_EVENT_HOME_DOC:
      return lineedit_move_history (state, state->hist_last->next);

    case LINEEDIT_EVENT_INSERT:
{ tty_char c = '*';/****************/
  return extensible_string_insert
             (edited,
              edited->length,
              1,
              &c);
}
      return ___FIX(___INVALID_OP_ERR);

    case LINEEDIT_EVENT_DELETE:
      if (edited->length == 0)
        return lineedit_line_done (state, 1);
      else
        return lineedit_delete_chars (state, state->current.edit_point+1, 0);

    case LINEEDIT_EVENT_DELETE_WORD:
      {
        int i;
        if (!lineedit_word_boundary (state, 0, state->current.edit_point, &i))
          return ___FIX(___INVALID_OP_ERR);
        return lineedit_delete_chars (state, i, 0);
      }

    case LINEEDIT_EVENT_END:
      return lineedit_move_edit_point (state, edited->length);

    case LINEEDIT_EVENT_END_DOC:
      return lineedit_move_history (state, state->hist_last);

    case LINEEDIT_EVENT_F1:
    case LINEEDIT_EVENT_META_F1:
    case LINEEDIT_EVENT_F2:
    case LINEEDIT_EVENT_META_F2:
    case LINEEDIT_EVENT_F3:
    case LINEEDIT_EVENT_META_F3:
    case LINEEDIT_EVENT_F4:
    case LINEEDIT_EVENT_META_F4:
#ifdef SUPPORT_F5_TO_F12
    case LINEEDIT_EVENT_F5:
    case LINEEDIT_EVENT_META_F5:
    case LINEEDIT_EVENT_F6:
    case LINEEDIT_EVENT_META_F6:
    case LINEEDIT_EVENT_F7:
    case LINEEDIT_EVENT_META_F7:
    case LINEEDIT_EVENT_F8:
    case LINEEDIT_EVENT_META_F8:
    case LINEEDIT_EVENT_F9:
    case LINEEDIT_EVENT_META_F9:
    case LINEEDIT_EVENT_F10:
    case LINEEDIT_EVENT_META_F10:
    case LINEEDIT_EVENT_F11:
    case LINEEDIT_EVENT_META_F11:
    case LINEEDIT_EVENT_F12:
    case LINEEDIT_EVENT_META_F12:
#endif
      break;
    }

  return ___FIX(___INVALID_OP_ERR);
}

___HIDDEN ___WORD lineedit_process_events
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
#ifdef LINEEDIT_WITH_NONBLOCKING_IO
#define AGGREGATION_BUFFER_SIZE (80*50)
#else
#define AGGREGATION_BUFFER_SIZE 1
#endif

  ___WORD e1;
  ___WORD e2;
  tty_char aggregation_buffer[AGGREGATION_BUFFER_SIZE];
  lineedit_event ev;
  int i;
  struct lineedit_state_undo_struct previous;

  /*
   * Process events only if no input line is currently available.
   */

  if (state->input.buffer != NULL)
    return ___FIX(___NO_ERR);

  /*
   * Process events only if no line editing output is pending.
   */

  if ((e1 = lineedit_output_drain (state)) != ___FIX(___NO_ERR))
    return e1;

  /*
   * Save state so that we can undo if there is an error.
   */

  previous = state->current;

  i = 0;

  ev.event_kind = LINEEDIT_EVENT_NONE;

  state->current.paren_balance_trigger = 0;

  while (1)
    {
      if ((e1 = lineedit_get_event (state, &ev)) != ___FIX(___NO_ERR) ||
          ev.event_kind == LINEEDIT_EVENT_NONE)
        break;

      if (ev.event_kind != LINEEDIT_EVENT_KEY ||
          ev.event_char < UNICODE_SPACE ||
          ev.event_char == UNICODE_RUBOUT)
        break;

      aggregation_buffer[i++] = ev.event_char;

      ev.event_kind = LINEEDIT_EVENT_NONE;

      if (i == AGGREGATION_BUFFER_SIZE)
        break;
    }

  e2 = lineedit_insert_chars (state, aggregation_buffer, i);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  e2 = lineedit_process_single_event (state, &ev);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  if (e1 == ___FIX(___INVALID_OP_ERR))
    {
      /*
       * Undo line editing operation.
       */

      tty_char c = UNICODE_BELL;

      extensible_string_set_length
        (&state->output,
         0,
         1); /* ignore error */

      state->output_wlo = state->output.length;

      state->current = previous;

      e1 = extensible_string_insert
             (&state->output,
              state->output.length,
              1,
              &c);
    }

  if (state->current.paren_balance_trigger)
    {
      /*
       * Parenthesis balancing started in this round of event
       * processing.
       */

      lineedit_time duration;

      lineedit_get_time (&state->current.paren_balance_end);
      lineedit_init_time (&duration, 0, state->paren_balance_duration_msecs);
      lineedit_add_time (&state->current.paren_balance_end, &duration);

      state->current.paren_balance_in_progress = 1;

#ifndef LINEEDIT_WITH_NONBLOCKING_IO
      lineedit_sleep_until (&state->current.paren_balance_end);
      lineedit_end_paren_balance (state); /* ignore error */
#endif
    }
  else if (state->current.paren_balance_in_progress)
    {
      /*
       * Parenthesis balancing did not start in this round of event
       * processing and parenthesis balancing is in progress, check if
       * parenthesis balancing should end.
       */

      if (i > 0 || ev.event_kind != LINEEDIT_EVENT_NONE)
        lineedit_end_paren_balance (state); /* ignore error */
#ifdef LINEEDIT_WITH_NONBLOCKING_IO
      else
        {
          lineedit_time time_after_end;

          lineedit_get_time (&time_after_end);

          lineedit_sub_time
            (&time_after_end,
             &state->current.paren_balance_end);

          if (!lineedit_negative_time (&time_after_end))
            lineedit_end_paren_balance (state); /* ignore error */
        }
#endif
    }

  e2 = lineedit_output_drain (state);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  return e1;
}

___HIDDEN ___WORD lineedit_read
   ___P((lineedit_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
lineedit_state *state;
tty_char *buf;
int len;)
{
  ___WORD e;
  int rlo;
  int line_length;
  int avail;

  if ((e = lineedit_process_events (state)) != ___FIX(___NO_ERR))
    return e;

  if (state->input.buffer == NULL)
    return ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));

  rlo = state->input_rlo;
  line_length = state->input.length;
  avail = line_length - rlo;

  if (avail > 0)
    {
      tty_char *p = &state->input.buffer[rlo];
      tty_char *end;

      if (len > avail)
        len = avail;

      end = p + len;

      while (p < end)
        *buf++ = *p++;

      rlo += len;
    }
  else
    len = 0;

  state->input_rlo = rlo;

  if (rlo >= line_length)
    {
      extensible_string_cleanup (&state->input); /* ignore error */
      state->input.buffer = NULL;
    }

  return ___FIX(len);
}

___HIDDEN ___WORD lineedit_module_setup ___PVOID
{
  int i;

  for (i=0; i<LINEEDIT_MAX_TAB_STOP; i++)
    lem.tab_spaces[i] = UNICODE_SPACE;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD lineedit_module_cleanup ___PVOID
{
  return ___FIX(___NO_ERR);
}

#endif

/*---------------------------------------------------------------------------*/

/* Tty management routines */

#define TTY_BIG_SCREEN_SIZE (80*50)

___HIDDEN void tty_broadcast_signal
   ___P((tty_signal s),
        (s)
tty_signal s;)
{
  tty_state *list = tm.list;

  if (list != NULL)
    {
      tty_state *state = list;
      do
        {
          state->received_signal[s] = 1;
          state = state->next;
        } while (state != list);
    }
}

___HIDDEN ___WORD tty_setup_common
   ___P((tty *t,
         tty_state *state),
        (t,
         state)
tty *t;
tty_state *state;)
{
  ___WORD e;
  tty_state *dest;

  state->refcount = 1;

  state->input_mode = CBREAK_LINEEDIT_INPUT_MODE;
  state->term_mode = INITIAL_TERM_MODE;

  state->rlo = 0;
  state->rhi = 0;
  state->wlo = 0;
  state->whi = 0;
  state->werr = ___FIX(___NO_ERR);

  state->received_signal[USER_INTERRUPT_SIGNAL] = 0;

#ifdef USE_LINEEDIT

  /*
   * Setup line editing state.
   */

  if ((e = lineedit_setup (&state->les, state)) != ___FIX(___NO_ERR))
    {
      ___free_mem (state);
      return e;
    }

  /*
   * Switch to CBREAK mode so that terminal driver does not perform
   * its own line editing.
   */

  if ((e = tty_set_term_mode (state, CBREAK_TERM_MODE))
      != ___FIX(___NO_ERR))
    return e;

#endif

  /* add tty_state to tail of circular list of tty_state's */

  dest = tm.list;

  if (dest == NULL)
    {
      state->prev = state;
      state->next = state;
      tm.list = state;
    }
  else
    {
      tty_state *before_dest = dest->prev;
      state->next = dest;
      state->prev = before_dest;
      dest->prev = state;
      before_dest->next = state;
    }

  t->state = state;

  return ___FIX(___NO_ERR);
}

#ifndef USE_UNIX
#ifndef USE_WINDOWS
#ifndef USE_MACOS

/**************************/

#endif
#endif
#endif

#ifdef USE_UNIX

___HIDDEN ___BOOL tty_equal
   ___P((int fd1,
         int fd2),
        (fd1,
         fd2)
int fd1;
int fd2;)
{
  struct stat fd1_stat;
  struct stat fd2_stat;

  if (fd1 == fd2)
    return 1;

  /* check if st_rdev fields are the same */

  if (fstat (fd1, &fd1_stat) == 0 &&
      fstat (fd2, &fd2_stat) == 0 &&
      fd1_stat.st_rdev == fd2_stat.st_rdev)
    return 1;

  /* check if the descriptors both refer to the controlling terminal */

  if (tcgetpgrp (fd1) >= 0 &&
      tcgetpgrp (fd2) >= 0)
    return 1;

  return 0;
}

___HIDDEN ___WORD tty_setup_os_specific
   ___P((tty *t,
         int fd),
        (t,
         fd)
tty *t;
int fd;)
{
  tty_state *state;

  t->state = NULL;

  /* check if the same terminal was previously opened */

  state = tm.list;

  if (state != NULL)
    {
      do
        {
          if (tty_equal (fd, state->fd))
            {
              close (fd); /* ignore error */
              state->refcount++;
              t->state = state;
              return ___FIX(___NO_ERR);
            }
          state = state->next;
        } while (state != tm.list);
    }

  /* need to allocate a new tty_state */

  state = ___alloc_mem (sizeof (tty_state));

  if (state == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  state->received_signal[WINDOW_SIZE_SIGNAL] = 1; /* query window size */
  state->received_signal[CONTINUE_SIGNAL] = 0;

  state->fd = fd;

  if (tcgetattr (state->fd, &state->initial_termios) < 0)
    {
      ___WORD e = ___FIX(err_code_from_errno ());
      ___free_mem (state);
      return e;
    }

  state->encoding = LATIN1_ENCODING;

  return tty_setup_common (t, state);
}

___HIDDEN ___WORD tty_cleanup_os_specific
   ___P((tty_state *state),
        (state)
tty_state *state;)
{
  ___WORD e1 = ___FIX(___NO_ERR);

  if (close (state->fd) < 0)
    e1 = ___FIX(err_code_from_errno ());

  return e1;
}

___HIDDEN ___WORD tty_force_open
   ___P((tty *t),
        (t)
tty *t;)
{
  ___WORD e;
  int fd;
  char term_name[L_ctermid];
  char *term_type;
  tty_state *state = t->state;

  if (state != NULL)
    return ___FIX(___NO_ERR);

  /* tty refers to the console */

  ctermid (term_name); /* get controlling terminal's name */

#if 0
  strcpy (term_name, "/dev/pts/3");/***********************/
#endif

#ifdef LINEEDIT_WITH_NONBLOCKING_IO
  fd = open (term_name, O_RDWR|O_NONBLOCK);
#else
  fd = open (term_name, O_RDWR);
#endif

  if (fd < 0)
    return ___FIX(err_code_from_errno ());

  if ((e = tty_setup_os_specific (t, fd)) != ___FIX(___NO_ERR))
    close (fd); /* ignore error */

  term_type = getenv (CONSOLE_TERMINAL_TYPE_ENV_VAR);

  if ((e = tty_set_terminal_type (t, term_type, DEFAULT_EMACS_BINDINGS))
      != ___FIX(___NO_ERR) ||
      (e = tty_set_encoding (t, NATIVE_ENCODING))
      != ___FIX(___NO_ERR))
    tty_cleanup (t); /* ignore error */

  return e;
}

___HIDDEN ___WORD tty_set_term_mode
   ___P((tty_state *state,
         tty_term_mode term_mode),
        (state,
         term_mode)
tty_state *state;
tty_term_mode term_mode;)
{
  struct termios new_termios;

  new_termios = state->initial_termios;

  switch (term_mode)
    {
    case CBREAK_TERM_MODE:
      new_termios.c_lflag &= ~(ECHO | ICANON);
      new_termios.c_cc[VMIN] = 0;
      new_termios.c_cc[VTIME] = 0;
      break;

    case RAW_TERM_MODE:
      new_termios.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
      new_termios.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON | IXOFF);
      new_termios.c_cflag &= ~(CSIZE | PARENB);
      new_termios.c_cflag |= CS8;
      new_termios.c_oflag &= ~(OPOST);
      new_termios.c_cc[VMIN] = 0;
      new_termios.c_cc[VTIME] = 0;
      break;
    }

  if (tcsetattr (state->fd, TCSAFLUSH, &new_termios) < 0)
    return ___FIX(err_code_from_errno ());

  state->term_mode = term_mode;

#ifdef USE_LINEEDIT
  state->les->linefeed_moves_to_left_margin =
    (new_termios.c_oflag & (OPOST | ONLCR) == (OPOST | ONLCR));
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN void tty_signal_handler
   ___P((int sig),
        (sig)
int sig;)
{
  signal (sig, tty_signal_handler);

  switch (sig)
    {
    case SIGINT:
      tty_broadcast_signal (USER_INTERRUPT_SIGNAL);
      if (tm.user_interrupt_handler != NULL)
        tm.user_interrupt_handler ();
      break;

    case SIGWINCH:
      tty_broadcast_signal (WINDOW_SIZE_SIGNAL);
      {/**************/
        struct winsize size;
        if (ioctl (0, TIOCGWINSZ, &size) < 0)
          ;/************/
        printf ("%d rows, %d columns\n", size.ws_row, size.ws_col);
        fflush (stdout);
      }
      break;

    case SIGCONT:
      tty_broadcast_signal (CONTINUE_SIGNAL);
      tty_set_term_mode (tm.console.state, tm.console.state->term_mode);
      printf ("continued\n");
      fflush (stdout);
      break;
    }
}

___HIDDEN ___WORD tty_read_os_specific
   ___P((tty_state *state,
         ___U8 *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
___U8 *buf;
int len;)
{
  int n = read (state->fd, buf, len);

  if (n < 0)
    return ___FIX(err_code_from_errno ());

  return ___FIX(n);
}

___HIDDEN ___WORD tty_write_os_specific
   ___P((tty_state *state,
         ___U8 *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
___U8 *buf;
int len;)
{
  int n;

#if 1
  n = write (state->fd, buf, len);
#else
  /*****************************/
  {
    static last_time = 0;
    int now;
    struct timeval tv;
    gettimeofday (&tv, NULL);
    now = tv.tv_usec / 33000; /* simulate 300 baud */
    if (0 && now == last_time)
      n = 0;
    else
      {
        int x = len;
        if (x >= 1)
          x = 1;
        n = write (state->fd, buf, x);
        if (n == 1)
          {
#if 1
            printf ("%c", *buf); fflush(stdout);
#endif
          }
        last_time = now;
      }
  }
#endif

  if (n < 0)
    return ___FIX(err_code_from_errno ());

  return ___FIX(n);
}

___HIDDEN ___WORD tty_module_setup_os_specific ___PVOID
{
  signal (SIGINT, tty_signal_handler);
  signal (SIGWINCH, tty_signal_handler);
  signal (SIGCONT, tty_signal_handler);

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_module_cleanup_os_specific ___PVOID
{
  signal (SIGINT, SIG_DFL);
  signal (SIGWINCH, SIG_DFL);
  signal (SIGCONT, SIG_DFL);

  return ___FIX(___NO_ERR);
}

#endif

#ifdef USE_WINDOWS

___HIDDEN ___WORD tty_setup_os_specific
   ___P((tty *t,
         HANDLE input,
         HANDLE output),
        (t,
         input,
         output)
tty *t;
HANDLE input;
HANDLE output;)
{
  tty_state *state;

  t->state = NULL;

  /* check if the same terminal was previously opened */

  state = tm.list;

  if (state != NULL)
    {
      /* Windows only allows one console per process. */

      CloseHandle (input); /* ignore error */
      CloseHandle (output); /* ignore error */
      state->refcount++;
      t->state = state;
      return ___FIX(___NO_ERR);
    }

  /* need to allocate a new tty_state */

  state = ___alloc_mem (sizeof (tty_state));

  if (state == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  state->input = input;
  state->output = output;
  state->input_rlo = 0;
  state->input_rhi = 0;

  {
    DWORD n;
    if (WriteConsoleW (output, NULL, 0, &n, NULL))
      state->encoding = UCS2LE_ENCODING;
    else
      state->encoding = LATIN1_ENCODING;
  }

  return tty_setup_common (t, state);
}

___HIDDEN ___WORD tty_cleanup_os_specific
   ___P((tty_state *state),
        (state)
tty_state *state;)
{
  ___WORD e1 = ___FIX(___NO_ERR);
  ___WORD e2;

  if (!CloseHandle (state->input))
    e1 = ___FIX(err_code_from_GetLastError ());

  if (!CloseHandle (state->output))
    {
      e2 = ___FIX(err_code_from_GetLastError ());
      if (e1 == ___FIX(___NO_ERR))
        e1 = e2;
    }

  return e1;
}

___HIDDEN ___WORD tty_force_open
   ___P((tty *t),
        (t)
tty *t;)
{
  tty_state *state = t->state;

  if (state != NULL)
    return ___FIX(___NO_ERR);

  /* tty refers to the console */

  /********************/

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_set_term_mode
   ___P((tty_state *state,
         tty_term_mode term_mode),
        (state,
         term_mode)
tty_state *state;
tty_term_mode term_mode;)
{
  state->term_mode = term_mode;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_read_os_specific
   ___P((tty_state *state,
         ___U8 *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
___U8 *buf;
int len;)
{
  int i = 0;

  while (state->input_rlo >= state->input_rhi)
    {
      DWORD n;

#ifdef LINEEDIT_WITH_NONBLOCKING_IO

      if (!GetNumberOfConsoleInputEvents (state->input, &n))
        return ___FIX(err_code_from_GetLastError ());

      if (n <= 0)
        goto done;

      if (n > INPUT_RBUF_SIZE)
        n = INPUT_RBUF_SIZE;

#else

      n = INPUT_RBUF_SIZE;

#endif

      if (!((state->encoding == UCS2LE_ENCODING)
            ? ReadConsoleInputW
                (state->input,
                 state->input_rbuf,
                 n,
                 &n)
            : ReadConsoleInputA
                (state->input,
                 state->input_rbuf,
                 n,
                 &n)))
        return ___FIX(err_code_from_GetLastError ());

      state->input_rlo = 0;
      state->input_rhi = n;
    }

  while (state->input_rlo < state->input_rhi)
    {
      INPUT_RECORD *ir = &state->input_rbuf[state->input_rlo];

      switch (ir->EventType)
        {
        case KEY_EVENT:
          {
            DWORD ctrl_keys;
            ___BOOL ctrl;
            ___BOOL alt;
            ___BOOL shift;
            char *seq;
            ___UCS4 c;

            if (!(ir->Event.KeyEvent.bKeyDown &&
                  ir->Event.KeyEvent.wRepeatCount > 0))
              {
                state->input_rlo++;
                break;
              }

            ctrl_keys = ir->Event.KeyEvent.dwControlKeyState;
            ctrl = ctrl_keys & (LEFT_CTRL_PRESSED|RIGHT_CTRL_PRESSED);
            alt = ctrl_keys & (LEFT_ALT_PRESSED|RIGHT_ALT_PRESSED);
            shift = ctrl_keys & SHIFT_PRESSED;

            if (state->encoding == UCS2LE_ENCODING)
              c = ir->Event.KeyEvent.uChar.UnicodeChar;
            else
              c = ir->Event.KeyEvent.uChar.AsciiChar;

            seq = NULL;

            if (!ctrl && !alt && !shift && c == 0)
              switch (ir->Event.KeyEvent.wVirtualKeyCode)
                {
                case VK_BACK:   seq = "\010";     break;
                case VK_UP:     seq = "\033[A";   break;
                case VK_DOWN:   seq = "\033[B";   break;
                case VK_RIGHT:  seq = "\033[C";   break;
                case VK_LEFT:   seq = "\033[D";   break;
                case VK_HOME:   seq = "\033[H";   break;
                case VK_INSERT: seq = "\033[2~";  break;
                case VK_DELETE: seq = "\177";     break;
                case VK_END:    seq = "\033[F";   break;
                case VK_PRIOR:  seq = "\033[5~";  break;
                case VK_NEXT:   seq = "\033[6~";  break;
                case VK_F1:     seq = "\033OP";   break;
                case VK_F2:     seq = "\033OQ";   break;
                case VK_F3:     seq = "\033OR";   break;
                case VK_F4:     seq = "\033OS";   break;
#ifdef SUPPORT_F5_TO_F12
                case VK_F5:     seq = "\033[15~"; break;
                case VK_F6:     seq = "\033[17~"; break;
                case VK_F7:     seq = "\033[18~"; break;
                case VK_F8:     seq = "\033[19~"; break;
                case VK_F9:     seq = "\033[20~"; break;
                case VK_F10:    seq = "\033[21~"; break;
                case VK_F11:    seq = "\033[23~"; break;
                case VK_F12:    seq = "\033[24~"; break;
#endif
                }

            if (seq == NULL)
              {
                if (c == 0)
                  {
                    ir->Event.KeyEvent.wRepeatCount = 0;
                    break;
                  }

                if (ctrl && c == UNICODE_SPACE)
                  c = UNICODE_NUL;

                if (state->encoding == UCS2LE_ENCODING)
                  {
                    if (len < i + 2)
                      goto done;
                    buf[i++] = c & 0xff;
                    buf[i++] = c >> 8;
                  }
                else
                  {
                    if (len < i + 1)
                      goto done;
                    buf[i++] = c & 0xff;
                  }
              }
            else
              {
                while (*seq != '\0')
                  {
                    if (state->encoding == UCS2LE_ENCODING)
                      {
                        if (len < i + 2)
                          goto done;
                        buf[i++] = *seq++;
                        buf[i++] = 0;
                      }
                    else
                      {
                        if (len < i + 1)
                          goto done;
                        buf[i++] = *seq++;
                      }
                  }
              }

            ir->Event.KeyEvent.wRepeatCount--;

            break;
          }

        case MOUSE_EVENT:
          {
            ___WORD e;

            if (i > 0)
              goto done; /* in case paste returns an error */

            state->input_rlo++;

            if (ir->Event.MouseEvent.dwEventFlags == 0 &&
                (ir->Event.MouseEvent.dwButtonState ==
                 (FROM_LEFT_1ST_BUTTON_PRESSED|RIGHTMOST_BUTTON_PRESSED) ||
                 ir->Event.MouseEvent.dwButtonState ==
                 FROM_LEFT_2ND_BUTTON_PRESSED))
              if ((e = lineedit_paste_from_clipboard (state->les))
                  != ___FIX(___NO_ERR))
                return e;

            goto done; /* pasted text is available to be read */
          }

        default:
          state->input_rlo++;
          break;
        }
    }

  done:

  if (i == 0)
    return ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));

  return ___FIX(i);
}

___HIDDEN ___WORD tty_write_os_specific
   ___P((tty_state *state,
         ___U8 *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
___U8 *buf;
int len;)
{
  DWORD n;

  if (state->encoding == UCS2LE_ENCODING)
    {
      len = len >> 1;
      if (!WriteConsoleW (state->output, (wchar_t*)buf, len, &n, NULL))
        return ___FIX(err_code_from_GetLastError ());
      n = n << 1;
    }
  else
    {
      if (!WriteConsoleA (state->output, (char*)buf, len, &n, NULL))
        return ___FIX(err_code_from_GetLastError ());
    }

  return ___FIX(n);
}

___HIDDEN BOOL WINAPI tty_console_event_handler
   ___P((DWORD dwCtrlType),
        (dwCtrlType)
DWORD dwCtrlType;)
{
  tty_state *state = tm.console.state;

  switch (dwCtrlType)
    {
    case CTRL_C_EVENT:
      {
        if (state != NULL &&
            state->term_mode != CBREAK_TERM_MODE)
          return TRUE;

        tty_broadcast_signal (USER_INTERRUPT_SIGNAL);

        if (tm.user_interrupt_handler != NULL)
          tm.user_interrupt_handler ();

        return TRUE;
      }

    case CTRL_BREAK_EVENT:
      {
        if (state != NULL &&
            state->term_mode == RAW_TERM_MODE)
          return TRUE;

        break;
      }

    case CTRL_CLOSE_EVENT:
    case CTRL_LOGOFF_EVENT:
    case CTRL_SHUTDOWN_EVENT:
      break;
    }

  return FALSE; /* let system handle the console event */
}

___HIDDEN ___WORD tty_module_setup_os_specific ___PVOID
{
  if (!SetConsoleCtrlHandler (tty_console_event_handler, TRUE))
    return ___FIX(err_code_from_GetLastError ());

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_module_cleanup_os_specific ___PVOID
{
  SetConsoleCtrlHandler (tty_console_event_handler, FALSE); /* ignore error */
  return ___FIX(___NO_ERR);
}

#endif

___HIDDEN ___WORD tty_setup_console
   ___P((tty *t),
        (t)
tty *t;)
{
  t->state = NULL;
  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_cleanup
   ___P((tty *t),
        (t)
tty *t;)
{
  ___WORD e1;
  ___WORD e2;
  tty_state *state = t->state;

  t->state = NULL; /* avoid dangling pointer */

  if (state == NULL || --state->refcount != 0)
    return ___FIX(___NO_ERR);

#ifdef USE_LINEEDIT
  e1 = lineedit_cleanup (state->les);
#else
  e1 = ___FIX(___NO_ERR);
#endif

  e2 = tty_set_term_mode (state, INITIAL_TERM_MODE);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  e2 = tty_cleanup_os_specific (state);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  /* remove tty_state from circular list of tty_state's */

  if (state->next == state) /* only tty_state in list? */
    tm.list = NULL;
  else
    {
      if (tm.list == state)
        tm.list = state->next;
      state->next->prev = state->prev;
      state->prev->next = state->next;
    }

  ___free_mem (state);

  return e1;
}

___HIDDEN ___WORD tty_set_terminal_type
   ___P((tty *t,
         char *terminal_type,
         ___BOOL emacs_bindings),
        (t,
         terminal_type,
         emacs_bindings)
tty *t;
char *terminal_type;
___BOOL emacs_bindings;)
{
  ___WORD e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT
  e = lineedit_set_terminal_type (state->les, terminal_type, emacs_bindings);
#endif

  return e;
}

___HIDDEN ___WORD tty_set_encoding
   ___P((tty *t,
         tty_encoding encoding),
        (t,
         encoding)
tty *t;
tty_encoding encoding;)
{
  ___WORD e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

  state->encoding = encoding;

  return e;
}

___HIDDEN ___WORD tty_set_input_mode
   ___P((tty *t,
         tty_input_mode input_mode),
        (t,
         input_mode)
tty *t;
tty_input_mode input_mode;)
{
  ___WORD e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

  switch (input_mode)
    {
    default:
    case CBREAK_LINEEDIT_INPUT_MODE:
    case CBREAK_INPUT_MODE:
      e = tty_set_term_mode (state, CBREAK_TERM_MODE);
      break;

    case RAW_LINEEDIT_INPUT_MODE:
    case RAW_INPUT_MODE:
      e = tty_set_term_mode (state, RAW_TERM_MODE);
      break;
    }

  if (e == ___FIX(___NO_ERR))
    state->input_mode = input_mode;

  return e;
}

___HIDDEN ___WORD tty_set_paren_balance_duration
   ___P((tty *t,
         int msecs),
        (t,
         msecs)
tty *t;
int msecs;)
{
  ___WORD e;

  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT
  e = lineedit_set_paren_balance_duration (state->les, msecs);
#endif

  return e;
}

___HIDDEN ___WORD tty_set_max_history_length
   ___P((tty *t,
         int max_history_length),
        (t,
         max_history_length)
tty *t;
int max_history_length;)
{
  ___WORD e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT
  lineedit_set_max_history_length (state->les, max_history_length);
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_set_text_attrs
   ___P((tty *t,
         tty_text_attrs input_attrs,
         tty_text_attrs output_attrs),
        (t,
         input_attrs,
         output_attrs)
tty *t;
tty_text_attrs input_attrs;
tty_text_attrs output_attrs;)
{
  ___WORD e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT
  state->les->input_attrs = input_attrs;
  state->les->output_attrs = output_attrs;
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_read_from_rbuf
   ___P((tty_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
tty_char *buf;
int len;)
{
  int rlo = state->rlo;
  int rhi = state->rhi;
  ___U8 *rbuf = state->rbuf;
  int n = len;

  switch (state->encoding)
    {
    case ASCII_ENCODING:
      {
        ___U8 byte;

        while (rlo < rhi && n > 0)
          {
            byte = rbuf[rlo++];
            if (byte > 0x7f)
              {
                rbuf[--rlo] = byte & 0x7f;
                *buf++ = UNICODE_ESCAPE;
              }
            else
              *buf++ = byte;
            n--;
          }
        break;
      }

    default: /* in case state->encoding is set incorrectly */
    case LATIN1_ENCODING:
      while (rlo < rhi && n > 0)
        {
          *buf++ = rbuf[rlo++];
          n--;
        }
      break;

    case UTF8_ENCODING:
      {
        ___U8 byte;
        ___UCS4 c;

        while (rlo < rhi && n > 0)
          {
            int start = rlo; /* in case we need to backtrack */

            byte = rbuf[rlo++];

            if (byte <= 0x7f)
              c = byte; /* ASCII character */
            else if (byte <= 0xbf || byte > 0xfd)
              c = TTY_CHAR_MAX+1; /* illegal first byte of a UTF-8 */
            else
              {
                ___U8 next;
                ___UCS4 range;

                c = byte; /* upper bits are removed later */
                range = 1<<6;

                loop:
                if (rlo >= rhi)
                  {
                    /*
                     * Not enough bytes to form a UTF-8 encoded
                     * character, leave them there for later.
                     */
                    rlo = start;
                    break; /* stop short */
                  }
                next = rbuf[rlo++];
                if (next <= 0x7f || next > 0xbf)
                  {
                    /*
                     * The byte is not a legal "after-first" byte.
                     * Leave it there so that we will try to process it
                     * as a first byte later.
                     */
                    rlo--;
                    c = TTY_CHAR_MAX+1;
                  }
                else
                  {
                    c = (c << 6) + (next & 0x3f);
                    byte <<= 1;
                    range <<= 5;
                    if (byte & 0x40)
                      goto loop;
                    c &= range-1;
                  }
              }

            if (c > TTY_CHAR_MAX)
              {
#ifdef LINEEDIT_REPORT_CHAR_ENCODING_ERRORS
                if (n == len)
                  {
                    state->rlo = rlo; /* step over the faulty character */
                    return ___FIX(___CTOS_UTF8STRING_ERR);
                  }
                rlo = start; /* error will be reported later */
                break; /* stop short */
#else
                c = UNICODE_REPLACEMENT; /* Unicode replacement character */
#endif
              }

            *buf++ = c;
            n--;
          }
        break;
      }

    case UCS2LE_ENCODING:
      while (rlo+1 < rhi && n > 0)
        {
          ___UCS4 c = rbuf[rlo++];
          c += rbuf[rlo++] << 8;
          *buf++ = c;
          n--;
        }
      break;

    case UCS2BE_ENCODING:
      while (rlo+1 < rhi && n > 0)
        {
          ___UCS4 c = rbuf[rlo++];
          c = (c << 8) + rbuf[rlo++];
          *buf++ = c;
          n--;
        }
      break;

    case UCS4LE_ENCODING:
      while (rlo+3 < rhi && n > 0)
        {
          ___UCS4 c = rbuf[rlo++];
          c += rbuf[rlo++] << 8;
          c += rbuf[rlo++] << 16;
          c += rbuf[rlo++] << 24;
          *buf++ = c;
          n--;
        }
      break;

    case UCS4BE_ENCODING:
      while (rlo+3 < rhi && n > 0)
        {
          ___UCS4 c = rbuf[rlo++];
          c = (c << 8) + rbuf[rlo++];
          c = (c << 8) + rbuf[rlo++];
          c = (c << 8) + rbuf[rlo++];
          *buf++ = c;
          n--;
        }
      break;
    }

  state->rlo = rlo;

  return ___FIX(len-n);
}

___HIDDEN ___WORD tty_rbuf_fill
   ___P((tty_state *state),
        (state)
tty_state *state;)
{
  int rlo = state->rlo;
  int rhi = state->rhi;
  int end = TTY_RBUF_SIZE;
  ___U8 *rbuf = state->rbuf;
  ___WORD n;

  if (rlo != 0)
    {
      int i = 0;
      while (rlo < rhi)
        rbuf[i++] = rbuf[rlo++];
      rlo = 0;
      rhi = i;
      state->rlo = rlo;
      state->rhi = rhi;
    }

  if ((n = tty_read_os_specific (state, &rbuf[rhi], end-rhi)) < ___FIX(0))
    return n;

  state->rhi = rhi + ___INT(n);

  return ___FIX(___NO_ERR);
}

___HIDDEN ___WORD tty_read_noncanonical
   ___P((tty_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
tty_char *buf;
int len;)
{
  ___WORD e;
  ___WORD n;

#ifdef USE_LINEEDIT

  lineedit_state *les = state->les;
  tty_char *p = les->paste_text;

  /*
   * Try to get characters from pasted text first.
   */

  if (les->paste_cancel)
    {
      if (p != NULL)
        {
          ___free_mem (p);
          p = NULL;
          les->paste_text = NULL;
        }
      les->paste_cancel = 0;
    }

  if (p != NULL)
    {
      int start = les->paste_index;
      int end = start + len;
      int i = start;
      tty_char c;

      while (i < end && (c = p[i]) != UNICODE_NUL)
        {
          *buf++ = c;
          i++;
        }

      if (i < end)
        {
          /* reached null character at end of pasted text */

          ___free_mem (p);
          les->paste_text = NULL;
        }

      if (i > start)
        {
          les->paste_index = i;
          return ___FIX(i-start);
        }
    }

#endif

  /*
   * Get characters from rbuf.
   */

  if ((n = tty_read_from_rbuf (state, buf, len)) == ___FIX(0))
    {
      if ((e = tty_rbuf_fill (state)) != ___FIX(___NO_ERR))
        return e;
      n = tty_read_from_rbuf (state, buf, len);
    }

  return n;
}

___HIDDEN ___WORD tty_read
   ___P((tty *t,
         tty_char *buf,
         int len),
        (t,
         buf,
         len)
tty *t;
tty_char *buf;
int len;)
{
  ___WORD e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT

  if (state->input_mode == CBREAK_LINEEDIT_INPUT_MODE ||
      state->input_mode == RAW_LINEEDIT_INPUT_MODE)
    return lineedit_read (state->les, buf, len);

#endif

  return tty_read_noncanonical (state, buf, len);
}

___HIDDEN ___WORD tty_write_to_wbuf
   ___P((tty_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
tty_state *state;
tty_char *buf;
int len;)
{
  int whi = state->whi;
  int end = TTY_WBUF_SIZE;
  ___U8 *wbuf = state->wbuf;
  int n = len;

  switch (state->encoding)
    {
    default: /* in case state->encoding is set incorrectly */
    case ASCII_ENCODING:
    case LATIN1_ENCODING:
      while (whi < end && n > 0)
        {
          ___UCS4 c = *buf++;
          if (c > 0xff)
            {
#ifdef LINEEDIT_REPORT_CHAR_ENCODING_ERRORS
              if (n == len)
                {
                  /* The caller must step over the faulty character. */
                  return ___FIX(___STOC_LATIN1STRING_ERR);
                }
              break; /* stop short */
#else
              c = UNICODE_BELL;
#endif
            }
          wbuf[whi++] = c;
          n--;
        }
      break;

    case UTF8_ENCODING:
      while (whi < end && n > 0)
        {
          ___UCS4 c = *buf++;
          ___U8 *p = &wbuf[whi];
          int bytes = 1;

          if (c <= 0x7f)
            *p = c; /* ASCII character */
          else
            {
              ___UCS4 x = c;
              if      (x <= 0x7ff)     bytes = 2;
              else if (x <= 0xffff)    bytes = 3;
              else if (x <= 0x1fffff)  bytes = 4;
              else if (x <= 0x3ffffff) bytes = 5;
              else                     bytes = 6;
              if (whi+bytes > end)
                break; /* stop short */
              p += bytes;
              switch (bytes)
                {
                case 6:  *--p = 0x80+(x&0x3f); x >>= 6;
                case 5:  *--p = 0x80+(x&0x3f); x >>= 6;
                case 4:  *--p = 0x80+(x&0x3f); x >>= 6;
                case 3:  *--p = 0x80+(x&0x3f); x >>= 6;
                default: *--p = 0x80+(x&0x3f); x >>= 6;
                }
              *--p = 0xff - (0xff>>bytes) + x;
            }
          whi += bytes;
          n--;
        }
      break;

    case UCS2LE_ENCODING:
      while (whi+1 < end && n > 0)
        {
          ___UCS4 c = *buf++;
          wbuf[whi++] = c & 0xff;
          wbuf[whi++] = (c >> 8) & 0xff;
          n--;
        }
      break;

    case UCS2BE_ENCODING:
      while (whi+1 < end && n > 0)
        {
          ___UCS4 c = *buf++;
          wbuf[whi++] = (c >> 8) & 0xff;
          wbuf[whi++] = c & 0xff;
          n--;
        }
      break;

    case UCS4LE_ENCODING:
      while (whi+3 < end && n > 0)
        {
          ___UCS4 c = *buf++;
          wbuf[whi++] = c & 0xff;
          wbuf[whi++] = (c >> 8) & 0xff;
          wbuf[whi++] = (c >> 16) & 0xff;
          wbuf[whi++] = (c >> 24) & 0xff;
          n--;
        }
      break;

    case UCS4BE_ENCODING:
      while (whi+3 < end && n > 0)
        {
          ___UCS4 c = *buf++;
          wbuf[whi++] = (c >> 24) & 0xff;
          wbuf[whi++] = (c >> 16) & 0xff;
          wbuf[whi++] = (c >> 8) & 0xff;
          wbuf[whi++] = c & 0xff;
          n--;
        }
      break;
    }

  state->whi = whi;

  return ___FIX(len-n);
}

___HIDDEN void tty_wbuf_drain
   ___P((tty_state *state),
        (state)
tty_state *state;)
{
  int wlo = state->wlo;
  int whi = state->whi;
  int end = TTY_WBUF_SIZE;
  ___U8 *wbuf = state->wbuf;
  ___WORD n;

  if (wlo < whi)
    {
      if ((n = tty_write_os_specific (state, &wbuf[wlo], whi-wlo)) < ___FIX(0))
        {
          state->werr = n;
          return;
        }

      wlo += ___INT(n);

      if (wlo >= whi || (wlo<<1) >= end)
        {
          /* shift buffer if first half is now empty */
          int i = 0;
          while (wlo < whi)
            wbuf[i++] = wbuf[wlo++];
          wlo = 0;
          state->whi = i;
        }

      state->wlo = wlo;
    }
}

___HIDDEN ___WORD tty_write
   ___P((tty *t,
         tty_char *buf,
         int len),
        (t,
         buf,
         len)
tty *t;
tty_char *buf;
int len;)
{
  ___WORD e;
  ___WORD n;
  tty_state *state;

#ifdef USE_LINEEDIT
  lineedit_state *les;
#endif

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT

  les = state->les;

  if ((e = lineedit_end_paren_balance (les)) != ___FIX(___NO_ERR))
    return e;

  if ((e = lineedit_output_set_text_attrs (les, les->output_attrs))
      != ___FIX(___NO_ERR))
    return e;

  if ((e = lineedit_output_drain (les)) != ___FIX(___NO_ERR))
    return e;

#endif

  if ((e = state->werr) != ___FIX(___NO_ERR))
    {
      state->werr = ___FIX(___NO_ERR);
      return e;
    }

  n = tty_write_to_wbuf (state, buf, len);

  tty_wbuf_drain (state); /* error code returned in state->werr */

#ifdef USE_LINEEDIT

  if (n >= ___FIX(0))
    {
      int pos = les->current.screen_cursor + les->current.cursor_to_point;
      lineedit_track_cursor_movement (les, buf, ___INT(n));
      les->current.cursor_to_point = pos - les->current.screen_cursor;
    }

#endif

  return n;
}

___HIDDEN ___WORD tty_module_setup
   ___P((void (*user_interrupt_handler) ___PVOID),
        (user_interrupt_handler)
void (*user_interrupt_handler) ___PVOID;)
{
  ___WORD e;

  tm.list = NULL;
  tm.user_interrupt_handler = user_interrupt_handler;

  if ((e = tty_module_setup_os_specific ()) != ___FIX(___NO_ERR))
    return e;

#ifdef USE_LINEEDIT
  if ((e = lineedit_module_setup ()) != ___FIX(___NO_ERR))
    {
      tty_module_cleanup_os_specific (); /* ignore error */
      return e;
    }
#endif

  return tty_setup_console (&tm.console);
}

___HIDDEN ___WORD tty_module_cleanup ___PVOID
{
  ___WORD e1;
  ___WORD e2;

  e1 = tty_cleanup (&tm.console);

#ifdef USE_LINEEDIT
  e2 = lineedit_module_cleanup ();
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;
#endif

  e2 = tty_module_cleanup_os_specific ();
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  return e1;
}

/*---------------------------------------------------------------------------*/

#ifndef ___VERSION_MAJOR

___HIDDEN void lineedit_user_interrupt_handler ___PVOID
{
  printf ("ctrl-c\n");
  fflush (stdout);
}

int main ()
{
#ifdef USE_UNIX
  pid_t pid;
  pid_t pgid, t;
  pid = getpid ();
  pgid = getpgid (pid);
  t = tcgetpgrp (0);
#if 0
  printf ("pid = %d  pgid = %d  t = %d\n", pid, pgid, t);
#endif
  if (setpgid (0, 0) < 0)
    exit (1);
  pid = getpid ();
  pgid = getpgid (pid);
#if 0
  printf ("pid = %d  pgid = %d\n", pid, pgid);
#endif
#endif

  if (tty_module_setup (lineedit_user_interrupt_handler) != ___FIX(___NO_ERR))
    exit (1);

#if 0
  if (tty_force_open (&tm.console))
    exit (1);
#endif

#if 0
printf ("isatty (tm.console.state->fd) = %d\n", isatty (tm.console.state->fd));
fflush (stdout);
printf ("ttyname (tm.console.state->fd) = %s\n", ttyname (tm.console.state->fd));
fflush (stdout);
printf ("ttyname (0) = %s\n", ttyname (0));
fflush (stdout);

  write (tm.console.state->fd, "haha\n", 5);
#endif

#if 0
  { int i = 1000000; char c; while (i>0) { if (read (tm.console.state->fd, &c, 1) == 1) {
    write (tm.console.state->fd, "(", 1);
    write (tm.console.state->fd, &c, 1);
    write (tm.console.state->fd, ")", 1);
  }
    i--;
  }
  }
#endif

#if 0
  {
    struct { int yes, no; } foo[10];
    int i;
    int j;
    char c1 = '1';
    char c2 = '2';
    j = 0;
    for (i=0; i<5; i++)
      {
        while (write (tm.console.state->fd, &c1, 1) == 1)
          j++;
        foo[i].yes = j;
        j = 1;
        while (write (tm.console.state->fd, &c2, 1) == -1)
          j++;
        foo[i].no = j;
        j = 1;
      }
    for (i=0; i<5; i++)
      printf ("yes=%d no=%d\n", foo[i].yes, foo[i].no);
    fflush(stdout);
  }

  write (tm.console.state->fd, "hihi\n", 5);
#endif

#if 0
  {
    tty_char c;
    while (1)
      {
        tty_char msg[100];
        int i = 0;
        while (tty_get_char (&tm.console, &c) == 0)
          {
            i++;
            if (0 && i % 10000 == 0)
              {
                tty_char x = '*';
                tty_write (&tm.console, &x, 1);
              }
          }
        msg[0] = ' ';
        msg[1] = c;
        tty_write (&tm.console, msg, 2);

        msg[0] = '(';
        msg[1] = '0'+(c/100);
        msg[2] = '0'+(c/10%10);
        msg[3] = '0'+(c%10);
        msg[4] = ' ';
        msg[5] = '0'+(tm.console.state->row/10);
        msg[6] = '0'+(tm.console.state->row%10);
        msg[7] = ',';
        msg[8] = '0'+(tm.console.state->col/10);
        msg[9] = '0'+(tm.console.state->col%10);
        msg[10] = ')';
        tty_write (&tm.console, msg, 11);

        if (c == 33)
          break;
      }
  }
#endif

#if 1
{
  int fd1;
  int fd2;
  tty tty1;
  tty tty2;

#define tty1 tm.console
#define tty2 tm.console

  ___WORD e;
  tty_char buf[5];

#ifdef USE_WINDOWS

    {
      HANDLE in;
      HANDLE out;

      in = CreateFile
             (_T("CONIN$"),
              GENERIC_READ | GENERIC_WRITE,
              FILE_SHARE_READ | FILE_SHARE_WRITE,
              NULL,
              OPEN_EXISTING,
              0,
              NULL);

      out = CreateFile
              (_T("CONOUT$"),
               GENERIC_READ | GENERIC_WRITE,
               FILE_SHARE_READ | FILE_SHARE_WRITE,
               NULL,
               OPEN_EXISTING,
               0,
               NULL);

      if ((e = tty_setup_os_specific (&tty1, in, out)) != ___FIX(___NO_ERR))
        exit (999);
    }

  if ((e = tty_set_text_attrs
             (&tty1,
              MAKE_TEXT_ATTRS(NORMAL_STYLE,DEFAULT_COLOR,COLOR_CYAN),
              MAKE_TEXT_ATTRS(NORMAL_STYLE,COLOR_BLUE,DEFAULT_COLOR)))
      != ___FIX(___NO_ERR))
    exit (999);

#else

#if 0

  fd1 = open_xterm ();

  if ((e = tty_setup_os_specific (&tty1, fd1)) != ___FIX(___NO_ERR))
    exit (999);

  if ((e = tty_set_terminal_type
             (&tty1,
              "xterm",
              DEFAULT_EMACS_BINDINGS))
      != ___FIX(___NO_ERR))
    exit (999);

  if ((e = tty_set_encoding
             (&tty1,
              NATIVE_ENCODING))
      != ___FIX(___NO_ERR))
    exit (999);

#if 0
  if ((e = tty_set_paren_balance_duration (&tty1, 5000))
      != ___FIX(___NO_ERR))
    exit (999);
#endif

#if 0
  fd2 = open_xterm ();

  if ((e = tty_setup_os_specific (&tty2, fd2)) != ___FIX(___NO_ERR))
    exit (999);

  if ((e = tty_set_terminal_type
             (&tty2,
              "xterm",
              DEFAULT_EMACS_BINDINGS))
      != ___FIX(___NO_ERR))
    exit (999);

  if ((e = tty_set_encoding
             (&tty2,
              NATIVE_ENCODING))
      != ___FIX(___NO_ERR))
    exit (999);
#endif

#endif

#endif

#if 0
  if ((e = tty_set_text_attrs
             (&tty2,
              MAKE_TEXT_ATTRS(NORMAL_STYLE,DEFAULT_COLOR,DEFAULT_COLOR),
              MAKE_TEXT_ATTRS(BOLD_STYLE,DEFAULT_COLOR,DEFAULT_COLOR)))
      != ___FIX(___NO_ERR))
    exit (999);

#endif

#if 0
  if ((e = tty_set_paren_balance_duration (&tty2, 5000))
      != ___FIX(___NO_ERR))
    exit (999);
#endif

  if ((e = tty_set_input_mode (&tty1, CBREAK_LINEEDIT_INPUT_MODE)) != ___FIX(___NO_ERR))
    exit (999);

#if 0
  if ((e = tty_set_input_mode (&tty1, CBREAK_INPUT_MODE)) != ___FIX(___NO_ERR))
    exit (999);

  if ((e = tty_set_input_mode (&tty2, CBREAK_INPUT_MODE)) != ___FIX(___NO_ERR))
    exit (999);
#endif

#if 0
  tty_set_max_history_length (&tm.console, 5);

  tty_set_paren_balance_duration (&tm.console, 500);
#endif

  {
    tty_char c;

    c = '>';
    tty_write (&tty1, &c, 1);
    c = '_';
    tty_write (&tty1, &c, 1);

    c = '>';
    tty_write (&tty2, &c, 1);
    c = '_';
    tty_write (&tty2, &c, 1);
  }

  while (1)
    {
      tty_char c;
      ___WORD n;

      n = tty_read (&tty1, buf, 1);
      if (n > ___FIX(0))
        {
          while (tty_write (&tty2, buf, ___INT(n)) != ___FIX(1))
            ;
          if (buf[0] == '\\')
            break;
          if (buf[0] == 10)
            {
              c = UNICODE_RETURN;
              tty_write (&tty2, &c, 1);
              c = '>';
              tty_write (&tty1, &c, 1);
              c = '_';
              tty_write (&tty1, &c, 1);
            }
        }
      else if (n == ___FIX(0))
        break;
      else if (n != ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)))
        { printf ("error %d\n", ___INT(n)); fflush (stdout);
          if (___INT(n) != ___CTOS_UTF8STRING_ERR) break; }

      n = tty_read (&tty2, buf, 1);
      if (n > ___FIX(0))
        {
          while (tty_write (&tty1, buf, ___INT(n)) != ___FIX(1))
            ;
          if (buf[0] == '\\')
            break;
          if (buf[0] == 10)
            {
              c = UNICODE_RETURN;
              tty_write (&tty1, &c, 1);
              c = '>';
              tty_write (&tty2, &c, 1);
              c = '_';
              tty_write (&tty2, &c, 1);
            }
        }
      else if (n == ___FIX(0))
        break;
      else if (n != ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)))
        { printf ("error %d\n", ___INT(n)); fflush (stdout);
          if (___INT(n) != ___CTOS_UTF8STRING_ERR) break; }
    }
}
#endif

  if (tty_module_cleanup () != ___FIX(___NO_ERR))
    exit (1);

  return 0;
}

#else






void terminal_setup ___PVOID
{
  tty_module_setup (NULL); /* ignore error */
}

void terminal_cleanup ___PVOID
{
  tty_module_cleanup (); /* ignore error */
}

tty *terminal_open ___P((int fd),(fd)
int fd;)
{
  tty *t;
  ___WORD e;

 if (!isatty (fd))
   return NULL;

 t = ___alloc_mem (sizeof (tty));

  if (t != NULL)
    {
      int fd2 = dup (fd);

#if 0
      int fl;

      fl = fcntl (fd2, F_GETFL);
      fcntl (fd2, F_SETFL, fl | O_NONBLOCK); /* set nonblocking mode */
#endif

      if ((e = tty_setup_os_specific (t, fd2)) != ___FIX(___NO_ERR))
        {
          ___free_mem (t);
          t = NULL;
        }
      else
        {
          if ((e = tty_set_terminal_type
                     (t,
                      getenv (CONSOLE_TERMINAL_TYPE_ENV_VAR),
                      DEFAULT_EMACS_BINDINGS))
              != ___FIX(___NO_ERR))
            ;
          if ((e = tty_set_encoding
                     (t,
                      UTF8_ENCODING))
              != ___FIX(___NO_ERR))
            ;
        }
    }

  return t;
}

void terminal_close ___P((tty *t),(t)
tty *t;)
{
  if (t != NULL)
    tty_cleanup (t); /* ignore error */
}

___WORD ___io_set_terminal_type
   ___P((___STREAM *stream, char *term_type, ___BOOL emacs_bindings),
        (stream, term_type, emacs_bindings)
___STREAM *stream;
char *term_type;
___BOOL emacs_bindings;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___WORD e;
      if ((e = tty_set_terminal_type (stream->dummy1, term_type, emacs_bindings))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___WORD ___io_set_encoding
   ___P((___STREAM *stream, int encoding),
        (stream, encoding)
___STREAM *stream;
int encoding;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___WORD e;
      if ((e = tty_set_encoding (stream->dummy1, encoding))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___WORD ___io_set_text_attributes
   ___P((___STREAM *stream, int input, int output),
        (stream, input, output)
___STREAM *stream;
int input;
int output;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___WORD e;
      if ((e = tty_set_text_attrs (stream->dummy1, input, output))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___WORD ___io_set_max_history_length
   ___P((___STREAM *stream, int max_history_length),
        (stream, max_history_length)
___STREAM *stream;
int max_history_length;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___WORD e;
      if ((e = tty_set_max_history_length (stream->dummy1, max_history_length))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___WORD ___io_set_paren_balance_duration
   ___P((___STREAM *stream, int msecs),
        (stream, msecs)
___STREAM *stream;
int msecs;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___WORD e;
      if (msecs >= 0)
        if ((e = tty_set_paren_balance_duration (stream->dummy1, msecs))
            != ___FIX(___NO_ERR))
          ;
    }

  return ___VOID;
}
___WORD ___io_set_input_mode
   ___P((___STREAM *stream, int input_mode),
        (stream, input_mode)
___STREAM *stream;
int input_mode;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___WORD e;
      if ((e = tty_set_input_mode (stream->dummy1, input_mode))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___WORD terminal_get ___P((tty *t),(t)
tty *t;)
{
  tty_char c;
  ___WORD n;

  do
    {
      n = tty_read (t, &c, 1);
      if (n == ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)))
        return ___FIX((EINTR));/******************/
    } while (n == ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)));

  if (n == ___FIX(0))
    return ___EOF;
  else if (n == ___FIX(1))
    return ___CHR(c);
  else
    return n;
}

___WORD terminal_put ___P((tty *t, ___UCS4 c),(t, c)
tty *t;
___UCS4 c;)
{
  tty_char ch = c;
  ___WORD n;

  do
    {
      n = tty_write (t, &ch, 1);
    } while (n == ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN)) ||
             n == ___FIX(0));

  if (n == ___FIX(1))
    return ___FAL;

  return n;
}

#endif
#endif

#if 0

___WORD ___io_set_terminal_type
   ___P((___STREAM *stream, char *term_type, ___BOOL emacs_bindings),
        (stream, term_type, emacs_bindings)
___STREAM *stream;
char *term_type;
___BOOL emacs_bindings;)
{
  return ___VOID;
}

___WORD ___io_set_encoding
   ___P((___STREAM *stream, int encoding),
        (stream, encoding)
___STREAM *stream;
int encoding;)
{
  return ___VOID;
}

___WORD ___io_set_text_attributes
   ___P((___STREAM *stream, int input, int output),
        (stream, input, output)
___STREAM *stream;
int input;
int output;)
{
  return ___VOID;
}

___WORD ___io_set_max_history_length
   ___P((___STREAM *stream, int max_history_length),
        (stream, max_history_length)
___STREAM *stream;
int max_history_length;)
{
  return ___VOID;
}

___WORD ___io_set_paren_balance_duration
   ___P((___STREAM *stream, int msecs),
        (stream, msecs)
___STREAM *stream;
int msecs;)
{
  return ___VOID;
}

___WORD ___io_set_input_mode
   ___P((___STREAM *stream, int input_mode),
        (stream, input_mode)
___STREAM *stream;
int input_mode;)
{
  return ___VOID;
}

___WORD terminal_get ___P((tty *t),(t)
tty *t;)
{
  return 0;
}

___WORD terminal_put ___P((tty *t, ___UCS4 c),(t, c)
tty *t;
___UCS4 c;)
{
  return 0;
}

/*---------------------------------------------------------------------------*/

/* Low-level I/O. */

___HIDDEN ___STREAM io_stdin, io_stdout, io_stderr;

___HIDDEN int default_io_encoding;

___HIDDEN void setup_io ___PVOID
{
  int force_tty = ___setup_params.force_tty;
  int force_unbuffered_io = ___setup_params.force_unbuffered_io;

  default_io_encoding = ___setup_params.default_io_encoding;

#define terminal_setup()
#define terminal_open(x)NULL
#define terminal_close(x)
#define terminal_cleanup()
  terminal_setup ();/***************zzzzzzzzzzzzzz*/

  io_stdin.stdio_stream  = stdin;
  io_stdin.io_encoding   = default_io_encoding;
  io_stdin.force_tty     = force_tty;
  io_stdin.lookahead     = ___FAL;
  io_stdin.dummy1 = terminal_open (fileno (stdin));/***************zzzzzzzzzzzzzz*/

  io_stdout.stdio_stream = stdout;
  io_stdout.io_encoding  = default_io_encoding;
  io_stdout.force_tty    = force_tty;
  io_stdout.lookahead    = ___FAL;
  io_stdout.dummy1 = terminal_open (fileno (stdout));/***************zzzzzzzzzzzzzz*/

  io_stderr.stdio_stream = stderr;
  io_stderr.io_encoding  = default_io_encoding;
  io_stderr.force_tty    = force_tty;
  io_stderr.lookahead    = ___FAL;
  io_stderr.dummy1 = terminal_open (fileno (stderr));/***************zzzzzzzzzzzzzz*/

  if (force_unbuffered_io)
    {
      setbuf (stdin, 0);
      setbuf (stdout, 0);
      setbuf (stderr, 0);
    }

#ifndef BUFFERED_STDOUT

  if ((force_tty || isatty (fileno (stdout))) && !force_unbuffered_io)
    setbuf (stdout, 0); /* unbuffered stdout, for running Gambit in emacs */

#endif

#ifdef USE_MACOS

#ifdef USE_mac_gui

  putchar ('\0');
  fflush (stdout); /* force call of mac_gui_setup */

  atexit (___cleanup); /* make sure cleanup is done if user quits */

#endif

#ifdef USE_SIOUX

  SIOUXSettings.initializeTB     = 1;
  SIOUXSettings.standalone       = 1;
  SIOUXSettings.setupmenus       = 1;
  SIOUXSettings.autocloseonquit  = 1;
  SIOUXSettings.asktosaveonclose = 0;
  SIOUXSettings.showstatusline   = 0;
  SIOUXSettings.tabspaces        = 8;

  putchar ('\0');
  fflush (stdout); /* force creation of window */

  SIOUXSetTitle ("\pInteraction");

  atexit (___cleanup); /* make sure cleanup is done if user quits */

#endif

#endif

#ifdef __WATCOMC__
#ifdef __SW_BW

  _dwSetAboutDlg ("About Gambit",
                  "Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.");
  _dwSetAppTitle ("Gambit");
  _dwSetConTitle (0, "Interaction");

  atexit (___cleanup); /* make sure cleanup is done if user quits */

#endif
#endif
}


#ifdef USE_mac_gui


___HIDDEN int mac_gui_getc (FILE *stdio_stream)
{
  if (stdio_stream == stdin)
    {
      char c;
      long i;
      i = mac_gui_read (&c, 1);
      if (i == -1) /* nothing to read? */
        {
          errno = EINTR;
          return EOF;
        }
      if (i == 0) /* end-of-file? */
        {
          stdio_stream->state.eof = 1;
          return EOF;
        }
      return c;
    }
  else
    return getc (stdio_stream);
}

#undef getc
#define getc(stream)mac_gui_getc(stream)


___HIDDEN void mac_gui_signal_interrupt (void)
{
  ___raise_interrupt (___INTR_USER);
}


short InstallConsole (short fd)
{
  mac_gui_set_signal_interrupt (mac_gui_signal_interrupt);
  return mac_gui_setup ();
}


void RemoveConsole (void)
{
  while (!mac_gui_quit ())
    while (!mac_gui_update ()) ;
  mac_gui_cleanup ();
}


long WriteCharsToConsole (char *buffer, long n)
{
  return mac_gui_write (buffer, n);
}


long ReadCharsFromConsole (char *buffer, long n)
{
  /*
   * This procedure should never get called in normal circumstances
   * because the I/O system uses "mac_gui_getc" to read characters from the console.
   * It is defined just in case there is a need to use other C library I/O functions
   * for debugging.
   */

  int i, m = 0;
  do
    {
      if (mac_gui_update () && mac_gui_quit ())
        ___exit_process (1);
      i = mac_gui_read (buffer, n-m);
      if (i < 1)
        if (m == 0) continue; else break;
      buffer += i;
      m += i;
    } while (m < n);
  return m;
}


char *__ttyname (long fd)
{
  if (fd >= 0 && fd <= 2)
    return "interaction";

  return NULL;
}


#endif


#ifdef USE_autocad_gui

#define STDIN_BUFFER_MAX_LENGTH 256
#define STDIO_EOF 1
#define STDIO_ERROR 2

___HIDDEN char stdin_buffer[STDIN_BUFFER_MAX_LENGTH+1];
___HIDDEN int stdin_buffer_len = 0;
___HIDDEN int stdin_buffer_pos = 0;
___HIDDEN int stdin_flags = 0;
___HIDDEN int stdout_flags = 0;

___HIDDEN void autocad_gui_clearerr (FILE *stdio_stream)
{
  if (stdio_stream == stdin)
    stdin_flags = 0;
  else if (stdio_stream == stdout || stdio_stream == stderr)
    stdout_flags = 0;
  else
    clearerr (stdio_stream);
}

___HIDDEN int autocad_gui_feof (FILE *stdio_stream)
{
  if (stdio_stream == stdin)
    return (stdin_flags & STDIO_EOF);
  else if (stdio_stream == stdout || stdio_stream == stderr)
    return (stdout_flags & STDIO_EOF);
  else
    return feof (stdio_stream);
}

___HIDDEN int autocad_gui_getc (FILE *stdio_stream)
{
  if (stdio_stream == stdin)
    {
      int err;
      if (stdin_buffer_pos >= stdin_buffer_len)
        {
          err = acedGetStringB (1, NULL, stdin_buffer, STDIN_BUFFER_MAX_LENGTH);
          switch (err)
            {
            case RTNORM:
              break;
            case RTCAN:
              ___raise_interrupt (___INTR_USER);
              stdin_flags = (stdin_flags | STDIO_ERROR);
              errno = EINTR;
              return EOF;
            default:
              stdin_flags = (stdin_flags | STDIO_ERROR);
              errno = EACCES;
              return EOF;
            }

          stdin_buffer_pos = 0;
          stdin_buffer_len = 0;
          while (stdin_buffer[stdin_buffer_len] != '\0')
            stdin_buffer_len++;
          stdin_buffer[stdin_buffer_len++] = '\n';
        }
      stdin_flags = 0;
      return (unsigned char)stdin_buffer[stdin_buffer_pos++];
    }
  else
    return getc (stdio_stream);
}

___HIDDEN int autocad_gui_putc (int c, FILE *stdio_stream)
{
  if (stdio_stream == stdout || stdio_stream == stderr)
    {
      char str[2];
      str[0] = c;
      str[1] = '\0';
      if (acedPrompt (str) == RTNORM)
        return c;
      else
        {
          stdout_flags = (stdout_flags | STDIO_ERROR);
          errno = EACCES;
          return EOF;
        }
    }
  else
    return putc (c, stdio_stream);
}

#undef clearerr
#define clearerr(stream)autocad_gui_clearerr(stream)

#undef feof
#define feof(stream)autocad_gui_feof(stream)

#undef getc
#define getc(stream)autocad_gui_getc(stream)

#undef putc
#define putc(c,stream)autocad_gui_putc(c,stream)

#endif


___HIDDEN void cleanup_io ___PVOID
{
  terminal_close (io_stdin.dummy1);/***************zzzzzzzzzzzzzz*/
  terminal_close (io_stdout.dummy1);/***************zzzzzzzzzzzzzz*/
  terminal_close (io_stderr.dummy1);/***************zzzzzzzzzzzzzz*/
  terminal_cleanup ();/***************zzzzzzzzzzzzzz*/
}



___STREAM *___io_stdin ___PVOID
{
  return &io_stdin;
}


___STREAM *___io_stdout ___PVOID
{
  return &io_stdout;
}


___STREAM *___io_stderr ___PVOID
{
  return &io_stderr;
}


___STREAM *___io_open_file
   ___P((char *path, int direction, int encoding),
        (path, direction, encoding)
char *path;
int direction;
int encoding;)
{
  ___STREAM *stream = (___STREAM*) ___alloc_mem (sizeof (___STREAM));
  FILE *stdio_stream;
  char *dir;
  int io_encoding = (encoding == ___IO_DEFAULT_ENCODING)
                    ? default_io_encoding
                    : encoding;
  int binary = (io_encoding > ___IO_UTF8_ENCODING);

  if (stream == 0) /* couldn't allocate space for stream? */
    return 0;

  switch (direction)
    {
    case ___IO_INPUT:
      if (binary) dir = "rb"; else dir = "r";
      break;
    case ___IO_OUTPUT:
      if (binary) dir = "wb"; else dir = "w";
      break;
    default: /* ___IO_INPUT_OUTPUT */
      if (binary) dir = "r+b"; else dir = "r+";
    }

  stdio_stream = fopen (path, dir);
  
  if (stdio_stream == 0) /* stream could not be opened? */
    {
      ___free_mem (stream);
      return 0;
    }

  stream->stdio_stream = stdio_stream;
  stream->kind = ___IO_FILE;
  stream->io_encoding = io_encoding;
  stream->lookahead = ___FAL;
  stream->dummy1 = NULL;/***************zzzzzzzzzzzzzz*/

  return stream;
}


___STREAM *___io_open_pipe
   ___P((char *command, int direction, int encoding),
        (command, direction, encoding)
char *command;
int direction;
int encoding;)
{
  ___STREAM *stream = (___STREAM*) ___alloc_mem (sizeof (___STREAM));
  FILE *stdio_stream;
  char *dir;
  int io_encoding = (encoding == ___IO_DEFAULT_ENCODING)
                    ? default_io_encoding
                    : encoding;
  int binary = (io_encoding > ___IO_UTF8_ENCODING);

  if (stream == 0) /* couldn't allocate space for stream? */
    return 0;

  switch (direction)
    {
    case ___IO_INPUT:
      if (binary) dir = "rb"; else dir = "r";
      break;
    case ___IO_OUTPUT:
      if (binary) dir = "wb"; else dir = "w";
      break;
    default: /* ___IO_INPUT_OUTPUT */
      ___free_mem (stream);
      return 0; /* bidirectional pipes not allowed */
    }

  stdio_stream = popen (command, dir);
  
  if (stdio_stream == 0) /* stream could not be opened? */
    {
      ___free_mem (stream);
      return 0;
    }

  stream->stdio_stream = stdio_stream;
  stream->kind = ___IO_PIPE;
  stream->io_encoding = io_encoding;
  stream->lookahead = ___FAL;
  stream->dummy1 = NULL;/***************zzzzzzzzzzzzzz*/

  return stream;
}


___WORD ___io_close ___P((___STREAM *stream),(stream)
___STREAM *stream;)
{
  FILE *stdio_stream;
  int x;

  if (stream == 0) /* stream has been freed? */
    return ___FAL;

  stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream has been closed? */
    return ___FAL;

  if (stream->kind == ___IO_FILE)
    x = fclose (stdio_stream);
  else
    x = pclose (stdio_stream);

  if (x == EOF)
    {
      ___WORD r = ___FIX(errno);
      errno = ESUCCESS;
      clearerr (stdio_stream);
      return r;
    }

  stream->stdio_stream = 0;
  return ___FAL;
}


___WORD ___io_free ___P((___STREAM *stream),(stream)
___STREAM *stream;)
{
  if (stream == 0) /* stream has been freed? */
    return ___FAL;

  ___free_mem (stream);

  return ___FAL;
}


___HIDDEN ___WORD io_get ___P((___STREAM *stream),(stream)
___STREAM *stream;)
{
  FILE *stdio_stream = stream->stdio_stream;
  int x;

  if (stdio_stream == 0) /* stream is closed? */
    return ___EOF;

  if (stream->dummy1 != NULL) return terminal_get (stream->dummy1);/***************zzzzzzzzzzzzzz*/

  if (feof (stdio_stream))
    {
      clearerr (stdio_stream);
      return ___EOF;
    }

  switch (stream->io_encoding)
    {
    case ___IO_CHAR_ENCODING:
      if ((x = getc (stdio_stream)) != EOF)
        return ___CHR(uchar_to_ucs4(x));
      break;

    case ___IO_LATIN1_ENCODING:
    case ___IO_BYTE_ENCODING:
      if ((x = getc (stdio_stream)) != EOF)
        return ___CHR(x);
      break;

    case ___IO_UTF8_ENCODING:
      {
        unsigned char byte;
        ___UCS4 c;
        int bits;
        if ((x = getc (stdio_stream)) == EOF)
          goto err;
        byte = x;
        if (byte <= 0x7f)
          return ___CHR(byte);
        if (byte <= 0xbf || byte > 0xfd)
          return ___FIX(EIO);
        c = byte; /* upper bits are removed later */
        bits = 6;
        while (byte & 0x40)
          {
            if ((x = getc (stdio_stream)) == EOF)
              goto err;
            if (x <= 0x7f || x > 0xbf)
              return ___FIX(EIO);
            c = (c << 6) + (x & 0x3f);
            byte <<= 1;
            bits += 5;
          }
        c &= ((unsigned long)1<<bits)-1;
        if ((c > 0xd7ff && c <= 0xdfff) ||
            (c > 0xfffd && c <= 0xffff))
          return ___FIX(EIO);
        if ((c & (~(unsigned long)0<<(bits-5))) == 0)
          return ___FIX(EIO);
        if (c > ___MAX_CHR)
          return ___FIX(EDOM);
        return ___CHR(c);
      err:
        break;
      }

    case ___IO_UCS2_ENCODING:
      if ((x = getc (stdio_stream)) != EOF)
        {
          ___UCS4 c = x;
          if ((x = getc (stdio_stream)) != EOF)
            {
              c = (c<<8) + x;
              if ((c > 0xd7ff && c <= 0xdfff) ||
                  (c > 0xfffd && c <= 0xffff))
                return ___FIX(EIO);
              if (c > ___MAX_CHR)
                return ___FIX(EDOM);
              return ___CHR(c);
            }
          if (feof (stdio_stream))
            {
              clearerr (stdio_stream);
              return ___FIX(EIO);
            }
          clearerr (stdio_stream);
          return ___FIX(errno);
        }
      break;

    default: /* ___IO_UCS4_ENCODING */
      if ((x = getc (stdio_stream)) != EOF)
        {
          ___UCS4 c = x;
          if ((x = getc (stdio_stream)) != EOF)
            {
              c = (c<<8) + x;
              if ((x = getc (stdio_stream)) != EOF)
                {
                  c = (c<<8) + x;
                  if ((x = getc (stdio_stream)) != EOF)
                    {
                      c = (c<<8) + x;
                      if ((c > 0xd7ff && c <= 0xdfff) ||
                          (c > 0xfffd && c <= 0xffff))
                        return ___FIX(EIO);
                      if (c > ___MAX_CHR)
                        return ___FIX(EDOM);
                      return ___CHR(c);
                    }
                }
            }
          if (feof (stdio_stream))
            {
              clearerr (stdio_stream);
              return ___FIX(EIO);
            }
          clearerr (stdio_stream);
          return ___FIX(errno);
        }
      break;
    }

  if (feof (stdio_stream)
#ifdef __BORLANDC__
      || errno == EPIPE
#else
#ifdef __WATCOMC__
      || errno < 0 /* bug in Watcom C??? apparently errno<0 at end of pipe */
#endif
#endif
     )
    {
      clearerr (stdio_stream);
      return ___EOF;
    }

  clearerr (stdio_stream);
  return ___FIX(errno);
}

___WORD ___io_getc ___P((___STREAM *stream),(stream)
___STREAM *stream;)
{
  ___WORD r = stream->lookahead;

  if (r == ___FAL)
    return io_get (stream);

  stream->lookahead = ___FAL;
  return r;
}

___WORD ___io_peek ___P((___STREAM *stream),(stream)
___STREAM *stream;)
{
  ___WORD r = stream->lookahead;

  if (r == ___FAL)
    {
      r = io_get (stream);
      if (!___FIXNUMP(r))
        stream->lookahead = r;
    }

  return r;
}

___WORD ___io_ready ___P((___STREAM *stream),(stream)
___STREAM *stream;)
{
  return ___TRU;
}

___WORD ___io_putc ___P((___STREAM *stream, ___UCS4 c),(stream, c)
___STREAM *stream;
___UCS4 c;)
{
  FILE *stdio_stream = stream->stdio_stream;
  int x;

  if (stdio_stream == 0) /* stream is closed? */
    return ___FAL;

  if (stream->dummy1 != NULL) return terminal_put (stream->dummy1, c);/***************zzzzzzzzzzzzzz*/

  switch (stream->io_encoding)
    {
    case ___IO_CHAR_ENCODING:
      {
        ___UCS4 uc = ucs4_to_uchar(c);
        if (uc > UCHAR_MAX)
          return ___FIX(EDOM);
        x = putc (uc, stdio_stream);
        break;
      }

    case ___IO_LATIN1_ENCODING:
    case ___IO_BYTE_ENCODING:
      if (c > 0xff)
        return ___FIX(EDOM);
      x = putc (c, stdio_stream);
      break;

    case ___IO_UTF8_ENCODING:
      {
        char buf[6];
        char *p1 = buf, *p2 = buf;
        ___utf8_put (&p2, c);
        x = EOF;
        while (p1 < p2 && (x = putc (*p1++, stdio_stream)) != EOF) ;
        break;
      }

    case ___IO_UCS2_ENCODING:
      if (c > 0xffff)
        return ___FIX(EDOM);
      (x = putc ((c>>8) & 0xff, stdio_stream)) != EOF &&
      (x = putc (c      & 0xff, stdio_stream));
      break;

    default: /* ___IO_UCS4_ENCODING */
      if (c > 0x7fffffff)
        return ___FIX(EDOM);
      (x = putc ((c>>24) & 0xff, stdio_stream)) != EOF &&
      (x = putc ((c>>16) & 0xff, stdio_stream)) != EOF &&
      (x = putc ((c>>8)  & 0xff, stdio_stream)) != EOF &&
      (x = putc (c       & 0xff, stdio_stream));
      break;
    }

  if (x == EOF)
    {
      ___WORD r = ___FIX(errno);
      errno = ESUCCESS;
      clearerr (stdio_stream);
      return r;
    }

  return ___FAL;
}

___WORD ___io_flush ___P((___STREAM *stream),(stream)
___STREAM *stream;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___FAL;

  if (fflush (stdio_stream) == EOF)
    {
      ___WORD r = ___FIX(errno);
      errno = ESUCCESS;
      clearerr (stdio_stream);
      return r;
    }

  return ___FAL;
}

___WORD ___io_width ___P((___STREAM *stream),(stream)
___STREAM *stream;)
{
  return ___FIX(79);
}

___WORD ___io_isatty ___P((___STREAM *stream),(stream)
___STREAM *stream;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___FAL;

  return ___BOOLEAN((stream->force_tty || isatty (fileno (stdio_stream))));
}

#endif

___HIDDEN void setup_io ___PVOID
{
#ifdef USE_MACOS

#ifdef USE_mac_gui

  putchar ('\0');
  fflush (stdout); /* force call of mac_gui_setup */

  atexit (___cleanup); /* make sure cleanup is done if user quits */

#endif

#ifdef USE_SIOUX

  SIOUXSettings.initializeTB     = 1;
  SIOUXSettings.standalone       = 1;
  SIOUXSettings.setupmenus       = 1;
  SIOUXSettings.autocloseonquit  = 1;
  SIOUXSettings.asktosaveonclose = 0;
  SIOUXSettings.showstatusline   = 0;
  SIOUXSettings.tabspaces        = 8;

  putchar ('\0');
  fflush (stdout); /* force creation of window */

  SIOUXSetTitle ("\pInteraction");

  atexit (___cleanup); /* make sure cleanup is done if user quits */

#endif

  has_IdleUpdate = trap_exists (_IdleUpdate);
  has_WaitNextEvent = trap_exists (_WaitNextEvent);
  has_OSDispatch = trap_exists (_OSDispatch);

  {
    long response;

    has_FindFolder = (Gestalt (gestaltFindFolderAttr, &response) == noErr &&
                      test_bit (response, gestaltFindFolderPresent));
    has_AliasMgr = (Gestalt (gestaltAliasMgrAttr, &response) == noErr &&
                    test_bit (response, gestaltAliasMgrPresent));
    has_AppleEvents = (Gestalt (gestaltAppleEventsAttr, &response) == noErr &&
                       test_bit (response, gestaltAppleEventsPresent));
  }

#endif

#ifdef __WATCOMC__
#ifdef __SW_BW

  _dwSetAboutDlg ("About Gambit",
                  "Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.");
  _dwSetAppTitle ("Gambit");
  _dwSetConTitle (0, "Interaction");

  atexit (___cleanup); /* make sure cleanup is done if user quits */

#endif
#endif
}


___HIDDEN void cleanup_io ___PVOID
{
}


/*---------------------------------------------------------------------------*/

void ___setup_os ___PVOID
{
  /*
   * To perform correct cleanup when the program terminates an
   * "atexit (___cleanup)" is performed in "setup_io" in certain
   * environments.  There must not be any possibility of program
   * termination through "exit (...)" between the "atexit (...)"
   * and the entry of "___setup_mem".  This guarantees that
   * "___cleanup" does not access dangling pointers.
   */

  setup_os_specific ();
  setup_fp ();
  setup_process_times ();
  setup_alloc_mem ();
  setup_io ();
  setup_dynamic_load ();
  setup_interrupt_handling ();
}


void ___cleanup_os ___PVOID
{
  cleanup_interrupt_handling ();
  cleanup_dynamic_load ();
  cleanup_io ();
  cleanup_alloc_mem ();
  cleanup_process_times ();
  cleanup_fp ();
  cleanup_os_specific ();
}


/*---------------------------------------------------------------------------*/
